#include <cppunit/extensions/HelperMacros.h>
#include <DAQChannelMap/ChannelMapException.h>
#include <DAQChannelMap/DAQChannelMap.h>

#define PATTERN1 0x01234567
#define PATTERN2 0x76543210
#define PATTERN3 0x89abcdef
#define PATTERN4 0xfedcba98
#define PATTERN5 0xf0f0f0f0
#define PATTERN6 0xe1e1e1e1
#define PATTERN7 0xd2d2d2d2
#define PATTERN8 0xc3c3c3c3
#define PATTERN9 0xb4b4b4b4
#define PATTERN0 0xa5a5a5a5

#define BUFFER_WORDS 1000
#define SMALL_BUFFER_WORDS 1
 
using namespace daqchannelmap;

class DAQChannelMapTests : public CppUnit::TestFixture
{
  CPPUNIT_TEST_SUITE(DAQChannelMapTests);
  CPPUNIT_TEST(testBasicChannelGetOnline);
  CPPUNIT_TEST(testDiBlockLookups);
  CPPUNIT_TEST(testViewLookups);


  CPPUNIT_TEST(testBasicOfflineChannelGet);
  CPPUNIT_TEST(testBasicChannelGetOffline);
  CPPUNIT_TEST(testDiBlockLookupsOffline);

  CPPUNIT_TEST(testPlaneLookupsOffline);
  CPPUNIT_TEST(testViewLookupsOffline);

  CPPUNIT_TEST_SUITE_END();

public:
  void setUp();
  void tearDown();

  void testBasicChannelGetOnline();
  void testDiBlockLookups();
  void testViewLookups();

  void testBasicChannelGetOffline();
  void testBasicOfflineChannelGet();
  void testDiBlockLookupsOffline();

  void testPlaneLookupsOffline();
  void testViewLookupsOffline();
  
private:

  int vector_count;

};

// Register all the tests
CPPUNIT_TEST_SUITE_NAMED_REGISTRATION(DAQChannelMapTests, "DAQChannelMap");
CPPUNIT_REGISTRY_ADD_TO_DEFAULT("DAQChannelMap");

void
DAQChannelMapTests::setUp()
{ 
}

void
DAQChannelMapTests::tearDown()
{ 
}

void
DAQChannelMapTests::testBasicChannelGetOnline()
{
  // Retrieve the channel map class (singleton)
  DAQChannelMap* chanMap;
  chanMap = DAQChannelMap::getInstance();
  
  dchan theChan;


  vector_count = 0;
  printf("\nTesting Online Channel encoding/decoding\n");
  // Test channel encoding/decoding
  for(int det             = IPND_ID; det < FARDET_ID;           ++det){
    for(int diblock       = 1; diblock < NUM_DIBLOCK_DET[det];  ++diblock){
      for(int dcm         = 1; dcm     < NUM_DCMS_DIBLOCK[det]; ++dcm    ){
	for(int port      = 1; port    < NUM_DCM_PORTS;         ++port   ){
	  for(int pixel   = 1; pixel   < (NUM_PIXELS+1);        ++pixel  ){
	    theChan = DAQChannelMap::encodeDChan(det,diblock,dcm,port,pixel);
	    vector_count++;
	    CPPUNIT_ASSERT(DAQChannelMap::getDetector(theChan) == det);
	    CPPUNIT_ASSERT(DAQChannelMap::getDiBlock(theChan)  == diblock);
	    CPPUNIT_ASSERT(DAQChannelMap::getDCMLocal(theChan) == dcm);
	    CPPUNIT_ASSERT(DAQChannelMap::getDCMPort(theChan)  == port);
	    CPPUNIT_ASSERT(DAQChannelMap::getPixel(theChan)    == pixel);
	  } // endfor pixel
	} // endfor port
      } // endfor dcm
    } // endfor diblock
  } // endfor det

  printf("Channel encoding/decoding complete\n");
  printf("Run Condition: %d test vectors run", vector_count);
    
}

void DAQChannelMapTests::testBasicChannelGetOffline()
{
  // Retrieve the channel map class (singleton)
  DAQChannelMap* chanMap;
  chanMap = DAQChannelMap::getInstance();
  
  lchan theChan;
  
  printf("\nTesting Offline Channel encoding/decoding\n");
  vector_count = 0;
  // Test channel encoding/decoding
  for(int det             = IPND_ID;              det   < FARDET_ID;                                ++det){
    for(int plane         = START_CHAN_IDX;       plane < NUM_PLANES_DIBLOCK*NUM_DIBLOCK_DET[det];  ++plane){
      for(int cell        = START_CHAN_IDX;       cell  < NUM_YCHAN_PLANE[det];                     ++cell    ){
	    theChan = DAQChannelMap::encodeLChan(det,plane,cell);
	    vector_count++;
	    CPPUNIT_ASSERT(DAQChannelMap::getDetectorOffline(theChan) == det);
	    CPPUNIT_ASSERT(DAQChannelMap::getPlaneOffline(theChan)    == plane);
	    CPPUNIT_ASSERT(DAQChannelMap::getCellOffline(theChan)     == cell);

#ifdef VERBOSE_DEBUG
	    printf("%2d %4d %3d\t0x%08x\t %d %d %d\n", det, plane, cell, theChan,
		   DAQChannelMap::getDetectorOffline(theChan), 
		   DAQChannelMap::getPlaneOffline(theChan),
		   DAQChannelMap::getCellOffline(theChan));
#endif /* VERBOSE_DEBUG */
      } // endfor dcm
    } // endfor diblock
  } // endfor det
  printf("Online Channel encoding/decoding complete\n");
  printf("Run Condition: %d test vectors run", vector_count);
    
}

void DAQChannelMapTests::testBasicOfflineChannelGet()
{
  // Retrieve the channel map class (singleton)
  DAQChannelMap* chanMap;
  chanMap = DAQChannelMap::getInstance();
  
  dchan theChan;
  
  printf("\nTesting Channel encoding/decoding\n");

  int max_planes;
  int max_cells;

  // Test channel encoding/decoding
  for(int det             = IPND_ID; det   < FARDET_ID;             ++det   ){
    max_planes = NUM_PLANES_DIBLOCK * NUM_DIBLOCK_DET[det];
      max_cells = NUM_XCHAN_PLANE[det] > NUM_YCHAN_PLANE[det] ? 
      NUM_XCHAN_PLANE[det] : NUM_YCHAN_PLANE[det];
    for(int plane         = 1;       plane < max_planes;            ++plane ){
      for(int cell        = 1;       cell  < max_cells;             ++cell  ){
	theChan = DAQChannelMap::encodeLChan(det,plane,cell);
	CPPUNIT_ASSERT(DAQChannelMap::getDetector(theChan)      == det);
	CPPUNIT_ASSERT(DAQChannelMap::getPlaneOffline(theChan)  == plane);
	CPPUNIT_ASSERT(DAQChannelMap::getCellOffline(theChan)   == cell);
      } // endfor cell
    } // endfor plane
  } // endfor det
  printf("Offline Channel encoding/decoding complete\n");
    
}

void
DAQChannelMapTests::testDiBlockLookups()
{
  // Retrieve the channel map class (singleton)
  DAQChannelMap* chanMap;
  chanMap = DAQChannelMap::getInstance();
  
  dchan theChan;
  printf("\n");

  printf("\nTesting Diblock Information retrieval\n");

  int CHAN_AB_NEAR = (NUM_X_PLANES_A_BLOCK + NUM_X_PLANES_B_BLOCK)*NUM_CHAN_X_PLANE_IPND + 
    (NUM_Y_PLANES_A_BLOCK + NUM_Y_PLANES_B_BLOCK)*NUM_CHAN_Y_PLANE_IPND; 
  int CHAN_AA_NEAR = (NUM_X_PLANES_A_BLOCK + NUM_X_PLANES_A_BLOCK)*NUM_CHAN_X_PLANE_IPND + 
    (NUM_Y_PLANES_A_BLOCK + NUM_Y_PLANES_A_BLOCK)*NUM_CHAN_Y_PLANE_IPND; 

  int CHAN_AB_FAR = (NUM_X_PLANES_A_BLOCK + NUM_X_PLANES_B_BLOCK)*NUM_CHAN_X_PLANE_FARDET + 
    (NUM_Y_PLANES_A_BLOCK + NUM_Y_PLANES_B_BLOCK)*NUM_CHAN_Y_PLANE_FARDET; 
  int CHAN_AA_FAR = (NUM_X_PLANES_A_BLOCK + NUM_X_PLANES_A_BLOCK)*NUM_CHAN_X_PLANE_FARDET + 
    (NUM_Y_PLANES_A_BLOCK + NUM_Y_PLANES_A_BLOCK)*NUM_CHAN_Y_PLANE_FARDET; 


  int NEAR_CHAN_VECTOR[] = {CHAN_IDX_0_1
			  CHAN_AB_NEAR, CHAN_AB_NEAR, CHAN_AA_NEAR,0x0};
  int FAR_CHAN_VECTOR[]  = {CHAN_IDX_0_1
			  CHAN_AB_FAR, CHAN_AB_FAR, CHAN_AA_FAR,
			  CHAN_AB_FAR, CHAN_AB_FAR, CHAN_AB_FAR,
			  CHAN_AB_FAR, CHAN_AA_FAR, CHAN_AB_FAR,
			  CHAN_AB_FAR, CHAN_AB_FAR, CHAN_AB_FAR,
			  CHAN_AA_FAR, CHAN_AB_FAR, CHAN_AB_FAR,
			  0x0};

  printf("-----\n");
  for(int diblock=START_CHAN_IDX; diblock <= NUM_DIBLOCK_IPND; diblock++){
    printf("IPND Detector, Diblock %2d = %02X ",diblock, 
	   DAQChannelMap::getDiBlockType(diblock, IPND_ID));
    CPPUNIT_ASSERT(DAQChannelMap::getDiBlockType(diblock, IPND_ID) == 
		   DetectorDiBlockConfig_IPND[diblock]);
    CPPUNIT_ASSERT(DAQChannelMap::getChanPerDiBlockType(DAQChannelMap::getDiBlockType(diblock, IPND_ID),
							IPND_ID) == NEAR_CHAN_VECTOR[diblock]);
    printf("(Channels: %3d)\n",DAQChannelMap::getChanPerDiBlockType(DAQChannelMap::getDiBlockType(diblock, IPND_ID),IPND_ID));
  } // endfor diblock

  printf("-----\n");
  for(int diblock=START_CHAN_IDX; diblock <= NUM_DIBLOCK_NDOS; diblock++){
    printf("NDOS Detector, Diblock %2d = %02X ",diblock, 
	   DAQChannelMap::getDiBlockType(diblock, NDOS_ID));
    CPPUNIT_ASSERT(DAQChannelMap::getDiBlockType(diblock, NDOS_ID) == 
		   DetectorDiBlockConfig_NDOS[diblock]);
    CPPUNIT_ASSERT(DAQChannelMap::getChanPerDiBlockType(DAQChannelMap::getDiBlockType(diblock, NDOS_ID),
							NDOS_ID) == NEAR_CHAN_VECTOR[diblock]);
    printf("(Channels: %3d)\n",DAQChannelMap::getChanPerDiBlockType(DAQChannelMap::getDiBlockType(diblock, NDOS_ID),NDOS_ID));
		   
  } // endfor diblock

  printf("-----\n");
  for(int diblock=START_CHAN_IDX; diblock <= NUM_DIBLOCK_NEARDET; diblock++){
    printf("Near Detector, Diblock %2d = %02X ",diblock, 
	   DAQChannelMap::getDiBlockType(diblock, NEARDET_ID));
    CPPUNIT_ASSERT(DAQChannelMap::getDiBlockType(diblock, NEARDET_ID) == 
		   DetectorDiBlockConfig_NEARDET[diblock]);
    CPPUNIT_ASSERT(DAQChannelMap::getChanPerDiBlockType(DAQChannelMap::getDiBlockType(diblock, NEARDET_ID),
							NEARDET_ID) == NEAR_CHAN_VECTOR[diblock]);
		   
    printf("(Channels: %3d)\n",DAQChannelMap::getChanPerDiBlockType(DAQChannelMap::getDiBlockType(diblock, NEARDET_ID),NEARDET_ID));
  } // endfor diblock

  printf("-----\n");
  for(int diblock=START_CHAN_IDX; diblock <= NUM_DIBLOCK_FARDET; diblock++){
    printf("Far  Detector, Diblock %2d = %02X ",diblock, 
	   DAQChannelMap::getDiBlockType(diblock, FARDET_ID));
    CPPUNIT_ASSERT(DAQChannelMap::getDiBlockType(diblock, FARDET_ID) == 
		   DetectorDiBlockConfig_FARDET[diblock]);
    CPPUNIT_ASSERT(DAQChannelMap::getChanPerDiBlockType(DAQChannelMap::getDiBlockType(diblock, FARDET_ID),
							FARDET_ID) == FAR_CHAN_VECTOR[diblock]);

    printf("(Channels: %3d)\n",DAQChannelMap::getChanPerDiBlockType(DAQChannelMap::getDiBlockType(diblock, FARDET_ID),FARDET_ID));
  } // endfor diblock

  printf("-----\n");

}

void DAQChannelMapTests::testDiBlockLookupsOffline(){
  // Retrieve the channel map class (singleton)
  DAQChannelMap* chanMap;
  chanMap = DAQChannelMap::getInstance();

  lchan theLChan;

  printf("\nTesting Offline Diblock decoding\n");

  // Define some known test vectors and see what they return
  uint32_t tvect[20];		// Test Vectors
  tvect[0] = DAQChannelMap::encodeLChan(IPND_ID, 1 + END_OFFSET_IDX, 5);
  tvect[1] = DAQChannelMap::encodeLChan(IPND_ID, 2 + END_OFFSET_IDX, 5);
  tvect[2] = DAQChannelMap::encodeLChan(IPND_ID, 30 + END_OFFSET_IDX, 37);
  tvect[3] = DAQChannelMap::encodeLChan(IPND_ID, 31 + END_OFFSET_IDX, 37);
  tvect[4] = DAQChannelMap::encodeLChan(IPND_ID, 32 + END_OFFSET_IDX, 37);
  tvect[5] = DAQChannelMap::encodeLChan(IPND_ID, 61 + END_OFFSET_IDX, 3);
  tvect[6] = DAQChannelMap::encodeLChan(IPND_ID, 62 + END_OFFSET_IDX, 3);
  tvect[7] = DAQChannelMap::encodeLChan(IPND_ID, 63 + END_OFFSET_IDX, 3);
  tvect[8] = DAQChannelMap::encodeLChan(IPND_ID, 93 + END_OFFSET_IDX, 3);
  tvect[9] = DAQChannelMap::encodeLChan(IPND_ID, 94 + END_OFFSET_IDX, 3);



  uint32_t avect1[] = {1,1,1,1,2,2,2,3,3,4};
  uint32_t avect2[] = {1,1,1,1,1,1,1,2,2,2};
  uint32_t avect3[] = {AB_BLOCK,AB_BLOCK,AB_BLOCK,AB_BLOCK,AB_BLOCK,
		       AB_BLOCK,AB_BLOCK,AB_BLOCK,AB_BLOCK,AB_BLOCK};

  for(int i=0; i< 10; i++){
    printf("0x%08x |\t%d %d %x | \t%d %d %x\n",tvect[i],avect1[i],avect2[i],avect3[i],
	   DAQChannelMap::getBlockOffline(tvect[i]),
	   DAQChannelMap::getDiBlockOffline(tvect[i]),
	   DAQChannelMap::getDiBlockTypeOffline(tvect[i])
	   );
    CPPUNIT_ASSERT(DAQChannelMap::getBlockOffline(tvect[i])           == avect1[i]);
    CPPUNIT_ASSERT(DAQChannelMap::getDiBlockOffline(tvect[i])         == avect2[i]);
    CPPUNIT_ASSERT(DAQChannelMap::getDiBlockTypeOffline(tvect[i])     == avect3[i]);
  }

  printf("---\n");

  tvect[0] = DAQChannelMap::encodeLChan(NDOS_ID, 1 + END_OFFSET_IDX, 5);
  tvect[1] = DAQChannelMap::encodeLChan(NDOS_ID, 2 + END_OFFSET_IDX, 5);
  tvect[2] = DAQChannelMap::encodeLChan(NDOS_ID, 30 + END_OFFSET_IDX, 37);
  tvect[3] = DAQChannelMap::encodeLChan(NDOS_ID, 31 + END_OFFSET_IDX, 37);
  tvect[4] = DAQChannelMap::encodeLChan(NDOS_ID, 32 + END_OFFSET_IDX, 37);
  tvect[5] = DAQChannelMap::encodeLChan(NDOS_ID, 61 + END_OFFSET_IDX, 3);
  tvect[6] = DAQChannelMap::encodeLChan(NDOS_ID, 62 + END_OFFSET_IDX, 3);
  tvect[7] = DAQChannelMap::encodeLChan(NDOS_ID, 63 + END_OFFSET_IDX, 3);
  tvect[8] = DAQChannelMap::encodeLChan(NDOS_ID, 93 + END_OFFSET_IDX, 3);
  tvect[9] = DAQChannelMap::encodeLChan(NDOS_ID, 94 + END_OFFSET_IDX, 3);
  tvect[10] = DAQChannelMap::encodeLChan(NDOS_ID, 123 + END_OFFSET_IDX, 3);
  tvect[11] = DAQChannelMap::encodeLChan(NDOS_ID, 124 + END_OFFSET_IDX, 3);
  tvect[12] = DAQChannelMap::encodeLChan(NDOS_ID, 125 + END_OFFSET_IDX, 3);
  tvect[13] = DAQChannelMap::encodeLChan(NDOS_ID, 126 + END_OFFSET_IDX, 3);
  tvect[14] = DAQChannelMap::encodeLChan(NDOS_ID, 125 + END_OFFSET_IDX, 3);
  tvect[15] = DAQChannelMap::encodeLChan(NDOS_ID, 126 + END_OFFSET_IDX, 3);

  uint32_t avect21[] = {1,1,1,1,2,2,2,3,3,4,4,4,5,5};
  uint32_t avect22[] = {1,1,1,1,1,1,1,2,2,2,2,2,3,3};
  uint32_t avect23[] = {AB_BLOCK,AB_BLOCK,AB_BLOCK,AB_BLOCK,AB_BLOCK,
			AB_BLOCK,AB_BLOCK,AB_BLOCK,AB_BLOCK,AB_BLOCK,
			AB_BLOCK,AB_BLOCK,AA_BLOCK,AA_BLOCK};


  for(int i=0; i< 14; i++){
    printf("0x%08x |\t%d %d %x | \t%d %d %x\n",tvect[i],avect21[i],avect22[i],avect23[i],
	   DAQChannelMap::getBlockOffline(tvect[i]),
	   DAQChannelMap::getDiBlockOffline(tvect[i]),
	   DAQChannelMap::getDiBlockTypeOffline(tvect[i])
	   );
    CPPUNIT_ASSERT(DAQChannelMap::getBlockOffline(tvect[i])           == avect21[i]);
    CPPUNIT_ASSERT(DAQChannelMap::getDiBlockOffline(tvect[i])         == avect22[i]);
    CPPUNIT_ASSERT(DAQChannelMap::getDiBlockTypeOffline(tvect[i])     == avect23[i]);
  }

  printf("---\n");

}

void DAQChannelMapTests::testViewLookups(){
  // Retrieve the channel map class (singleton)
  DAQChannelMap* chanMap;
  chanMap = DAQChannelMap::getInstance();

  dchan theDChan;
  uint32_t view;
  
  printf("\nTesting Online View decoding\n");
  // Test channel encoding/decoding
  for(int det             = IPND_ID; det <= FARDET_ID;           ++det){
    for(int diblock       = 1; diblock <= NUM_DIBLOCK_DET[det];  ++diblock){
      for(int dcm         = 1; dcm     <= NUM_DCMS_DIBLOCK[det]; ++dcm    ){
	for(int port      = 1; port    <= NUM_DCM_PORTS;         ++port   ){
	  for(int pixel   = 1; pixel   <= (NUM_PIXELS+1);        ++pixel  ){
	    theDChan = DAQChannelMap::encodeDChan(det,diblock,dcm,port,pixel);

	    if(dcm <= NUM_X_DCMS[det]){
	      CPPUNIT_ASSERT(DAQChannelMap::getView(theDChan) == X_VIEW);
	    }
#ifdef VERBOSE_DEBUG
	    printf("%2d %2d %2d\t0x%08x\t %x \n",
		   DAQChannelMap::getDetector(theDChan), 
		   DAQChannelMap::getDiBlock(theDChan),  
		   DAQChannelMap::getDCMLocal(theDChan),
		   theDChan,
		   DAQChannelMap::getView(theDChan)); 
	    
#endif /* VERBOSE_DEBUG */
	  } // endfor pixel
	} // endfor port
      } // endfor dcm
    } // endfor diblock
  } // endfor det
  printf("Online View decoding complete\n");

  
  lchan theLChan;
  
  printf("\nTesting Offline View decoding\n");

  for(int det_ID = IPND_ID; det_ID <= FARDET_ID; det_ID++){
  // Define some known test vectors and see what they return
  uint32_t tvect[20];		// Test Vectors
  uint32_t avect[20];		// Answer Vectors
  tvect[0]  = DAQChannelMap::encodeLChan(det_ID, 1 + END_OFFSET_IDX, 5);
  avect[0]  = X_VIEW;
  tvect[1]  = DAQChannelMap::encodeLChan(det_ID, 2 + END_OFFSET_IDX, 45);
  avect[1]  = Y_VIEW;
  tvect[2]  = DAQChannelMap::encodeLChan(det_ID, 3 + END_OFFSET_IDX, 5);
  avect[2]  = X_VIEW;
  tvect[3]  = DAQChannelMap::encodeLChan(det_ID, 4 + END_OFFSET_IDX, 45);
  avect[3]  = Y_VIEW;
  tvect[4]  = DAQChannelMap::encodeLChan(det_ID, 5 + END_OFFSET_IDX, 37);
  avect[4]  = X_VIEW;
  tvect[5]  = DAQChannelMap::encodeLChan(det_ID, 6 + END_OFFSET_IDX, 37);
  avect[5]  = Y_VIEW;
  tvect[6]  = DAQChannelMap::encodeLChan(det_ID, 7 + END_OFFSET_IDX, 37);
  avect[6]  = X_VIEW;
  tvect[7]  = DAQChannelMap::encodeLChan(det_ID, 62 + END_OFFSET_IDX, 3);
  avect[7]  = Y_VIEW;
  tvect[8]  = DAQChannelMap::encodeLChan(det_ID, 62 + END_OFFSET_IDX + 1, 3);
  avect[8]  = X_VIEW;
  tvect[9]  = DAQChannelMap::encodeLChan(det_ID, 62 + END_OFFSET_IDX + 2, 3);
  avect[9]  = Y_VIEW;
  tvect[10] = DAQChannelMap::encodeLChan(det_ID, 62 + END_OFFSET_IDX + 3, 3);
  avect[10] = X_VIEW;
  tvect[11] = DAQChannelMap::encodeLChan(det_ID, 62 + END_OFFSET_IDX + 4, 3);
  avect[11] = Y_VIEW;
  tvect[12] = DAQChannelMap::encodeLChan(det_ID, 62 + END_OFFSET_IDX + 30, 3);
  avect[12] = Y_VIEW;
  tvect[13] = DAQChannelMap::encodeLChan(det_ID, 62 + END_OFFSET_IDX + 31, 3);
  avect[13] = X_VIEW;
  tvect[14] = DAQChannelMap::encodeLChan(det_ID, 124 + END_OFFSET_IDX, 3);
  avect[14] = Y_VIEW;
  tvect[15] = DAQChannelMap::encodeLChan(det_ID, 124 + END_OFFSET_IDX + 1, 3);
  avect[15] = X_VIEW;
  tvect[16] = DAQChannelMap::encodeLChan(det_ID, 124 + END_OFFSET_IDX + 2, 3);
  avect[16] = Y_VIEW;
  tvect[17] = DAQChannelMap::encodeLChan(det_ID, 124 + END_OFFSET_IDX + 31, 3);
  avect[17] = X_VIEW;
  tvect[18] = DAQChannelMap::encodeLChan(det_ID, 124 + END_OFFSET_IDX + 32, 3);
  avect[18] = X_VIEW;
  tvect[19] = DAQChannelMap::encodeLChan(det_ID, 124 + END_OFFSET_IDX + 33, 3);
  avect[19] = Y_VIEW;

  for(int i=0; i< 20; i++){
    if( (det_ID != IPND_ID) || (i < 10)){
      printf("%d 0x%08x\t| %4d \t| %4d\n",i, tvect[i],avect[i],DAQChannelMap::getViewOffline(tvect[i]));
    CPPUNIT_ASSERT(DAQChannelMap::getViewOffline(tvect[i])     == avect[i]);
    }
  }

  }
#ifdef FULL_TESTING
  // Test channel encoding/decoding
  for(int det             = IPND_ID;              det   < FARDET_ID;                                ++det){
    for(int plane         = START_CHAN_IDX;       plane < NUM_PLANES_DIBLOCK*NUM_DIBLOCK_DET[det];  ++plane){
      for(int cell        = START_CHAN_IDX;       cell  < NUM_YCHAN_PLANE[det];                     ++cell    ){
	    theLChan = DAQChannelMap::encodeLChan(det,plane,cell);
	    CPPUNIT_ASSERT(DAQChannelMap::getViewOffline(theLChan)     == X_VIEW);
	    CPPUNIT_ASSERT(DAQChannelMap::getPlaneOffline(theLChan)    == plane);
	    CPPUNIT_ASSERT(DAQChannelMap::getCellOffline(theLChan)     == cell);
#ifdef VERBOSE_DEBUG
	    printf("%2d %4d %3d\t0x%08x\t %d %d %d\n", det, plane, cell, theLChan,
		   DAQChannelMap::getDetectorOffline(theLChan), 
		   DAQChannelMap::getPlaneOffline(theLChan),
		   DAQChannelMap::getCellOffline(theLChan));
#endif /* VERBOSE_DEBUG */

      } // endfor dcm
    } // endfor diblock
  } // endfor det
#endif
  printf("Offline View decoding complete\n");

}


 void DAQChannelMapTests::testPlaneLookupsOffline(){
  // Retrieve the channel map class (singleton)
  DAQChannelMap* chanMap;
  chanMap = DAQChannelMap::getInstance();

  lchan theLChan;
  int thePlane;
  
  printf("\nTesting Offline Plane (diblock level) Decoding\n");
  // Test plane number decoding

  vector_count = 0;
  // Test channel encoding/decoding


  for(int det        = IPND_ID;        det     <= FARDET_ID;            ++det){
    for(int diblock  = START_CHAN_IDX; diblock < NUM_DIBLOCK_DET[det]; ++diblock){
      for(int plane  = START_CHAN_IDX; plane   < NUM_PLANES_DIBLOCK;   ++plane){
	if(START_CHAN_IDX){
	  thePlane = (diblock-1)*NUM_PLANES_DIBLOCK + plane;
	}else{
	  thePlane = diblock*NUM_PLANES_DIBLOCK + plane;
	}	
	for(int cell = START_CHAN_IDX; cell    < DAQChannelMap::getMaxChanPlane((DetectorIdent)det,thePlane); ++cell    ){
	  theLChan = DAQChannelMap::encodeLChan(det,thePlane,cell);	  
	  vector_count++;
	  CPPUNIT_ASSERT(DAQChannelMap::getPlaneDiBlockLocalOffline(theLChan) == plane);
	} // endfor cell
      }	// endfor plane
    } // endfor diblock    
  } // endfor det

  printf("Offline Plane decoding (block level) complete\n");
  printf("Run Condition: %d test vectors run.\n", vector_count);

  vector_count = 0;

  // Test channel encoding/decoding
  for(int det        = IPND_ID;        det     <= FARDET_ID;            ++det){
    printf("Detector = %s\n",DetectorLabels[det]);
    for(int block    = START_CHAN_IDX; block   < NUM_DIBLOCK_DET[det]*2; ++block){
      // printf("Block Number: %d\n",block);
      for(int plane  = START_CHAN_IDX; plane   < NUM_PLANES_BLOCK;   ++plane){
	// printf("Plane Number: %d \n",plane);
	if(START_CHAN_IDX){
	  thePlane = (block-1)*NUM_PLANES_BLOCK + plane;
	}else{
	  thePlane = block*NUM_PLANES_BLOCK + plane;
	}
	for(int cell = START_CHAN_IDX; cell    < DAQChannelMap::getMaxChanPlane((DetectorIdent)det,thePlane); ++cell    ){
	  theLChan = DAQChannelMap::encodeLChan(det,thePlane,cell);	  
	  vector_count++;
	  CPPUNIT_ASSERT(DAQChannelMap::getPlaneBlockLocalOffline(theLChan) == plane);
	} // endfor cell
      }	// endfor plane
    } // endfor diblock    
  } // endfor det

  printf("Offline Plane decoding complete\n");
  printf("Run Condition: %d test vectors run.\n", vector_count);

}

void DAQChannelMapTests::testViewLookupsOffline(){
  // Retrieve the channel map class (singleton)
  DAQChannelMap* chanMap;
  chanMap = DAQChannelMap::getInstance();

  lchan theLChan;
  lchan testchan;
  int thePlane;
  
  printf("\nTesting View (diblock level) Decoding\n");
  // Test view decoder

  vector_count = 0;

  int xcount=0;
  int ycount=0;

  int theView;

  for(int det        = IPND_ID;        det     <= FARDET_ID;            ++det){
    printf("Detector = %s\n",DetectorLabels[det]);
    for(int diblock    = START_CHAN_IDX; diblock   < NUM_DIBLOCK_DET[det]; ++diblock){
      // printf("Block Number: %d\n",block);

      xcount=0;
      ycount=0;
      for(int plane  = START_CHAN_IDX; plane   < NUM_PLANES_BLOCK;   ++plane){
	// printf("Plane Number: %d \n",plane);

	if(START_CHAN_IDX){
	  thePlane = (diblock-1)*NUM_PLANES_DIBLOCK + plane;
	}else{
	  thePlane = diblock*NUM_PLANES_DIBLOCK + plane;
	}
	
	theView = DAQChannelMap::getViewOffline(DAQChannelMap::encodeLChan(det,thePlane,START_CHAN_IDX));

	switch(theView){
	case X_VIEW:
	  xcount++;
	  break;
	case Y_VIEW:
	  ycount++;
	  break;
	default:
	  printf("ERROR\n");
	  break;
	}
	
	for(int cell = START_CHAN_IDX; cell    < DAQChannelMap::getMaxChanPlane((DetectorIdent)det,thePlane); ++cell    ){
	  theLChan = DAQChannelMap::encodeLChan(det,thePlane,cell);	  
	  vector_count++;
	 
	  switch(theView){
	  case X_VIEW:
	    CPPUNIT_ASSERT(DAQChannelMap::getPlaneViewDiBlockLocal(theLChan) == xcount);
	    break;
	  case Y_VIEW:
	    CPPUNIT_ASSERT(DAQChannelMap::getPlaneViewDiBlockLocal(theLChan) == ycount);
	    break;
	  default:
	    printf("ERROR\n");
	    break;
	  }

	} // endfor cell
      }	// endfor plane
    } // endfor diblock    
  } // endfor det
  

  printf("Offline View decoding complete\n");
  printf("Run Condition: %d test vectors run.\n", vector_count);


  printf("\nTesting Plane View (global level) Decoding\n");

  vector_count = 0;

  for(int det        = IPND_ID;        det     <= FARDET_ID;            ++det){
    printf("Detector = %s\n",DetectorLabels[det]);

    xcount = 0;
    ycount = 0;

    for(int diblock = START_CHAN_IDX; diblock <= NUM_DIBLOCK_DET[det]; ++diblock){
      printf("%s\tDiblock %d \t0x%02X\tXp= %d\tYp = %d\n",DetectorLabels[det],diblock,DetectorDiBlockConfig[det][diblock],
	     DAQChannelMap::getPlanesDiBlockView((DiBlock_TYPE)DetectorDiBlockConfig[det][diblock],(DetView_TYPE)X_VIEW),
	     DAQChannelMap::getPlanesDiBlockView((DiBlock_TYPE)DetectorDiBlockConfig[det][diblock],(DetView_TYPE)Y_VIEW));
    }

    for(int plane    = START_CHAN_IDX; plane   < NUM_DIBLOCK_DET[det]*NUM_PLANES_DIBLOCK; ++plane){
      // printf("Plane Number: %d\n",plane);

      theLChan = DAQChannelMap::encodeLChan(det,plane,START_CHAN_IDX);
	theView = DAQChannelMap::getViewOffline(theLChan);

	switch(theView){
	case X_VIEW:
	  xcount++;
	  break;
	case Y_VIEW:
	  ycount++;
	  break;
	default:
	  printf("ERROR\n");
	  break;
	}

	vector_count++;
	


	DAQChannelMap::getPlaneViewLocal(theLChan);

	switch(theView){
	  case X_VIEW:
	    CPPUNIT_ASSERT(DAQChannelMap::getPlaneViewLocal(theLChan) == xcount);
	    break;
	  case Y_VIEW:
	    CPPUNIT_ASSERT(DAQChannelMap::getPlaneViewLocal(theLChan) == ycount);
	    break;
	  default:
	    printf("ERROR\n");
	    break;
	}

    } // endfor plane
  } // endfor det

  printf("Offline Plane View decoding complete\n");
  printf("Run Condition: %d test vectors run.\n", vector_count);


}
