#include <cppunit/extensions/HelperMacros.h>
#include <DAQChannelMap/DAQChannelMap.h>
#include "NovaDAQConventions/DAQConventions.h"
#include "NovaDAQUtilities/LocationUtils.h"

#include <cppunit/TestResultCollector.h>
#include <cppunit/TestRunner.h>
#include <cppunit/TestResult.h>
#include <cppunit/TextTestProgressListener.h>
#include <cppunit/TextOutputter.h>
#include <cppunit/extensions/TestFactoryRegistry.h>
#include <cppunit/TestSuite.h>
#include <cppunit/XmlOutputter.h>


#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <fcntl.h>
#include <sys/stat.h>

#include <algorithm>
#include <vector>

#include <string>

/// File containing unittest data
#define CHANMAP_FILE "daqchannelmap_data.txt"

#if DAQCHANNELMAP_VERSION == 2
#define GETDETECTOR getDetector
#define GETDIBLOCK  getDiBlock
#define GETDCM      getDCM
#define GETFEB      getDCMLink
#define GETPIXEL    getPixel
#else
#define GETDETECTOR getDetector
#define GETDIBLOCK  getDiBlock
#define GETDCM      getDCMLocal
#define GETFEB      getDCMPort
#define GETPIXEL    getPixel
#endif

using namespace daqchannelmap;


typedef std::vector<int>     ItemVector;
typedef ItemVector::iterator ItemIterator;


    class DAQChannelMapUnitTest: public CPPUNIT_NS::TestFixture {
public:
    CPPUNIT_TEST_SUITE(DAQChannelMapUnitTest);
    //CPPUNIT_TEST(testSimple);
    CPPUNIT_TEST(testFromFile);
    CPPUNIT_TEST_SUITE_END();

  //DAQChannelMapUnitTest(){};
  //~DAQChannelMapUnitTest(){};

public:
  
  /// Some setup/teardown methods that Unittest require
  void setUp()   {}
  void tearDown(){}

  void createUnittestTable();
  void testSimple();
  void testFromFile();

  bool findUnittestFile();
  bool findUnittestDirectory();

  int  run(int argc, char* argv[]);

private:

  /// Add tested detector to the list
  bool addDetector(int);
  
  /// Check whether we tested all available detectors
  bool checkTestedDetectors() const;
  
public:

    /// File name for the ChannelMap data
    std::string            _fileName;
    std::string            _dirName;

    novadaq::LocationUtils _locationUtils;

    uint32_t               _numberOfDetectorsTested;

    ItemVector                   _detectorsTested;

};

//------------------------------------------------------------------------------
bool DAQChannelMapUnitTest::addDetector(int detector){

    ItemIterator itemItr = std::find(_detectorsTested.begin(), _detectorsTested.end(), detector);

    if (itemItr == _detectorsTested.end()) {
        _detectorsTested.push_back(detector);

    }

    return true;
}

//------------------------------------------------------------------------------
/// Check whether we tested all available detectors
bool DAQChannelMapUnitTest::checkTestedDetectors() const{

    std::cout<<"\n";

    bool is_everything_ok = true;
    
    for(int idet = (int)novadaq::cnv::kUNKNOWN_DET+1; idet < (int)novadaq::cnv::kNDetector; ++idet){

        if(idet == novadaq::cnv::kNEARDET) continue;

        //ItemIterator itemItr = std::find(_detectorsTested.begin(), _detectorsTested.end(), idet);

        bool foundit = false;
        for(uint i=0; i<_detectorsTested.size();++i){
            if(_detectorsTested[i] == idet){
                foundit = true;
                break;
            }
        }

        if (!foundit) {
            std::cout<<"Error! We did not test detector "<<novadaq::cnv::DetInfo::GetName(idet)<<" (id="<<idet<<")!!!!!\n";
            is_everything_ok = false;
        }
        else{
            std::cout<<"Successfully tested detector "<<novadaq::cnv::DetInfo::GetName(idet)<<" (id="<<idet<<")\n";
        }

    }

    if(!is_everything_ok){
        std::cout<<"Need to make a new unittest data by executing DAQChannelMapUnitTest::createUnittestTable\n";
    }

    return is_everything_ok;
}



//------------------------------------------------------------------------------
bool DAQChannelMapUnitTest::findUnittestFile(){

    if(_dirName.size() == 0)
        findUnittestDirectory();

    if(_dirName.size() == 0)
        return false;

    std::string file_try = _dirName;
    file_try.append("/");
    file_try.append(CHANMAP_FILE);

    if(!_locationUtils.findFile(file_try, _fileName)){
       _fileName = file_try;
       return false;
    }
    
    return true;
}

//------------------------------------------------------------------------------
bool DAQChannelMapUnitTest::findUnittestDirectory(){

    std::string dir_try = "DAQChannelMap/cxx/unittest/";
    return _locationUtils.findFile(dir_try, _dirName);
}

//------------------------------------------------------------------------------
void DAQChannelMapUnitTest::testSimple(){

    DAQChannelMap* chanMap = DAQChannelMap::getInstance();

    int             idet   = 3;
    offline_plane_t iplane = 155;
    offline_cell_t  icell  = 64;

    lchan lChan   = chanMap->encodeLChan(idet, iplane, icell);
    dchan daqChan = chanMap->encodeDChan(lChan);
    lchan lchan_check   = chanMap->encodeLChan(daqChan);

    uint32_t det     = chanMap->getDetector(daqChan);
    uint32_t diblock = chanMap->getDiBlock(daqChan);
    uint32_t dcm     = chanMap->getDCM(daqChan);
    uint32_t port    = chanMap->getDCMLink(daqChan);
    uint32_t pixel   = chanMap->getPixel(daqChan);

        int      det2     = chanMap->getDetectorOffline(lChan);
        uint32_t diblock2 = chanMap->getDiBlockOffline(lChan);
        uint32_t dcm2     = chanMap->getDCMOffline(lChan);
        uint32_t port2    = chanMap->getDCMLinkOffline(lChan);
        uint32_t pixel2   = chanMap->getPixelOffline(lChan);


        if(lchan_check != lChan){
            std::cout<<"Error0 in dchan_check2 != daqChan \n";
            std::cout<<"Det="<<det<<" DB="<<diblock<<" dcm="<<dcm<<" port="<<port<<" pixel="<<pixel
                    <<" Det2="<<det2<<" DB2="<<diblock2<<" dcm2="<<dcm2<<" port2="<<port2<<" pixel2="<<pixel2
                     <<" daqChan="<<daqChan
                     //<<" lchan="<<lChan<<" Dchan2="<<dchan_check2
                     <<" plane="<<iplane<<" cell="<<icell
                     //<<" plane2="<<plane2<<" cell2="<<cell2
                     <<"\n";
        }

        CPPUNIT_ASSERT(lchan_check == lChan);

}


//------------------------------------------------------------------------------
void DAQChannelMapUnitTest::createUnittestTable() {

  std::cout<<"Creating a table\n";
  
  // Retrieve the channel map class (singleton)
  DAQChannelMap* chanMap = DAQChannelMap::getInstance();

  
  findUnittestFile();
  
  FILE * fp = fopen(_fileName.c_str(), "wb");
  
  // Test channel encoding/decoding
  for(int idet = chanMap->getDetectorIDMinimum(); idet < chanMap->getDetectorIDMaximum(); ++idet){
      offline_plane_t Nplanes = chanMap->getTotalNumberOfPlanes(idet);

      /// Loop over planes in diblock
      for(offline_plane_t iplane = 0; iplane < Nplanes; ++iplane){

                const uint32_t num_cells =  chanMap->getNumberOfChannelsInPlane(idet, iplane);

                for(uint32_t icell = 0; icell < num_cells; ++icell){

                        lchan lChan   = chanMap->encodeLChan(idet, iplane, icell);
                        dchan daqChan = chanMap->encodeDChan(lChan);

                        uint32_t det     = chanMap->getDetector(daqChan);
                        uint32_t diblock = chanMap->getDiBlock(daqChan);
                        uint32_t dcm     = chanMap->getDCM(daqChan);
                        uint32_t port    = chanMap->getDCMLink(daqChan);
                        uint32_t pixel   = chanMap->getPixel(daqChan);

                        uint32_t block[7];
                        block[0] = det;
                        block[1] = diblock;
                        block[2] = dcm;
                        block[3] = port;
                        block[4] = pixel;
                        block[5] = daqChan;
                        block[6] = lChan;

                        fwrite (&(block[0]), 4 , 7 , fp);
            
                } // endfor icell
        } // endfor plane
  } // endfor det
  fclose(fp);

  

}

//------------------------------------------------------------------------------
void DAQChannelMapUnitTest::testFromFile(){


    /// clear the list of tested detectors
    _detectorsTested.clear();

    DAQChannelMap* chanMap = DAQChannelMap::getInstance();

    if(!findUnittestFile())
      return;

    struct stat file_status;
    stat(_fileName.c_str(), &file_status);

    int file_handler = open(_fileName.c_str(), O_RDONLY);

    uint32_t size_of_file_bytes = file_status.st_size;


    uint32_t block[7];

    uint32_t number_of_bytes_read = 0;

    std::cout<<"Starting the test\n";
    while(number_of_bytes_read < size_of_file_bytes){
        number_of_bytes_read += read(file_handler, &block[0], 7*4);

        /// Stuff read from file
        const int      det     = block[0];
        const uint32_t diblock = block[1];
        const uint32_t dcm     = block[2];
        const uint32_t port    = block[3];
        const uint32_t pixel   = block[4];
        const dchan    daqChan = block[5];
        const lchan    lChan   = block[6];

        /// Add detector to the list of tested detectors
        addDetector(det);

        /// Calculate DAQChannel
        const dchan dchan_check = chanMap->encodeDChan(det, diblock, dcm, port, pixel);

        if(det == novadaq::cnv::kNEARDET) continue;


        if(dchan_check != daqChan){
            std::cout<<"Error in First dchan_check != daqChan \n";
            std::cout<<"Ch0="<<daqChan<<" ch1="<<dchan_check<<"\n";
        }
        CPPUNIT_ASSERT(dchan_check == daqChan);

        lchan lchan_check = chanMap->encodeLChan(daqChan);

        dchan dchan_check2 = chanMap->encodeDChan(lChan);


        int      plane     = chanMap->getPlane(daqChan);
        int      cell      = chanMap->getCell(daqChan);
        int      plane2    = chanMap->getPlaneOffline(lChan);
        int      cell2     = chanMap->getCellOffline(lChan);


        int      det2     = chanMap->getDetectorOffline(lChan);
        uint32_t diblock2 = chanMap->getDiBlockOffline(lChan);
        uint32_t dcm2     = chanMap->getDCMOffline(lChan);
        uint32_t port2    = chanMap->getDCMLinkOffline(lChan);
        uint32_t pixel2   = chanMap->getPixelOffline(lChan);


        if(dchan_check2 != daqChan || lchan_check != lChan){
            std::cout<<"Error in dchan_check2 != daqChan \n";
            std::cout<<"Det="<<det<<" DB="<<diblock<<" dcm="<<dcm<<" port="<<port<<" pixel="<<pixel
                    <<" Det2="<<det2<<" DB2="<<diblock2<<" dcm2="<<dcm2<<" port2="<<port2<<" pixel2="<<pixel2
                     <<" daqChan="<<daqChan<<" Dchan2="<<dchan_check2
                     <<" lchan="<<lChan<<" lchan_check="<<lchan_check
                     <<" plane="<<plane<<" cell="<<cell
                     <<" plane2="<<plane2<<" cell2="<<cell2
                     <<" Numchannels="<<chanMap->getNumberOfChannelsInPlane(det, plane2)
                     <<"\n";
        }

        CPPUNIT_ASSERT(lchan_check == lChan);
        CPPUNIT_ASSERT(dchan_check2 == daqChan);

    }


    CPPUNIT_ASSERT(checkTestedDetectors());

}

//------------------------------------------------------------------------------
int  DAQChannelMapUnitTest::run(int argc, char* argv[]){

      // Create the test controller
  CPPUNIT_NS::TestResult controller;
  // Add a listener that collects test results
  CPPUNIT_NS::TestResultCollector result;
  controller.addListener(&result);
  // Add a listener that prints a dot for each test as they run
  CPPUNIT_NS::TextTestProgressListener progress;
  controller.addListener(&progress);

  // Add the top suite to the test runner
  CPPUNIT_NS::TestRunner runner;
  runner.addTest(CPPUNIT_NS::TestFactoryRegistry::getRegistry().makeTest());

  // Command line argument is test path. Defaults to "", which resolves to the top level suite.
  std::string testPath = (argc > 1) ? std::string(argv[1]) : std::string("");

  try {
      std::cout << "Running "
                << (testPath.size() > 0 ? testPath.c_str() : "all tests")
                << std::endl;

      runner.run (controller, testPath);

      /// Make sure that all log messages are printed
      usleep(10);

      std::cout<<"Unittest Results:\n";
      // Output the results as text to std::cout
      CPPUNIT_NS::TextOutputter textOutputter(&result, std::cout);
      textOutputter.write();

      // Output the results in xml format to a file
      // This is temporarily commented out until we can add user control
      // of whether this file gets generated.
      //std::ofstream xmlOutputFile("cppUnitTestResult.xml");
      //CPPUNIT_NS::XmlOutputter xmlOutputter(&result,xmlOutputFile);
      //xmlOutputter.write();
      //xmlOutputFile.close();
  }
  catch (std::invalid_argument &e) { // Test path not resolved

    std::cerr     << "\n" << "ERROR: " << e.what() << std::endl;
    return 1;
  }

  return result.wasSuccessful() ? 0 : 1;

}

//------------------------------------------------------------------------------
int main(int argc, char* argv[]){


    DAQChannelMapUnitTest bm_test;
    //bm_test.createUnittestTable();

    if(bm_test.run(argc, argv) == 1){

        //std::cout<<"Creating a table\n";
        //bm_test.createUnittestTable();
        //return bm_test.run(argc, argv);
    }


    return 0;
}


///Register
CPPUNIT_TEST_SUITE_NAMED_REGISTRATION(DAQChannelMapUnitTest, "DAQChannelMapUnitTest");
CPPUNIT_REGISTRY_ADD_TO_DEFAULT("DAQChannelMapUnitTest");
