#ifndef RAWTIMINGMARKER_H
#define RAWTIMINGMARKER_H
#include "DAQDataFormats/RawDAQData.h"
#include "DAQDataFormats/Macros.h"

namespace daqdataformats {

  static const uint32_t RAWTIMINGMARKER_SIZE = 2;


  //! Data structure to hold the parts of the timing word
  struct TimingMarker {
    uint32_t MarkerLowWord;
    uint32_t MarkerHighWord;
  };
     namespace timmarker {
        enum TimingMarkerMASKS{
            MASK_Lo = 0xFFFFFFFF,
            MASK_Hi = 0x00FFFFFF
        };

        enum TimingMarkerWord{
            WORD_Lo = 0,
            WORD_Hi = 1
        };

        DEFINE_SHIFT_FROM_MASK(Lo)
        DEFINE_SHIFT_FROM_MASK(Hi)

    };// end of namespace timmarker


  //! Class to hold the timing markers used in the construction of Microslices in the DCMs
  /*!
    The RawTimingMarker class impliments the data formats used to create proper timing marker
    which is used in the construction of a data microslice in the Data concentrator modules 
    data collection process.  These markers are designed to separate blocks of data coming 
    out of the front end boards at roughly 5us intervals.  All the data between time markers
    is assume to have occured after the first marker and before the second marker.

    Each marker is 56 bits (0-55) long divided into a high and low word.

    This class provides a set of generator methods which will produce a valid time marker
    based off of the current system clock of the host system on which it is running.  This marker
    is generated using the CPU specific real time clock system and may not be available 
    on all platforms.
   */
  class RawTimingMarker: public RawDAQData {
        public:
            // Constructors
            RawTimingMarker();		               //! The default constructor
            RawTimingMarker(const RawTimingMarker& copy_in); //! Deep copy constructor
            RawTimingMarker(const void* timingmarker_buffer);//! Shallow copy constructor
            /// Destructor
            ~RawTimingMarker();

            uint32_t  getLowWord() const;  ///! Get method for lower half of timing word
            uint32_t  getHighWord() const; ///! Get method for upper half of timing word
            uint64_t  getTime() const;     ///! Get the time-stamp as a single 64-bit value

            bool setLowWord (const uint32_t); ///! Set method for lower half of timing word
            bool setHighWord(const uint32_t); ///! Set method for upper half of timing word
            bool setTime    (const uint64_t); ///! Set the time-stamp as a single 64-bit value

            /// Method to generate a valid timing marker from the system clocks
            void generate();

            uint32_t sizeofdata      () const;

            using RawDAQData::readData;	// unhide the reset of the readData methods
            void*    readData(const void* buffer){return readDataGeneral(buffer);}

  protected:
            void     printStandard     (std::ostream& os=std::cout) const;
            void*    readDataFromBuffer(const void* buffer) {return readDataGeneral(buffer);}
            bool     setFunctionPointers() { return true;}
            
        private:
            /// Read the TimingMarker version is meaningless, because we don't have versioning for it
            version_t figureOutTheVersion(const void*) const {return 0;}

  };
  
  
} // END NAMESPCE daqdataformats
#endif /* RAWTIMINGMARKER_H */
