#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <error.h>

extern "C"{
#include <getopt.h>
}

#include "DAQDataFormats/DAQDataFormats.h"
#include "DAQDataFormats/RawRun.h"
#include "DAQDataFormats/RawRunHeader.h"
#include "DAQDataFormats/RawEvent.h"
#include "DAQDataFormats/RawEventHeader.h"
#include "DAQDataFormats/RawTrigger.h"
#include "DAQDataFormats/RawMicroBlock.h"
#include "DAQDataFormats/RawNanoSlice.h"
#include "DAQQualityCheck/QualityCheck.h"

using namespace daqdataformats;
using namespace daqqc;

#define PROGRAM_VERSION 1.0

#ifndef TRUE
#define TRUE   1
#endif /* TRUE */

#ifndef FALSE
#define FALSE  0
#endif /* FALSE */

#ifndef FAIL
#define FAIL  -1
#endif /* FAIL */

using namespace std;


#define SHM_MODE (SHM_R | SHM_W | IPC_CREAT)
// #define READ_SIZE 8094
#define READ_SIZE 65535

/////////////////////////////////////////////
// VERSION INFORMATION FUNCTION
int show_version(char** argv)
{
   printf("NOvA-DAQ Utility: Run quality and integrity checks on raw data files\nVersion %f\n",PROGRAM_VERSION);
  return TRUE;
}

/////////////////////////////////////////////
// Program help and usage function
int show_help(char** argv)
{
  show_version(argv);
  printf("Usage: %s [options] \n",argv[0]);
  printf("-h\t\t--help\t\t\tShow program help\n");
  printf("-v\t\t--version\t\tShow version information\n");
  printf("-q\t\t--quiet\t\t\tQuiet (non-interactive)mode\n");
  printf("-f file\t\t--file=filename\t\tRead input from specified file\n");
  printf("-t type\t\t--type=type\t\tSpecify reading run (-t 1) or event (-t 2) files\n");

  exit(TRUE);
  return TRUE;
}

void prompt(int flag){
  if(flag){
    printf("---Enter to continue---\n");
    getchar();
  }
}


// #define num_lines 2
int main(int argc, char** argv){

  // Comandline options variables
  int option_number;                  /* Command line option number for processing      */
  int option_index     =0;            /* Index of command option processed              */
  int help_flag        = FALSE;       /* Flag for users requesting help                 */
  int version_flag     = FALSE;       /* Flag for users version info                    */
  int quiet_flag       = FALSE;       /* Flag for users requesting non interactive mode */
  int file_flag        = FALSE;       /* Flag indicating we want to read from a file    */
  int type             = 2;           /* Index for what the input type is: 1=run, 2=events */

  // File related variables
  char* filename;                     /* Pointer to the buffer for the filename         */
  int infile =0;                      /* File descriptor                                */

  int   bytes_read;
  int   read_flag;
  int   total_bytes;

  // Data Format Specific
  RawEvent theEvent;
  RawEventHeader theHeader;
#ifndef RAWRUN_IS_NOT_IMPLEMENTED    
  RawRun theRun;
#endif
  RawRunHeader theRunHeader;

  QualityCheck qc;

 /* Set up the structure for the long options flags */ 
  struct option long_options[] =
    {
      /* These options set a flag.   */
      {"help",        no_argument,           &help_flag,       TRUE},
      {"version",     no_argument,           &version_flag,    TRUE},
      {"quiet",       no_argument,           &quiet_flag,      TRUE},
      {"interactive", no_argument,           &quiet_flag,      FALSE},
      /* These options don't set a flag. We distinguish them by their values.  */
      {"file",        required_argument,     0,                'f'},
      {"type",        required_argument,     0,                't'},
      {0, 0, 0, 0}
    };

  /*************************************/
  /* Parse the commandline for options */
  while(TRUE){
    option_number = getopt_long(argc,argv,"h?vqf:m:s:t:",long_options,&option_index);

    if(option_number == -1){    /* We are out of commandline options */
      break;                    /* Jump out of the options loops     */
    }
    
    switch(option_number){
    case 0:
      /* If this option set a flag, do nothing else now.   */
      if (long_options[option_index].flag != 0){break;}
      break;
    case 'h':
    case '?':
      help_flag = TRUE;                                  /* Set the help flag */
      break;
    case 'v':
      version_flag = TRUE;                               /* Set the version flag */
    case 'q':
      quiet_flag = TRUE;                                 /* Set the quiet flag */
      break;
    case 'f':
      file_flag = TRUE;                                /* We will need to send output to this file  */
      filename = (char*)malloc(strlen(optarg)+1);      /* Allocate a buffer for the file name       */
      strcpy(filename,optarg);                         /* Copy the arg into the output filename     */
      printf("Got filename: %s\n",filename);
      break;
    case 't':
      type = atoi(optarg);
      break;
    default:
      printf("Invalid or Unimplimented Options or arguement: --%s %s\nSee Documentation\n\n\n",long_options[option_index].name,optarg);
      exit(FAIL);
    } /* endswitch option_number */
  } /* endwhile option_number !=-1 */
  /* We are done with commandline options */
  /****************************************/

  if(version_flag){show_version(argv); exit(FAIL);}
  if(help_flag)   {show_help(argv);    exit(FAIL);}

  if(file_flag){
    infile = open(filename,O_RDONLY);
    if(!infile){
      printf("Failed to open file for reading: %s\n",filename);
      exit(FAIL);
    }else{
      printf("Opened file: %s for reading.\n",filename);
      struct stat stbuf;
      stat(filename, &stbuf);
      if (0 == stbuf.st_size) {
	printf("File is empty!\n");
	exit(FAIL);
      }
    }
  } /* endif file_flag */

  // the last argument should be the filename
  filename = argv[optind++];
  infile = open(filename, O_RDONLY);
  if (!infile) {
    printf("Failed to open file for reading: %s\n",filename);
    exit(FAIL);
  }else{
    printf("Opened file: %s for reading.\n",filename);
    struct stat stbuf;
    stat(filename, &stbuf);
    if (0 == stbuf.st_size) {
      printf("File is empty!\n");
      exit(FAIL);
    }
  }
  
  /* For efficiency we read our file in a default 64k chunks */
  total_bytes = 0;
  read_flag = TRUE;

  // Allocate two buffer, one for reading in the header
  // The other for the actual event data

  char* buff;
  char* header_buff;

  // The header is a fixed size, we allocate it
  if (type == 1)
    header_buff = (char*) malloc(theRunHeader.sizeofdata()*sizeof(uint32_t));
  else if (type == 2)
    header_buff = (char*) malloc(theHeader.sizeofdata()*sizeof(uint32_t));

  // The event is a variable size, we allocate it up to a maximum size
  // and reuse the buffer on successive reads for efficiency
  buff = (char*)  malloc(READ_SIZE);

  uint32_t event_size;

  int i=0;
  while(read_flag){
    i++;

    // We must assume that the current seek point of the 
    // file descriptor points to the start of a RawEvent
    if (type == 1)
      bytes_read = read(infile, header_buff, theRunHeader.sizeofdata()*sizeof(uint32_t));
    else if (type == 2)
      bytes_read = read(infile, header_buff, theHeader.sizeofdata()*sizeof(uint32_t));

    // Check for end of file or read error
    if(bytes_read <= 0){ 
      read_flag = FALSE;
      continue;
    }
    
    if (type == 1) {
      if(bytes_read != int(theRunHeader.sizeofdata()*sizeof(uint32_t))){
	perror("Incomplete header read");
	read_flag = FALSE;
	continue;
      }
    } else if (type == 2) {
      if(bytes_read != int(theHeader.sizeofdata()*sizeof(uint32_t))){
	perror("Incomplete header read");
	read_flag = FALSE;
	continue;
      }
    }
  
    // Point the header to the input buffer we just filled
    if (type == 1) 
      theRunHeader.setBufferSource(header_buff);
    else if (type == 2)
      theHeader.setBufferSource(header_buff);

    // We copy the header into the Run/Event buffer
    if (type == 1)
      memcpy(buff, header_buff, theRunHeader.sizeofdata()*sizeof(uint32_t));
    else if (type == 2)
      memcpy(buff, header_buff, theHeader.sizeofdata()*sizeof(uint32_t));

    uint32_t read_size;
    // We extract the full run/event size
    if (type == 1) {
      event_size = theRunHeader.getRunSize();
      read_size = (event_size - theRunHeader.sizeofdata())*(sizeof(uint32_t));
    } else {
      event_size = theHeader.getEventSize();
      read_size = (event_size - theHeader.sizeofdata())*(sizeof(uint32_t));
    }

    // We now read the rest of the event in from the file
    void* buff_ins;
    if (type == 1) 
      buff_ins = (uint32_t*)buff + theRunHeader.sizeofdata(); 
    else if (type == 2)
      buff_ins = (uint32_t*)buff + theHeader.sizeofdata(); 

    bytes_read = read(infile,buff_ins, read_size);
    if(bytes_read <=0){
      read_flag = FALSE;
      continue;
    }

    if((type == 1) && (event_size*sizeof(uint32_t)) != (bytes_read + theRunHeader.sizeofdata()*sizeof(uint32_t))){
      // We did not get the full read
      perror("Incomplete run read");
      read_flag = FALSE;
      continue;
    }
    if((type == 2) && (event_size*sizeof(uint32_t)) != (bytes_read + theHeader.sizeofdata()*sizeof(uint32_t))){
      // We did not get the full read
      perror("Incomplete event read");
      read_flag = FALSE;
      continue;
    }
    
    // We point the buffer for the event and header to point at the 
    // read buffer.
    if (type == 1) {
#ifndef RAWRUN_IS_NOT_IMPLEMENTED            
      theRun.setBufferSource(buff);
      printf("Run buffered\n");
#endif
    } else if (type == 2) {
      theEvent.setBufferSource(buff);
      printf("Event %d buffered\n",i);
    }
    // A this point we have populated the memory segment
    // We sleep before advancing to the next event

    bool result = false;
    if (type == 1){
#ifndef RAWRUN_IS_NOT_IMPLEMENTED            
      result = qc.checkRun(theRun);
#endif
    }
    else
      result = qc.checkEvent(theEvent);

    printf("\nChecks %s!\n", result ? "Passed" : "Failed");
    prompt(quiet_flag);   // Pause prompt
    if (type == 1)
      read_flag = FALSE;
  }

  // Deallocate the local buffer
  free(buff);

  return 1;
}
