#include <DatabaseUtils/DAQConfig/DCMRunConfig.h>
#include <DatabaseUtils/DAQConfig/ConfigDataTree.h>
#include <DatabaseUtils/DAQConfig/DBColumnParams.h>
#include <DatabaseUtils/DAQConfig/NamedConfigUtils.h>

namespace dbutils {
namespace daqconfig {

/**
 * Creates a DCM run configuration object from the specified
 * list of subsystem configuration IDs.
 *
 * @param subsystemConfigIdList The list of subsystem
 *        configuration IDs.
 * @param dcmName The name of the DCM of interest.
 */
DCMRunConfig::
DCMRunConfig(const std::vector<int64_t>& subsystemConfigIdList,
             const std::string& dcmName)
{
    _primarySubsystemConfigId = -1;
    if (subsystemConfigIdList.size() > 0) {
        _primarySubsystemConfigId = subsystemConfigIdList[0];
        _hasValidData = _fetchAllData(subsystemConfigIdList, dcmName);
    }
    else {
        _hasValidData = false;
    }
}

/**
 * Creates a DCM run configuration object from the specified
 * subsystem configuration specification.
 *
 * @param nameSpec The named configuration of interest.
 * @param dcmName The name of the DCM of interest.
 */
DCMRunConfig::
DCMRunConfig(const SubsystemNameSpec& nameSpec,
             const std::string& dcmName)
{
    _primarySubsystemConfigId = -1;
    std::vector<int64_t> subsystemConfigIdList;
    if (NamedConfigUtils::
        getSubsystemConfigIdAncestry(nameSpec, subsystemConfigIdList)) {
        _primarySubsystemConfigId = subsystemConfigIdList[0];
        _hasValidData = _fetchAllData(subsystemConfigIdList, dcmName);
    }
    else {
        _hasValidData = false;
    }
}

/**
 * Creates a DCM run configuration object from the specified
 * global configuration specification.
 *
 * @param nameSpec The named configuration of interest.
 * @param dcmName The name of the DCM of interest.
 */
DCMRunConfig::
DCMRunConfig(const GlobalNameSpec& nameSpec,
             const std::string& dcmName)
{
    _hasValidData = false;
    _primarySubsystemConfigId = -1;
    int64_t subsystemConfigId;
    if (NamedConfigUtils::
        getSubsystemConfigIdFromGlobalName(nameSpec,
                                           SubsystemNameSpec::
                                           DCM_SUBSYSTEM_NAME,
                                           SubsystemNameSpec::
                                           RUN_CONFIG_TYPE,
                                           subsystemConfigId)) {
        std::vector<int64_t> subsystemConfigIdList;
        IDSpec idSpec(subsystemConfigId, IDSpec::SUBSYSTEM, IDSpec::DAQ);
        if (NamedConfigUtils::
            getConfigIdAncestry(idSpec, subsystemConfigIdList)) {
            _primarySubsystemConfigId = subsystemConfigIdList[0];
            _hasValidData = _fetchAllData(subsystemConfigIdList, dcmName);
        }
    }
}

/**
 * Creates a DCM run configuration object from the specified
 * global configuration ID.
 *
 * @param globalConfigId The ID of the global configuration of interest.
 * @param dcmName The name of the DCM of interest.
 */
DCMRunConfig::
DCMRunConfig(const int64_t& globalConfigId,
             const std::string& dcmName)
{
    IDSpec idSpec(globalConfigId, IDSpec::GLOBAL, IDSpec::DAQ);
    _primarySubsystemConfigId = -1;
    std::vector<int64_t> subsystemConfigIdList;
    if (NamedConfigUtils::
        getSubsystemConfigIdAncestryFromGlobalId(idSpec,
                                                 SubsystemNameSpec::
                                                 DCM_SUBSYSTEM_NAME,
                                                 SubsystemNameSpec::
                                                 RUN_CONFIG_TYPE,
                                                 subsystemConfigIdList)) {
        _primarySubsystemConfigId = subsystemConfigIdList[0];
        _hasValidData = _fetchAllData(subsystemConfigIdList, dcmName);
    }
    else {
        _hasValidData = false;
    }
}

bool DCMRunConfig::
_fetchAllData(const std::vector<int64_t>& subsystemConfigIdList,
              const std::string& dcmName)
{
    bool status = true;

    // system-wide data
    if (status) {
        status = _fetchDCMSystemParams(subsystemConfigIdList);
    }

    // dcm-specific data
    if (status) {
        status = _fetchDCMDataDevValues(subsystemConfigIdList, dcmName);
    }
    if (status) {
        status = _fetchDCMApplicationParams(subsystemConfigIdList, dcmName);
    }

    return status;
}

bool DCMRunConfig::
_fetchDCMSystemParams(const std::vector<int64_t>& subsystemConfigIdList)
{
    ConfigDataSelector cfgSelector;

    std::vector<std::string> dataColumnsOfInterest;
    dataColumnsOfInterest.push_back("milliSliceDT");

    ConfigDataTree
        dcmSystemParamDataSet("NovaDatabase/tables/DAQConfig/DCMSystemRunParameters.xml",
                              subsystemConfigIdList, cfgSelector,
                              dataColumnsOfInterest);

    if (! dcmSystemParamDataSet.getTrueValue("milliSliceDT",
                                             _millisliceTimeLength)) {
        return false;
    }
    return true;
}

bool DCMRunConfig::
_fetchDCMDataDevValues(const std::vector<int64_t>& subsystemConfigIdList,
                       const std::string& dcmName)
{
    ConfigDataSelector cfgSelector;
    cfgSelector.addSelectionColumn(DBColumnParams::DCMID_COLUMN_NAME,
                                   dcmName, dcmName);

    std::vector<std::string> dataColumnsOfInterest;
    dataColumnsOfInterest.push_back("numberOfReadAttempts");

    ConfigDataTree
        dcmDataDevDataSet("NovaDatabase/tables/DAQConfig/DCMDataDevRunParameters.xml",
                          subsystemConfigIdList, cfgSelector,
                          dataColumnsOfInterest);

    if (! dcmDataDevDataSet.getEffectiveValueFor1Key(dcmName,
                                                     "numberOfReadAttempts",
                                                     _numberOfReadAttempts)) {
        return false;
    }
    return true;
}

bool DCMRunConfig::
_fetchDCMApplicationParams(const std::vector<int64_t>& subsystemConfigIdList,
                           const std::string& dcmName)
{
    ConfigDataSelector cfgSelector;
    cfgSelector.addSelectionColumn(DBColumnParams::DCMID_COLUMN_NAME,
                                   dcmName, dcmName);

    std::vector<std::string> dataColumnsOfInterest;
    dataColumnsOfInterest.push_back("milliSliceMode");
    dataColumnsOfInterest.push_back("milliSliceCalculateCheckSum");
    dataColumnsOfInterest.push_back("corruptMessageLimit");
    dataColumnsOfInterest.push_back("corruptMessageResetCount");
    dataColumnsOfInterest.push_back("monitorMetricSendTimeInterval");
    dataColumnsOfInterest.push_back("monitorCtrlRegGeneralStatus");
    dataColumnsOfInterest.push_back("maxAllowedDelayInMilliSec");
    dataColumnsOfInterest.push_back("monitorFEBStatusTimeInterval");
    dataColumnsOfInterest.push_back("monitorFEBStatusLimit");

    ConfigDataTree
        dcmParamDataSet("NovaDatabase/tables/DAQConfig/DCMApplicationRunParameters.xml",
                        subsystemConfigIdList, cfgSelector,
                        dataColumnsOfInterest);

    if (! dcmParamDataSet.getEffectiveValueFor1Key(dcmName,
                                                   "milliSliceMode",
                                                   _milliSliceMode)) {
        return false;
    }
    if (! dcmParamDataSet.getEffectiveValueFor1Key(dcmName,
                                                   "milliSliceCalculateCheckSum",
                                                   _milliSliceCalculateCheckSum)) {
        return false;
    }
    if (! dcmParamDataSet.getEffectiveValueFor1Key(dcmName,
                                                   "corruptMessageLimit",
                                                   _corruptMessageLimit)) {
        return false;
    }
    if (! dcmParamDataSet.getEffectiveValueFor1Key(dcmName,
                                                   "corruptMessageResetCount",
                                                   _corruptMessageResetCount)) {
        return false;
    }
    if (! dcmParamDataSet.getEffectiveValueFor1Key(dcmName,
                                                   "monitorMetricSendTimeInterval",
                                                   _monitorMetricSendTimeInterval)) {
        return false;
    }
    if (! dcmParamDataSet.getEffectiveValueFor1Key(dcmName,
                                                   "monitorCtrlRegGeneralStatus",
                                                   _monitorCtrlRegGeneralStatus)) {
        return false;
    }
    if (! dcmParamDataSet.getEffectiveValueFor1Key(dcmName,
                                                   "maxAllowedDelayInMilliSec",
                                                   _maxAllowedDelayInMilliSec)) {
        return false;
    }
    if (! dcmParamDataSet.getEffectiveValueFor1Key(dcmName,
                                                   "monitorFEBStatusTimeInterval",
                                                   _monitorFEBStatusTimeInterval)) {
        return false;
    }
    if (! dcmParamDataSet.getEffectiveValueFor1Key(dcmName,
                                                   "monitorFEBStatusLimit",
                                                   _monitorFEBStatusLimit)) {
        return false;
    }
    return true;
}

};  // end of namespace daqconfig
};  // end of namespace dbutils
