#include <DatabaseUtils/GUI/BasicEditorDisplay.h>
#include <DatabaseUtils/GUI/ConfigSelectionDialog.h>
#include <DatabaseUtils/DAQConfig/NamedConfigUtils.h>
#include <NovaDAQUtilities/EnvVarCache.h>
#include <boost/lexical_cast.hpp>
#include <QtGui/QInputDialog>
#include <QtGui/QMenu>
#include <QtGui/QMenuBar>
#include <QtGui/QMessageBox>

using namespace dbutils::daqconfig;

namespace dbutils {
namespace gui {

BasicEditorDisplay::
BasicEditorDisplay(QWidget *parent) : QMainWindow(parent)
{
 
    this->setGeometry(0, 0, 750, 468);

    QMenu* fileMenu = menuBar()->addMenu(tr("&File"));

    QAction* reloadAction = fileMenu->addAction(tr("&Reload"));
    QObject::connect(reloadAction, SIGNAL(triggered()),
                     this, SLOT(reloadCurrentDataset()));

    QAction* openAction = fileMenu->addAction(tr("&Open"));
    QObject::connect(openAction, SIGNAL(triggered()),
                     this, SLOT(loadNewDataset()));

    _saveAction = fileMenu->addAction(tr("&Save"));
    QObject::connect(_saveAction, SIGNAL(triggered()),
                     this, SLOT(saveChanges()));
    _saveAction->setEnabled(false);

    _saveAsAction = fileMenu->addAction(tr("Save &As"));
    QObject::connect(_saveAsAction, SIGNAL(triggered()),
                     this, SLOT(saveChangesWithNewName()));
    _saveAsAction->setEnabled(false);

    QAction* exitAction = fileMenu->addAction(tr("E&xit"));
    QObject::connect(exitAction, SIGNAL(triggered()),
                     this, SLOT(close()));
}

void BasicEditorDisplay::reloadCurrentDataset()
{
    if (hasChanges()) {
        QMessageBox confirmBox(this);
        std::string msgText = "There are changes that haven't been saved yet.";
        confirmBox.setText(msgText.c_str());

        msgText = "Are you sure that you want to reload this dataset without ";
        msgText.append("saving the existing changes to the database?");
        confirmBox.setInformativeText(msgText.c_str());

        confirmBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
        confirmBox.setDefaultButton(QMessageBox::No);
        int retCode = confirmBox.exec();

        if (retCode == QMessageBox::No) {return;}
    }

    int64_t globalConfigId = _currentGlobalConfigId;
    if (! _globalNameSpecPtr->hasTimestamp()) {
        if (! NamedConfigUtils::getGlobalConfigId((*_globalNameSpecPtr),
                                                  globalConfigId)) {
            std::cerr << "Error fetching globalConfigId." << std::endl;
            return;
        }
    }

    IDSpec idSpec(globalConfigId, IDSpec::GLOBAL,
                  (_globalNameSpecPtr->isSlowControlsConfiguration() ?
                   IDSpec::SLOWCONTROLS : IDSpec::DAQ));

    if (_doReloadCurrentDataset(idSpec)) {
        //_globalNameSpecPtr = nameSpecPtr;
        _currentGlobalConfigId = globalConfigId;
        //if (_settingsDisplay != 0) {
        //    _settingsDisplay->updateNamedConfiguration(*_globalNameSpecPtr);
        //}
    }
}

// 13-Dec-2011, KAB - the loadNewDataset and reloadCurrentDataset methods
// should probably be refactored to pull out common elements.

void BasicEditorDisplay::loadNewDataset()
{
    if (hasChanges()) {
        QMessageBox confirmBox(this);
        std::string msgText = "There are changes that haven't been saved yet.";
        confirmBox.setText(msgText.c_str());

        msgText = "Are you sure that you want to load a new dataset without ";
        msgText.append("saving the existing changes to the database?");
        confirmBox.setInformativeText(msgText.c_str());

        confirmBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
        confirmBox.setDefaultButton(QMessageBox::No);
        int retCode = confirmBox.exec();

        if (retCode == QMessageBox::No) {return;}
    }

    ConfigSelectionDialog selectDialog((_globalNameSpecPtr->
                                        isSlowControlsConfiguration() ?
                                        IDSpec::SLOWCONTROLS : IDSpec::DAQ),
                                       _globalNameSpecPtr->getDetectorName(), this);
    int retCode = selectDialog.exec();
    if (retCode == QDialog::Rejected) {
        return;
    }

    boost::shared_ptr<dbutils::daqconfig::GlobalNameSpec> nameSpecPtr;
    if (selectDialog.getSelectedTime()==NamedConfigUtils::LATEST_TIME_STRING) {
        nameSpecPtr.
            reset(new GlobalNameSpec(selectDialog.getSelectedConfigName(),
                                     selectDialog.getSelectedDCMOpMode(),
                                     _globalNameSpecPtr->getDetectorName()));
    }
    else {
        nameSpecPtr.
            reset(new GlobalNameSpec(selectDialog.getSelectedConfigName(),
                                     selectDialog.getSelectedDCMOpMode(),
                                     _globalNameSpecPtr->getDetectorName(),
                                     selectDialog.getSelectedTime()));
    }
    nameSpecPtr->isSlowControlsConfiguration(_globalNameSpecPtr->
                                             isSlowControlsConfiguration());
    int64_t globalConfigId = -1;
    if (! NamedConfigUtils::getGlobalConfigId((*nameSpecPtr),
                                              globalConfigId)) {
        std::cerr << "Error fetching globalConfigId." << std::endl;
        return;
    }
    IDSpec idSpec(globalConfigId, IDSpec::GLOBAL,
                  (_globalNameSpecPtr->isSlowControlsConfiguration() ?
                   IDSpec::SLOWCONTROLS : IDSpec::DAQ));

    if (_doLoadNewDataset(idSpec)) {
        _globalNameSpecPtr = nameSpecPtr;
        _currentGlobalConfigId = globalConfigId;
        if (_settingsDisplay != 0) {
            _settingsDisplay->updateNamedConfiguration(*_globalNameSpecPtr);
        }
    }
    else {
        std::string msgText = "The loading of the new dataset failed. ";
        msgText.append("Depending on when the failure occurred, you may have an inconsistent set of data displayed in this editor. Please retry the loading of the new dataset until success is achieved before editing the data in the display.");
        QMessageBox::critical(this, "Data Loading Error",
                              msgText.c_str());
        return;
    }
}

void BasicEditorDisplay::saveChanges()
{
    if (! _confirmSave()) {return;}

    IDSpec oldGlobalIdSpec(_currentGlobalConfigId, IDSpec::GLOBAL,
                           (_globalNameSpecPtr->isSlowControlsConfiguration() ?
                            IDSpec::SLOWCONTROLS : IDSpec::DAQ));

    dbutils::SimpleParameterSet paramSet;
    paramSet.put(NamedConfigUtils::DESCRIPTION_COLUMN_NAME,
                 "New values from the Configuration Editor GUI.");
    int64_t globalConfigId;
    novadaq::Status resultStatus =
        NamedConfigUtils::copyGlobalConfig(oldGlobalIdSpec, true, paramSet,
                                           "", false, globalConfigId);
    if (resultStatus.code != NamedConfigUtils::COPY_SUCCEEDED) {
        bool userChoice = _getUserResponseToGlobalCopyFailure(resultStatus);
        if (! userChoice) {return;}

        resultStatus =
            NamedConfigUtils::copyGlobalConfig(oldGlobalIdSpec, true, paramSet,
                                               "", true, globalConfigId);
        if (resultStatus.code != NamedConfigUtils::COPY_SUCCEEDED) {
            std::string msgText = "Unable to create a copy of the named ";
            msgText.append("global configuration with ID = ");
            msgText.append(boost::lexical_cast<std::string>
                           (_currentGlobalConfigId));
            msgText.append(": ");
            msgText.append(resultStatus.message);
            QMessageBox::critical(this, "Save Error", msgText.c_str());
            return;
        }
    }

    // 12-Dec-2011, KAB - this copying of the GlobalNameSpec is simply
    // to get rid of any timestamp that may be present in the existing
    // _globalNameSpecPtr (since the act of saving will create a new latest)
    boost::shared_ptr<dbutils::daqconfig::GlobalNameSpec>
        nameSpecPtr(new GlobalNameSpec(_globalNameSpecPtr->getConfigName(),
                                       _globalNameSpecPtr->getDCMOpMode(),
                                       _globalNameSpecPtr->getDetectorName()));
    nameSpecPtr->
        isSlowControlsConfiguration(_globalNameSpecPtr->
                                    isSlowControlsConfiguration());
    IDSpec globalIdSpec(globalConfigId, IDSpec::GLOBAL,
                        (_globalNameSpecPtr->isSlowControlsConfiguration() ?
                         IDSpec::SLOWCONTROLS : IDSpec::DAQ));
    if (_doSaveChanges(oldGlobalIdSpec, globalIdSpec, (*nameSpecPtr))) {
        if (NamedConfigUtils::markConfigDone(globalIdSpec, true)) {
            _globalNameSpecPtr = nameSpecPtr;
            _currentGlobalConfigId = globalConfigId;
            if (_settingsDisplay != 0) {
                _settingsDisplay->updateNamedConfiguration(*nameSpecPtr);
            }
        }
    }
    else {
        bool status = NamedConfigUtils::markConfigDone(globalIdSpec, false);

        std::string msgText = "An error occurred when saving changes to ";
        msgText.append("the database.<br/><br/>");

        if (status) {
            msgText.append("However, the new configuration was ");
            msgText.append("successfully marked as Done (changes have ");
            msgText.append("finished). Please try saving again in a few ");
            msgText.append("minutes, but realize that you may need to ");
            msgText.append("over-write the incomplete save that just ");
            msgText.append("failed (the GUI will guide you through that).");
        }
        else {
            msgText.append("In addition, the new configuration was not ");
            msgText.append("successfully marked as Done (so it still ");
            msgText.append("appears that there is a change in progress ");
            msgText.append("in the DB table). If so, you'll need ");
            msgText.append("to use the ");
            msgText.append("<a href=\"https://cdcvs.fnal.gov/redmine/projects/novadaq-config-mgmt/wiki/How_To_recover_from_a_failed_pedestal_upload_-_December_2011\">MarkConfigChangesDone</a> utility to mark the change as Done.");
            msgText.append("In any case, please try saving again in a few ");
            msgText.append("minutes - the GUI will guide you through any ");
            msgText.append("extra steps that you need to take.");
        }

        QMessageBox::critical(this, "Save Error", msgText.c_str());
    }
}

void BasicEditorDisplay::saveChangesWithNewName()
{
    if (! _confirmSave()) {return;}

    NamedConfigUtils::GlobalNameSpecMap globalNameMap;
    NamedConfigUtils::
        getAvailableGlobalConfigs((_globalNameSpecPtr->
                                   isSlowControlsConfiguration() ?
                                   IDSpec::SLOWCONTROLS : IDSpec::DAQ),
                                  _globalNameSpecPtr->getDetectorName(),
                                  globalNameMap);

    bool status;
    std::string msg = "Please enter the new configuration name.";
    if (globalNameMap.size() > 0) {
        msg.append("<br/><br/>Recent choices for the name and the DCM ");
        msg.append("operating mode include the following:<ul>");
        NamedConfigUtils::GlobalNameSpecMap::const_iterator iter;
        NamedConfigUtils::GlobalNameSpecMap::const_iterator iterEnd =
            globalNameMap.end();
        for (iter = globalNameMap.begin(); iter != iterEnd; ++iter) {
            msg.append("<li>");
            msg.append((*iter).first.getConfigName());
            msg.append(" ");
            msg.append((*iter).first.getDCMOpMode());
            msg.append("</li>");
        }
        msg.append("</ul>");
    }
    QString newConfigName =
        QInputDialog::getText(this, tr("Configuration Name"), msg.c_str(),
                              QLineEdit::Normal,
                              _globalNameSpecPtr->getConfigName().c_str(),
                              &status);
    if (! status) {return;}

    msg = "Please enter the new DCM operating mode.";
    if (globalNameMap.size() > 0) {
        msg.append("<br/><br/>Recent choices for the name and the DCM ");
        msg.append("operating mode include the following:<ul>");
        NamedConfigUtils::GlobalNameSpecMap::const_iterator iter;
        NamedConfigUtils::GlobalNameSpecMap::const_iterator iterEnd =
            globalNameMap.end();
        for (iter = globalNameMap.begin(); iter != iterEnd; ++iter) {
            msg.append("<li>");
            msg.append((*iter).first.getConfigName());
            msg.append(" ");
            msg.append((*iter).first.getDCMOpMode());
            msg.append("</li>");
        }
        msg.append("</ul>");
    }
    QString newDCMOpMode =
        QInputDialog::getText(this, tr("DCM Operating Mode"), msg.c_str(),
                              QLineEdit::Normal,
                              _globalNameSpecPtr->getDCMOpMode().c_str(),
                              &status);
    if (! status) {return;}

    IDSpec oldGlobalIdSpec(_currentGlobalConfigId, IDSpec::GLOBAL,
                           (_globalNameSpecPtr->isSlowControlsConfiguration() ?
                            IDSpec::SLOWCONTROLS : IDSpec::DAQ));

    dbutils::SimpleParameterSet paramSet;
    paramSet.put(NamedConfigUtils::NAME_COLUMN_NAME,
                 newConfigName.toStdString());
    paramSet.put(NamedConfigUtils::DCMOPMODE_COLUMN_NAME,
                 newDCMOpMode.toStdString());
    paramSet.put(NamedConfigUtils::DESCRIPTION_COLUMN_NAME,
                 "New named configuration from the Configuration Editor GUI.");
    int64_t globalConfigId;
    novadaq::Status resultStatus =
        NamedConfigUtils::copyGlobalConfig(oldGlobalIdSpec, true, paramSet,
                                           "", false, globalConfigId);
    if (resultStatus.code != NamedConfigUtils::COPY_SUCCEEDED) {
        bool userChoice = _getUserResponseToGlobalCopyFailure(resultStatus);
        if (! userChoice) {return;}

        resultStatus =
            NamedConfigUtils::copyGlobalConfig(oldGlobalIdSpec, true, paramSet,
                                               "", true, globalConfigId);
        if (resultStatus.code != NamedConfigUtils::COPY_SUCCEEDED) {
            std::string msgText = "Unable to create a copy of the named ";
            msgText.append("global configuration with ID = ");
            msgText.append(boost::lexical_cast<std::string>
                           (_currentGlobalConfigId));
            msgText.append(": ");
            msgText.append(resultStatus.message);
            QMessageBox::critical(this, "Save Error", msgText.c_str());
            return;
        }
    }

    boost::shared_ptr<dbutils::daqconfig::GlobalNameSpec>
        nameSpecPtr(new GlobalNameSpec(newConfigName.toStdString(),
                                       newDCMOpMode.toStdString(),
                                       _globalNameSpecPtr->getDetectorName()));
    nameSpecPtr->
        isSlowControlsConfiguration(_globalNameSpecPtr->
                                    isSlowControlsConfiguration());
    IDSpec globalIdSpec(globalConfigId, IDSpec::GLOBAL,
                        (_globalNameSpecPtr->isSlowControlsConfiguration() ?
                         IDSpec::SLOWCONTROLS : IDSpec::DAQ));
    if (_doSaveChanges(oldGlobalIdSpec, globalIdSpec, (*nameSpecPtr))) {
        if (NamedConfigUtils::markConfigDone(globalIdSpec, true)) {
            _globalNameSpecPtr = nameSpecPtr;
            _currentGlobalConfigId = globalConfigId;
            if (_settingsDisplay != 0) {
                _settingsDisplay->updateNamedConfiguration(*nameSpecPtr);
            }
        }
    }
    else {
        NamedConfigUtils::markConfigDone(globalIdSpec, false);
    }
}

void BasicEditorDisplay::showHelpDialog()
{
    std::string msgText = "Helpful hints:";
    msgText.append("<ul>");
    msgText.append("<li>To edit a value, double-click on it [left mouse button]. (A single mouse click simply selects the value, it doesn't enable editing.)</li>");
    msgText.append("<li>Once you have edited a value, you should single-click on a different cell in the table to ensure that the editor captures your change. (Pressing the Tab or Enter key after you enter your change will also work.)</li>");
    msgText.append("<li>For \"mask\" values, a bitmap editor is provided. To open the bitmap editor, right-mouse-click on the value that you want to edit, and select the \"Open bitmap editor\" option from the popup menu.</li>");
    msgText.append("</ul>");
    msgText.append(_customHelpText);
    QMessageBox::information(this, "Configuration editor help",
                             msgText.c_str());
}

void BasicEditorDisplay::reactToDataChange()
{
    if (hasChanges()) {
        std::string dbUser =
            novadaq::EnvVarCache::getInstance().getEnvVar("NOVADBUSER");
        if (dbUser != "nova_reader") {
            _saveAction->setEnabled(true);
            _saveAsAction->setEnabled(true);
        }
    }
}

void BasicEditorDisplay::reactToDataSave()
{
    if (! hasChanges()) {
        _saveAction->setEnabled(false);
        _saveAsAction->setEnabled(false);
    }
}

void BasicEditorDisplay::closeEvent(QCloseEvent* event)
{
    if (! hasChanges()) {event->accept(); return;}

    QMessageBox confirmBox(this);

    std::string msgText = "There are changes that haven't been saved yet.";
    confirmBox.setText(msgText.c_str());

    msgText = "Are you sure that you want to exit without saving ";
    msgText.append("your changes to the database?");
    confirmBox.setInformativeText(msgText.c_str());

    confirmBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
    confirmBox.setDefaultButton(QMessageBox::No);
    int retCode = confirmBox.exec();

    if (retCode == QMessageBox::Yes) {event->accept(); return;}
    else {event->ignore(); return;}
}

bool BasicEditorDisplay::_confirmSave()
{
    if (! hasChanges()) {return false;}

    std::string dbString =
        novadaq::EnvVarCache::getInstance().getEnvVar("NOVADBNAME");
    QMessageBox confirmBox(this);

    std::string msgText = "You are about to update the \"";
    msgText.append(dbString);
    msgText.append("\" database.");
    confirmBox.setText(msgText.c_str());

    msgText = "Are you sure that you want to save your changes to the \"";
    msgText.append(dbString);
    msgText.append("\" database?");
    confirmBox.setInformativeText(msgText.c_str());

    confirmBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
    confirmBox.setDefaultButton(QMessageBox::No);
    int retCode = confirmBox.exec();

    return (retCode == QMessageBox::Yes);
}

bool BasicEditorDisplay::
_getUserResponseToGlobalCopyFailure(const novadaq::Status& statusObj)
{
    if (statusObj.code ==
        NamedConfigUtils::COPY_FOUND_UPDATE_IN_PROGRESS) {
        std::string msgText = "A problem was found when saving changes to ";
        msgText.append("the database.<br/><br/>");
        msgText.append("The creation of a configuration with the same ");
        msgText.append("name and DCM operating mode is currently in ");
        msgText.append("progress. (Attempting to copy global ");
        msgText.append("configuration with ID ");
        msgText.append(boost::lexical_cast<std::string>
                       (_currentGlobalConfigId));
        msgText.append(".)");
        msgText.append("<ul><li>Detailed message: \"");
        msgText.append(statusObj.message);
        msgText.append("\"</li></ul>");
        msgText.append("Please try again in a few minutes, or if ");
        msgText.append("this problem persists, use the ");
        msgText.append("<a href=\"https://cdcvs.fnal.gov/redmine/projects/novadaq-config-mgmt/wiki/How_To_recover_from_a_failed_pedestal_upload_-_December_2011\">MarkConfigChangesDone</a> utility to fix the issue.");
        msgText.append("<br/><br/>Another alternative is to save ");
        msgText.append("this configuration dataset with a different ");
        msgText.append("name (use File-&gt;Save As).");
        QMessageBox::warning(this, "Save Warning", msgText.c_str());

        return false;
    }
    else if (statusObj.code ==
             NamedConfigUtils::COPY_FOUND_INTERMEDIATE_UPDATE) {
        QMessageBox messageBox(this);
        messageBox.setIcon(QMessageBox::Warning);
        messageBox.setWindowTitle("Save Warning");

        std::string msgText = "A problem was found when saving changes to ";
        msgText.append("the database.<br/><br/>");
        msgText.append("Intermediate changes will be obscured by ");
        msgText.append("this change. (Attempting to copy global ");
        msgText.append("configuration with ID ");
        msgText.append(boost::lexical_cast<std::string>
                       (_currentGlobalConfigId));
        msgText.append(".)");
        msgText.append("<ul><li>Detailed message: \"");
        msgText.append(statusObj.message);
        msgText.append("\"</li></ul>");
        messageBox.setText(msgText.c_str());

        msgText = "Do you want to save anyway? (<b>Please Note</b> that ";
        msgText.append("if you say Yes, the dataset that you are ");
        msgText.append("currently saving will take precedence over the ");
        msgText.append("existing one with the same name and DCM operating ");
        msgText.append("mode <b>and</b> differences will <b>not</b> be ");
        msgText.append("automatically brought forward.)");
        messageBox.setInformativeText(msgText.c_str());

        messageBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
        messageBox.setDefaultButton(QMessageBox::No);
        int retCode = messageBox.exec();

        return (retCode == QMessageBox::Yes);
    }
    else if (statusObj.code != NamedConfigUtils::COPY_SUCCEEDED) {
        std::cerr << "Unable to create a copy of the named "
                  << "global configuration with ID = "
                  << _currentGlobalConfigId << ": " << statusObj.message
                  << std::endl;
        return false;
    }

    return true;
}

void BasicEditorDisplay::_addHelpMenu(std::string customText)
{
    _customHelpText = customText;

    QMenu* helpMenu = menuBar()->addMenu(tr("&Help"));
    QAction* helpAction =
        helpMenu->addAction(tr("&Help"));
    QObject::connect(helpAction, SIGNAL(triggered()),
                     this, SLOT(showHelpDialog()));
}

};  // end of namespace gui
};  // end of namespace dbutils
