#include <DatabaseUtils/GUI/BitmapEditor.h>
#include <NovaDAQUtilities/HexUtils.h>
#include <QtGui/QPushButton>
#include <QtGui/QDialogButtonBox>
#include <QtGui/QHBoxLayout>
#include <QtGui/QMessageBox>
#include <QtGui/QLabel>
#include <QtGui/QVBoxLayout>
#include <boost/lexical_cast.hpp>
#include <iomanip>
#include <stdint.h>
#include <iostream>

namespace dbutils {
namespace gui {

BitmapEditor::
BitmapEditor(const std::string& valueString, int bitCount, QWidget *parent) :
    QDialog(parent)
{
    setWindowTitle(tr("Bitmap Editor"));

    // ********************************
    // first, parse the input value
    // ********************************
    int64_t initialValue;
    if (! novadaq::HexUtils::parse(valueString, initialValue)) {
        _inputWasSpecifiedInHex = false;
        try {
            initialValue = boost::lexical_cast<int64_t>(valueString);
        }
        catch (...) {
            std::string msgText = "Data Parsing Error: ";
            msgText.append("Unable to convert value \"");
            msgText.append(valueString);
            msgText.append("\" to an integer.");
            QLabel* label = new QLabel(msgText.c_str());

            QDialogButtonBox* buttonBox =
                new QDialogButtonBox(QDialogButtonBox::Cancel);
            buttonBox->setCenterButtons(true);
            connect(buttonBox, SIGNAL(rejected()), this, SLOT(reject()));

            QVBoxLayout *mainLayout = new QVBoxLayout;
            mainLayout->addWidget(label);
            mainLayout->addWidget(buttonBox);
            setLayout(mainLayout);

            return;
        }
    }
    else {
        _inputWasSpecifiedInHex = true;
    }

    // ********************************
    // next, create the checkboxes
    // ********************************
    int bitsPerGroup = 8;
    int nGroups = ((int) ((bitCount - 1) / bitsPerGroup)) + 1;
    //int nGroups = 4;
    //int bitsPerGroup = ((int) ((bitCount - 1) / nGroups)) + 1;

    int64_t _64BitOne = 1;
    QHBoxLayout* fullBitBox = new QHBoxLayout;
    int totalCount = 0;
    for (int igrp = 0; igrp < nGroups; ++igrp) {
        QVBoxLayout* groupBox = new QVBoxLayout;
        for (int ibit = 0; ibit < bitsPerGroup; ++ibit) {
            std::string idxString =
                boost::lexical_cast<std::string>(totalCount);
            QCheckBox* cbox = new QCheckBox(idxString.c_str());

            if ((initialValue & (_64BitOne << totalCount)) != 0) {
                cbox->setChecked(true);
            }

            groupBox->addWidget(cbox);
            _checkboxList.push_back(cbox);

            ++totalCount;
            if (totalCount >= bitCount) {break;}
        }
        QWidget* tmpWidget = new QWidget();
        tmpWidget->setLayout(groupBox);
        fullBitBox->addWidget(tmpWidget);
    }
    QWidget* cboxWidget = new QWidget();
    cboxWidget->setLayout(fullBitBox);

    // ********************************
    // create the select/deselect buttons
    // ********************************
    QHBoxLayout* selectButtonBox = new QHBoxLayout;
    selectButtonBox->addStretch();
    QPushButton* button = new QPushButton(tr("Select All"));
    connect(button, SIGNAL(clicked()), this, SLOT(selectAll()));
    selectButtonBox->addWidget(button);
    button = new QPushButton(tr("Deselect All"));
    connect(button, SIGNAL(clicked()), this, SLOT(deselectAll()));
    selectButtonBox->addWidget(button);
    selectButtonBox->addStretch();

    QWidget* selButtonWidget = new QWidget();
    selButtonWidget->setLayout(selectButtonBox);

    // ********************************
    // create the accept/reject buttons
    // ********************************
    QDialogButtonBox* buttonBox =
        new QDialogButtonBox(QDialogButtonBox::Ok |
                             QDialogButtonBox::Cancel);
    buttonBox->setCenterButtons(true);

    connect(buttonBox, SIGNAL(accepted()), this, SLOT(accept()));
    connect(buttonBox, SIGNAL(rejected()), this, SLOT(reject()));

    // ********************************
    // put it all together
    // ********************************
    QVBoxLayout *mainLayout = new QVBoxLayout;
    mainLayout->addWidget(selButtonWidget);
    mainLayout->addWidget(cboxWidget);
    mainLayout->addWidget(buttonBox);
    setLayout(mainLayout);
}

std::string BitmapEditor::getResult() const
{
    int64_t _64BitOne = 1;
    int64_t workVal = 0;
    for (int idx = 0; idx < (int) _checkboxList.size(); ++idx) {
        if (_checkboxList[idx]->isChecked()) {
            workVal |= (_64BitOne << idx);
        }
    }

    if (_inputWasSpecifiedInHex) {
        if (_checkboxList.size() > 32) {
            return novadaq::HexUtils::format(workVal, 16);
        }
        else {
            return novadaq::HexUtils::format(workVal, 8);
        }
    }
    else {
        return boost::lexical_cast<std::string>(workVal);
    }
}

void BitmapEditor::selectAll()
{
    for (int idx = 0; idx < (int) _checkboxList.size(); ++idx) {
        _checkboxList[idx]->setChecked(true);
    }
}

void BitmapEditor::deselectAll()
{
    for (int idx = 0; idx < (int) _checkboxList.size(); ++idx) {
        _checkboxList[idx]->setChecked(false);
    }
}

};  // end of namespace gui
};  // end of namespace dbutils
