#include <DatabaseUtils/GUI/ConfigDataTableEditor.h>
#include <DatabaseUtils/GUI/ConfigDataTableModel.h>
#include <QtGui/QInputDialog>

namespace dbutils {
namespace gui {

/**
 * Creates the table display for editing configuration data from
 * a single database table.
 */
ConfigDataTableEditor::
ConfigDataTableEditor(const std::string& tableDefinitionFileName,
                      QWidget *parent) : QTableView(parent)
{
    _tableDefinitionFileName = tableDefinitionFileName;

    this->setAlternatingRowColors(true);
    this->setSelectionBehavior(QAbstractItemView::SelectItems);
    this->setHorizontalScrollMode(QAbstractItemView::ScrollPerPixel);
}

void ConfigDataTableEditor::mousePressEvent(QMouseEvent* event)
{
    // first, do the standard processing for the mouse event
    QTableView::mousePressEvent(event);

    // next, check if we should show the popup menu (right-mouse)
    if (event->type() == QEvent::MouseButtonPress &&
        event->button() == Qt::RightButton) {
        _popupMenu->popup(event->globalPos());
    }
}

bool ConfigDataTableEditor::
loadData(const std::vector<int64_t>& subsystemConfigIdList)
{
    ConfigDataTableModel *model =
        static_cast<ConfigDataTableModel *>(this->model());

    if (model == 0) {
        model = new ConfigDataTableModel(_tableDefinitionFileName);
        if (! model->loadData(subsystemConfigIdList)) {return false;}

        connect(model, SIGNAL(aChangeWasMade()),
                this, SLOT(reactToDataModelChange()));
        connect(model, SIGNAL(dataWasSaved()),
                this, SLOT(reactToDataSave()));

        this->setModel(model);

        _popupMenu = new QMenu(this);
        connect(this->selectionModel(),
                SIGNAL(currentChanged(const QModelIndex &,
                                      const QModelIndex &)),
                this, SLOT(updatePopupMenu(const QModelIndex &)));
    }
    else {
        if (! model->loadData(subsystemConfigIdList)) {return false;}
    }

    this->resizeColumnsToContents();
    this->resizeRowsToContents();
    return true;
}

bool ConfigDataTableEditor::hasChanges() const
{
    ConfigDataTableModel *model =
        static_cast<ConfigDataTableModel *>(this->model());

    if (model == 0) {return false;}

    return model->hasChanges();
}

bool ConfigDataTableEditor::saveChanges(const int64_t& oldSubsystemConfigId,
                                        const int64_t& subsystemConfigId)
{
    ConfigDataTableModel *model =
        static_cast<ConfigDataTableModel *>(this->model());

    if (model == 0) {return false;}

    return model->saveChanges(subsystemConfigId);
}

bool ConfigDataTableEditor::addInitialRow()
{
    ConfigDataTableModel *model =
        static_cast<ConfigDataTableModel *>(this->model());

    if (model == 0) {return true;}
    if (! isVisible()) {return true;}
    if (model->rowCount() > 0) {return true;}

    return model->addInitialRow();
}

std::string ConfigDataTableEditor::getTableName() const
{
    ConfigDataTableModel *model =
        static_cast<ConfigDataTableModel *>(this->model());

    if (model == 0) {return "unknown table";}

    return model->getTableName();
}

void ConfigDataTableEditor::updatePopupMenu(const QModelIndex& currentIndex)
{
    _popupMenu->clear();

    // always allow a row to be added
    _addRowAction = _popupMenu->addAction(tr("Add Row"));
    QObject::connect(_addRowAction, SIGNAL(triggered()),
                     this, SLOT(addRow()));

    // only allow a valid row to be deleted
    if (currentIndex.isValid()) {
        _addRowAction = _popupMenu->addAction(tr("Delete Row"));
        QObject::connect(_addRowAction, SIGNAL(triggered()),
                         this, SLOT(deleteRow()));
    }
}

void ConfigDataTableEditor::reactToDataModelChange()
{
    emit aChangeWasMade();
}

void ConfigDataTableEditor::reactToDataSave()
{
    emit dataWasSaved();
}

void ConfigDataTableEditor::addRow()
{
    ConfigDataTableModel *model =
        static_cast<ConfigDataTableModel *>(this->model());
    if (model == 0) {return;}

    QModelIndex currentIndex = this->selectionModel()->currentIndex();
    if (currentIndex.isValid()) {
        model->addRowAfter(currentIndex);
    }
    else {
        model->addInitialRow();
    }
}

void ConfigDataTableEditor::deleteRow()
{
    ConfigDataTableModel *model =
        static_cast<ConfigDataTableModel *>(this->model());
    if (model == 0) {return;}

    QModelIndex currentIndex = this->selectionModel()->currentIndex();
    if (currentIndex.isValid()) {
        model->deleteRow(currentIndex);
    }
}

};  // end of namespace gui
};  // end of namespace dbutils
