#include <DatabaseUtils/GUI/ConfigDataTableModel.h>
#include <DatabaseUtils/DAQConfig/DBColumnParams.h>
#include <DatabaseUtils/DAQConfig/DBColumnUtils.h>
#include <DatabaseUtils/DAQConfig/NamedConfigUtils.h>
#include <DatabaseUtils/SimpleParameterSet.h>
#include <NovaDAQUtilities/EnvVarCache.h>
#include <NovaDAQUtilities/HexUtils.h>
#include <NovaDAQUtilities/EnvVarCache.h>
#include <QtGui/QtGui>
#include <boost/lexical_cast.hpp>
#include <boost/algorithm/string.hpp>
#include <iomanip>

// ************************************************************
// PLEASE NOTE that this class is very specific to the DAQ configuration
// system.  In particular, it assumes that the configuration data that it
// will be used with A) will not use inheritance and B) will allow
// multiple rows per named subsystem configuration.
// ************************************************************

namespace NOVADB = nova::database;
using namespace dbutils::daqconfig;

namespace dbutils {
namespace gui {

// ************************************************************
// constructor/destructor
// ************************************************************

ConfigDataTableModel::
ConfigDataTableModel(const std::string& tableDefinitionFileName,
                     QObject *parent)
    : QAbstractItemModel(parent)
{
    _dbTablePtr.reset(new NOVADB::Table(tableDefinitionFileName.c_str()));
    std::string detName = 
      novadaq::EnvVarCache::getInstance().getEnvVar("NOVADAQ_ENVIRONMENT");
    _dbTablePtr->SetDetector(detName);

    std::string dbUser =
        novadaq::EnvVarCache::getInstance().getEnvVar("CFGMGR_DB_USER");
    if (dbUser.length() > 0) {
        _dbTablePtr->SetUser(dbUser);
    }
}


// ************************************************************
// the following methods do *not* have implementations in
// QAbstractItemModel, so they need to be implemented here
// (or in subclasses)
// ************************************************************

int ConfigDataTableModel::columnCount(const QModelIndex & /* parent */) const
{
    return _rootItem->columnCount();
}

QVariant ConfigDataTableModel::data(const QModelIndex &index, int role) const
{
    if (!index.isValid()) {return QVariant();}
    //std::cout << "Role = " << role << std::endl;

    if (role == Qt::BackgroundRole) {
        TreeItem *item = _getItem(index);
        if (item->isDirty() || item->isNew()) {
            QColor yellow("yellow");
            QBrush brush(yellow);
            return brush;
        }
    }

    if (role != Qt::DisplayRole && role != Qt::EditRole)
        return QVariant();

    TreeItem *item = _getItem(index);

    return item->data(index.column());
}

QModelIndex ConfigDataTableModel::index(int row, int column,
                                        const QModelIndex &parent) const
{
    if (parent.isValid() && parent.column() != 0) {
        return QModelIndex();
    }

    TreeItem *parentItem = _getItem(parent);

    TreeItem *childItem = parentItem->child(row);
    if (childItem)
        return createIndex(row, column, childItem);
    else
        return QModelIndex();
}

QModelIndex ConfigDataTableModel::parent(const QModelIndex &index) const
{
    if (!index.isValid()) {return QModelIndex();}

    TreeItem *childItem = _getItem(index);
    TreeItem *parentItem = childItem->parent();

    if (parentItem == _rootItem.get())
        return QModelIndex();

    return createIndex(parentItem->childNumber(), 0, parentItem);
}

int ConfigDataTableModel::rowCount(const QModelIndex &parent) const
{
    TreeItem *parentItem = _getItem(parent);

    return parentItem->childCount();
}


// ************************************************************
// the following methods *do* have implementations in QAbstractItemModel,
// but want to customize the implementation
// ************************************************************

Qt::ItemFlags ConfigDataTableModel::flags(const QModelIndex &index) const
{
    if (!index.isValid())
        return 0;

    return Qt::ItemIsEditable | Qt::ItemIsEnabled | Qt::ItemIsSelectable;
}

QVariant ConfigDataTableModel::
headerData(int section, Qt::Orientation orientation, int role) const
{
    if (orientation == Qt::Horizontal && role == Qt::DisplayRole)
        return _rootItem->data(section);

    return QVariant();
}

bool ConfigDataTableModel::insertRows(int position, int rows, const QModelIndex &parent)
{
    TreeItem *parentItem = _getItem(parent);

    beginInsertRows(parent, position, position + rows - 1);

    bool status = parentItem->insertChildren(position, rows,
                                             _rootItem->columnCount());
    if (status) {
        emit aChangeWasMade();
    }

    endInsertRows();

    return status;
}

bool ConfigDataTableModel::removeRows(int position, int rows, const QModelIndex &parent)
{
    TreeItem *parentItem = _getItem(parent);

    beginRemoveRows(parent, position, position + rows - 1);

    bool status = parentItem->removeChildren(position, rows);
    if (status) {
        emit aChangeWasMade();
    }

    endRemoveRows();

    return status;
}

bool ConfigDataTableModel::setData(const QModelIndex &index,
                                  const QVariant &value,
                                  int role)
{
    if (role != Qt::EditRole)
        return false;

    TreeItem *item = _getItem(index);
    QVariant oldValue = item->data(index.column());
    if (value == oldValue) {
        return false;
    }

    bool result = item->setData(index.column(), value);

    if (result) {
        item->markDirty();

        if (item->childCount() > 0) {
            this->removeRows(0, item->childCount(), index);
        }

        emit dataChanged(index, index);
        emit aChangeWasMade();
    }

    return result;
}

bool ConfigDataTableModel::setHeaderData(int section, Qt::Orientation orientation,
                                          const QVariant &value, int role)
{
    if (role != Qt::EditRole || orientation != Qt::Horizontal)
        return false;

    bool result = _rootItem->setData(section, value);

    if (result)
        emit headerDataChanged(orientation, section, section);

    return result;
}


// ************************************************************
// the following methods are unique to this class and its children
// ************************************************************

bool ConfigDataTableModel::
loadData(const std::vector<int64_t>& subsystemConfigIdList)
{
    if (! _dbTablePtr->GetConnection()) {return false;}
    if (! _dbTablePtr->ExistsInDB()) {return false;}

    // populate the root item with the DB table column names
    _dataColumnNames.clear();
    _dataColumnTypes.clear();
    QVector<QVariant> rootData;
    std::vector<std::string> columnNames = _dbTablePtr->GetColNames();
    foreach (std::string colName, columnNames) {
        if (DBColumnUtils::isInfrastructureColumn(colName)) {
            continue;
        }
        if (DBColumnUtils::
            isSubsystemConfigReferenceColumn(colName)) {
            continue;
        }

        _dataColumnNames.push_back(colName);
        QString tmpQString(colName.c_str());
        rootData << tmpQString;
        _dataColumnTypes.push_back("unknown");
    }

    if (_rootItem.get() != 0) {
        QModelIndex dummyIndex;
        this->removeRows(0, _rootItem->childCount(), dummyIndex);
    }
    _rootItem.reset(new TreeItem(rootData));
    QModelIndex rootIndex = this->index(0, 0);

    // find the relevant DB entries and populate the rest of the data model
    for (uint32_t cfgIdx = 0; cfgIdx<subsystemConfigIdList.size(); ++cfgIdx) {
        _dbTablePtr->Clear();
        _dbTablePtr->SetValidityRange(NamedConfigUtils::SUBSYSCFGID_COLUMN_NAME,
                                      subsystemConfigIdList[cfgIdx]);
        if (! _dbTablePtr->LoadFromDB()) {return false;}

        if (_dbTablePtr->NRow() > 0) {
            int initialRowCount = _dbTablePtr->NRow();
            insertRows(0, initialRowCount, rootIndex);

            for (int iRow = 0; iRow < initialRowCount; ++iRow) {
                NOVADB::Row* rowPtr = _dbTablePtr->GetRow(iRow);

                // add the requested item
                TreeItem* newItem = _rootItem->child(iRow);

                for (uint32_t iCol=0; iCol<_dataColumnNames.size(); ++iCol) {
                    std::string colName =
                        boost::to_lower_copy(_dataColumnNames[iCol]);

                    std::string value;
                    NOVADB::Column* colPtr = rowPtr->Col(colName);
                    if (colPtr->IsNull()) {continue;}
                    if (! colPtr->Get(value)) {return false;}
                    std::string colType = colPtr->Type();
                    _dataColumnTypes[iCol] = colType;

                    if (colType == "bool" || colType == "boolean") {
                        if (value == "t" || value == "1") {
                            value = "true";
                        }
                        else {
                            value = "false";
                        }
                    }

                    if (colName == "mask") {
                        try {
                            if (colType == "bigint") {
                                value = novadaq::HexUtils::format(value, 16);
                            }
                            else {
                                value = novadaq::HexUtils::format(value, 8);
                            }
                        }
                        catch (...) {}
                    }

                    newItem->setData(iCol, QVariant(value.c_str()));
                }
            }
        }

        break;  // TODO - handle a single subsysCfgId better
    }

    _dbTablePtr->CloseConnection();
    return true;
}

bool ConfigDataTableModel::hasChanges() const
{
    if (_rootItem->isDirty()) {return true;}

    // loop over children, if any
    for (int idx = 0; idx < _rootItem->childCount(); ++idx) {
        TreeItem* childItem = _rootItem->child(idx);
        if (childItem->isDirty() || childItem->isNew()) {return true;}
    }

    return false;
}

bool ConfigDataTableModel::saveChanges(const int64_t& subsystemConfigId)
{
    if (! _dbTablePtr->GetConnection()) {return false;}
    if (! _dbTablePtr->ExistsInDB()) {return false;}
    _dbTablePtr->SetVerbosity(1);

    for (int idx = 0; idx < _rootItem->childCount(); ++idx) {
        TreeItem* childItem = _rootItem->child(idx);

        boost::shared_ptr<NOVADB::Row> newRow(_dbTablePtr->NewRow());
        newRow->Set(NamedConfigUtils::SUBSYSCFGID_COLUMN_NAME,
                    subsystemConfigId);

        for (uint32_t iCol = 0; iCol < _dataColumnNames.size(); ++iCol) {
            std::string colName = boost::to_lower_copy(_dataColumnNames[iCol]);
            std::string colType = boost::to_lower_copy(_dataColumnTypes[iCol]);

            std::string value = childItem->data(iCol).toString().toStdString();

            if (value.find("0x") != std::string::npos) {
                try {
                    if (colType == "bigint") {
                        std::istringstream iss(value);
                        uint64_t u64Value;
                        iss >> std::hex >> u64Value;
                        int64_t i64Value = (int64_t) u64Value;
                        value = boost::lexical_cast<std::string>(i64Value);
                    }
                    else if (colType == "integer" || colType == "int") {
                        std::istringstream iss(value);
                        uint32_t u32Value;
                        iss >> std::hex >> u32Value;
                        int32_t i32Value = (int32_t) u32Value;
                        value = boost::lexical_cast<std::string>(i32Value);
                    }
                }
                catch (...) {}
            }

            if (colType == "bool" || colType == "boolean") {
                std::string valueCopy = boost::to_lower_copy(value);
                if (valueCopy == "t" || valueCopy == "true" || valueCopy == "1") {
                    newRow->Set(colName, true);
                }
                else {
                    newRow->Set(colName, false);
                }
            }
            else {
                newRow->Set(colName, value);
            }
        }

        _dbTablePtr->AddRow(newRow.get());
    }

    if (! _dbTablePtr->WriteToDB()) {return false;}

    _rootItem->markClean();
    for (int idx = 0; idx < _rootItem->childCount(); ++idx) {
        TreeItem* childItem = _rootItem->child(idx);
        childItem->markClean();
        childItem->markOld();
    }
    emit dataWasSaved();

    _dbTablePtr->CloseConnection();

    return true;
}

bool ConfigDataTableModel::addInitialRow()
{
    QModelIndex parentIndex = this->index(0, 0);
    beginInsertRows(parentIndex, 0, 0);

    bool status = _rootItem->insertChildren(0, 1, _rootItem->columnCount());
    if (status) {
        TreeItem* childItem = _rootItem->child(0);
        childItem->markNew();
        emit aChangeWasMade();
    }

    endInsertRows();

    return status;
}

bool ConfigDataTableModel::
addRowAfter(const QModelIndex &baseIndex)
{
    if (! baseIndex.isValid()) {return false;}

    TreeItem *baseItem = this->_getItem(baseIndex);
    int baseOffset = baseItem->childNumber();
    int position = baseOffset + 1;

    QModelIndex parentIndex = this->parent(baseIndex);
    TreeItem *parentItem = this->_getItem(parentIndex);

    beginInsertRows(parentIndex, position, position);

    bool status = parentItem->insertChildren(position, 1,
                                             _rootItem->columnCount());
    if (status) {
        QModelIndex childIndex = this->index(position, 0, parentIndex);
        if (childIndex.isValid()) {
            TreeItem* childItem = this->_getItem(childIndex);

            for (int idx = 0; idx < parentItem->columnCount(); ++idx) {
                childItem->setData(idx, baseItem->data(idx));
            }

            childItem->markNew();
        }
        emit aChangeWasMade();
    }

    endInsertRows();

    return status;
}

bool ConfigDataTableModel::
deleteRow(const QModelIndex &baseIndex)
{
    if (! baseIndex.isValid()) {return false;}

    TreeItem *baseItem = this->_getItem(baseIndex);
    int baseOffset = baseItem->childNumber();
    int position = baseOffset;

    QModelIndex parentIndex = this->parent(baseIndex);
    TreeItem *parentItem = this->_getItem(parentIndex);

    beginRemoveRows(parentIndex, position, position);

    bool status = parentItem->removeChildren(position, 1);
    if (status) {
        _rootItem->markDirty();
        emit aChangeWasMade();
    }

    endRemoveRows();

    return status;
}

std::string ConfigDataTableModel::getTableName() const
{
    return _dbTablePtr->Name();
}


// ************************************************************
// protected methods
// ************************************************************

TreeItem* ConfigDataTableModel::_getItem(const QModelIndex &index) const
{
    if (index.isValid()) {
        TreeItem* item = static_cast<TreeItem*>(index.internalPointer());
        if (item) return item;
    }
    return _rootItem.get();
}

void ConfigDataTableModel::_getModifiedItems(TreeItem* baseItem,
                                            std::vector<TreeItem*>& list) const
{
    // add this item to the list if it has been modified
    if (baseItem->isDirty() || baseItem->isNew()) {list.push_back(baseItem);}

    // loop over children, if any
    for (int idx = 0; idx < baseItem->childCount(); ++idx) {
        TreeItem* childItem = baseItem->child(idx);
        this->_getModifiedItems(childItem, list);
    }
}

};  // end of namespace gui
};  // end of namespace dbutils
