#include <DatabaseUtils/GUI/ConfigDataTreeEditor.h>
#include <DatabaseUtils/GUI/ConfigDataTreeModel.h>
#include <DatabaseUtils/GUI/BitmapEditor.h>
#include <DatabaseUtils/DAQConfig/DBColumnParams.h>
#include <QtGui/QInputDialog>

using namespace dbutils::daqconfig;

namespace dbutils {
namespace gui {

/**
 * Creates the tree display for editing configuration data from
 * a single database table.
 */
ConfigDataTreeEditor::
ConfigDataTreeEditor(QWidget *parent) : QTreeView(parent)
{
    this->setAlternatingRowColors(true);
    this->setSelectionBehavior(QAbstractItemView::SelectItems);
    this->setHorizontalScrollMode(QAbstractItemView::ScrollPerPixel);
    this->setAnimated(false);
    this->setAllColumnsShowFocus(true);
}

void ConfigDataTreeEditor::mousePressEvent(QMouseEvent* event)
{
    // first, do the standard processing for the mouse event
    QTreeView::mousePressEvent(event);

    // next, check if we should show the popup menu (right-mouse)
    if (event->type() == QEvent::MouseButtonPress &&
        event->button() == Qt::RightButton) {

        if (! _popupMenu->isEmpty()) {
            _popupMenu->popup(event->globalPos());
        }
    }
}

bool ConfigDataTreeEditor::
loadData(boost::shared_ptr<daqconfig::ConfigDataTree> dataTree)
{
    ConfigDataTreeModel *model =
        static_cast<ConfigDataTreeModel *>(this->model());

    if (model == 0) {
        model = new ConfigDataTreeModel(this);
        if (! model->loadData(dataTree)) {return false;}

        connect(model, SIGNAL(aChangeWasMade()),
                this, SLOT(reactToDataModelChange()));
        connect(model, SIGNAL(dataWasSaved()),
                this, SLOT(reactToDataSave()));
        connect(model, SIGNAL(dataChanged(const QModelIndex&,
                                          const QModelIndex&)),
                this, SLOT(refreshDisplayOfChildren(const QModelIndex&,
                                                    const QModelIndex&)));

        this->setModel(model);

        _popupMenu = new QMenu(this);
        connect(this->selectionModel(),
                SIGNAL(currentChanged(const QModelIndex &,
                                      const QModelIndex &)),
                this, SLOT(updatePopupMenu(const QModelIndex &)));
    }
    else {
        if (! model->loadData(dataTree)) {return false;}
    }

    this->expandAll();
    for (int column = 0; column < model->columnCount(); ++column) {
        this->resizeColumnToContents(column);
    }
    this->collapseAll();
    this->expandToDepth(0);
    return true;
}

bool ConfigDataTreeEditor::hasChanges() const
{
    ConfigDataTreeModel *model =
        static_cast<ConfigDataTreeModel *>(this->model());

    if (model == 0) {return false;}

    return model->hasChanges();
}

bool ConfigDataTreeEditor::saveChanges(const int64_t& oldSubsystemConfigId,
                                       const int64_t& subsystemConfigId)
{
    ConfigDataTreeModel *model =
        static_cast<ConfigDataTreeModel *>(this->model());

    if (model == 0) {return false;}

    return model->saveChanges(subsystemConfigId);
}

void ConfigDataTreeEditor::updatePopupMenu(const QModelIndex& currentIndex)
{
    _popupMenu->clear();
    if (! currentIndex.isValid()) {return;}

    ConfigDataTreeModel *model =
        static_cast<ConfigDataTreeModel *>(this->model());
    if (model == 0) {return;}

    // handle "mask" data columns
    std::string dataColumnName;
    bool dcnLookupWorked =
        model->dataColumnNameForIndex(currentIndex, dataColumnName);
    if (dcnLookupWorked && dataColumnName == "mask") {
        QAction* obeAction = _popupMenu->addAction(tr("Open bitmap editor"));
        QObject::connect(obeAction, SIGNAL(triggered()),
                         this, SLOT(editBitmap()));
        return;
    }

    // handle selector columns
    std::string peerType = "";
    std::string childType = "";
    std::string selectorColumn;
    if (model->selectorColumnNameForIndex(currentIndex, selectorColumn)) {
        if (selectorColumn == DBColumnParams::DCMNAME_COLUMN_NAME) {
            peerType = "DCM";
        }
        if (selectorColumn == DBColumnParams::FEBID_COLUMN_NAME) {
            peerType = "FEB";
        }
        if (selectorColumn == DBColumnParams::PIXELID_COLUMN_NAME) {
            peerType = "Pixel";
        }
        if (selectorColumn == DBColumnParams::HOSTNAME_COLUMN_NAME) {
            peerType = "Host";
        }
    }
    if (model->selectorColumnNameForChild(currentIndex, selectorColumn)) {
        if (selectorColumn == DBColumnParams::DCMNAME_COLUMN_NAME) {
            childType = "DCM";
        }
        if (selectorColumn == DBColumnParams::FEBID_COLUMN_NAME) {
            childType = "FEB";
        }
        if (selectorColumn == DBColumnParams::PIXELID_COLUMN_NAME) {
            childType = "Pixel";
        }
        if (selectorColumn == DBColumnParams::HOSTNAME_COLUMN_NAME) {
            childType = "Host";
        }
    }

    _candidatesForInsertBefore = model->allowsInsertBefore(currentIndex);
    _candidatesForInsertAfter = model->allowsInsertAfter(currentIndex);
    _candidatesForChildInsert = model->allowsChildInsert(currentIndex);

    if (_candidatesForInsertBefore.size() > 0 &&
        _candidatesForInsertAfter.size() == 0) {
        std::string text = "Add " + peerType;
        _addRowBeforeAction = _popupMenu->addAction(tr(text.c_str()));
        QObject::connect(_addRowBeforeAction, SIGNAL(triggered()),
                         this, SLOT(addRowBefore()));
    }
    else if (_candidatesForInsertBefore.size() == 0 &&
             _candidatesForInsertAfter.size() > 0) {
        std::string text = "Add " + peerType;
        _addRowAfterAction = _popupMenu->addAction(tr(text.c_str()));
        QObject::connect(_addRowAfterAction, SIGNAL(triggered()),
                         this, SLOT(addRowAfter()));
    }
    else if (_candidatesForInsertBefore.size() > 0 &&
             _candidatesForInsertAfter.size() > 0) {
        std::string text = "Add " + peerType + " before this";
        _addRowBeforeAction = _popupMenu->addAction(tr(text.c_str()));
        QObject::connect(_addRowBeforeAction, SIGNAL(triggered()),
                         this, SLOT(addRowBefore()));
        text = "Add " + peerType + " after this";
        _addRowAfterAction = _popupMenu->addAction(tr(text.c_str()));
        QObject::connect(_addRowAfterAction, SIGNAL(triggered()),
                         this, SLOT(addRowAfter()));
    }

    if (_candidatesForChildInsert.size() > 0) {
        std::string text = "Add " + childType;
        _addChildRowAction = _popupMenu->addAction(tr(text.c_str()));
        QObject::connect(_addChildRowAction, SIGNAL(triggered()),
                         this, SLOT(addChildRow()));
    }
}

bool ConfigDataTreeEditor::addInitialRow()
{
    ConfigDataTreeModel *model =
        static_cast<ConfigDataTreeModel *>(this->model());

    if (model == 0) {return true;}

    // TODO - implement the adding of an inital tree row
    //return model->hasChanges();
    return true;
}

void ConfigDataTreeEditor::reactToDataModelChange()
{
    emit aChangeWasMade();
}

void ConfigDataTreeEditor::reactToDataSave()
{
    emit dataWasSaved();
}

void ConfigDataTreeEditor::
refreshDisplayOfChildren(const QModelIndex& topLeft,
                         const QModelIndex& bottomRight)
{
    // this is a hack to get the correct set of children to be displayed
    // when we update a default value and remove some, but not all, of
    // the children.  Oddly, this->isExpanded(topLeft) doesn't seem to
    // work, otherwise we would use it here...
    this->collapse(topLeft);
    this->expand(topLeft);
}

void ConfigDataTreeEditor::addRowBefore()
{
    bool status;
    ConfigDataTreeModel *model =
        static_cast<ConfigDataTreeModel *>(this->model());
    if (model == 0) {return;}

    QModelIndex currentIndex = this->selectionModel()->currentIndex();
    std::string peerType;
    if (model->selectorColumnNameForIndex(currentIndex, peerType)) {
        if (peerType == DBColumnParams::FEBID_COLUMN_NAME) {
            QString responseText =
                QInputDialog::getItem(this, tr("Add FEB"), tr("FEB ID:"),
                                      _candidatesForInsertBefore, 0, false,
                                      &status);
            if (status) {
                model->addRowBefore(currentIndex, responseText);
            }
        }
        else if (peerType == DBColumnParams::PIXELID_COLUMN_NAME) {
            QString responseText =
                QInputDialog::getItem(this, tr("Add Pixel"), tr("Pixel ID:"),
                                      _candidatesForInsertBefore, 0, false,
                                      &status);
            if (status) {
                model->addRowBefore(currentIndex, responseText);
            }
        }
        else {
        }
    }

    this->updatePopupMenu(currentIndex);
    this->resizeColumnToContents(0);
}

void ConfigDataTreeEditor::addRowAfter()
{
    bool status;
    ConfigDataTreeModel *model =
        static_cast<ConfigDataTreeModel *>(this->model());
    if (model == 0) {return;}

    QModelIndex currentIndex = this->selectionModel()->currentIndex();
    std::string peerType;
    if (model->selectorColumnNameForIndex(currentIndex, peerType)) {
        if (peerType == DBColumnParams::DCMNAME_COLUMN_NAME) {
            QString responseText =
                QInputDialog::getText(this, tr("Add DCM"), tr("New DCM name:"),
                                      QLineEdit::Normal,
                                      _candidatesForInsertAfter[0], &status);
            if (status) {
                model->addRowAfter(currentIndex, responseText);
            }
        }
        else if (peerType == DBColumnParams::FEBID_COLUMN_NAME) {
            QString responseText =
                QInputDialog::getItem(this, tr("Add FEB"), tr("FEB ID:"),
                                      _candidatesForInsertAfter, 0, false,
                                      &status);
            if (status) {
                model->addRowAfter(currentIndex, responseText);
            }
        }
        else if (peerType == DBColumnParams::PIXELID_COLUMN_NAME) {
            QString responseText =
                QInputDialog::getItem(this, tr("Add Pixel"), tr("Pixel ID:"),
                                      _candidatesForInsertAfter, 0, false,
                                      &status);
            if (status) {
                model->addRowAfter(currentIndex, responseText);
            }
        }
        else if (peerType == DBColumnParams::HOSTNAME_COLUMN_NAME) {
            QString responseText =
                QInputDialog::getText(this, tr("Add Host"), tr("New Host Name:"),
                                      QLineEdit::Normal,
                                      _candidatesForInsertAfter[0], &status);
            if (status) {
                model->addRowAfter(currentIndex, responseText);
            }
        }
        else {
        }
    }

    this->updatePopupMenu(currentIndex);
    this->resizeColumnToContents(0);
}

void ConfigDataTreeEditor::addChildRow()
{
    bool status;
    ConfigDataTreeModel *model =
        static_cast<ConfigDataTreeModel *>(this->model());
    if (model == 0) {return;}

    QModelIndex currentIndex = this->selectionModel()->currentIndex();
    std::string childType;
    if (model->selectorColumnNameForChild(currentIndex, childType)) {
        if (childType == DBColumnParams::DCMNAME_COLUMN_NAME) {
            QString responseText =
                QInputDialog::getText(this, tr("Add DCM"), tr("New DCM name:"),
                                      QLineEdit::Normal,
                                      _candidatesForChildInsert[0], &status);
            if (status) {
                model->addChildRow(currentIndex, responseText);
                this->expand(currentIndex);
            }
        }
        else if (childType == DBColumnParams::FEBID_COLUMN_NAME) {
            QString responseText =
                QInputDialog::getItem(this, tr("Add FEB"), tr("FEB ID:"),
                                      _candidatesForChildInsert, 0, false,
                                      &status);
            if (status) {
                model->addChildRow(currentIndex, responseText);
                this->expand(currentIndex);
            }
        }
        else if (childType == DBColumnParams::PIXELID_COLUMN_NAME) {
            QString responseText =
                QInputDialog::getItem(this, tr("Add Pixel"), tr("Pixel ID:"),
                                      _candidatesForChildInsert, 0, false,
                                      &status);
            if (status) {
                model->addChildRow(currentIndex, responseText);
                this->expand(currentIndex);
            }
        }
        else if (childType == DBColumnParams::HOSTNAME_COLUMN_NAME) {
            QString responseText =
                QInputDialog::getText(this, tr("Add Host"), tr("New Host Name:"),
                                      QLineEdit::Normal,
                                      _candidatesForChildInsert[0], &status);
            if (status) {
                model->addChildRow(currentIndex, responseText);
                this->expand(currentIndex);
            }
        }
        else {
        }
    }

    this->updatePopupMenu(currentIndex);
    this->resizeColumnToContents(0);
}

void ConfigDataTreeEditor::editBitmap()
{
    ConfigDataTreeModel *model =
        static_cast<ConfigDataTreeModel *>(this->model());
    if (model == 0) {return;}

    QModelIndex currentIndex = this->selectionModel()->currentIndex();
    QVariant value = model->data(currentIndex);

    int bitCount = 32;
    std::string dataColumnType;
    bool dctLookupWorked =
        model->dataColumnTypeForIndex(currentIndex, dataColumnType);
    if (dctLookupWorked && dataColumnType == "bigint") {
        bitCount = 64;
    }

    BitmapEditor editor(value.toString().toStdString(), bitCount, this);
    int result = editor.exec();

    if (result == QDialog::Accepted) {
        model->setData(currentIndex, editor.getResult().c_str());
    }
}

};  // end of namespace gui
};  // end of namespace dbutils
