#include <DatabaseUtils/GUI/CustomTabWidget.h>
#include <DatabaseUtils/GUI/CustomTabBehaviors.h>
#include <DatabaseUtils/GUI/Changeable.h>
#include <DatabaseUtils/GUI/Storeable.h>
#include <boost/shared_ptr.hpp>
#include <QtGui/QProgressDialog>
#include <QtGui/QTabBar>

namespace dbutils {
namespace gui {

CustomTabWidget::
CustomTabWidget(QWidget *parent) : QTabWidget(parent), _isTopLevel(false)
{
}

int CustomTabWidget::addTab(QWidget* page, const QString& label)
{
    connect(page, SIGNAL(aChangeWasMade()), this, SLOT(reactToDataChange()));
    connect(page, SIGNAL(dataWasSaved()), this, SLOT(reactToDataSave()));

    return QTabWidget::addTab(page, label);
}

bool CustomTabWidget::loadData(const dbutils::daqconfig::IDSpec& idSpec)
{
    bool status = true;

    boost::shared_ptr<QProgressDialog> progressBar;
    if (_isTopLevel) {
        progressBar.reset(new QProgressDialog("Loading data...", "Close", 0,
                                              this->count(), this));
        progressBar->setWindowModality(Qt::WindowModal);
        progressBar->setMinimumDuration(250);
    }

    for (int idx = 0; idx < this->count(); ++idx) {
        if (_isTopLevel) {progressBar->setValue(idx);}
        QWidget* theWidget = widget(idx);
        Storeable* storeableObject = dynamic_cast<Storeable*>(theWidget);
        if (! storeableObject->loadData(idSpec)) {status = false;}
    }
    if (_isTopLevel) {progressBar->setValue(this->count());}

    return status;
}

bool CustomTabWidget::hasChanges() const
{
    for (int idx = 0; idx < this->count(); ++idx) {
        QWidget* theWidget = widget(idx);
        Changeable* changeableObject = dynamic_cast<Changeable*>(theWidget);
        if (changeableObject->hasChanges()) {return true;}
    }
    return false;
}

bool CustomTabWidget::addInitialRow()
{
    bool status = true;
    for (int idx = 0; idx < this->count(); ++idx) {
        QWidget* theWidget = widget(idx);
        Changeable* changeableObject = dynamic_cast<Changeable*>(theWidget);
        if (! changeableObject->addInitialRow()) {status = false;}
    }
    return status;
}

bool CustomTabWidget::
saveChanges(const dbutils::daqconfig::IDSpec& oldGlobalIdSpec,
            const dbutils::daqconfig::IDSpec& newGlobalIdSpec,
            const dbutils::daqconfig::GlobalNameSpec& newGlobalNameSpec)
{
    if (! hasChanges()) {return true;}

    boost::shared_ptr<QProgressDialog> progressBar;
    if (_isTopLevel) {
        progressBar.reset(new QProgressDialog("Loading data...", "Close", 0,
                                              this->count(), this));
        progressBar->setWindowModality(Qt::WindowModal);
        progressBar->setMinimumDuration(250);
    }

    bool status = true;
    for (int idx = 0; idx < this->count(); ++idx) {
        if (_isTopLevel) {progressBar->setValue(idx);}
        QWidget* theWidget = widget(idx);
        Storeable* storeableObject = dynamic_cast<Storeable*>(theWidget);
        if (! storeableObject->saveChanges(oldGlobalIdSpec,
                                           newGlobalIdSpec,
                                           newGlobalNameSpec)) {
            status = false;
        }
    }
    if (_isTopLevel) {progressBar->setValue(this->count());}
    return status;
}

void CustomTabWidget::reactToDataChange()
{
    CustomTabBehaviors::updateTabTextColors(this, this->tabBar());
    emit aChangeWasMade();
}

void CustomTabWidget::reactToDataSave()
{
    CustomTabBehaviors::updateTabTextColors(this, this->tabBar());
    emit dataWasSaved();
}

};  // end of namespace gui
};  // end of namespace dbutils
