#include <DatabaseUtils/GUI/DAQConfigEditorDisplay.h>
#include <DatabaseUtils/GUI/CustomTabBehaviors.h>
#include <DatabaseUtils/GUI/TabbedSetOfTreeEditors.h>
#include <DatabaseUtils/GUI/TabbedSetOfTableEditors.h>
#include <DatabaseUtils/DAQConfig/GlobalNameSpec.h>
#include <DatabaseUtils/DAQConfig/NamedConfigUtils.h>
#include <NovaDAQUtilities/EnvVarCache.h>
#include <QtCore/QCoreApplication>
#include <QtCore/QProcess>
#include <QtGui/QVBoxLayout>
#include <QtGui/QInputDialog>
#include <QtGui/QMenu>
#include <QtGui/QMenuBar>
#include <QtGui/QMessageBox>
#include <QtGui/QProgressDialog>
#include <QtGui/QFileDialog>
#include <boost/assign/std/vector.hpp>  // for 'operator+=()'
#include <boost/filesystem.hpp>  // for directory parsing
#include <algorithm>
#include <sstream>
#include <iomanip>
#include <locale>

using namespace boost::assign;  // bring 'operator+=()' into scope
using namespace dbutils::daqconfig;

namespace dbutils {
namespace gui {

DAQConfigEditorDisplay::
DAQConfigEditorDisplay(QWidget *parent) : BasicEditorDisplay(parent)
{
    std::string detectorString =
        novadaq::EnvVarCache::getInstance().getEnvVar("NOVADAQ_ENVIRONMENT");
    if (detectorString.length() == 0) {
        detectorString = "NOVADAQ_ENVIRONMENT_EnvVar_NeedsToBeSet";
    }
    
    // Menu option to call the scripts that
    // loads the threshold and masks generated in pedestal runner
    QMenu* actionMenu = menuBar()->addMenu(tr("&Action"));
    QAction* genScriptAction =
        actionMenu->addAction(tr("&Load Thresholds and Masks"));
    QObject::connect(genScriptAction, SIGNAL(triggered()),
                     this, SLOT(loadThresholdsAndMasks()));
    QAction* addInitialRowAction =
        actionMenu->addAction(tr("&Add Initial Row To Displayed Table"));
    QObject::connect(addInitialRowAction, SIGNAL(triggered()),
                     this, SLOT(addInitialRow()));

    _addHelpMenu();



    _globalNameSpecPtr.
      reset(new GlobalNameSpec("InitialGlobalConfig", "DCS",
                              detectorString));
    _currentGlobalConfigId = -1;
    if (! NamedConfigUtils::getGlobalConfigId((*_globalNameSpecPtr),
                                              _currentGlobalConfigId)) {
        std::cerr << "Error fetching globalConfigId." << std::endl;
    }
//    IDSpec idSpec(_currentGlobalConfigId, IDSpec::GLOBAL,
//                  (_globalNameSpecPtr->isSlowControlsConfiguration() ?
//                   IDSpec::SLOWCONTROLS : IDSpec::DAQ));

    std::vector<std::string> tableDefinitions;
    _tabWidget = new CustomTabWidget(this);
    _tabWidget->isTopLevel(true);
    QObject::connect(_tabWidget, SIGNAL(aChangeWasMade()),
                     this, SLOT(reactToDataChange()));
    QObject::connect(_tabWidget, SIGNAL(dataWasSaved()),
                     this, SLOT(reactToDataSave()));

    // DCM Connections
    tableDefinitions.clear();
    tableDefinitions +=
        "NovaDatabase/tables/DAQConfig/DCMApplicationConnectParameters.xml";

    TabbedSetOfTreeEditors* editorSet =
        new TabbedSetOfTreeEditors("DCMApplication", "connections",
                                   tableDefinitions,this);
    //editorSet->loadData(idSpec);
    _tabWidget->addTab(editorSet, "DCM Connection");

    // DDT Connect
    tableDefinitions.clear();
    tableDefinitions +=
      "NovaDatabase/tables/DAQConfig/DDTManagerConnectParameters.xml";

    editorSet =
      new TabbedSetOfTreeEditors("DDTManager", "connections",
				 tableDefinitions, this);
    //editorSet->loadData(idSpec);
    _tabWidget->addTab(editorSet, "DDT Manager");


    // DCM Hardware
    tableDefinitions.clear();
    tableDefinitions +=
        "NovaDatabase/tables/DAQConfig/FEBEnableMasks.xml",
        "NovaDatabase/tables/DAQConfig/PixelEnableMasks.xml",
        "NovaDatabase/tables/DAQConfig/PixelThresholds.xml",
        "NovaDatabase/tables/DAQConfig/DCMFPGAFirmwareLocations.xml",
        "NovaDatabase/tables/DAQConfig/FEBFPGAFirmwareLocations.xml",
        "NovaDatabase/tables/DAQConfig/DCMApplicationParameters.xml",
        "NovaDatabase/tables/DAQConfig/TimingSystemSettings.xml",
        "NovaDatabase/tables/DAQConfig/DCMDataDevParameters.xml",
        "NovaDatabase/tables/DAQConfig/DCMFPGAParameters.xml",
        "NovaDatabase/tables/DAQConfig/ASICRegisterSettings.xml",
        "NovaDatabase/tables/DAQConfig/DSODataRegulatorSettings.xml",
        "NovaDatabase/tables/DAQConfig/FEBPulserParameters.xml",
        "NovaDatabase/tables/DAQConfig/PixelOffsets.xml",
        "NovaDatabase/tables/DAQConfig/DCMSystemParameters.xml";

    editorSet =
        new TabbedSetOfTreeEditors("DCMApplication", "hardware",
                                   tableDefinitions,this);
    //editorSet->loadData(idSpec);
    _tabWidget->addTab(editorSet, "DCM Hardware");

    // DCM Run
    tableDefinitions.clear();
    tableDefinitions +=
        "NovaDatabase/tables/DAQConfig/DCMApplicationRunParameters.xml",
        "NovaDatabase/tables/DAQConfig/DCMDataDevRunParameters.xml",
        "NovaDatabase/tables/DAQConfig/DCMSystemRunParameters.xml";

    editorSet =
        new TabbedSetOfTreeEditors("DCMApplication", "run",
                                   tableDefinitions,this);
    //editorSet->loadData(idSpec);
    _tabWidget->addTab(editorSet, "DCM Run");

    // DaqMonitor Run
    tableDefinitions.clear();
    tableDefinitions +=
      "NovaDatabase/tables/DAQConfig/DaqMonitorRunParameters.xml";

    editorSet =
        new TabbedSetOfTreeEditors("DaqMonitor", "run",
                                   tableDefinitions,this);
    //editorSet->loadData(idSpec);
    _tabWidget->addTab(editorSet, "DaqMonitor Run");

    // BNEVB Run
    tableDefinitions.clear();
    tableDefinitions +=
        "NovaDatabase/tables/DAQConfig/BNEVBRunParameters.xml";

    editorSet =
        new TabbedSetOfTreeEditors("BufferNodeEVB", "run",
                                   tableDefinitions, this);
    //editorSet->loadData(idSpec);
    _tabWidget->addTab(editorSet, "BNEVB Run");

    // GT Run
    tableDefinitions.clear();
    tableDefinitions +=
        "NovaDatabase/tables/DAQConfig/CalibrationTriggers.xml",
        "NovaDatabase/tables/DAQConfig/SpillTriggers.xml",
	"NovaDatabase/tables/DAQConfig/DataDrivenTriggers.xml",
        "NovaDatabase/tables/DAQConfig/ExternalTriggers.xml",
        "NovaDatabase/tables/DAQConfig/ManualTriggers.xml",
        "NovaDatabase/tables/DAQConfig/SNEWSTriggers.xml",
        "NovaDatabase/tables/DAQConfig/SuperNovaTrigger.xml",
        "NovaDatabase/tables/DAQConfig/SuperNovaFilter.xml",
        "NovaDatabase/tables/DAQConfig/GlobalThrottle.xml",
        "NovaDatabase/tables/DAQConfig/DDTThrottle.xml",
        "NovaDatabase/tables/DAQConfig/TriggerOffsets.xml";

    TabbedSetOfTableEditors* editorSet2 =
        new TabbedSetOfTableEditors("GlobalTrigger", "run",
                                    tableDefinitions,this);
    //editorSet2->loadData(idSpec);
    _tabWidget->addTab(editorSet2, "Global Trigger Run");

    // DL Run
    tableDefinitions.clear();
    tableDefinitions +=
        "NovaDatabase/tables/DAQConfig/DataLoggerStreams.xml",
        "NovaDatabase/tables/DAQConfig/DataLoggerSystemParameters.xml";

    editorSet2 =
        new TabbedSetOfTableEditors("DataLogger", "run",
                                    tableDefinitions,this);
    //editorSet2->loadData(idSpec);
    _tabWidget->addTab(editorSet2, "Data Logger Run");


    _settingsDisplay = new SettingsDisplay();
    this->show();
    loadNewDataset();
//    _settingsDisplay->updateNamedConfiguration(*_globalNameSpecPtr);

    QVBoxLayout* thisDisplayLayout = new QVBoxLayout;
    thisDisplayLayout->addWidget(_tabWidget);
    thisDisplayLayout->addWidget(_settingsDisplay);
    QWidget* tmpWidget = new QWidget();
    tmpWidget->setContentsMargins(0,5,0,0);
    tmpWidget->setLayout(thisDisplayLayout);

    setCentralWidget(tmpWidget);

}

bool DAQConfigEditorDisplay::hasChanges() const
{
    return _tabWidget->hasChanges();
}

bool DAQConfigEditorDisplay::
_doReloadCurrentDataset(const dbutils::daqconfig::IDSpec& idSpec)
{
    return _tabWidget->loadData(idSpec);
}

bool DAQConfigEditorDisplay::
_doLoadNewDataset(const dbutils::daqconfig::IDSpec& idSpec)
{
    return _tabWidget->loadData(idSpec);
}

bool DAQConfigEditorDisplay::
_doSaveChanges(const dbutils::daqconfig::IDSpec& oldIdSpec,
               const dbutils::daqconfig::IDSpec& newIdSpec,
               const dbutils::daqconfig::GlobalNameSpec& newNameSpec)
{
    return _tabWidget->saveChanges(oldIdSpec, newIdSpec, newNameSpec);
}

void DAQConfigEditorDisplay::addInitialRow()
{
    _tabWidget->addInitialRow();
}

void DAQConfigEditorDisplay::loadThresholdsAndMasks()
{
  ///////////////////////////////////
  // Prompt for location of the csv
  std::string detName=_globalNameSpecPtr->getDetectorName();
  std::string detDir;
  std::string pedDir =
        novadaq::EnvVarCache::getInstance().getEnvVar("NOVADAQ_PEDESTAL_DIR").insert(5,"1");
  if(detName=="FarDet")
    if(pedDir != "")
      detDir = pedDir + "/PedestalData/PedScans";
    else
      detDir = "/data1-a/PedestalData/PedScans";
  else if(detName=="NDOS")
    detDir = "/data/PedestalData/PedScans";
  else if(detName=="NearDet")
    detDir = "/data/PedestalData/PedScans";
  else if(detName=="NDSBTest")
    detDir = "/data/PedestalData/PedScans";
  else
    detDir = "";

  QFileDialog peddialog(this);
  peddialog.setViewMode(QFileDialog::Detail);
  QString folderName = peddialog.getExistingDirectory(this, tr("Select Pedestal Directory."),
							 detDir.c_str(),
							 QFileDialog::ShowDirsOnly|QFileDialog::DontUseNativeDialog);
  if(folderName==""){return;}

  /////////////////////////////////
  // DB load confirmation box
  QMessageBox confirmBox(this);
  int count;
  std::ostringstream pixels;
  std::ostringstream febs;
  std::ostringstream dcms;
  
  boost::filesystem::path p(folderName.toUtf8().constData());
  boost::filesystem::path uniquestamp = p.filename();
  
  std::string threshFileName = folderName.toUtf8().constData();
  threshFileName.append("/thresholds_");
  threshFileName.append(uniquestamp.string());
  threshFileName.append(".csv");
  std::string pixelFileName = folderName.toUtf8().constData();
  pixelFileName.append("/pixelmask_"); 
  pixelFileName.append(uniquestamp.string());
  pixelFileName.append(".csv");
  std::string febFileName = folderName.toUtf8().constData();
  febFileName.append("/febmask_"); 
  febFileName.append(uniquestamp.string());
  febFileName.append(".csv");
  

  std::ifstream threshFile(threshFileName.c_str()); 
  count=std::count(std::istreambuf_iterator<char>(threshFile), 
  		    std::istreambuf_iterator<char>(), '\n');
  pixels << count;

  std::ifstream pixelFile(pixelFileName.c_str()); 
  count=std::count(std::istreambuf_iterator<char>(pixelFile), 
		  std::istreambuf_iterator<char>(), '\n');
  febs << count;

  std::ifstream febFile(febFileName.c_str()); 
  count=std::count(std::istreambuf_iterator<char>(febFile), 
		  std::istreambuf_iterator<char>(), '\n');
  dcms << count;
  
  std::string msgText = "You are about to load threholds and masks from  ";
  msgText.append(folderName.toUtf8().constData());
  msgText.append("\n\n Found results from ");
  msgText.append(pixels.str());
  msgText.append(" pixels, ");
  msgText.append(febs.str());
  msgText.append(" febs and ");
  msgText.append(dcms.str());
  msgText.append(" dcms.");
  confirmBox.setText(msgText.c_str());

  msgText = "Are you sure that you want to load these values? ";
  confirmBox.setInformativeText(msgText.c_str());
  confirmBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
  confirmBox.setDefaultButton(QMessageBox::No);
  int retCode = confirmBox.exec();
  if (retCode != QMessageBox::Yes) {return;}

  //////////
  // Ask if user want to overide all setting in the config or simply append
  // apply proper defualts if they want to override
  QMessageBox loadBox(this);
  msgText = "Do you want to override ALL existing exceptions underneath the defaults or append the results?";
  loadBox.setText(msgText.c_str());
  loadBox.addButton(tr("Override"),QMessageBox::AcceptRole);
  loadBox.addButton(tr("Append"),QMessageBox::NoRole);
  loadBox.addButton(QMessageBox::Cancel);
  loadBox.setDefaultButton(QMessageBox::Cancel);
  retCode = loadBox.exec();
  std::string configName = _globalNameSpecPtr->getConfigName();
  std::string systemCommand = "LoadEnableMasksAndThresholds NovaDatabase/tables/DAQConfig ";
  if (retCode == QMessageBox::Cancel){return;}
  else if (retCode == QMessageBox::AcceptRole) {
    std::string tmpstring;
    tmpstring = folderName.toUtf8().constData();
    tmpstring.append("/febXXXXXX");
    char *febTmpName = new char[tmpstring.length() + 1]; 
    strcpy(febTmpName, tmpstring.c_str());  // Get temp name

    tmpstring = folderName.toUtf8().constData();
    tmpstring.append("/pixelXXXXXX");  
    char *pixelTmpName =  new char[tmpstring.length() + 1];  // Get temp name
    strcpy(pixelTmpName, tmpstring.c_str());  // Get temp name

    tmpstring =folderName.toUtf8().constData();
    tmpstring.append("/thresholdXXXXXX");
    char *threshTmpName =  new char[tmpstring.length() + 1];  // Get temp name
    strcpy(threshTmpName, tmpstring.c_str());  // Get temp name

    mkstemp(febTmpName);
    mkstemp(pixelTmpName);
    mkstemp(threshTmpName);
    FILE * febFileTmp = fopen (febTmpName, "w");
    FILE * pixelFileTmp = fopen (pixelTmpName, "w");
    FILE * threshFileTmp = fopen (threshTmpName, "w");
    
    std::string systemCommandtmp = "";
    
    //defaults
    fputs ("1,default,0,0,1,2,3,4\n", febFileTmp);
    fputs ("1,default,-1,0,0,1,2,3,4\n",pixelFileTmp);
    fputs ("1,default,-1,-1,200,0,1,2,3,4\n",threshFileTmp);
    fclose(febFileTmp);
    fclose(pixelFileTmp);
    fclose(threshFileTmp);
    systemCommandtmp.append(systemCommand);
    systemCommandtmp.append(febTmpName);
    systemCommandtmp.append(" ");
    systemCommandtmp.append(pixelTmpName);
    systemCommandtmp.append(" ");
    systemCommandtmp.append(threshTmpName);
    systemCommandtmp.append(" ");
    systemCommandtmp.append(configName);
    system(systemCommandtmp.c_str());
  
    remove(febTmpName);
    remove(pixelTmpName);
    remove(threshTmpName);
  }

  /////////////////////////////////
  // Prompt user with status ans load into the database
   //#Output the database environment
  //echo
  //echo 'Loading into' $NOVADBNAME 'on' $NOVADBHOST 'at port' $NOVADBPORT '....'

  //#Load the Thesholds and Masks
 
  systemCommand.append(febFileName);
  systemCommand.append(" ");
  systemCommand.append(pixelFileName);
  systemCommand.append(" ");
  systemCommand.append(threshFileName);
  systemCommand.append(" ");
  systemCommand.append(configName);
  system(systemCommand.c_str());

  ///////////////////////////////
  //Refresh the editor
  reloadCurrentDataset();
  
}; // end of loadThresholdsAndMasks

};  // end of namespace gui
};  // end of namespace dbutils

