#include <DatabaseUtils/DAQConfig/DBColumnParams.h>
#include <DatabaseUtils/DAQConfig/GlobalNameSpec.h>
#include <DatabaseUtils/DAQConfig/NamedConfigUtils.h>
#include <DatabaseUtils/DAQConfig/SubsystemNameSpec.h>
#include <DatabaseUtils/DAQConfig/PixelEnableMasks.h>
#include <DatabaseUtils/DAQConfig/PixelThresholds.h>
#include <NovaDAQUtilities/EnvVarCache.h>
#include <NovaDAQUtilities/HexUtils.h>
#include <boost/lexical_cast.hpp>
#include <boost/shared_ptr.hpp>
#include <fstream>
#include <iomanip>

namespace DBUTILS = dbutils::daqconfig;

/**
 * Command-line utility for printing enable masks and thresholds.
 */

int main(int argc, char **argv)
{
    // parse command line arguments and print usage hints, if needed.
    std::string timestamp = "";
    bool printUsage = false;
    int c;
    std::string febmaskFilename("febmask_dump.csv");
    std::string pixelmaskFilename("pixelmask_dump.csv");
    std::string thresholdFilename("thresholds_dump.csv");
    while ((c = getopt (argc, argv, "f:p:r:t:h")) != -1) {
        switch (c)
        {
        case 't':
            timestamp = optarg;
            break;
        case 'f':
            febmaskFilename = optarg;
            break;
        case 'p':
            pixelmaskFilename = optarg;
            break;
        case 'r':
            thresholdFilename = optarg;
            break;
        case 'h':
            printUsage = true;
            break;
        default:
            printUsage = true;
            break;
        }
    }

    int64_t globalConfigId = 0;
    std::string configName;
    std::string dcmOpMode;
    std::string detectorName;
    std::vector<std::string> dcmList;
    if (! printUsage) {
        if ((argc - optind) < 1) {
            printUsage = true;
        }
        else {
            try {
                std::string firstArg(argv[optind]);
                globalConfigId = boost::lexical_cast<int64_t>(firstArg);
                ++optind;

                if ((argc - optind) < 1) {
                    printUsage = true;
                }
            }
            catch (...) {
                if ((argc - optind) < 4) {
                    printUsage = true;
                }
                else {
                    {
                        std::string tmpString(argv[optind]);
                        configName = tmpString;
                        ++optind;
                    }
                    {
                        std::string tmpString(argv[optind]);
                        dcmOpMode = tmpString;
                        ++optind;
                    }
                    {
                        std::string tmpString(argv[optind]);
                        detectorName = tmpString;
                        ++optind;
                    }
                }
            }
        }
    }
    if (! printUsage) {
        while (optind < argc) {
            std::string tmpString(argv[optind]);
            dcmList.push_back(tmpString);
            ++optind;
        }
    }

    // print usage hints, if needed.
    if (printUsage) {
        std::cout << std::endl;
        std::cout << "Usage: " << argv[0] << " [-t '<timestamp>'] "
                  << "[-f febmaskFilename] "
                  << "[-p pixelmaskFilename] "
                  << "[-r thresholdFilename] "
                  << "<globalCfgName> "
                  << "<dcmOpMode> <detectorName> <dcmList> "
                  << std::endl;
        std::cout << " OR " << std::endl;
        std::cout << "       " << argv[0] << " [-f febmaskFilename] "
                  << "[-p pixelmaskFilename] "
                  << "[-r thresholdFilename] "
                  << "<globalCfgId> "
                  << "<dcmList>" << std::endl;
        std::cout << "where" << std::endl;
        std::cout << " * globalCfgName specifies the "
                  << "named configuration to dump" << std::endl;
        std::cout << " * dcmOpMode specifies the DCM operating mode "
                  << "(e.g. DCS)" << std::endl;
        std::cout << " * detectorName specifies the detector name "
                  << "(e.g. NDOS)" << std::endl;
        std::cout << " * dcmList specifies the DCMs to focus on"
                  << std::endl;
        std::cout << " * timestamp specifies the time validity "
                  << "context to use [fmt=\"YYYYMMDD HH:MM:SS\"]" << std::endl
                  << "        (Remember that time should be UTC)"
                  << std::endl;
        std::cout << " * febmaskFilename specifies the name of the csv file "
                  << "to generate for FEB masks (defaults to "
                  << "febmask_dump.csv)" << std::endl;
        std::cout << " * pixelmaskFilename specifies the name of the csv file "
                  << "to generate for pixel masks (defaults to "
                  << "pixelmask_dump.csv)" << std::endl;
        std::cout << " * thresholdFilename specifies the name of the csv file "
                  << "to generate for pixel thresholds (defaults to "
                  << "thresholds_dump.csv)" << std::endl;
        std::cout << " * globalCfgId specifies a particular "
                  << "named global configuration to dump" << std::endl;
        std::cout << "Examples:" << std::endl;
        std::cout << "   " << argv[0] << " InitialGlobalConfig "
                  << "DCS NDOS dcm-3-01-01" << std::endl;
        std::cout << "   " << argv[0] << " -t '20110222 23:59:59' "
                  << "InitialGlobalConfig "
                  << "DCS NDOS dcm-3-01-01" << std::endl;
        std::cout << "   " << argv[0] << " 62 dcm-3-01-01" << std::endl;
        return -1;
    }

    // double-check that the user really wants to do this
    std::string dbName =
        novadaq::EnvVarCache::getInstance().getEnvVar("NOVADBNAME");
    std::cout << std::endl;
    std::cout << "Are you sure that you want to dump the enable masks and "
              << "thresholds from the " << dbName << " database (yes/no): ";
    std::string response;
    std::cin >> response;
    if (response.length() == 0 ||
        (response[0] != 'y' && response[0] != 'Y')) {
        exit(1);
    }

    // fetch the config data and generate the files
    if (globalConfigId == 0) {
        boost::shared_ptr<DBUTILS::GlobalNameSpec> globalNameSpec;
        if (timestamp.length() < 14) {
            globalNameSpec.reset(new DBUTILS::GlobalNameSpec(configName,
                                                             dcmOpMode,
                                                             detectorName));
        }
        else {
            globalNameSpec.reset(new DBUTILS::GlobalNameSpec(configName,
                                                             dcmOpMode,
                                                             detectorName,
                                                             timestamp));
        }
        if (! DBUTILS::NamedConfigUtils::getGlobalConfigId(*globalNameSpec,
                                                           globalConfigId)) {
            std::cout << std::endl << "No global configuration was found "
                      << "for the parameters that you specified!"
                      << std::endl;
            std::cout << "Please double-check that the values for the input "
                      << "parameters are correct..."
                      << std::endl;
            exit(1);
        }
    }

    DBUTILS::PixelEnableMasks enableMasks(globalConfigId, dcmList);
    DBUTILS::PixelThresholds thresholds(globalConfigId, dcmList);

    if (enableMasks.hasValidData() && thresholds.hasValidData()) {


        std::ofstream febmaskStream(febmaskFilename.c_str(),
                                    std::ios_base::out);
        if (! febmaskStream.good()) {
            std::cout << "ERROR: unable to open output file "
                      << febmaskFilename << "." << std::endl;
            exit(1);
        }
        for (uint32_t iDCM = 0; iDCM < dcmList.size(); ++iDCM) {
            std::string dcmName(dcmList[iDCM]);
            int64_t febMask;
            if (! enableMasks.getFEBEnableMask(dcmName, febMask)) {
                std::cout << "Error fetching the FEB mask for DCM "
                          << dcmName << std::endl;
                exit(1);
            }
            febmaskStream << "1," << dcmName << ","
                          << novadaq::HexUtils::format(febMask, 16)
                          << ",2,0,0,0,0" << std::endl;
        }
        febmaskStream.close();


        std::ofstream pixelmaskStream(pixelmaskFilename.c_str(),
                                      std::ios_base::out);
        if (! pixelmaskStream.good()) {
            std::cout << "ERROR: unable to open output file "
                      << pixelmaskFilename << "." << std::endl;
            exit(1);
        }
        for (uint32_t iDCM = 0; iDCM < dcmList.size(); ++iDCM) {
            std::string dcmName(dcmList[iDCM]);
            for (uint16_t febId = DBUTILS::DBColumnParams::LOW_FEB_ID;
                 febId <= DBUTILS::DBColumnParams::HIGH_FEB_ID; ++febId) {
                int32_t pixelMask;
                if (! enableMasks.getPixelEnableMask(dcmName, febId,
                                                     pixelMask)) {
                    std::cout << "Error fetching the pixel mask for DCM "
                              << dcmName << ", FEB " << febId << std::endl;
                    exit(1);
                }
                pixelmaskStream << "1," << dcmName << "," << febId << ","
                                << novadaq::HexUtils::format(pixelMask, 8)
                                << ",2,0,0,0,0" << std::endl;
            }
        }
        pixelmaskStream.close();


        std::ofstream thresholdStream(thresholdFilename.c_str(),
                                      std::ios_base::out);
        if (! thresholdStream.good()) {
            std::cout << "ERROR: unable to open output file "
                      << thresholdFilename << "." << std::endl;
            exit(1);
        }
        for (uint32_t iDCM = 0; iDCM < dcmList.size(); ++iDCM) {
            std::string dcmName(dcmList[iDCM]);
            for (uint16_t febId = DBUTILS::DBColumnParams::LOW_FEB_ID;
                 febId <= DBUTILS::DBColumnParams::HIGH_FEB_ID; ++febId) {
                for (uint16_t pixelId = DBUTILS::DBColumnParams::LOW_PIXEL_ID;
                     pixelId <= DBUTILS::DBColumnParams::HIGH_PIXEL_ID;
                     ++pixelId) {
                    //if (enableMasks.pixelIsEnabled(dcmName, febId, pixelId)) {
                    int16_t threshold;
                    if (! thresholds.getThreshold(dcmName, febId,
                                                  pixelId, threshold)) {
                        std::cout << "Error fetching the pixel threshold "
                                  << "for DCM " << dcmName << ", FEB "
                                  << febId << ", pixel "
                                  << pixelId << std::endl;
                        exit(1);
                    }
                    thresholdStream << "1," << dcmName << "," << febId
                                    << "," << pixelId << "," << threshold
                                    << ",2,0,0,0,0" << std::endl;
                    //}
                }
            }
        }
        thresholdStream.close();


    }
    else {
        std::cout << std::endl << "No configuration data was found "
                  << "for the parameters that you specified!"
                  << std::endl;
        std::cout << "Please double-check that the values for the input "
                  << "parameters are correct..."
                  << std::endl;
    }

    return 0;
}
