#include <DatabaseUtils/DAQConfig/NamedConfigUtils.h>
#include <NovaDAQUtilities/EnvVarCache.h>

using namespace dbutils::daqconfig;

/**
 * Command-line utility for printing out the available subsystem configurations.
 */

int main(int argc, char **argv)
{
    // parse command line arguments and print usage hints, if needed.
    bool printUsage = false;
    int c;
    while ((c = getopt (argc, argv, "h")) != -1) {
        switch (c)
        {
        case 'h':
            printUsage = true;
            break;
        default:
            printUsage = true;
            break;
        }
    }

    std::string daqOrSlowControls;
    std::string detectorName;
    std::string subsystemName;
    std::string configType;
    if (! printUsage) {
        if ((argc - optind) < 4) {
            printUsage = true;
        }
        else {
            {
                std::string tmpString(argv[optind]);
                if (tmpString == "DAQ" || tmpString == "SlowControls") {
                    daqOrSlowControls = tmpString;
                }
                else {
                    printUsage = true;
                }
                ++optind;
            }
            {
                std::string tmpString(argv[optind]);
                detectorName = tmpString;
                ++optind;
            }
            {
                std::string tmpString(argv[optind]);
                subsystemName = tmpString;
                ++optind;
            }
            {
                std::string tmpString(argv[optind]);
                configType = tmpString;
                ++optind;
            }
        }
    }

    // print usage hints, if needed.
    if (printUsage) {
        std::cout << std::endl;
        std::cout << "Usage: " << argv[0] << " <\"DAQ\"|\"SlowControls\"> "
                  << "<detectorName> <subsystemName> <configurationType>"
                  << std::endl;
        std::cout << "Where: " << std::endl;
        std::cout << " * \"DAQ\"|\"SlowControls\" selects which set "
                  << "of database tables to use" << std::endl;
        std::cout << " * detectorName specifies the detector name "
                  << "(e.g. NDOS)" << std::endl;
        std::cout << " * subsystemName specifies the subsystem name "
                  << "(e.g. DCMApplication)" << std::endl;
        std::cout << " * configurationType specifies the config type "
                  << "(e.g. hardware)" << std::endl;
        std::cout << "Examples:" << std::endl;
        std::cout << "   " << argv[0] << " DAQ NDOS DCMApplication hardware"
                  << std::endl;
        std::cout << "   " << argv[0] << " DAQ NDOS DCMApplication run"
                  << std::endl;
        std::cout << "   " << argv[0] << " DAQ NDOS GlobalTrigger run"
                  << std::endl;
        std::cout << "   " << argv[0] << " DAQ NDOS DataLogger run"
                  << std::endl;
        return -1;
    }

    // tell the user which database he/she is using
    std::string dbName =
        novadaq::EnvVarCache::getInstance().getEnvVar("NOVADBNAME");
    std::cout << "*** You are currently using the " << dbName
              << " database. ***" << std::endl;

    // fetch the configurations
    NamedConfigUtils::SubsystemNameSpecMap subsystemNameMap;
    if (! NamedConfigUtils::
        getAvailableSubsystemConfigs((daqOrSlowControls == "DAQ" ?
                                      IDSpec::DAQ : IDSpec::SLOWCONTROLS),
                                     detectorName, subsystemName,
                                     configType, subsystemNameMap)) {
        std::cerr << "*** ERROR: Unable to fetch the named subsystem "
                  << "configurations for detector \"" << detectorName
                  << "\"."<< std::endl;
        exit(2);
    }

    // fetch the configurations
    if (! NamedConfigUtils::getAvailableSubsystemConfigs(IDSpec::DAQ,
                                                         subsystemName,
                                                         configType,
                                                         detectorName,
                                                         subsystemNameMap)) {
        std::cerr << "*** ERROR: Unable to fetch the named subsystem "
                  << "configurations from the database! (File = "
                  << __FILE__ << ", Line = " << __LINE__ << ")" << std::endl;
        exit(2);
    }

    // print out the results
    if (subsystemNameMap.size() > 0) {
        NamedConfigUtils::SubsystemNameSpecMap::const_iterator iter;
        NamedConfigUtils::SubsystemNameSpecMap::const_iterator iterEnd =
            subsystemNameMap.end();
        for (iter = subsystemNameMap.begin(); iter != iterEnd; ++iter) {
            std::cout << (*iter).first.toString() << std::endl;

            std::cout << "  *** Details:" << std::endl;
            std::vector<SubsystemNameSpec> detailsList = (*iter).second;
            for (int idx = 0; idx < (int) detailsList.size(); ++idx) {
                std::cout << detailsList[idx].toString(4) << std::endl;
            }
        }
    }
    else {
        std::cout << std::endl << "No subsystem configurations were found "
                  << "for the detector, etc. that you specified!"
                  << std::endl;
        std::cout << "Please double-check that the values for the input "
                  << "parameters are correct..."
                  << std::endl;
    }

    return 0;
}
