#include "DatabaseUtils/DAQConfig/DBColumnParams.h"
#include "DatabaseUtils/DAQConfig/GeneralDataReader.h"
#include "DatabaseUtils/DAQConfig/GlobalNameSpec.h"
#include "DatabaseUtils/DAQConfig/NamedConfigUtils.h"
#include "DatabaseUtils/DAQConfig/SubsystemNameSpec.h"
#include "NovaDatabase/Table.h"
#include "NovaDatabase/Util.h"
#include <iomanip>

using namespace nova::database;
using namespace dbutils::daqconfig;

/**
 * Function to format an integer value as a hex string.
 */
std::string _hexFmt(const int64_t& value, const uint32_t& ndigits = 8)
{
    std::ostringstream formatter;
    formatter << "0x" << std::hex << std::setfill('0')
              << std::setw(ndigits) << std::uppercase << value;
    return formatter.str();
}
std::string _hexFmt(const int32_t& value, const uint32_t& ndigits = 8)
{
    std::ostringstream formatter;
    formatter << "0x" << std::hex << std::setfill('0')
              << std::setw(ndigits) << std::uppercase << value;
    return formatter.str();
}

/**
 * Command-line utility for displaying DCM hardware configuration information.
 */

int main(int argc, char **argv)
{
    // initialization
    std::string dcmSubsystemName = "DCMApplication";
    std::string hwConfigType = "hardware";
    std::string fname;
    boost::shared_ptr<Table> dbt;

    bool printAllData = false;
    bool printDCMData = false;
    bool printFEBData = false;
    bool printPixelData = false;

    // parse command line arguments and print usage hints, if needed.
    bool printUsage = false;
    int c;
    while ((c = getopt (argc, argv, "dfp")) != -1) {
        switch (c)
        {
        case 'd':
            printDCMData = true;
            break;
        case 'f':
            printFEBData = true;
            break;
        case 'p':
            printPixelData = true;
            break;
        default:
            printUsage = true;
        }
    }

    std::string configName;
    std::string dcmOpMode;
    std::string detectorName;
    std::string dcmName;
    std::string tmpDefaultString;
    int16_t selectedFebId;
    if (DBColumnParams::getInstance().
        columnKeyToGetDefaultValue(DBColumnParams::FEBID_COLUMN_NAME,
                                   tmpDefaultString)) {
        selectedFebId = boost::lexical_cast<int16_t>(tmpDefaultString);
    }
    else {
        std::cerr << "WARNING: Unable to determine default FEB ID"
                  << std::endl;
        selectedFebId = -1;
    }
    int16_t selectedPixelId;
    if (DBColumnParams::getInstance().
        columnKeyToGetDefaultValue(DBColumnParams::PIXELID_COLUMN_NAME,
                                   tmpDefaultString)) {
        selectedPixelId = boost::lexical_cast<int16_t>(tmpDefaultString);
    }
    else {
        std::cerr << "WARNING: Unable to determine default PIXEL ID"
                  << std::endl;
        selectedPixelId = -1;
    }
    if (! printUsage) {
        if ((argc - optind) < 4) {
            printUsage = true;
        }
        else {
            {
                std::string tmpString(argv[optind]);
                configName = tmpString;
                ++optind;
            }
            {
                std::string tmpString(argv[optind]);
                dcmOpMode = tmpString;
                ++optind;
            }
            {
                std::string tmpString(argv[optind]);
                detectorName = tmpString;
                ++optind;
            }
            {
                std::string tmpString(argv[optind]);
                dcmName = tmpString;
                ++optind;
            }
            if ((argc - optind) >= 1) {
                {
                    std::string tmpString(argv[optind]);
                    selectedFebId = boost::lexical_cast<int16_t>(tmpString);
                    ++optind;
                }
                if ((argc - optind) >= 1) {
                    std::string tmpString(argv[optind]);
                    selectedPixelId = boost::lexical_cast<int16_t>(tmpString);
                    ++optind;
                }
            }
        }
    }

    // print usage hints, if needed.
    if (printUsage) {
        std::cerr << "Usage: " << argv[0] << " [-d] [-f] [-p] <globalCfgName|"
                  << "dcmHwCfgName> <dcmOpMode> <detectorName> <dcmName> "
                  << "[febId [pixelId]]" << std::endl;
        std::cerr << "where  -d limits the output to DCM-related parameters"
                  << std::endl;
        std::cerr << "       -f limits the output to FEB-related parameters"
                  << std::endl;
        std::cerr << "       -p limits the output to pixel-related parameters"
                  << std::endl;
        std::cerr << "and globalCfgName (or dcmHwCfgName) specifies the named "
                  << "configuration to show" << std::endl;
        std::cerr << "    dcmOpMode specifies the DCM operating mode (e.g. DCS)"
                  << std::endl;
        std::cerr << "    detectorName specifies the detector name (e.g. NDOS)"
                  << std::endl;
        std::cerr << "    dcmName specifies the DCM to focus on"
                  << std::endl;
        std::cerr << "    febId   specifies a particular FEB channel"
                  << std::endl;
        std::cerr << "    pixelId specifies a particular pixel"
                  << std::endl;
        return -1;
    }
    if (! printDCMData && ! printFEBData && ! printPixelData) {
        printAllData = true;
    }

    // look up the specified named configuration
    // initially, assume that it's a subsystem configuration
    SubsystemNameSpec subsystemNameSpec(configName, dcmOpMode, detectorName,
                                        dcmSubsystemName, hwConfigType);
    int64_t subsystemConfigId;
    bool success =
        NamedConfigUtils::getSubsystemConfigId(subsystemNameSpec,
                                               subsystemConfigId);
    if (! success) {
        GlobalNameSpec globalNameSpec(configName, dcmOpMode, detectorName);
        success = NamedConfigUtils::
            getSubsystemConfigIdFromGlobalName(globalNameSpec,
                                               dcmSubsystemName,
                                               hwConfigType,
                                               subsystemConfigId);

        // couldn't find any configuration with the expected name
        if (! success) {
            std::cerr << "*** ERROR: Unable to find named configuration \""
                      << configName << "\", with DCM operating mode \""
                      << dcmOpMode << "\" and detector name \""
                      << detectorName << "\", in either the "
                      << "named_subsystem_configurations or "
                      << "named_global_configurations "
                      << "database tables!" << std::endl;
            exit(2);
        }
    }

    // fetch the system-wide data that we need
    int32_t microsliceTimeLength = 0;
    std::string dcmFPGAMode = "";
    if (printAllData || printDCMData) {
        GeneralDataReader
            dcmSystemParamReader("NovaDatabase/tables/DAQConfig/DCMSystemParameters.xml",
                                 subsystemConfigId);
        if (! dcmSystemParamReader.getValue("microSliceDT", microsliceTimeLength)) {
            std::cerr << "*** WARNING: Unable to fetch the microslice time length!"
                      << std::endl;
        }
        if (! dcmSystemParamReader.getValue("fpgaMode", dcmFPGAMode)) {
            std::cerr << "*** WARNING: Unable to fetch the DCM FPGA mode!"
                      << std::endl;
        }
    }

    // fetch the DCM-specific data that we need
    int32_t dcmReadTimeout = 0;
    int16_t useMmapBuffer = 0;
    if (printAllData || printDCMData) {
        GeneralDataReader
            dcmDataDevReader("NovaDatabase/tables/DAQConfig/DCMDataDevParameters.xml",
                             subsystemConfigId,
                             DBColumnParams::DCMID_COLUMN_NAME);
        if (! dcmDataDevReader.getValueFor1Key("readTimeout", dcmReadTimeout,
                                               dcmName)) {
            std::cerr << "*** WARNING: Unable to fetch the DCM data dev "
                      << "read timeout!" << std::endl;
        }
        if (! dcmDataDevReader.getValueFor1Key("useMemoryMappedBuffer",
                                               useMmapBuffer,
                                               dcmName)) {
            std::cerr << "*** WARNING: Unable to fetch the (DCM data dev) "
                      << "flag for using the memory mapped buffer!"
                      << std::endl;
        }
    }

    std::string dcmFirmwareLocation = "";
    if (printAllData || printDCMData) {
        GeneralDataReader
            dcmFirmwareReader("NovaDatabase/tables/DAQConfig/DCMFPGAFirmwareLocations.xml",
                              subsystemConfigId,
                              DBColumnParams::DCMID_COLUMN_NAME);
        if (! dcmFirmwareReader.getValueFor1Key("location", dcmFirmwareLocation,
                                                dcmName)) {
            std::cerr << "*** WARNING: Unable to fetch the DCM FPGA firmware location!"
                      << std::endl;
        }
    }

    std::string febFirmwareLocation;
    if (printAllData || printDCMData) {
        GeneralDataReader
            febFirmwareReader("NovaDatabase/tables/DAQConfig/FEBFPGAFirmwareLocations.xml",
                              subsystemConfigId,
                              DBColumnParams::DCMID_COLUMN_NAME);
        if (! febFirmwareReader.getValueFor1Key("location", febFirmwareLocation,
                                                dcmName)) {
            std::cerr << "*** WARNING: Unable to fetch the FEB FPGA firmware location!"
                      << std::endl;
        }
    }

    int64_t febEnableMask = 0;
    if (printAllData || printDCMData) {
        GeneralDataReader
            febEnableMaskReader("NovaDatabase/tables/DAQConfig/FEBEnableMasks.xml",
                                subsystemConfigId,
                                DBColumnParams::DCMID_COLUMN_NAME);
        if (! febEnableMaskReader.getValueFor1Key("mask", febEnableMask,
                                                  dcmName)) {
            std::cerr << "*** WARNING: Unable to fetch the FEB enable mask!"
                      << std::endl;
        }
    }

    int32_t fpgaLinkTimeout = 0;
    if (printAllData || printDCMData) {
        GeneralDataReader
            dcmParamReader("NovaDatabase/tables/DAQConfig/DCMFPGAParameters.xml",
                              subsystemConfigId,
                              DBColumnParams::DCMID_COLUMN_NAME);
        if (! dcmParamReader.getValueFor1Key("linkTimeout", fpgaLinkTimeout,
                                             dcmName)) {
            std::cerr << "*** WARNING: Unable to fetch the FPGA link timeout!"
                      << std::endl;
        }
    }

    int16_t numberOfHwConfigAttempts = 0;
    if (printAllData || printDCMData) {
        GeneralDataReader
            dcmParamReader("NovaDatabase/tables/DAQConfig/DCMApplicationParameters.xml",
                              subsystemConfigId,
                              DBColumnParams::DCMID_COLUMN_NAME);
        if (! dcmParamReader.getValueFor1Key("numberOfHwConfigAttempts",
                                             numberOfHwConfigAttempts,
                                             dcmName)) {
            std::cerr << "*** WARNING: Unable to fetch the "
                      << "number of hardware config attempts!"
                      << std::endl;
        }
    }

    int16_t timingSystemEnable = 0;
    int16_t timingSystemPort = 0;
    int32_t tduFirmwareVersion = 0;
    int16_t timingHistoryEnable = 0;
    if (printAllData || printDCMData) {
        GeneralDataReader
            timingSystemReader("NovaDatabase/tables/DAQConfig/TimingSystemSettings.xml",
                               subsystemConfigId,
                               DBColumnParams::DCMID_COLUMN_NAME);
        if (! timingSystemReader.getValueFor1Key("enable", timingSystemEnable,
                                                 dcmName)) {
            std::cerr << "*** WARNING: Unable to fetch the timing system enable!"
                      << std::endl;
        }
        if (! timingSystemReader.getValueFor1Key("port", timingSystemPort,
                                                 dcmName)) {
            std::cerr << "*** WARNING: Unable to fetch the timing system port!"
                      << std::endl;
        }
        if (! timingSystemReader.getValueFor1Key("tduFirmwareVersion",
                                                 tduFirmwareVersion,
                                                 dcmName)) {
            std::cerr << "*** WARNING: Unable to fetch the TDU firmware version!"
                      << std::endl;
        }
        if (! timingSystemReader.getValueFor1Key("historyEnable",
                                                 timingHistoryEnable,
                                                 dcmName)) {
            std::cerr << "*** WARNING: Unable to fetch the timing history enable!"
                      << std::endl;
        }
    }

    // fetch the FEB-specific data that we need
    uint16_t lowFebId = 0;
    uint16_t highFebId = 63;
    if (selectedFebId >= 0) {
        lowFebId = selectedFebId;
        highFebId = selectedFebId;
    }

    std::vector<int16_t> asicSpareValueList;
    std::vector<int16_t> asicVtSelValueList;
    std::vector<int16_t> asicRefSelValueList;
    std::vector<int16_t> asicISelValueList;
    std::vector<int16_t> asicBwSelValueList;
    std::vector<int16_t> asicGSelValueList;
    std::vector<int16_t> asicTfSelValueList;
    std::vector<int16_t> asicMux2to1ValueList;
    std::vector<int16_t> asicMux8to1ValueList;
    std::vector<int32_t> asicQInjPixelMaskList;
    if (printAllData || printFEBData) {
        GeneralDataReader
            asicRegisterReader("NovaDatabase/tables/DAQConfig/ASICRegisterSettings.xml",
                               subsystemConfigId,
                               DBColumnParams::DCMID_COLUMN_NAME,
                               DBColumnParams::FEBID_COLUMN_NAME);
        if (! asicRegisterReader.getListOfValuesFor2Keys("spare", asicSpareValueList,
                                                         dcmName, lowFebId, highFebId)) {
            std::cerr << "*** WARNING: Unable to fetch the ASIC spare values!"
                      << std::endl;
        }
        if (! asicRegisterReader.getListOfValuesFor2Keys("vtsel", asicVtSelValueList,
                                                         dcmName, lowFebId, highFebId)) {
            std::cerr << "*** WARNING: Unable to fetch the ASIC VtSel values!"
                      << std::endl;
        }
        if (! asicRegisterReader.getListOfValuesFor2Keys("refsel", asicRefSelValueList,
                                                         dcmName, lowFebId, highFebId)) {
            std::cerr << "*** WARNING: Unable to fetch the ASIC RefSel values!"
                      << std::endl;
        }
        if (! asicRegisterReader.getListOfValuesFor2Keys("isel", asicISelValueList,
                                                         dcmName, lowFebId, highFebId)) {
            std::cerr << "*** WARNING: Unable to fetch the ASIC ISel values!"
                      << std::endl;
        }
        if (! asicRegisterReader.getListOfValuesFor2Keys("bwsel", asicBwSelValueList,
                                                         dcmName, lowFebId, highFebId)) {
            std::cerr << "*** WARNING: Unable to fetch the ASIC BwSel values!"
                      << std::endl;
        }
        if (! asicRegisterReader.getListOfValuesFor2Keys("gsel", asicGSelValueList,
                                                         dcmName, lowFebId, highFebId)) {
            std::cerr << "*** WARNING: Unable to fetch the ASIC GSel values!"
                      << std::endl;
        }
        if (! asicRegisterReader.getListOfValuesFor2Keys("tfsel", asicTfSelValueList,
                                                         dcmName, lowFebId, highFebId)) {
            std::cerr << "*** WARNING: Unable to fetch the ASIC TfSel values!"
                      << std::endl;
        }
        if (! asicRegisterReader.getListOfValuesFor2Keys("mux2to1", asicMux2to1ValueList,
                                                         dcmName, lowFebId, highFebId)) {
            std::cerr << "*** WARNING: Unable to fetch the ASIC Mux2to1 values!"
                      << std::endl;
        }
        if (! asicRegisterReader.getListOfValuesFor2Keys("mux8to1", asicMux8to1ValueList,
                                                         dcmName, lowFebId, highFebId)) {
            std::cerr << "*** WARNING: Unable to fetch the ASIC Mux8to1 values!"
                      << std::endl;
        }
        if (! asicRegisterReader.getListOfValuesFor2Keys("qInjPixelMask", asicQInjPixelMaskList,
                                                         dcmName, lowFebId, highFebId)) {
            std::cerr << "*** WARNING: Unable to fetch the ASIC charge "
                      << "injection pixel masks!" << std::endl;
        }
    }

    std::vector<int16_t> dsoPointsPerPixelList;
    if (printAllData || printFEBData) {
        GeneralDataReader
            dsoDataRegulatorReader("NovaDatabase/tables/DAQConfig/DSODataRegulatorSettings.xml",
                                   subsystemConfigId,
                                   DBColumnParams::DCMID_COLUMN_NAME,
                                   DBColumnParams::FEBID_COLUMN_NAME);
        if (! dsoDataRegulatorReader.getListOfValuesFor2Keys("pointsPerPixel", dsoPointsPerPixelList,
                                                             dcmName, lowFebId, highFebId)) {
            std::cerr << "*** WARNING: Unable to fetch the DSO points per pixel!"
                      << std::endl;
        }
    }

    std::vector<int16_t> pulserEnableList;
    std::vector<int32_t> pulserPeriodList;
    std::vector<int32_t> pulserWidthList;
    if (printAllData || printFEBData) {
        GeneralDataReader
            febPulserReader("NovaDatabase/tables/DAQConfig/FEBPulserParameters.xml",
                            subsystemConfigId,
                            DBColumnParams::DCMID_COLUMN_NAME,
                            DBColumnParams::FEBID_COLUMN_NAME);
        if (! febPulserReader.getListOfValuesFor2Keys("pulserEnable", pulserEnableList,
                                                      dcmName, lowFebId, highFebId)) {
            std::cerr << "*** WARNING: Unable to fetch the pulser enable values!"
                      << std::endl;
        }
        if (! febPulserReader.getListOfValuesFor2Keys("pulserPeriod", pulserPeriodList,
                                                      dcmName, lowFebId, highFebId)) {
            std::cerr << "*** WARNING: Unable to fetch the pulser period values!"
                      << std::endl;
        }
        if (! febPulserReader.getListOfValuesFor2Keys("pulserWidth", pulserWidthList,
                                                      dcmName, lowFebId, highFebId)) {
            std::cerr << "*** WARNING: Unable to fetch the pulser width values!"
                      << std::endl;
        }
    }

    std::vector<int32_t> pixelEnableList;
    if (printAllData || printFEBData) {
        GeneralDataReader
            pixelEnableReader("NovaDatabase/tables/DAQConfig/PixelEnableMasks.xml",
                              subsystemConfigId,
                              DBColumnParams::DCMID_COLUMN_NAME,
                              DBColumnParams::FEBID_COLUMN_NAME);
        if (! pixelEnableReader.getListOfValuesFor2Keys("mask", pixelEnableList,
                                                        dcmName, lowFebId, highFebId)) {
            std::cerr << "*** WARNING: Unable to fetch the pixel enable masks!"
                      << std::endl;
        }
    }

    // fetch the Pixel-specific data that we need
    uint16_t lowPixelId = 0;
    uint16_t highPixelId = 31;
    if (selectedPixelId >= 0) {
        lowPixelId = selectedPixelId;
        highPixelId = selectedPixelId;
    }

    std::vector<boost::shared_ptr<std::vector<int16_t> > > pixelOffsetMatrix;
    if (printAllData || printPixelData) {
        GeneralDataReader
            pixelOffsetReader("NovaDatabase/tables/DAQConfig/PixelOffsets.xml",
                              subsystemConfigId,
                              DBColumnParams::DCMID_COLUMN_NAME,
                              DBColumnParams::FEBID_COLUMN_NAME,
                              DBColumnParams::PIXELID_COLUMN_NAME);
        if (! pixelOffsetReader.getSetOfValuesFor3Keys("offsetValue", pixelOffsetMatrix,
                                                       dcmName, lowFebId, highFebId,
                                                       lowPixelId, highPixelId)) {
            std::cerr << "*** WARNING: Unable to fetch the pixel offsets!"
                      << std::endl;
        }
    }

    std::vector<boost::shared_ptr<std::vector<int16_t> > > pixelThresholdMatrix;
    if (printAllData || printPixelData) {
        GeneralDataReader
            pixelThresholdReader("NovaDatabase/tables/DAQConfig/PixelThresholds.xml",
                                 subsystemConfigId,
                                 DBColumnParams::DCMID_COLUMN_NAME,
                                 DBColumnParams::FEBID_COLUMN_NAME,
                                 DBColumnParams::PIXELID_COLUMN_NAME);
        if (! pixelThresholdReader.getSetOfValuesFor3Keys("threshold", pixelThresholdMatrix,
                                                          dcmName, lowFebId, highFebId,
                                                          lowPixelId, highPixelId)) {
            std::cerr << "*** WARNING: Unable to fetch the pixel thresholds!"
                      << std::endl;
        }
    }


    // print out header
    std::cout << "### Hardware configuration for " << dcmName
              << " ###" << std::endl;

    // print system-wide and DCM-specific params
    if (printAllData || printDCMData) {
        std::cout << "  DCM FPGA mode = " << dcmFPGAMode << std::endl;
        std::cout << "  MilliSliceDT = " << microsliceTimeLength << std::endl;

        std::cout << "  DCM FPGA firmware = " << dcmFirmwareLocation
                  << std::endl;
        std::cout << "  FEB FPGA firmware = " << febFirmwareLocation
                  << std::endl;
        std::cout << "  FEB enable mask = " << _hexFmt(febEnableMask, 16)
                  << std::endl;

        std::cout << "  DCM data dev read timeout = " << dcmReadTimeout
                  << std::endl;
        std::cout << "  DCM number of harware configuration attempts = "
                  << numberOfHwConfigAttempts << std::endl;
        std::cout << "  DCM data dev mmap buffer flag = " << useMmapBuffer
                  << std::endl;
        std::cout << "  FPGA link timeout = " << fpgaLinkTimeout << std::endl;
        std::cout << "  Timing system enable = " << timingSystemEnable
                  << std::endl;
        std::cout << "  Timing system port = " << timingSystemPort << std::endl;
        std::cout << "  TDU firmware version = "
                  << _hexFmt(tduFirmwareVersion, 4) << std::endl;
        std::cout << "  Timing history enable = " << timingHistoryEnable
                  << std::endl;
    }

    // get the FEB-specific params
    uint16_t febCount = 1 + highFebId - lowFebId;
    for (uint16_t febId = lowFebId; febId <= highFebId; ++febId) {
        if (printAllData || printFEBData || printPixelData) {
            std::cout << "    FEB channel " << febId << ":" << std::endl;
        }
        uint16_t febIndex = febId - lowFebId;

        if (printAllData || printFEBData) {
            if (pixelEnableList.size() == febCount) {
                std::cout << "      Pixel enable mask = "
                          << _hexFmt(pixelEnableList[febIndex]) << std::endl;
            }

            if (asicSpareValueList.size() == febCount) {
                std::cout << "      ASIC Spare = "
                          << _hexFmt(asicSpareValueList[febIndex], 4)
                          << std::endl;
            }
            if (asicVtSelValueList.size() == febCount) {
                std::cout << "      ASIC VtSel = "
                          << _hexFmt(asicVtSelValueList[febIndex], 4)
                          << std::endl;
            }
            if (asicRefSelValueList.size() == febCount) {
                std::cout << "      ASIC RefSel = "
                          << _hexFmt(asicRefSelValueList[febIndex], 4)
                          << std::endl;
            }
            if (asicISelValueList.size() == febCount) {
                std::cout << "      ASIC ISel = "
                          << _hexFmt(asicISelValueList[febIndex], 4)
                          << std::endl;
            }
            if (asicBwSelValueList.size() == febCount) {
                std::cout << "      ASIC BwSel = "
                          << _hexFmt(asicBwSelValueList[febIndex], 4)
                          << std::endl;
            }
            if (asicGSelValueList.size() == febCount) {
                std::cout << "      ASIC GSel = "
                          << _hexFmt(asicGSelValueList[febIndex], 4)
                          << std::endl;
            }
            if (asicTfSelValueList.size() == febCount) {
                std::cout << "      ASIC TfSel = "
                          << _hexFmt(asicTfSelValueList[febIndex], 4)
                          << std::endl;
            }
            if (asicMux2to1ValueList.size() == febCount) {
                std::cout << "      ASIC Mux2to1 = "
                          << _hexFmt(asicMux2to1ValueList[febIndex], 4)
                          << std::endl;
            }
            if (asicMux8to1ValueList.size() == febCount) {
                std::cout << "      ASIC Mux8to1 = "
                          << _hexFmt(asicMux8to1ValueList[febIndex], 4)
                          << std::endl;
            }
            if (asicQInjPixelMaskList.size() == febCount) {
                std::cout << "      ASIC Charge Injection Pixel Mask = "
                          << _hexFmt(asicQInjPixelMaskList[febIndex], 8)
                          << std::endl;
            }

            if (pulserEnableList.size() == febCount) {
                std::cout << "      Pulser enable = "
                          << _hexFmt(pulserEnableList[febIndex], 4)
                          << std::endl;
            }

            if (pulserPeriodList.size() == febCount) {
                std::cout << "      Pulser period = "
                          << pulserPeriodList[febIndex] << std::endl;
            }

            if (pulserWidthList.size() == febCount) {
                std::cout << "      Pulser width = "
                          << pulserWidthList[febIndex] << std::endl;
            }

            if (dsoPointsPerPixelList.size() == febCount) {
                std::cout << "      DSO points per pixel = "
                          << dsoPointsPerPixelList[febIndex] << std::endl;
            }
        }

        // get the pixel-specific params
        uint16_t pixelCount = 1 + highPixelId - lowPixelId;
        for (uint16_t pixelId = lowPixelId; pixelId <= highPixelId; ++pixelId) {
            uint16_t pixelIndex = pixelId - lowPixelId;

            if (printAllData || printPixelData) {
                std::cout << "        Pixel " << pixelId << ":" << std::endl;

                if (pixelOffsetMatrix.size() == febCount) {
                    boost::shared_ptr<std::vector<int16_t> > offsetListPtr =
                        pixelOffsetMatrix[febIndex];

                    if (offsetListPtr->size() == pixelCount) {
                        std::cout << "          Offset = "
                                  << offsetListPtr->at(pixelIndex) << std::endl;
                    }
                }

                if (pixelThresholdMatrix.size() == febCount) {
                    boost::shared_ptr<std::vector<int16_t> > thresholdListPtr =
                        pixelThresholdMatrix[febIndex];

                    if (thresholdListPtr->size() == pixelCount) {
                        std::cout << "          Threshold = "
                                  << thresholdListPtr->at(pixelIndex) << std::endl;
                    }
                }
            }
        }
    }

    return 0;
}
