#include <cppunit/extensions/HelperMacros.h>
#include <NovaDAQUtilities/CppUnitTestDriver.h>
#include <DatabaseUtils/DAQAppMgr/ApplicationType.h>
#include <DatabaseUtils/DAQAppMgr/ApplicationInstance.h>
#include <DatabaseUtils/DAQAppMgr/ConfiguredApplicationTypes.h>
#include <DatabaseUtils/DAQAppMgr/ConfiguredProcessList.h>
#include <boost/shared_ptr.hpp>

/**
 * Unit tests for the DatabaseUtils/DAQAppMgr classes.
 *
 * @author Kurt Biery
 */

using namespace dbutils::daqappmgr;

class AppMgrUtilsUnitTest : public CppUnit::TestFixture
{
    CPPUNIT_TEST_SUITE(AppMgrUtilsUnitTest);
    CPPUNIT_TEST(testAppType);
    CPPUNIT_TEST(testAppInstance);
    CPPUNIT_TEST(testTypeList);
    CPPUNIT_TEST(testProcessList);
    CPPUNIT_TEST_SUITE_END();

public:
    void setUp();
    void tearDown();

    void testAppType();
    void testAppInstance();
    void testTypeList();
    void testProcessList();

private:
};

void
AppMgrUtilsUnitTest::setUp()
{
    // flush stdout to get progress markers to display in a timely way
    std::cout.flush();
}

void
AppMgrUtilsUnitTest::tearDown()
{
    // flush stdout to get progress markers to display in a timely way
    std::cout.flush();
}

void
AppMgrUtilsUnitTest::testAppType()
{
    boost::shared_ptr<ApplicationType> appType1;
    boost::shared_ptr<ApplicationType> appType2;

    std::vector<std::string> exeList1;
    exeList1.push_back("A");
    std::vector<std::string> exeList2;
    exeList2.push_back("B");
    std::vector<std::string> exeList3;
    exeList3.push_back("C");
    std::vector<std::string> exeList4;
    exeList4.push_back("Z");

    appType1.reset(new ApplicationType("A", exeList2, "C", "D", "E", "F",
                                       true, true));
    appType2.reset(new ApplicationType("A", exeList2, "C", "D", "E", "F",
                                       true, true));
    CPPUNIT_ASSERT(   *appType1 == *appType2);
    CPPUNIT_ASSERT(! (*appType1 != *appType2));
    CPPUNIT_ASSERT(! (*appType1 <  *appType2));

    CPPUNIT_ASSERT(appType1->getTypeName() == "A");
    CPPUNIT_ASSERT(appType1->getExeNameList().size() == 1);
    CPPUNIT_ASSERT((appType1->getExeNameList())[0] == "B");
    CPPUNIT_ASSERT(appType1->getStartCommand() == "C");
    CPPUNIT_ASSERT(appType1->getRshCommandForStart() == "D");
    CPPUNIT_ASSERT(appType1->getStopCommand() == "E");
    CPPUNIT_ASSERT(appType1->getRshCommandForStop() == "F");
    CPPUNIT_ASSERT(appType1->supportsStatusRequest());
    CPPUNIT_ASSERT(appType1->hasAppNameInProcessString());

    appType1.reset(new ApplicationType("A", exeList2, "C", "D", "E", "F",
                                       true, true));
    appType2.reset(new ApplicationType("Z", exeList2, "C", "D", "E", "F",
                                       true, true));
    CPPUNIT_ASSERT(! (*appType1 == *appType2));
    CPPUNIT_ASSERT(   *appType1 != *appType2);
    CPPUNIT_ASSERT(   *appType1 <  *appType2);
    CPPUNIT_ASSERT(! (*appType2 <  *appType1));

    appType1.reset(new ApplicationType("A", exeList2, "C", "D", "E", "F",
                                       true, true));
    appType2.reset(new ApplicationType("A", exeList4, "C", "D", "E", "F",
                                       true, true));
    CPPUNIT_ASSERT(! (*appType1 == *appType2));
    CPPUNIT_ASSERT(   *appType1 != *appType2);
    CPPUNIT_ASSERT(   *appType1 <  *appType2);
    CPPUNIT_ASSERT(! (*appType2 <  *appType1));

    appType1.reset(new ApplicationType("A", exeList2, "C", "D", "E", "F",
                                       true, true));
    appType2.reset(new ApplicationType("A", exeList2, "Z", "D", "E", "F",
                                       true, true));
    CPPUNIT_ASSERT(! (*appType1 == *appType2));
    CPPUNIT_ASSERT(   *appType1 != *appType2);
    CPPUNIT_ASSERT(   *appType1 <  *appType2);
    CPPUNIT_ASSERT(! (*appType2 <  *appType1));

    appType1.reset(new ApplicationType("A", exeList2, "C", "D", "E", "F",
                                       true, true));
    appType2.reset(new ApplicationType("A", exeList2, "C", "Z", "E", "F",
                                       true, true));
    CPPUNIT_ASSERT(! (*appType1 == *appType2));
    CPPUNIT_ASSERT(   *appType1 != *appType2);
    CPPUNIT_ASSERT(   *appType1 <  *appType2);
    CPPUNIT_ASSERT(! (*appType2 <  *appType1));

    appType1.reset(new ApplicationType("A", exeList2, "C", "D", "E", "F",
                                       true, true));
    appType2.reset(new ApplicationType("A", exeList2, "C", "D", "Z", "F",
                                       true, true));
    CPPUNIT_ASSERT(! (*appType1 == *appType2));
    CPPUNIT_ASSERT(   *appType1 != *appType2);
    CPPUNIT_ASSERT(   *appType1 <  *appType2);
    CPPUNIT_ASSERT(! (*appType2 <  *appType1));

    appType1.reset(new ApplicationType("A", exeList2, "C", "D", "E", "F",
                                       true, true));
    appType2.reset(new ApplicationType("A", exeList2, "C", "D", "E", "Z",
                                       true, true));
    CPPUNIT_ASSERT(! (*appType1 == *appType2));
    CPPUNIT_ASSERT(   *appType1 != *appType2);
    CPPUNIT_ASSERT(   *appType1 <  *appType2);
    CPPUNIT_ASSERT(! (*appType2 <  *appType1));

    appType1.reset(new ApplicationType("A", exeList2, "C", "D", "E", "F",
                                       false, true));
    appType2.reset(new ApplicationType("A", exeList2, "C", "D", "E", "F",
                                       true, true));
    CPPUNIT_ASSERT(! (*appType1 == *appType2));
    CPPUNIT_ASSERT(   *appType1 != *appType2);
    CPPUNIT_ASSERT(   *appType1 <  *appType2);
    CPPUNIT_ASSERT(! (*appType2 <  *appType1));

    appType1.reset(new ApplicationType("A", exeList2, "C", "D", "E", "F",
                                       true, false));
    appType2.reset(new ApplicationType("A", exeList2, "C", "D", "E", "F",
                                       true, true));
    CPPUNIT_ASSERT(! (*appType1 == *appType2));
    CPPUNIT_ASSERT(   *appType1 != *appType2);
    CPPUNIT_ASSERT(   *appType1 <  *appType2);
    CPPUNIT_ASSERT(! (*appType2 <  *appType1));

    appType1.reset(new ApplicationType("A", exeList2, "C", "D", "E", "F",
                                       true, true));
    appType2.reset(new ApplicationType("B", exeList1, "A", "A", "E", "F",
                                       false, false));
    CPPUNIT_ASSERT(! (*appType1 == *appType2));
    CPPUNIT_ASSERT(   *appType1 != *appType2);
    CPPUNIT_ASSERT(   *appType1 <  *appType2);
    CPPUNIT_ASSERT(! (*appType2 <  *appType1));

    appType1.reset(new ApplicationType("A", exeList1, "C", "D", "E", "F",
                                       true, true));
    appType2.reset(new ApplicationType("A", exeList3, "A", "A", "E", "F",
                                       false, false));
    CPPUNIT_ASSERT(! (*appType1 == *appType2));
    CPPUNIT_ASSERT(   *appType1 != *appType2);
    CPPUNIT_ASSERT(   *appType1 <  *appType2);
    CPPUNIT_ASSERT(! (*appType2 <  *appType1));

    appType1.reset(new ApplicationType("A", exeList2, "C", "D", "E", "F",
                                       true, true));
    appType2.reset(new ApplicationType("A", exeList2, "D", "A", "E", "F",
                                       false, false));
    CPPUNIT_ASSERT(! (*appType1 == *appType2));
    CPPUNIT_ASSERT(   *appType1 != *appType2);
    CPPUNIT_ASSERT(   *appType1 <  *appType2);
    CPPUNIT_ASSERT(! (*appType2 <  *appType1));

    appType1.reset(new ApplicationType("A", exeList2, "C", "D", "E", "F",
                                       true, true));
    appType2.reset(new ApplicationType("A", exeList2, "C", "E", "E", "F",
                                       false, false));
    CPPUNIT_ASSERT(! (*appType1 == *appType2));
    CPPUNIT_ASSERT(   *appType1 != *appType2);
    CPPUNIT_ASSERT(   *appType1 <  *appType2);
    CPPUNIT_ASSERT(! (*appType2 <  *appType1));
}

void
AppMgrUtilsUnitTest::testAppInstance()
{
    boost::shared_ptr<ApplicationInstance> appInstance1;
    boost::shared_ptr<ApplicationInstance> appInstance2;

    appInstance1.reset(new ApplicationInstance("A", "B", "C"));
    appInstance2.reset(new ApplicationInstance("A", "B", "C"));
    CPPUNIT_ASSERT(   *appInstance1 == *appInstance2);
    CPPUNIT_ASSERT(! (*appInstance1 != *appInstance2));
    CPPUNIT_ASSERT(! (*appInstance1 <  *appInstance2));

    CPPUNIT_ASSERT(appInstance1->getAppName() == "A");
    CPPUNIT_ASSERT(appInstance1->getTypeName() == "B");
    CPPUNIT_ASSERT(appInstance1->getHostName() == "C");
    CPPUNIT_ASSERT(appInstance1->getTypePtr().get() == 0);
    CPPUNIT_ASSERT(appInstance1->isEnabled());

    boost::shared_ptr<ApplicationType> appType1;
    boost::shared_ptr<ApplicationType> appType2;

    appInstance1.reset(new ApplicationInstance("A", "B", "C", appType1));
    appInstance2.reset(new ApplicationInstance("A", "B", "C", appType2));
    CPPUNIT_ASSERT(   *appInstance1 == *appInstance2);
    CPPUNIT_ASSERT(! (*appInstance1 != *appInstance2));
    CPPUNIT_ASSERT(! (*appInstance1 <  *appInstance2));

    boost::shared_ptr<Host> host1;
    boost::shared_ptr<Host> host2;

    appInstance1.reset(new ApplicationInstance("A", "B", "C", appType1, host1));
    appInstance2.reset(new ApplicationInstance("A", "B", "C", appType2, host2));
    CPPUNIT_ASSERT(   *appInstance1 == *appInstance2);
    CPPUNIT_ASSERT(! (*appInstance1 != *appInstance2));
    CPPUNIT_ASSERT(! (*appInstance1 <  *appInstance2));

    appInstance1.reset(new ApplicationInstance("A", "B", "C", appType1, host1, true));
    appInstance2.reset(new ApplicationInstance("A", "B", "C", appType2, host2, true));
    CPPUNIT_ASSERT(   *appInstance1 == *appInstance2);
    CPPUNIT_ASSERT(! (*appInstance1 != *appInstance2));
    CPPUNIT_ASSERT(! (*appInstance1 <  *appInstance2));

    appInstance1.reset(new ApplicationInstance("A", "B", "C", appType1, host1, true));
    appInstance2.reset(new ApplicationInstance("Z", "B", "C", appType2, host2, true));
    CPPUNIT_ASSERT(! (*appInstance1 == *appInstance2));
    CPPUNIT_ASSERT(   *appInstance1 != *appInstance2);
    CPPUNIT_ASSERT(   *appInstance1 <  *appInstance2);
    CPPUNIT_ASSERT(! (*appInstance2 <  *appInstance1));

    appInstance1.reset(new ApplicationInstance("A", "B", "C", appType1, host1, true));
    appInstance2.reset(new ApplicationInstance("A", "Z", "C", appType2, host2, true));
    CPPUNIT_ASSERT(! (*appInstance1 == *appInstance2));
    CPPUNIT_ASSERT(   *appInstance1 != *appInstance2);
    CPPUNIT_ASSERT(   *appInstance1 <  *appInstance2);
    CPPUNIT_ASSERT(! (*appInstance2 <  *appInstance1));

    appInstance1.reset(new ApplicationInstance("A", "B", "C", appType1, host1, true));
    appInstance2.reset(new ApplicationInstance("A", "B", "Z", appType2, host2, true));
    CPPUNIT_ASSERT(! (*appInstance1 == *appInstance2));
    CPPUNIT_ASSERT(   *appInstance1 != *appInstance2);
    CPPUNIT_ASSERT(   *appInstance1 <  *appInstance2);
    CPPUNIT_ASSERT(! (*appInstance2 <  *appInstance1));

    appInstance1.reset(new ApplicationInstance("A", "B", "C", appType1, host1, false));
    appInstance2.reset(new ApplicationInstance("A", "B", "C", appType2, host2, true));
    CPPUNIT_ASSERT(! (*appInstance1 == *appInstance2));
    CPPUNIT_ASSERT(   *appInstance1 != *appInstance2);
    CPPUNIT_ASSERT(   *appInstance1 <  *appInstance2);
    CPPUNIT_ASSERT(! (*appInstance2 <  *appInstance1));
}

void
AppMgrUtilsUnitTest::testTypeList()
{
    ConfiguredApplicationTypes typeList("BOGUS_DET_NAME");
    CPPUNIT_ASSERT(! typeList.hasValidData());

    std::vector<std::string> exeList2;
    exeList2.push_back("B");

    boost::shared_ptr<ApplicationType> appType1;
    appType1.reset(new ApplicationType("A", exeList2, "C", "D", "E", "F",
                                       true, true));
    typeList._addApplicationType(appType1);
    CPPUNIT_ASSERT(typeList.hasValidData());

    boost::shared_ptr<ApplicationType> appType2 = typeList.getTypeByName("A");
    CPPUNIT_ASSERT(*appType2 == *appType1);
}

void
AppMgrUtilsUnitTest::testProcessList()
{
    ConfiguredProcessList processList("BOGUS_DET_NAME");
    CPPUNIT_ASSERT(! processList.hasValidData());

    boost::shared_ptr<ApplicationInstance> appInstance1;
    appInstance1.reset(new ApplicationInstance("app001", "type001", "host001"));
    processList._addApplicationInstance(*appInstance1);
    CPPUNIT_ASSERT(processList.hasValidData());

    appInstance1.reset(new ApplicationInstance("app002", "type001", "host001"));
    processList._addApplicationInstance(*appInstance1);
    appInstance1.reset(new ApplicationInstance("app001", "type002", "host001"));
    processList._addApplicationInstance(*appInstance1);
    appInstance1.reset(new ApplicationInstance("app001", "type001", "host002"));
    processList._addApplicationInstance(*appInstance1);

    appInstance1.reset(new ApplicationInstance("app002", "type002", "host002"));
    processList._addApplicationInstance(*appInstance1);
    appInstance1.reset(new ApplicationInstance("app001", "type002", "host002"));
    processList._addApplicationInstance(*appInstance1);
    appInstance1.reset(new ApplicationInstance("app002", "type001", "host002"));
    processList._addApplicationInstance(*appInstance1);
    appInstance1.reset(new ApplicationInstance("app002", "type002", "host001"));
    processList._addApplicationInstance(*appInstance1);

    std::vector<std::string> appTypeList = processList.getUniqueAppTypes();
    CPPUNIT_ASSERT(appTypeList.size() == 2);
    CPPUNIT_ASSERT(appTypeList[0] == "type001");
    CPPUNIT_ASSERT(appTypeList[1] == "type002");

    std::vector<std::string> hostList = processList.getUniqueHosts();
    CPPUNIT_ASSERT(hostList.size() == 2);
    CPPUNIT_ASSERT(hostList[0] == "host001");
    CPPUNIT_ASSERT(hostList[1] == "host002");

    std::vector<ApplicationInstance> processSubset;
    processSubset.clear();
    CPPUNIT_ASSERT(processList.getProcessListByType("type001", processSubset));
    CPPUNIT_ASSERT(processSubset.size() == 4);

    processSubset.clear();
    std::vector<std::string> typeNameList;
    typeNameList.push_back("type001");
    CPPUNIT_ASSERT(processList.getProcessListByExcludedType(typeNameList,
                                                            processSubset));
    CPPUNIT_ASSERT(processSubset.size() == 4);

    processSubset.clear();
    typeNameList.push_back("type002");
    CPPUNIT_ASSERT(processList.getProcessListByExcludedType(typeNameList,
                                                            processSubset));
    CPPUNIT_ASSERT(processSubset.size() == 0);
}

CPPUNIT_TEST_SUITE_REGISTRATION(AppMgrUtilsUnitTest);
