#include <cppunit/extensions/HelperMacros.h>
#include <NovaDAQUtilities/CppUnitTestDriver.h>
#include <DatabaseUtils/DAQConfig/GlobalNameSpec.h>
#include <DatabaseUtils/DAQConfig/NamedConfigUtils.h>
#include <DatabaseUtils/DAQConfig/SubsystemNameSpec.h>

/**
 * Unit tests for the DAQConfig/NamedConfigUtils class.
 *
 * @author Kurt Biery
 */

using namespace dbutils::daqconfig;
using namespace nova::database;

const std::string NAMEDCFG_TABLE_DEFINITION =
    "DatabaseUtils/cxx/unittest/FakeNamedSubsystemConfigs.xml";

const std::string DCS_OP_MODE = "DCS";
const std::string DSO_OP_MODE = "DSO";

const std::string FAKE_SUBSYSTEM_NAME = "DCMApplication";
const std::string FAKE_CONFIG_TYPE = "hardware";

const std::string NAMED_CONFIG_A = "NamedConfigA";
const std::string NAMED_CONFIG_B = "NamedConfigB";
const std::string NAMED_CONFIG_C = "NamedConfigC";
const std::string NAMED_CONFIG_D = "NamedConfigD";

const std::string NAMED_CONFIG_E = "NamedConfigE";
const std::string NAMED_CONFIG_F = "NamedConfigF";
const std::string NAMED_CONFIG_G = "NamedConfigG";
const std::string NAMED_CONFIG_H = "NamedConfigH";

const std::string NDOS_DETECTOR_NAME = NamedConfigUtils::NDOS_DETECTOR_NAME;

class NamedConfigUnitTest : public CppUnit::TestFixture
{
    CPPUNIT_TEST_SUITE(NamedConfigUnitTest);
    CPPUNIT_TEST(populateDBTables);
    CPPUNIT_TEST(testGlobalNameSpec);
    CPPUNIT_TEST(testSubsystemNameSpec);
    CPPUNIT_TEST(testFetchingofIds);
    CPPUNIT_TEST(testNameReuse);
    CPPUNIT_TEST(testInvalidValues);
    CPPUNIT_TEST(testFetchOfAvailableSubsystemNames);
    CPPUNIT_TEST(cleanupDBTables);
    CPPUNIT_TEST_SUITE_END();

public:
    void setUp();
    void tearDown();

    void populateDBTables();
    void testGlobalNameSpec();
    void testSubsystemNameSpec();
    void testFetchingofIds();
    void testNameReuse();
    void testInvalidValues();
    void testFetchOfAvailableSubsystemNames();
    void cleanupDBTables();

private:

    bool _addTableRow(const std::string& configName, const int16_t& parentId);

    static boost::shared_ptr<Table> _namedCfgTablePtr;
};

boost::shared_ptr<Table> NamedConfigUnitTest::_namedCfgTablePtr;

void
NamedConfigUnitTest::setUp()
{
    // flush stdout to get progress markers to display in a timely way
    std::cout.flush();
}

void
NamedConfigUnitTest::tearDown()
{
    // flush stdout to get progress markers to display in a timely way
    std::cout.flush();
}

void
NamedConfigUnitTest::testGlobalNameSpec()
{
    boost::shared_ptr<GlobalNameSpec> globalNameSpec;

    globalNameSpec.reset(new GlobalNameSpec(NAMED_CONFIG_A, DCS_OP_MODE,
                                            NDOS_DETECTOR_NAME));
    CPPUNIT_ASSERT(globalNameSpec->getConfigName() == NAMED_CONFIG_A);
    CPPUNIT_ASSERT(globalNameSpec->getDCMOpMode() == DCS_OP_MODE);
    CPPUNIT_ASSERT(globalNameSpec->getDetectorId() == NDOS_DETECTOR_NAME);
    CPPUNIT_ASSERT(! globalNameSpec->hasTimestamp());

    globalNameSpec.reset(new GlobalNameSpec(NAMED_CONFIG_B, DSO_OP_MODE,
                                            NDOS_DETECTOR_NAME));
    CPPUNIT_ASSERT(globalNameSpec->getConfigName() == NAMED_CONFIG_B);
    CPPUNIT_ASSERT(globalNameSpec->getDCMOpMode() == DSO_OP_MODE);
    CPPUNIT_ASSERT(globalNameSpec->getDetectorId() == NDOS_DETECTOR_NAME);
    CPPUNIT_ASSERT(! globalNameSpec->hasTimestamp());

    globalNameSpec.reset(new GlobalNameSpec(NAMED_CONFIG_C, DCS_OP_MODE,
                                            NDOS_DETECTOR_NAME,
                                            "2010-12-01 01:02:03"));
    CPPUNIT_ASSERT(globalNameSpec->getConfigName() == NAMED_CONFIG_C);
    CPPUNIT_ASSERT(globalNameSpec->getDCMOpMode() == DCS_OP_MODE);
    CPPUNIT_ASSERT(globalNameSpec->getDetectorId() == NDOS_DETECTOR_NAME);
    CPPUNIT_ASSERT(globalNameSpec->hasTimestamp());

    globalNameSpec.reset(new GlobalNameSpec(NAMED_CONFIG_D, DSO_OP_MODE,
                                            NDOS_DETECTOR_NAME,
                                            "2010-12-01 01:02:03"));
    CPPUNIT_ASSERT(globalNameSpec->getConfigName() == NAMED_CONFIG_D);
    CPPUNIT_ASSERT(globalNameSpec->getDCMOpMode() == DSO_OP_MODE);
    CPPUNIT_ASSERT(globalNameSpec->getDetectorId() == NDOS_DETECTOR_NAME);
    CPPUNIT_ASSERT(globalNameSpec->hasTimestamp());

    globalNameSpec.reset(new GlobalNameSpec(NAMED_CONFIG_C, "BOGUS_OP_MODE",
                                            NDOS_DETECTOR_NAME));
    CPPUNIT_ASSERT(globalNameSpec->getConfigName() == NAMED_CONFIG_C);
    CPPUNIT_ASSERT(! globalNameSpec->hasTimestamp());
}

void
NamedConfigUnitTest::testSubsystemNameSpec()
{
    boost::shared_ptr<SubsystemNameSpec> subsystemNameSpec;

    subsystemNameSpec.reset(new SubsystemNameSpec(NAMED_CONFIG_A, DCS_OP_MODE,
                                                  NDOS_DETECTOR_NAME,
                                                  FAKE_SUBSYSTEM_NAME,
                                                  FAKE_CONFIG_TYPE));
    CPPUNIT_ASSERT(subsystemNameSpec->getConfigName() == NAMED_CONFIG_A);
    CPPUNIT_ASSERT(subsystemNameSpec->getDCMOpMode() == DCS_OP_MODE);
    CPPUNIT_ASSERT(subsystemNameSpec->getDetectorId() == NDOS_DETECTOR_NAME);
    CPPUNIT_ASSERT(! subsystemNameSpec->hasTimestamp());

    subsystemNameSpec.reset(new SubsystemNameSpec(NAMED_CONFIG_B, DSO_OP_MODE,
                                                  NDOS_DETECTOR_NAME,
                                                  FAKE_SUBSYSTEM_NAME,
                                                  FAKE_CONFIG_TYPE));
    CPPUNIT_ASSERT(subsystemNameSpec->getConfigName() == NAMED_CONFIG_B);
    CPPUNIT_ASSERT(subsystemNameSpec->getDCMOpMode() == DSO_OP_MODE);
    CPPUNIT_ASSERT(subsystemNameSpec->getDetectorId() == NDOS_DETECTOR_NAME);
    CPPUNIT_ASSERT(! subsystemNameSpec->hasTimestamp());

    subsystemNameSpec.reset(new SubsystemNameSpec(NAMED_CONFIG_C, DCS_OP_MODE,
                                                  NDOS_DETECTOR_NAME,
                                                  FAKE_SUBSYSTEM_NAME,
                                                  FAKE_CONFIG_TYPE,
                                                  "2010-12-01 01:02:03"));
    CPPUNIT_ASSERT(subsystemNameSpec->getConfigName() == NAMED_CONFIG_C);
    CPPUNIT_ASSERT(subsystemNameSpec->getDCMOpMode() == DCS_OP_MODE);
    CPPUNIT_ASSERT(subsystemNameSpec->getDetectorId() == NDOS_DETECTOR_NAME);
    CPPUNIT_ASSERT(subsystemNameSpec->hasTimestamp());

    subsystemNameSpec.reset(new SubsystemNameSpec(NAMED_CONFIG_D, DSO_OP_MODE,
                                                  NDOS_DETECTOR_NAME,
                                                  FAKE_SUBSYSTEM_NAME,
                                                  FAKE_CONFIG_TYPE,
                                                  "2010-12-01 01:02:03"));
    CPPUNIT_ASSERT(subsystemNameSpec->getConfigName() == NAMED_CONFIG_D);
    CPPUNIT_ASSERT(subsystemNameSpec->getDCMOpMode() == DSO_OP_MODE);
    CPPUNIT_ASSERT(subsystemNameSpec->getDetectorId() == NDOS_DETECTOR_NAME);
    CPPUNIT_ASSERT(subsystemNameSpec->hasTimestamp());

    subsystemNameSpec.reset(new SubsystemNameSpec(NAMED_CONFIG_C,
                                                  "BOGUS_OP_MODE",
                                                  NDOS_DETECTOR_NAME,
                                                  FAKE_SUBSYSTEM_NAME,
                                                  FAKE_CONFIG_TYPE));
    CPPUNIT_ASSERT(subsystemNameSpec->getConfigName() == NAMED_CONFIG_C);
    CPPUNIT_ASSERT(! subsystemNameSpec->hasTimestamp());
}

void
NamedConfigUnitTest::testFetchingofIds()
{
    boost::shared_ptr<SubsystemNameSpec> subsystemNameSpec;
    int64_t configId;

    CPPUNIT_ASSERT(_namedCfgTablePtr.get() != 0);
    CPPUNIT_ASSERT(_namedCfgTablePtr->ExistsInDB());

    // test the lookup of IDs that we expect to be there
    subsystemNameSpec.reset(new SubsystemNameSpec(NAMED_CONFIG_A,
                                                  DCS_OP_MODE,
                                                  NDOS_DETECTOR_NAME,
                                                  FAKE_SUBSYSTEM_NAME,
                                                  FAKE_CONFIG_TYPE));
    CPPUNIT_ASSERT(NamedConfigUtils::
                   getSubsystemConfigId(*subsystemNameSpec,
                                        configId,
                                        NAMEDCFG_TABLE_DEFINITION));
    CPPUNIT_ASSERT(configId == 1);
    subsystemNameSpec.reset(new SubsystemNameSpec(NAMED_CONFIG_B,
                                                  DCS_OP_MODE,
                                                  NDOS_DETECTOR_NAME,
                                                  FAKE_SUBSYSTEM_NAME,
                                                  FAKE_CONFIG_TYPE));
    CPPUNIT_ASSERT(NamedConfigUtils::
                   getSubsystemConfigId(*subsystemNameSpec,
                                        configId,
                                        NAMEDCFG_TABLE_DEFINITION));
    CPPUNIT_ASSERT(configId == 2);
    subsystemNameSpec.reset(new SubsystemNameSpec(NAMED_CONFIG_C,
                                                  DCS_OP_MODE,
                                                  NDOS_DETECTOR_NAME,
                                                  FAKE_SUBSYSTEM_NAME,
                                                  FAKE_CONFIG_TYPE));
    CPPUNIT_ASSERT(NamedConfigUtils::
                   getSubsystemConfigId(*subsystemNameSpec,
                                        configId,
                                        NAMEDCFG_TABLE_DEFINITION));
    CPPUNIT_ASSERT(configId == 3);
    subsystemNameSpec.reset(new SubsystemNameSpec(NAMED_CONFIG_D,
                                                  DCS_OP_MODE,
                                                  NDOS_DETECTOR_NAME,
                                                  FAKE_SUBSYSTEM_NAME,
                                                  FAKE_CONFIG_TYPE));
    CPPUNIT_ASSERT(NamedConfigUtils::
                   getSubsystemConfigId(*subsystemNameSpec,
                                        configId,
                                        NAMEDCFG_TABLE_DEFINITION));
    CPPUNIT_ASSERT(configId == 4);

    // test the lookup of an ID that we don't expect to be there
    subsystemNameSpec.reset(new SubsystemNameSpec("ZZZ_QQQ_RRR",
                                                  DCS_OP_MODE,
                                                  NDOS_DETECTOR_NAME,
                                                  FAKE_SUBSYSTEM_NAME,
                                                  FAKE_CONFIG_TYPE));
    CPPUNIT_ASSERT(! NamedConfigUtils::
                   getSubsystemConfigId(*subsystemNameSpec,
                                        configId,
                                        NAMEDCFG_TABLE_DEFINITION));

    // test the lookup of ancestor IDs by name spec
    std::vector<int64_t> idList;
    idList.clear();
    CPPUNIT_ASSERT(idList.size() == 0);
    subsystemNameSpec.reset(new SubsystemNameSpec(NAMED_CONFIG_A,
                                                  DCS_OP_MODE,
                                                  NDOS_DETECTOR_NAME,
                                                  FAKE_SUBSYSTEM_NAME,
                                                  FAKE_CONFIG_TYPE));
    CPPUNIT_ASSERT(NamedConfigUtils::
                   getSubsystemConfigIdAncestry(*subsystemNameSpec, idList,
                                                NAMEDCFG_TABLE_DEFINITION));
    CPPUNIT_ASSERT(idList.size() == 1);
    CPPUNIT_ASSERT(idList[0] == 1);

    idList.clear();
    CPPUNIT_ASSERT(idList.size() == 0);
    subsystemNameSpec.reset(new SubsystemNameSpec(NAMED_CONFIG_B,
                                                  DCS_OP_MODE,
                                                  NDOS_DETECTOR_NAME,
                                                  FAKE_SUBSYSTEM_NAME,
                                                  FAKE_CONFIG_TYPE));
    CPPUNIT_ASSERT(NamedConfigUtils::
                   getSubsystemConfigIdAncestry(*subsystemNameSpec, idList,
                                                NAMEDCFG_TABLE_DEFINITION));
    CPPUNIT_ASSERT(idList.size() == 2);
    CPPUNIT_ASSERT(idList[0] == 2);
    CPPUNIT_ASSERT(idList[1] == 1);

    idList.clear();
    CPPUNIT_ASSERT(idList.size() == 0);
    subsystemNameSpec.reset(new SubsystemNameSpec(NAMED_CONFIG_C,
                                                  DCS_OP_MODE,
                                                  NDOS_DETECTOR_NAME,
                                                  FAKE_SUBSYSTEM_NAME,
                                                  FAKE_CONFIG_TYPE));
    CPPUNIT_ASSERT(NamedConfigUtils::
                   getSubsystemConfigIdAncestry(*subsystemNameSpec, idList,
                                                NAMEDCFG_TABLE_DEFINITION));
    CPPUNIT_ASSERT(idList.size() == 3);
    CPPUNIT_ASSERT(idList[0] == 3);
    CPPUNIT_ASSERT(idList[1] == 2);
    CPPUNIT_ASSERT(idList[2] == 1);

    idList.clear();
    CPPUNIT_ASSERT(idList.size() == 0);
    subsystemNameSpec.reset(new SubsystemNameSpec(NAMED_CONFIG_D,
                                                  DCS_OP_MODE,
                                                  NDOS_DETECTOR_NAME,
                                                  FAKE_SUBSYSTEM_NAME,
                                                  FAKE_CONFIG_TYPE));
    CPPUNIT_ASSERT(NamedConfigUtils::
                   getSubsystemConfigIdAncestry(*subsystemNameSpec, idList,
                                                NAMEDCFG_TABLE_DEFINITION));
    CPPUNIT_ASSERT(idList.size() == 4);
    CPPUNIT_ASSERT(idList[0] == 4);
    CPPUNIT_ASSERT(idList[1] == 3);
    CPPUNIT_ASSERT(idList[2] == 2);
    CPPUNIT_ASSERT(idList[3] == 1);

    // test the lookup of ancestor IDs for a bogus name
    idList.clear();
    CPPUNIT_ASSERT(idList.size() == 0);
    subsystemNameSpec.reset(new SubsystemNameSpec("ZZZ_QQQ_RRR",
                                                  DCS_OP_MODE,
                                                  NDOS_DETECTOR_NAME,
                                                  FAKE_SUBSYSTEM_NAME,
                                                  FAKE_CONFIG_TYPE));
    CPPUNIT_ASSERT(! NamedConfigUtils::
                   getSubsystemConfigIdAncestry(*subsystemNameSpec, idList,
                                                NAMEDCFG_TABLE_DEFINITION));

    // test the lookup of ancestor IDs by ID
    {
        idList.clear();
        CPPUNIT_ASSERT(idList.size() == 0);
        IDSpec idSpec(1, IDSpec::SUBSYSTEM, IDSpec::DAQ);
        CPPUNIT_ASSERT(NamedConfigUtils::
                       getConfigIdAncestry(idSpec, idList,
                                           NAMEDCFG_TABLE_DEFINITION));
        CPPUNIT_ASSERT(idList.size() == 1);
        CPPUNIT_ASSERT(idList[0] == 1);
    }
    {
        idList.clear();
        CPPUNIT_ASSERT(idList.size() == 0);
        IDSpec idSpec(2, IDSpec::SUBSYSTEM, IDSpec::DAQ);
        CPPUNIT_ASSERT(NamedConfigUtils::
                       getConfigIdAncestry(idSpec, idList,
                                           NAMEDCFG_TABLE_DEFINITION));
        CPPUNIT_ASSERT(idList.size() == 2);
        CPPUNIT_ASSERT(idList[0] == 2);
        CPPUNIT_ASSERT(idList[1] == 1);
    }
    {
        idList.clear();
        CPPUNIT_ASSERT(idList.size() == 0);
        IDSpec idSpec(3, IDSpec::SUBSYSTEM, IDSpec::DAQ);
        CPPUNIT_ASSERT(NamedConfigUtils::
                       getConfigIdAncestry(idSpec, idList,
                                           NAMEDCFG_TABLE_DEFINITION));
        CPPUNIT_ASSERT(idList.size() == 3);
        CPPUNIT_ASSERT(idList[0] == 3);
        CPPUNIT_ASSERT(idList[1] == 2);
        CPPUNIT_ASSERT(idList[2] == 1);
    }
    {
        idList.clear();
        CPPUNIT_ASSERT(idList.size() == 0);
        IDSpec idSpec(4, IDSpec::SUBSYSTEM, IDSpec::DAQ);
        CPPUNIT_ASSERT(NamedConfigUtils::
                       getConfigIdAncestry(idSpec, idList,
                                           NAMEDCFG_TABLE_DEFINITION));
        CPPUNIT_ASSERT(idList.size() == 4);
        CPPUNIT_ASSERT(idList[0] == 4);
        CPPUNIT_ASSERT(idList[1] == 3);
        CPPUNIT_ASSERT(idList[2] == 2);
        CPPUNIT_ASSERT(idList[3] == 1);
    }

    // test the lookup of ancestor IDs for a bogus ID
    {
        idList.clear();
        CPPUNIT_ASSERT(idList.size() == 0);
        IDSpec idSpec(1010101, IDSpec::SUBSYSTEM, IDSpec::DAQ);
        CPPUNIT_ASSERT(! NamedConfigUtils::
                       getConfigIdAncestry(idSpec, idList,
                                           NAMEDCFG_TABLE_DEFINITION));
    }
}

void
NamedConfigUnitTest::testNameReuse()
{
    boost::shared_ptr<SubsystemNameSpec> subsystemNameSpec;
    int64_t configId;

    CPPUNIT_ASSERT(_namedCfgTablePtr.get() != 0);
    CPPUNIT_ASSERT(_namedCfgTablePtr->ExistsInDB());

    // test the lookup of IDs that we expect to be there
    subsystemNameSpec.reset(new SubsystemNameSpec(NAMED_CONFIG_E,
                                                  DCS_OP_MODE,
                                                  NDOS_DETECTOR_NAME,
                                                  FAKE_SUBSYSTEM_NAME,
                                                  FAKE_CONFIG_TYPE));
    CPPUNIT_ASSERT(NamedConfigUtils::
                   getSubsystemConfigId(*subsystemNameSpec,
                                        configId,
                                        NAMEDCFG_TABLE_DEFINITION));
    CPPUNIT_ASSERT(configId == 9);
    subsystemNameSpec.reset(new SubsystemNameSpec(NAMED_CONFIG_F,
                                                  DCS_OP_MODE,
                                                  NDOS_DETECTOR_NAME,
                                                  FAKE_SUBSYSTEM_NAME,
                                                  FAKE_CONFIG_TYPE));
    CPPUNIT_ASSERT(NamedConfigUtils::
                   getSubsystemConfigId(*subsystemNameSpec,
                                        configId,
                                        NAMEDCFG_TABLE_DEFINITION));
    CPPUNIT_ASSERT(configId == 10);
    subsystemNameSpec.reset(new SubsystemNameSpec(NAMED_CONFIG_G,
                                                  DCS_OP_MODE,
                                                  NDOS_DETECTOR_NAME,
                                                  FAKE_SUBSYSTEM_NAME,
                                                  FAKE_CONFIG_TYPE));
    CPPUNIT_ASSERT(NamedConfigUtils::
                   getSubsystemConfigId(*subsystemNameSpec,
                                        configId,
                                        NAMEDCFG_TABLE_DEFINITION));
    CPPUNIT_ASSERT(configId == 11);
    subsystemNameSpec.reset(new SubsystemNameSpec(NAMED_CONFIG_H,
                                                  DCS_OP_MODE,
                                                  NDOS_DETECTOR_NAME,
                                                  FAKE_SUBSYSTEM_NAME,
                                                  FAKE_CONFIG_TYPE));
    CPPUNIT_ASSERT(NamedConfigUtils::
                   getSubsystemConfigId(*subsystemNameSpec,
                                        configId,
                                        NAMEDCFG_TABLE_DEFINITION));
    CPPUNIT_ASSERT(configId == 12);

    // test the lookup of subsystem that we don't expect to be there
    subsystemNameSpec.reset(new SubsystemNameSpec(NAMED_CONFIG_E,
                                                  DCS_OP_MODE,
                                                  NDOS_DETECTOR_NAME,
                                                  "blahblahblah",
                                                  FAKE_CONFIG_TYPE));
    CPPUNIT_ASSERT(! NamedConfigUtils::
                   getSubsystemConfigId(*subsystemNameSpec,
                                        configId,
                                        NAMEDCFG_TABLE_DEFINITION));

    // test the lookup of ancestor IDs
    std::vector<int64_t> idList;
    idList.clear();
    CPPUNIT_ASSERT(idList.size() == 0);
    subsystemNameSpec.reset(new SubsystemNameSpec(NAMED_CONFIG_E,
                                                  DCS_OP_MODE,
                                                  NDOS_DETECTOR_NAME,
                                                  FAKE_SUBSYSTEM_NAME,
                                                  FAKE_CONFIG_TYPE));
    CPPUNIT_ASSERT(NamedConfigUtils::
                   getSubsystemConfigIdAncestry(*subsystemNameSpec, idList,
                                                NAMEDCFG_TABLE_DEFINITION));
    CPPUNIT_ASSERT(idList.size() == 4);
    CPPUNIT_ASSERT(idList[0] == 9);
    CPPUNIT_ASSERT(idList[1] == 7);
    CPPUNIT_ASSERT(idList[2] == 6);
    CPPUNIT_ASSERT(idList[3] == 5);

    idList.clear();
    CPPUNIT_ASSERT(idList.size() == 0);
    subsystemNameSpec.reset(new SubsystemNameSpec(NAMED_CONFIG_F,
                                                  DCS_OP_MODE,
                                                  NDOS_DETECTOR_NAME,
                                                  FAKE_SUBSYSTEM_NAME,
                                                  FAKE_CONFIG_TYPE));
    CPPUNIT_ASSERT(NamedConfigUtils::
                   getSubsystemConfigIdAncestry(*subsystemNameSpec, idList,
                                                NAMEDCFG_TABLE_DEFINITION));
    CPPUNIT_ASSERT(idList.size() == 2);
    CPPUNIT_ASSERT(idList[0] == 10);
    CPPUNIT_ASSERT(idList[1] == 5);

    idList.clear();
    CPPUNIT_ASSERT(idList.size() == 0);
    subsystemNameSpec.reset(new SubsystemNameSpec(NAMED_CONFIG_G,
                                                  DCS_OP_MODE,
                                                  NDOS_DETECTOR_NAME,
                                                  FAKE_SUBSYSTEM_NAME,
                                                  FAKE_CONFIG_TYPE));
    CPPUNIT_ASSERT(NamedConfigUtils::
                   getSubsystemConfigIdAncestry(*subsystemNameSpec, idList,
                                                NAMEDCFG_TABLE_DEFINITION));
    CPPUNIT_ASSERT(idList.size() == 3);
    CPPUNIT_ASSERT(idList[0] == 11);
    CPPUNIT_ASSERT(idList[1] == 10);
    CPPUNIT_ASSERT(idList[2] == 5);

    idList.clear();
    CPPUNIT_ASSERT(idList.size() == 0);
    subsystemNameSpec.reset(new SubsystemNameSpec(NAMED_CONFIG_H,
                                                  DCS_OP_MODE,
                                                  NDOS_DETECTOR_NAME,
                                                  FAKE_SUBSYSTEM_NAME,
                                                  FAKE_CONFIG_TYPE));
    CPPUNIT_ASSERT(NamedConfigUtils::
                   getSubsystemConfigIdAncestry(*subsystemNameSpec, idList,
                                                NAMEDCFG_TABLE_DEFINITION));
    CPPUNIT_ASSERT(idList.size() == 1);
    CPPUNIT_ASSERT(idList[0] == 12);

    // test the lookup of ancestor IDs for a bogus subsystem
    idList.clear();
    CPPUNIT_ASSERT(idList.size() == 0);
    subsystemNameSpec.reset(new SubsystemNameSpec(NAMED_CONFIG_E,
                                                  DCS_OP_MODE,
                                                  NDOS_DETECTOR_NAME,
                                                  "blahblahblah",
                                                  FAKE_CONFIG_TYPE));
    CPPUNIT_ASSERT(! NamedConfigUtils::
                   getSubsystemConfigIdAncestry(*subsystemNameSpec, idList,
                                                NAMEDCFG_TABLE_DEFINITION));
}

void
NamedConfigUnitTest::testInvalidValues()
{
    CPPUNIT_ASSERT(_namedCfgTablePtr.get() != 0);
    CPPUNIT_ASSERT(_namedCfgTablePtr->ExistsInDB());

    // test invalid DCM operating mode
    _namedCfgTablePtr->Clear();
    boost::shared_ptr<Row> rowPtr(_namedCfgTablePtr->NewRow());
    CPPUNIT_ASSERT(rowPtr->Set(NamedConfigUtils::NAME_COLUMN_NAME, "AAA"));
    CPPUNIT_ASSERT(rowPtr->Set(NamedConfigUtils::DCMOPMODE_COLUMN_NAME,
                               "BOGUS_DCM_OP_MODE"));
    CPPUNIT_ASSERT(rowPtr->Set(NamedConfigUtils::DETECTORID_COLUMN_NAME,
                               NDOS_DETECTOR_NAME));
    CPPUNIT_ASSERT(rowPtr->Set(NamedConfigUtils::SUBSYSTEMNAME_COLUMN_NAME,
                               FAKE_SUBSYSTEM_NAME));
    CPPUNIT_ASSERT(rowPtr->Set(NamedConfigUtils::CONFIGTYPE_COLUMN_NAME,
                               FAKE_CONFIG_TYPE));
    _namedCfgTablePtr->AddRow(rowPtr.get());
    CPPUNIT_ASSERT(! _namedCfgTablePtr->WriteToDB());

    // test invalid subsystem name
    _namedCfgTablePtr->Clear();
    rowPtr.reset(_namedCfgTablePtr->NewRow());
    CPPUNIT_ASSERT(rowPtr->Set(NamedConfigUtils::NAME_COLUMN_NAME, "BBB"));
    CPPUNIT_ASSERT(rowPtr->Set(NamedConfigUtils::DCMOPMODE_COLUMN_NAME,
                               DCS_OP_MODE));
    CPPUNIT_ASSERT(rowPtr->Set(NamedConfigUtils::DETECTORID_COLUMN_NAME,
                               NDOS_DETECTOR_NAME));
    CPPUNIT_ASSERT(rowPtr->Set(NamedConfigUtils::SUBSYSTEMNAME_COLUMN_NAME,
                               "BogusSubsystemName"));
    CPPUNIT_ASSERT(rowPtr->Set(NamedConfigUtils::CONFIGTYPE_COLUMN_NAME,
                               FAKE_CONFIG_TYPE));
    _namedCfgTablePtr->AddRow(rowPtr.get());
    CPPUNIT_ASSERT(! _namedCfgTablePtr->WriteToDB());

    // test invalid configuration type
    _namedCfgTablePtr->Clear();
    rowPtr.reset(_namedCfgTablePtr->NewRow());
    CPPUNIT_ASSERT(rowPtr->Set(NamedConfigUtils::NAME_COLUMN_NAME, "CCC"));
    CPPUNIT_ASSERT(rowPtr->Set(NamedConfigUtils::DCMOPMODE_COLUMN_NAME,
                               DCS_OP_MODE));
    CPPUNIT_ASSERT(rowPtr->Set(NamedConfigUtils::DETECTORID_COLUMN_NAME,
                               NDOS_DETECTOR_NAME));
    CPPUNIT_ASSERT(rowPtr->Set(NamedConfigUtils::SUBSYSTEMNAME_COLUMN_NAME,
                               FAKE_SUBSYSTEM_NAME));
    CPPUNIT_ASSERT(rowPtr->Set(NamedConfigUtils::CONFIGTYPE_COLUMN_NAME,
                               "Bogus_config_type"));
    _namedCfgTablePtr->AddRow(rowPtr.get());
    CPPUNIT_ASSERT(! _namedCfgTablePtr->WriteToDB());
}

void
NamedConfigUnitTest::testFetchOfAvailableSubsystemNames()
{
    NamedConfigUtils::SubsystemNameSpecMap subsystemNameSpecMap;

    CPPUNIT_ASSERT(_namedCfgTablePtr.get() != 0);
    CPPUNIT_ASSERT(_namedCfgTablePtr->ExistsInDB());

    CPPUNIT_ASSERT(subsystemNameSpecMap.size() == 0);
    CPPUNIT_ASSERT(NamedConfigUtils::
                   getAvailableSubsystemConfigs(IDSpec::DAQ,
                                                NDOS_DETECTOR_NAME,
                                                FAKE_SUBSYSTEM_NAME,
                                                FAKE_CONFIG_TYPE,
                                                subsystemNameSpecMap, 10,
                                                NAMEDCFG_TABLE_DEFINITION));
    CPPUNIT_ASSERT(subsystemNameSpecMap.size() == 8);
}

void
NamedConfigUnitTest::populateDBTables()
{
    _namedCfgTablePtr.reset(new Table(NAMEDCFG_TABLE_DEFINITION.c_str()));
    CPPUNIT_ASSERT(_namedCfgTablePtr->GetConnection());
    CPPUNIT_ASSERT(_namedCfgTablePtr->ExistsInDB());
    _namedCfgTablePtr->Clear();
    CPPUNIT_ASSERT(_namedCfgTablePtr->LoadFromDB());

    // populate initial named configurations, if needed
    if (_namedCfgTablePtr->NRow() == 0) {

        // id=1, name=NamedConfigA, parentId=null
        // id=2, name=NamedConfigB, parentId=1
        // id=3, name=NamedConfigC, parentId=2
        // id=4, name=NamedConfigD, parentId=3
        CPPUNIT_ASSERT(_addTableRow(NAMED_CONFIG_A, -1));
        CPPUNIT_ASSERT(_addTableRow(NAMED_CONFIG_B,  1));
        CPPUNIT_ASSERT(_addTableRow(NAMED_CONFIG_C,  2));
        CPPUNIT_ASSERT(_addTableRow(NAMED_CONFIG_D,  3));

        // id=5, name=NamedConfigE, parentId=null
        // id=6, name=NamedConfigF, parentId=5
        // id=7, name=NamedConfigG, parentId=6
        // id=8, name=NamedConfigH, parentId=7
        // sleep()
        // id= 9, name=NamedConfigE, parentId=7
        // id=10, name=NamedConfigF, parentId=5
        // id=11, name=NamedConfigG, parentId=10
        // id=12, name=NamedConfigH, parentId=null
        CPPUNIT_ASSERT(_addTableRow(NAMED_CONFIG_E, -1));
        CPPUNIT_ASSERT(_addTableRow(NAMED_CONFIG_F,  5));
        CPPUNIT_ASSERT(_addTableRow(NAMED_CONFIG_G,  6));
        CPPUNIT_ASSERT(_addTableRow(NAMED_CONFIG_H,  7));

        usleep(1250000);

        CPPUNIT_ASSERT(_addTableRow(NAMED_CONFIG_E,  7));
        CPPUNIT_ASSERT(_addTableRow(NAMED_CONFIG_F,  5));
        CPPUNIT_ASSERT(_addTableRow(NAMED_CONFIG_G, 10));
        CPPUNIT_ASSERT(_addTableRow(NAMED_CONFIG_H, -1));
    }

    // clear the Table objects as a nicety
    _namedCfgTablePtr->Clear();
}

void
NamedConfigUnitTest::cleanupDBTables()
{
    _namedCfgTablePtr.reset();
}

bool
NamedConfigUnitTest::_addTableRow(const std::string& configName,
                                  const int16_t& parentId)
{
    boost::shared_ptr<Row> rowPtr(_namedCfgTablePtr->NewRow());
    if (! rowPtr->Set(NamedConfigUtils::NAME_COLUMN_NAME, configName)) {
        return false;
    }
    if (! rowPtr->Set(NamedConfigUtils::DCMOPMODE_COLUMN_NAME, DCS_OP_MODE)) {
        return false;
    }
    if (! rowPtr->Set(NamedConfigUtils::DETECTORID_COLUMN_NAME,
                      NDOS_DETECTOR_NAME)) {
        return false;
    }
    if (! rowPtr->Set(NamedConfigUtils::ISOBSOLETE_COLUMN_NAME, false)) {
        return false;
    }
    if (! rowPtr->Set(NamedConfigUtils::ISCHANGING_COLUMN_NAME, false)) {
        return false;
    }
    if (! rowPtr->Set(NamedConfigUtils::ISCOMPLETE_COLUMN_NAME, true)) {
        return false;
    }
    if (! rowPtr->Set(NamedConfigUtils::SUBSYSTEMNAME_COLUMN_NAME,
                      FAKE_SUBSYSTEM_NAME)) {
        return false;
    }
    if (! rowPtr->Set(NamedConfigUtils::CONFIGTYPE_COLUMN_NAME,
                      FAKE_CONFIG_TYPE)) {
        return false;
    }
    if (parentId > 0) {
        if (! rowPtr->Set(NamedConfigUtils::PARENTID_COLUMN_NAME, parentId)) {
            return false;
        }
    }
    _namedCfgTablePtr->AddRow(rowPtr.get());
    if (! _namedCfgTablePtr->WriteToDB()) {
        return false;
    }
    return true;
}

CPPUNIT_TEST_SUITE_REGISTRATION(NamedConfigUnitTest);
