import psycopg2
import csv
import datetime
import time
import argparse
import os

# This function does the final step of writing the output file for the different temperature 
# settings of each apd used in the detector
# apddict is a dictionary that maps APD serial number to the feb it is connected to
# and the module it is mounted on. There is a dictionary for each temperature setting
# filesuffix is the identifier for low or high temperature running
def settings(apddict, filesuffix) :

    fileName    = "hv_"+filesuffix+".csv"
    if(useSafeMode) :
        fileName    = "hv_Safe"+filesuffix+".csv"
    settingFile = open(fileName, "w")
    fileName    = "apd_installation.txt"
    locFile     = open(fileName, "w")

    default     = '1,default,-1,cold,1000.0,0,1,2,3,4\n'
    scaleFactor = 1.

    # according to Leon M, warm voltages should be 0.976 the value recorded because we 
    # operate at 15C, not the warm test value of 25C
    if(filesuffix == "HighT") :
        default     = '1,default,-1,warm,2000.0,0,1,2,3,4\n'
        scaleFactor = 0.976

    # put in the default line
    settingFile.write(default)

    # make a dictionary of apd to location on the detector
    # additional information is diblock, dcm, febport, then voltage
    keyval = 'dcm-2-999-999-999'
    locationToAPD = dict( keyval = ( 'stupidapd', (-999, -999, 999., -999, now), ('stupidfeb', -999, now, -999., -999), -999, (-999, -999, -999), -999.))

    # loop over the apds
    for apd in apddict.iterkeys():
        data = apddict.get(apd)
        apdinfo = data[0]
        febdata = data[1]
        mod     = data[2]

        if(apdinfo[0] == -999) : continue

        # grab the location information for the apd from block/plane/module to diblock/dcm/feb port
        SQL = "select CAST(btrim(fcom.location1, 'DIBLK:') AS INT) diblk, CAST(btrim(fcom.position1, 'DCM:') AS INT) dcm , CAST(btrim(fcom.port1,'PRT:') AS INT) feb, mod.block, mod.layer, mod.position_in_layer from fnalprod_public.feb_comm_cables fcom, module mod, barcoded_item bi where mod.barcoded_item_id={} and (mod.block=CAST(btrim(fcom.location2,'BLK:') AS INT) and mod.layer=CAST(btrim(fcom.position2,'PLN:') AS INT) and mod.position_in_layer=CAST(btrim(fcom.port2,'POS:') AS INT)) and bi.barcoded_item_id={} ORDER BY diblk ASC, dcm ASC, feb ASC".format(mod,apdinfo[3])
        curAR.execute(SQL)    
        loc = curAR.fetchone()
        location = (loc[0], loc[1], loc[2])
        keyval = "dcm-2-{}-{}-{}".format(loc[0], loc[1], loc[2])
        
        # have all the pieces we need now
        # - apd voltage and location in apdinfo
        # - feb to apd connection in febtoapd
        # - feb voltage slope and intercept, remember abscisa is 4095 steps
        # find nearest integer to (apdinfo[2] - febdata[3])/febdata[4]
        # 1,<dcmName>,<febId>,<coolingState>,<voltage>,2,0,0,0,0
        intercept = febdata[3] 
        slope     = febdata[4]
        if slope == 0. : 
            slope = 0.02586

        voltage = (scaleFactor*apdinfo[2] - intercept)/slope
        if(useSafeMode) :
            voltage = (apdinfo[2] - 30 - intercept)/slope

        #voltage = (350 - intercept)/slope

        # test the voltage setting 
        if apdinfo[1] > 0 and (voltage < 2500 or voltage > 4094) :
            #print "Setting for loc: {} with APD: {} at temp {} is out of range {}".format(keyval, apd, apdinfo[1], voltage)
            if   voltage > 4094: voltage = 4094
            elif voltage < 2500: voltage = 2500
        elif apdinfo[1] < 0 and (voltage < 1000 or voltage > 4000) :
            #print "Setting for loc: {} with APD: {} at temp {} is out of range {}".format(keyval, apd, apdinfo[1], voltage)
            if   voltage > 4000: voltage = 4000
            elif voltage < 1000: voltage = 1000

        # check to see if the electronics location has already been used.  If not, then 
        # insert it into the dictionary.  If it has been used, then make sure the most
        # recent entry is in the dictionary
        if keyval not in locationToAPD :
            locationToAPD[keyval] = (apd, apdinfo, febdata, mod, location, voltage)
        else :
            data = locationToAPD.get(keyval)
            prevAPD = data[0]
            prevAPDInfo = data[1]
            prevFEB = data[2]
            debugFile.write("dcm-2-{}-{} port {} assigned twice: \n".format(location[0], location[1], location[2]))
            debugFile.write("\t Currently assigned APD: {} {} {}\n".format(prevAPD, prevAPDInfo, prevFEB))
            debugFile.write("\t Alternate assigned APD: {} {} {}\n".format(apd, apdinfo, febdata))
            if apdinfo[4] > prevAPDInfo[4] : 
                locationToAPD[keyval] = (apd, apdinfo, febdata, mod, location, voltage)
                debugFile.write("\t\t Using APD: {} {} {}\n".format(apd, apdinfo, febdata))            
            else : 
                debugFile.write("\t\t Using APD: {} {} {}\n".format(prevAPD, prevAPDInfo, prevFEB))
            
    # loop over all electronics locations in the dictionary and 
    # and write them out to the file.  Also write out the location used 
    # and the relevant information about the APD installed there
    for keyval in locationToAPD :
        
        data = locationToAPD.get(keyval)
        apd      = data[0]
        apdinfo  = data[1]
        febdata  = data[2]
        mod      = data[3]
        location = data[4]
        voltage  = data[5]
        
        if location[0] == -999 : continue

        diblock     = '{}'.format(location[0])
        if location[0] < 10 : diblock     = '0{}'.format(location[0])
        dcmposition = '{}'.format(location[1])
        if location[1] < 10 : dcmposition = '0{}'.format(location[1])
        temp = "warm"
        if apdinfo[1] < 0 : temp = "cold"

        dcmloc      =  'dcm-2-{}-{}'.format(diblock, dcmposition)
        
        if dcm in dcmloc :
            settingInfo =  '1,{},{},{},{},2,0,0,0,0\n'.format(dcmloc, location[2], temp, round(voltage) )
            locInfo     =  '{},{},{},{}\n'.format(dcmloc, location[2], apd, apdinfo[4])
            settingFile.write(settingInfo)
            locFile.write(locInfo)

    # close the files.
    settingFile.close()
    locFile.close()


#read in the password for the db
pswdloc  = os.environ['NOVADBPWDFILE']
pswdfile = open(pswdloc, "r")
pswd     = pswdfile.read()

SQL = "dbname=nova_hardware host=ifdbprod.fnal.gov user=nova_reader password={} port=5432".format(pswd)
 
# set up the argument parser, --safe tells the program to make the safe voltage files
parser      = argparse.ArgumentParser()
parser.add_argument("--safe", help="set voltages 30V below nominal", action="store_true")
parser.add_argument("--dcm",  help="specify which dcm to set the HV for, e.g. dcm-2-01", default="dcm-2")
args        = parser.parse_args()
useSafeMode = args.safe
dcm         = args.dcm

if useSafeMode : print "HV will be set in safe mode for {}".format(dcm)
else           : print "HV will be set for {}".format(dcm)

# try connecting to the database
# use the Fermilab (FL) database for testing information and the Ash River (AR)
# for the most up-to-date installation information
try:
    connFL = psycopg2.connect(SQL)
except:
    print "Unable to connect to the FL database"
    exit(0)
try:
    SQL = "dbname=novaqa host=198.124.69.52 user=nova_reader password={} port=5432".format(pswd)
    connAR = psycopg2.connect(SQL)
except:
    print "Unable to connect to the AR database"
    exit(0)

curFL = connFL.cursor()
curAR = connAR.cursor()
    
SQL = "set search_path to ashriverprod_factory;"
curFL.execute(SQL)
SQL = "set search_path to nova;"
curAR.execute(SQL)

# file to store debug/error information
debugName   = "debuginfo.txt"
debugFile   = open(debugName, "w")

# have to be sure we are only looking at APDs and FEBs from the far detector.  Use 
# the Hardware_position database to make dictionaries for the FEBs and APDs
dummy = 'dummyhw'
nonFDHW = dict(dummy = (dummy, dummy))
SQL = "select hw_id, hw_type, detector from public.Hardware_position where hw_type='FEB' OR hw_type='APD'"
curFL.execute(SQL)
rows = curFL.fetchall()
for row in rows:
    hw_type = str(row[1])
    if hw_type == 'APD' :
      chars = len(row[0])
      sn = row[0][chars-7:]
    elif hw_type == 'FEB' :
      feb = str(row[0])
      febsn1 = feb.replace('_', '.')
      sn     = febsn1.replace('-', '.')

    nonFDHW[sn] = hw_type, str(row[2])        

# get the feb slope and intercept information
# using a csv file for now, change to the database when available.
dummy = 'dummyfeb'
febinfo = dict(dummy = (-999., -999.))

SQL = "select febtest.febsn, febtest.hvfitintercept, febtest.hvfitslope from public.feb_test_summary febtest"
curFL.execute(SQL)
rows = curFL.fetchall()

for row in rows:
    feb = str(row[0])
    # change the format of the SN from having "_" or "-" to "." instead to be consistent with the db
    febsn1 = feb.replace('_', '.')
    febsn  = febsn1.replace('-', '.')
    # just put the last entry for the feb into the dictionary
    if febsn not in nonFDHW :
      febinfo[febsn] =  float(row[1]), float(row[2]) 

# get the relevant information from the APD table
# create a separate dictionary for each temperature range, > 0 and < 0
# if the current APD serial number exists in the dictionary, check the
# batch number and take the highest and overwrite the current entry if necessary
SQL = "select astat.apd_id, astat.batch_no, astat.target_degc, astat.Vr from public.apd_stats astat"
curFL.execute(SQL)
rows  = curFL.fetchall()
dummy = 'dummyapd1'
apdStatsLowT  = dict(dummy = (-999, -999, 999.) )
apdStatsHighT = dict(dummy = (-999,  999, 999.) )
for row in rows:
    # grab the last 7 characters of the apd SN only
    # first determine how many characters are in the string
    # then only keep the last 7
    chars = len(row[0])
    apd   = row[0][chars-7:]
    batch = row[1]
    temp  = row[2]
    vr    = row[3]
    stats = (batch, temp, vr)

    # decide which temperature we are looking at 
    # and also use the most recent test
    if apd not in nonFDHW :
        if temp > 0:
            if apd not in apdStatsHighT: 
                apdStatsHighT[apd] = stats
            else :
                prevStats = apdStatsHighT.get(apd)
                prevBatch = prevStats[0]
                if batch > prevBatch: 
                    apdStatsHighT[apd] = stats
        else:
            if apd not in apdStatsLowT: 
                apdStatsLowT[apd] = stats
            else :
                prevStats = apdStatsLowT.get(apd)
                prevBatch = prevStats[0]
                if batch > prevBatch: 
                    apdStatsLowT[apd] = stats

# now have all the APDs, go through and query for the FEB each is attached to
# first do a query to get all the installed APD serial numbers
SQL = "select bi.barcoded_item_id, bi.barcode, bi.date_added from barcoded_item bi where bi.glassfish_discriminator='apd' and date_removed is NULL"
curAR.execute(SQL)
rows = curAR.fetchall()
now = time.localtime()
# dictionary has (apdinfo, barcode, added date), (feb SN, barcode, added date, intercept, slope), module barcode
apdToFEBLowT  = dict(dummy = ((-999, -999, 999., -999, now), ('stupidfeb', -999, now, -999., -999), -999))
apdToFEBHighT = dict(dummy = ((-999, -999, 999., -999, now), ('stupidfeb', -999, now, -999., -999), -999))

# loop over the rows returned by the query
for row in rows:
    chars    = len(row[1])
    apd      = row[1][chars-7:]
    bcid     = row[0]
    apdadded = row[2] 

    # now grab the module connection information for the apd
    # if the apd is connected to 0 modules just skip it and 
    # go to the next one in the list
    # the query ignores any entry where the nova connection entry removed date is not null
    # if the apd is connected to multiple modules, then check
    # if they are duplicates. if they are not, skip this apd
    SQL = 'select mac.module, mac.connection_id, nc.date_removed from module_apd_connection mac, nova_connection nc where mac.apd={} and nc.connection_id=mac.connection_id and nc.date_removed is NULL;'.format(bcid)
    curAR.execute(SQL)
    numrows = curAR.rowcount

    if numrows < 1 :
        continue
 
    apdcon  = curAR.fetchone()
    mod     = apdcon[0]
    modcon  = apdcon[1]
    
    if numrows > 1 :
        debugFile.write('APD {} {} appears to be connected to multiple modules, it will be skipped\n'.format(apd, bcid))
        macrows    = curAR.fetchall()
        debugFile.write('\t The following connections are still valid:\n')
        for macrow in macrows :
            debugFile.write('\t\t Module ID {}, Connection ID {}\n'.format(macrow[0], macrow[1]))
        continue
    
    # grab the feb connected to the module
    SQL = 'select bi.barcode, mfc.feb, bi.barcoded_item_id, bi.date_added, mfc.connection_id, mfc.module, nc.date_removed from barcoded_item bi, module_feb_connection mfc, nova_connection nc where mfc.module={} and bi.barcoded_item_id=mfc.feb and nc.connection_id=mfc.connection_id and nc.date_removed is NULL;'.format(mod)
    curAR.execute(SQL)
    numrows = curAR.rowcount

    # if the feb is not connected to a module, just go to the next apd
    if numrows < 1 :
        debugFile.write('Module {} is not connected to a feb for APD {} \n'.format(mod,apd))
        continue

    # if a module has multiple febs, assume the last one added is still there
    febcon  = curAR.fetchone()
    febsn   = str(febcon[0])
    modcon  = febcon[4]
    added   = febcon[3] 

    if numrows > 1 :
        debugFile.write('Module {} appears to be connected to multiple febs\n'.format(mod))
        febrows = curAR.fetchall()
        debugFile.write('\t The following connections are still valid:\n')
        prevDate = added
        for febrow in febrows :
            debugFile.write('\t\t Module ID {}, Connection ID {} added on {}\n'.format(febrow[5], febrow[4], febrow[3]))
            if febrow[3] > prevDate :
                modcon   = febrow[4]
                removed  = febrow[6]
                prevDate = febrow[3]
                febsn    = str(febrow[0])
                added    = febrow[3]
        debugFile.write('\t using connection {} for added date of {}, removed date {}.\n'.format(modcon, added, removed))
    
    # look for the feb in the febinfo dictionary
    if febsn in febinfo :
       febmb   = febinfo.get(febsn)
       febdata = (febsn, febcon[1], added, febmb[0], febmb[1])
    else :
       debugFile.write('FEB {} installed but slope and intercept information not in database, use average values\n'.format(febsn))
       febdata = (febsn, febcon[1], added, 293.53, 0.02586)
       #exit()

    # now connect the FEB SN to the apd info
    # find the apd in the apdInfo dictionary
    if apd in apdStatsLowT :
        apdstats = apdStatsLowT.get(apd)
        apdinfo = (apdstats[0], apdstats[1], apdstats[2], bcid, apdadded) 
        apdToFEBLowT[apd]  = (apdinfo, febdata, mod)
    if apd in apdStatsHighT :
        apdstats = apdStatsHighT.get(apd)
        apdinfo = (apdstats[0], apdstats[1], apdstats[2], bcid, apdadded) 
        apdToFEBHighT[apd] = (apdinfo, febdata, mod)

settings(apdToFEBHighT, "HighT")
settings(apdToFEBLowT,  "LowT")

debugFile.close()

if(useSafeMode) :
    print "Finished making safe APD HV settings files"
else :
    print "Finished making nominal APD HV settings files"
