#include "NovaDAQMonitorClient/NdmcMetric.h"
#include "NovaDAQMonitorClient/NdmcClient.h"
#include "NovaDAQMonitorClient/NdmcClientService.h"

#include "NdmcClientTest.h"
using namespace NDMC_NAMESPACE;

/**
 * Register the NdmcClientTest suite in the registry of suites in a "folder"
 * of name "NovaDAQMonitorClient".  Store the folder in the default registry path.
 * $Id: NdmcClientTest.cpp,v 1.13 2014/02/03 02:21:31 kasahara Exp $
 */
CPPUNIT_TEST_SUITE_NAMED_REGISTRATION( NdmcClientTest, "NovaDAQMonitorClient" );
CPPUNIT_REGISTRY_ADD_TO_DEFAULT("NovaDAQMonitorClient");

/**
 * A public method to setUp the environment for the test.  This will be invoked before
 * every test in the fixture to get a fresh start.
 */
void NdmcClientTest::setUp() {

  _configFile = getConfigFullFilePathName("unittest/gmond.conf");
  NdmcClientService::setCONFIGFILE(_configFile);

}
/**
 * A public method to tearDown the environment for the test.  This will be invoked after
 * every test in the fixture to clean up allocated memory.
 */
void NdmcClientTest::tearDown() {
  std::cout.flush();
}

/**
 * A test method to test that NdmcClient ctor works correctly.
 */
void NdmcClientTest::testCtor() {

  int32_t partition = 1;
  std::string appName = "dcmapp";
  std::string groupName = "dcmapplication";
  uint32_t sendTimeInterval = 10;
  NdmcClient client(partition,appName,groupName,sendTimeInterval);

  CPPUNIT_ASSERT(client.getPartition() == partition);
  CPPUNIT_ASSERT(client.getApplicationName() == appName);
  CPPUNIT_ASSERT(client.getGroupName() == "dcmapplication P1");
  CPPUNIT_ASSERT(client.getSendTimeInterval() == sendTimeInterval);
  CPPUNIT_ASSERT(client.getNumMetric() == 0);

}

/**
 * A test method to test that NdmcClient setSendTimeInterval works correctly.
 */
void NdmcClientTest::testSetSendTimeInterval() {

  int32_t partition = 0;
  std::string appName = "";
  std::string groupName = "dcmapp";
  uint32_t sendTimeInterval = 10;
  NdmcClient client(partition,appName,groupName,sendTimeInterval);
  CPPUNIT_ASSERT(client.getSendTimeInterval() == sendTimeInterval);
  sendTimeInterval = 5;
  client.setSendTimeInterval(sendTimeInterval);
  CPPUNIT_ASSERT(client.getSendTimeInterval() == sendTimeInterval);

}

/**
 * A test method to test that NdmcClient setPartition works correctly.
 */
void NdmcClientTest::testSetPartition() {

  int32_t partition = 1;
  NdmcClient client(partition);
  CPPUNIT_ASSERT(client.getPartition() == partition);
  partition = 3;
  client.setPartition(partition);
  CPPUNIT_ASSERT(client.getPartition() == partition);

}

/**
 * A test method to test that NdmcClient setApplicationName works correctly.
 */
void NdmcClientTest::testSetApplicationName() {

  int32_t partition = 0;
  std::string appName = "dcmapp";
  NdmcClient client(partition, appName);
  CPPUNIT_ASSERT(client.getApplicationName() == appName);

  appName = "";
  client.setApplicationName(appName);
  CPPUNIT_ASSERT(client.getApplicationName() == appName);

}

/**
 * A test method to test that NdmcClient setGroupName works correctly.
 */
void NdmcClientTest::testSetGroupName() {

  int32_t partition = 0;
  std::string appName = "dcmapp";
  NdmcClient client(partition,appName);
  CPPUNIT_ASSERT(client.getGroupName() == "dcmapp P0");

  partition = -1;
  client.setPartition(partition);
  CPPUNIT_ASSERT(client.getGroupName() == "dcmapp");

  appName = "";
  client.setApplicationName(appName);
  CPPUNIT_ASSERT(client.getGroupName() == "");

  partition = 2;
  client.setPartition(partition);
  CPPUNIT_ASSERT(client.getGroupName() == "P2");

  std::string groupName = "GT";
  client.setGroupName(groupName);
  CPPUNIT_ASSERT(client.getGroupName() == "GT P2");

  groupName = "";
  client.setGroupName(groupName);
  CPPUNIT_ASSERT(client.getGroupName() == "P2");

}

/**
 * A test method to test that NdmcClient will adopt a metric correctly.
 */
void NdmcClientTest::testAdoptMetric() {

  int32_t partition = 0;
  std::string appName = "dcmapp";
  std::string groupName = "";
  uint32_t sendTimeInterval = 10;
  NdmcClient client(partition,appName,groupName,sendTimeInterval);

  CPPUNIT_ASSERT(client.getNumMetric()==0);

  std::string name1 = "If you thought science was certain";
  CPPUNIT_ASSERT(client._adoptMetric(new NdmcMetric<uint32_t>(name1))==0);
  CPPUNIT_ASSERT(client.getNumMetric() == 1);

  std::string name2 = "well that's just";
  CPPUNIT_ASSERT(client._adoptMetric(new NdmcMetric<double>(name2))==1);
  CPPUNIT_ASSERT(client.getNumMetric() == 2);

  std::string name3 = "an error on your part";
  CPPUNIT_ASSERT(client._adoptMetric(new NdmcMetric<int32_t>(name3))==2);
  CPPUNIT_ASSERT(client.getNumMetric() == 3);

  std::string name4 = "an error on your part";
  CPPUNIT_ASSERT(client._adoptMetric(new NdmcMetric<int32_t>(name4))==-1);
  CPPUNIT_ASSERT(client.getNumMetric() == 3);

}

/**
 * A test method to test that NdmcClient can retrieve a metric correctly by metricId
 */
void NdmcClientTest::testGetMetric() {

  int32_t partition = 0;
  std::string appName = "dcmapp";
  std::string groupName = "";
  uint32_t sendTimeInterval = 10;
  NdmcClient client(partition,appName,groupName,sendTimeInterval);

  CPPUNIT_ASSERT(client.getNumMetric()==0);

  std::string name1 = "I,";
  CPPUNIT_ASSERT(client._adoptMetric(new NdmcMetric<uint32_t>(name1))==0);
  CPPUNIT_ASSERT(client.getNumMetric() == 1);

  std::string name2 = "a universe of atoms,";
  CPPUNIT_ASSERT(client._adoptMetric(new NdmcMetric<double>(name2))==1);
  CPPUNIT_ASSERT(client.getNumMetric() == 2);

  std::string name3 = "an atom in the universe";
  CPPUNIT_ASSERT(client._adoptMetric(new NdmcMetric<int32_t>(name3))==2);
  CPPUNIT_ASSERT(client.getNumMetric() == 3);

  NdmcMetricABC* metric = NULL;
  metric = client._getMetric(2);
  CPPUNIT_ASSERT(metric);
  CPPUNIT_ASSERT(metric->getName() == name3);
  CPPUNIT_ASSERT(client.getNumMetric() == 3);
;
  metric = client._getMetric(0);
  CPPUNIT_ASSERT(metric);
  CPPUNIT_ASSERT(metric->getName() == name1);
  CPPUNIT_ASSERT(client.getNumMetric() == 3);

  metric = client._getMetric(1);
  CPPUNIT_ASSERT(metric);
  CPPUNIT_ASSERT(metric->getName() == name2);
  CPPUNIT_ASSERT(client.getNumMetric() == 3);

  metric = client._getMetric(3);
  CPPUNIT_ASSERT(!metric);
  CPPUNIT_ASSERT(client.getNumMetric() == 3);
  
}

/**
 * A test method to test that NdmcClient will create a metric correctly.
 */
void NdmcClientTest::testCreateMetric() {

  int32_t partition = 0;
  std::string appName = "dcmapp";
  std::string groupName = "";
  uint32_t sendTimeInterval = 10;
  NdmcClient client(partition,appName,groupName,sendTimeInterval);

  CPPUNIT_ASSERT(client.getNumMetric()==0);

  std::string name1 = "If you thought science was certain";
  CPPUNIT_ASSERT(client.createMetric<uint32_t>(name1,"your",VALUE)==0);
  CPPUNIT_ASSERT(client.getNumMetric() == 1);

  std::string name2 = "well that's just";
  CPPUNIT_ASSERT(client.createMetric<double>(name2,"tax dollars",COUNTER)==1);
  CPPUNIT_ASSERT(client.getNumMetric() == 2);

  std::string name3 = "an error on your part";
  CPPUNIT_ASSERT(client.createMetric<int32_t>(name3,"at work",RATE)==2);
  CPPUNIT_ASSERT(client.getNumMetric() == 3);

  std::string name4 = "an error on your part";
  CPPUNIT_ASSERT(client.createMetric<int32_t>(name4,"for you",AVERAGE)==-1);
  CPPUNIT_ASSERT(client.getNumMetric() == 3);

}

/**
 * A test method to test that NdmcClient will create a metric correctly.
 */
void NdmcClientTest::testGetLastValueSent() {

  int32_t partition = 0;
  std::string appName = "dcmapp";
  std::string groupName = "";
  uint32_t sendTimeInterval = 10;
  NdmcClient client(partition,appName,groupName,sendTimeInterval);

  CPPUNIT_ASSERT(client.getNumMetric()==0);

  std::string name1 = "If you thought science was certain";
  CPPUNIT_ASSERT(client.createMetric<uint32_t>(name1,"your",VALUE)==0);
  CPPUNIT_ASSERT(client.getNumMetric() == 1);
  NdmcMetric<uint32_t>* metric = NULL;
  metric = dynamic_cast<NdmcMetric<uint32_t>*>(client._getMetric(0));
  CPPUNIT_ASSERT(metric);
  CPPUNIT_ASSERT(metric->getName() == name1);
  uint32_t lastValueSent=0;
  metric->_lastValueSent = 10;
  metric->_valueHasBeenSent = 1;
  CPPUNIT_ASSERT(client.getLastValueSent(0,lastValueSent)==true);
  CPPUNIT_ASSERT(lastValueSent==10);

  std::string name2 = "well that's just";
  CPPUNIT_ASSERT(client.createMetric<double>(name2,"tax dollars",COUNTER)==1);
  CPPUNIT_ASSERT(client.getNumMetric() == 2);
  NdmcMetric<double>* metricd = dynamic_cast<NdmcMetric<double>*>(client._getMetric(1));
  CPPUNIT_ASSERT(metricd);
  CPPUNIT_ASSERT(metricd->getName() == name2);
  metricd->_lastValueSent = 20.5;
  metricd->_valueHasBeenSent = 1;
  double lastValueSentDouble;
  CPPUNIT_ASSERT(client.getLastValueSent(1,lastValueSentDouble)==true);
  CPPUNIT_ASSERT(lastValueSentDouble==20.5);

}

/**
 * A test method to test that NdmcClient::destroyMetric will destroy a metric correctly.
 */
void NdmcClientTest::testDestroyMetric() {

  int32_t partition = 0;
  std::string appName = "dcmapp";
  std::string groupName = "";
  uint32_t sendTimeInterval = 10;
  NdmcClient client(partition,appName,groupName,sendTimeInterval);

  CPPUNIT_ASSERT(client.getNumMetric()==0);

  std::string name1 = "If you thought science was certain";
  CPPUNIT_ASSERT(client.createMetric<uint32_t>(name1,"your",VALUE)==0);
  CPPUNIT_ASSERT(client.getNumMetric() == 1);
  client.destroyMetric(0);
  CPPUNIT_ASSERT(client.getNumMetric() == 0);

  std::string name2 = "well that's just";
  CPPUNIT_ASSERT(client.createMetric<double>(name2,"tax dollars",COUNTER)==0);
  CPPUNIT_ASSERT(client.getNumMetric() == 1);

  std::string name3 = "an error on your part";
  CPPUNIT_ASSERT(client.createMetric<int32_t>(name3,"at work",RATE)==1);
  CPPUNIT_ASSERT(client.getNumMetric() == 2);

  client.destroyMetric(2);
  CPPUNIT_ASSERT(client.getNumMetric() == 2);
  client.destroyMetric(-1); 
  CPPUNIT_ASSERT(client.getNumMetric() == 2);

  client.destroyMetric(0);
  CPPUNIT_ASSERT(client.getNumMetric() == 1);
  CPPUNIT_ASSERT(client._getMetric(1));

  CPPUNIT_ASSERT(client.createMetric<double>(name2,"tax dollars",COUNTER) == 2);
  CPPUNIT_ASSERT(client.getNumMetric() == 2);
  CPPUNIT_ASSERT(!client._getMetric(0));
  CPPUNIT_ASSERT(client._getMetric(1));
  CPPUNIT_ASSERT(client._getMetric(2));
  CPPUNIT_ASSERT(!client._getMetric(3));


}
