#include <cppunit/extensions/HelperMacros.h>
#include <NovaDAQUtilities/CacheManager.h>
#include <NovaDAQUtilities/EnvVarCache.h>
#include <boost/lexical_cast.hpp>
#include <stdlib.h>

using namespace novadaq;

class CacheTests : public CppUnit::TestFixture
{
    CPPUNIT_TEST_SUITE(CacheTests);
    CPPUNIT_TEST(testEnvVarCacheBasic);
    CPPUNIT_TEST(testEnvVarCacheNumbers);
    CPPUNIT_TEST_SUITE_END();

public:
    void setUp();
    void tearDown();

    void testEnvVarCacheBasic();
    void testEnvVarCacheNumbers();

private:

};

void
CacheTests::setUp()
{ 
}

void
CacheTests::tearDown()
{ 
}

void
CacheTests::testEnvVarCacheBasic()
{
    std::string sampleName("PHONY_BALONEY_NAME");
    std::string sampleValue1("Fred Flintstone");
    std::string sampleValue2("Barney Rubble");
    std::string currentValue;
    EnvVarCache& theCache = EnvVarCache::getInstance();

    // verify that there is initially no env var with our sample name
    currentValue = theCache.getEnvVar(sampleName);
    CPPUNIT_ASSERT(currentValue == "");

    // create an env var with our sample name and value
    CPPUNIT_ASSERT(setenv(sampleName.c_str(), sampleValue1.c_str(), 1) == 0);

    // verify that we still get an empty string from the cache
    // (because we haven't cleared the cache)
    currentValue = theCache.getEnvVar(sampleName);
    CPPUNIT_ASSERT(currentValue == "");

    // clear the cache and see if we get the new value from a new request
    theCache.flush();
    currentValue = theCache.getEnvVar(sampleName);
    CPPUNIT_ASSERT(currentValue != "");
    CPPUNIT_ASSERT(currentValue == sampleValue1);

    // asking a second time shouldn't change anything
    currentValue = theCache.getEnvVar(sampleName);
    CPPUNIT_ASSERT(currentValue != "");
    CPPUNIT_ASSERT(currentValue == sampleValue1);

    // verify that clearing the cache through the cache manager also works
    CPPUNIT_ASSERT(setenv(sampleName.c_str(), sampleValue2.c_str(), 1) == 0);
    CacheManager::getInstance().flushCaches();
    currentValue = theCache.getEnvVar(sampleName);
    CPPUNIT_ASSERT(currentValue != "");
    CPPUNIT_ASSERT(currentValue == sampleValue2);

    // remove the env var from the environment and verify that it still
    // remains in the cache
    CPPUNIT_ASSERT(unsetenv(sampleName.c_str()) == 0);
    currentValue = theCache.getEnvVar(sampleName);
    CPPUNIT_ASSERT(currentValue == sampleValue2);

    // clear the cache for later tests
    theCache.flush();
}

void
CacheTests::testEnvVarCacheNumbers()
{
    std::string sampleName("PHONY_NUMBERIC_ENVVAR");
    long sampleLong = 1492;
    double sampleDouble = 3.141592;
    std::string longString = boost::lexical_cast<std::string>(sampleLong);
    std::string doubleString = boost::lexical_cast<std::string>(sampleDouble);
    long defaultLong = -99;
    double defaultDouble = -123.0;
    std::string currentValue;
    long longValue;
    double doubleValue;
    EnvVarCache& theCache = EnvVarCache::getInstance();

    // verify that there is initially no env var with our sample name
    currentValue = theCache.getEnvVar(sampleName);
    CPPUNIT_ASSERT(currentValue == "");
    longValue = theCache.getEnvVarAsLong(sampleName, defaultLong);
    CPPUNIT_ASSERT(longValue == defaultLong);
    doubleValue = theCache.getEnvVarAsDouble(sampleName, defaultDouble);
    CPPUNIT_ASSERT(doubleValue == defaultDouble);

    // long value tests
    CPPUNIT_ASSERT(setenv(sampleName.c_str(), longString.c_str(), 1) == 0);
    theCache.flush();
    currentValue = theCache.getEnvVar(sampleName);
    CPPUNIT_ASSERT(currentValue != "");
    longValue = theCache.getEnvVarAsLong(sampleName, defaultLong);
    CPPUNIT_ASSERT(longValue == sampleLong);

    // double value tests
    CPPUNIT_ASSERT(setenv(sampleName.c_str(), doubleString.c_str(), 1) == 0);
    theCache.flush();
    currentValue = theCache.getEnvVar(sampleName);
    CPPUNIT_ASSERT(currentValue != "");
    doubleValue = theCache.getEnvVarAsDouble(sampleName, defaultDouble);
    CPPUNIT_ASSERT(doubleValue == sampleDouble);

    // cleanup
    CPPUNIT_ASSERT(unsetenv(sampleName.c_str()) == 0);
    theCache.flush();
}

CPPUNIT_TEST_SUITE_NAMED_REGISTRATION(CacheTests, "NovaDAQUtilities");
CPPUNIT_REGISTRY_ADD_TO_DEFAULT("NovaDAQUtilities");
