#include <cppunit/extensions/HelperMacros.h>
#include <NovaDAQUtilities/LocationUtils.h>
#include <NovaDAQUtilities/ProcessUtils.h>
#include <NovaDAQUtilities/TimeUtils.h>
#include <boost/filesystem.hpp>
#include <vector>
#include <sys/types.h>
#include <unistd.h>

namespace BFS = boost::filesystem;

using namespace novadaq;

class GeneralTests : public CppUnit::TestFixture
{
    CPPUNIT_TEST_SUITE(GeneralTests);
    CPPUNIT_TEST(testProcessUtils);
    CPPUNIT_TEST(testTimeUtils);
    CPPUNIT_TEST(testFindFile);
    CPPUNIT_TEST(testEnvVarReplacement);
    CPPUNIT_TEST_SUITE_END();

public:
    void setUp();
    void tearDown();

    void testProcessUtils();
    void testTimeUtils();
    void testFindFile();
    void testEnvVarReplacement();

private:

};

void
GeneralTests::setUp()
{
    // flush stdout to get progress markers to display in a timely way
    std::cout.flush();
}

void
GeneralTests::tearDown()
{
    // flush stdout to get progress markers to display in a timely way
    std::cout.flush();
}

void
GeneralTests::testProcessUtils()
{
    pid_t thisPid = getpid();
    pid_t parentPid = getppid();
    std::vector<int> childPidList;
    int result =
        ProcessUtils::getDescendantProcessIds(parentPid, childPidList);
    CPPUNIT_ASSERT(result == 0);
    if (result == 0) {
        bool found = false;
        for (std::vector<int>::size_type jdx = 0;
             jdx < childPidList.size(); ++jdx) {
            if (childPidList[jdx] == thisPid) {
                found = true;
                break;
            }
        }
        CPPUNIT_ASSERT(found);
    }
}

void
GeneralTests::testTimeUtils()
{
    // test the conversion of a timeval struct to a double
    struct timeval unixTime;
    gettimeofday(&unixTime, 0);
    double doubleTime = convertTimevalToDouble(unixTime);
    CPPUNIT_ASSERT( ((int64_t) (doubleTime * 1000000)) ==
                    ((((int64_t) unixTime.tv_sec) * 1000000) +
                     ((int64_t) unixTime.tv_usec)));
}

void
GeneralTests::testFindFile()
{
    bool foundFile;
    std::string locatedFileURL;

    // test a couple of expected full paths
    foundFile = LocationUtils::findFile("/etc/passwd", locatedFileURL);
    CPPUNIT_ASSERT(foundFile);
    CPPUNIT_ASSERT(locatedFileURL == "/etc/passwd");

    foundFile = LocationUtils::findFile("/nova/novadaq", locatedFileURL);
    CPPUNIT_ASSERT(foundFile);
    CPPUNIT_ASSERT(locatedFileURL == "/nova/novadaq");

    // test tilde expansion
    std::string pathString1, pathString2;
    foundFile = LocationUtils::findFile("~", pathString1);
    CPPUNIT_ASSERT(foundFile);
    foundFile = LocationUtils::findFile("$HOME", pathString2);
    CPPUNIT_ASSERT(foundFile);
    CPPUNIT_ASSERT(pathString2 == pathString1);

    // test that a file in the current working directory is found
    std::string localFile;
    BFS::path currentPath = BFS::current_path();
    BFS::directory_iterator endDirIter;
    for (BFS::directory_iterator dirIter(currentPath.file_string());
         dirIter != endDirIter; ++dirIter) {
        if (BFS::is_regular_file(dirIter->status())) {
            std::string filename(dirIter->path().filename());
            localFile = filename;
            break;
        }
    }
    if (localFile.length() == 0) {
        for (BFS::directory_iterator dirIter(currentPath.file_string());
             dirIter != endDirIter; ++dirIter) {
            std::string filename(dirIter->path().filename());
            localFile = filename;
            break;
        }
    }
    if (localFile.length() > 0) {
        foundFile = LocationUtils::findFile(localFile, locatedFileURL);
        CPPUNIT_ASSERT(foundFile);
        std::string tmpString(currentPath.file_string());
        tmpString.append("/" + localFile);
        CPPUNIT_ASSERT(locatedFileURL == tmpString);
    }

    // test that an expected file in our SRT directories is found
    foundFile = LocationUtils::findFile("NovaDAQUtilities/GNUmakefile",
                                        locatedFileURL);
    CPPUNIT_ASSERT(foundFile);

    // test that a garbage file is not found
    foundFile = LocationUtils::findFile("a/file/which/should/not/be/found",
                                        locatedFileURL);
    CPPUNIT_ASSERT(! foundFile);

    // try prefixes
    std::vector<std::string> prefixes;
    prefixes.push_back("NovaRunControl/cmd/");
    foundFile = LocationUtils::findFile("SimMode1.cmd",
                                        locatedFileURL,
                                        prefixes);
    CPPUNIT_ASSERT(foundFile);
}

void
GeneralTests::testEnvVarReplacement()
{
    {
        std::string userString = LocationUtils::replaceEnvVars("${USER}");
        char* envPtr = getenv("USER");
        CPPUNIT_ASSERT(envPtr != 0);
        std::string tmpString(envPtr);
        CPPUNIT_ASSERT(tmpString == userString);
    }

    {
        std::string pathString =
            LocationUtils::replaceEnvVars("${HOME}/${HOSTNAME}");
        char* envPtr = getenv("HOME");
        CPPUNIT_ASSERT(envPtr != 0);
        std::string tmpString(envPtr);
        tmpString.append("/");
        envPtr = getenv("HOSTNAME");
        if (envPtr != 0) {
            tmpString.append(envPtr);
            CPPUNIT_ASSERT(tmpString == pathString);
        }
    }

    {
        std::string userString = LocationUtils::wordExpansion("${USER}");
        char* envPtr = getenv("USER");
        CPPUNIT_ASSERT(envPtr != 0);
        std::string tmpString(envPtr);
        CPPUNIT_ASSERT(tmpString == userString);
    }

    {
        std::string pathString1 = LocationUtils::wordExpansion("${HOME}");
        std::string pathString2 = LocationUtils::wordExpansion("~");
        CPPUNIT_ASSERT(pathString2 == pathString1);
    }
}

CPPUNIT_TEST_SUITE_NAMED_REGISTRATION(GeneralTests, "NovaDAQUtilities");
CPPUNIT_REGISTRY_ADD_TO_DEFAULT("NovaDAQUtilities");
