package novadaq;

/**
 * A base class to be used when implementing a cache of data.
 * That is, a set of data that is not expected to change often,
 * but which may need to be updated at some point (either through
 * some automatic means or via user or application request).
 *
 * @author Qiming Lu
 * @version $Revision: 1.1 $ $Date: 2009/12/28 20:56:09 $
 */

public abstract class Cache {
    
    protected CachePolicy _policy;
    protected boolean _cacheIsEmpty;
    
    protected abstract void _loadDataIntoCache();
    protected abstract void _clearCachedData();

    /**
     * Flushes the cache. (Removes any saved data.)
     */
    public synchronized void flush() {
        _clearCachedData();
        _cacheIsEmpty = true;
    }

    /**
     * Constructor.
     *
     * @param policy the CachePolicy to use for automatically expiring
     *               the cache.
     */
    protected Cache(CachePolicy policy) {
        _policy = policy;
        _cacheIsEmpty = true;
    }

    /**
     * Reloads the cache if needed.
     */
    protected void _reloadIfNeeded() {
        if(_cacheIsEmpty || _policy.isExpired()) {
            _clearCachedData();
            _loadDataIntoCache();
            _cacheIsEmpty = false;
            _policy.cacheUpdated();
        }
    }
}
