package novadaq;

import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.StringReader;
import java.io.StringWriter;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Hashtable;
import org.exolab.castor.mapping.Mapping;
import org.exolab.castor.mapping.MappingException;
import org.exolab.castor.util.ObjectFactory;
import org.exolab.castor.xml.Marshaller;
import org.exolab.castor.xml.MarshalException;
import org.exolab.castor.xml.Unmarshaller;
import org.exolab.castor.xml.ValidationException;

/**
 * This class has methods to help translate Java objects to XML strings
 * using the Castor product available from http://www.castor.org.
 *
 * @author Kurt Biery
 * @version $Revision: 1.1 $ $Date: 2009/11/24 22:15:46 $
 */
public class CastorXMLHelper {
    /**
     * String to indicate that no map file was specified.
     */
    private static final String _NO_MAP_SPECIFIED = "NoMapSpecified";

    /**
     * Simple inner class to group a marshaller instance with its
     * associated string writer.
     */
    static class MarshalTeam {
        Marshaller marshaller;
        StringWriter writer;
    }

    /**
     * Cache of marshaller teams, keyed by the name of the map file used
     * by the marshaller.
     */
    private static Hashtable<String, MarshalTeam> _teamTable =
        new Hashtable<String, MarshalTeam>();

    /**
     * Cache of unmarshaller objects, keyed by the class name of the
     * object to be deserialized *and* the map file used.
     */
    private static Hashtable<String, Unmarshaller> _unmarshallerTable =
        new Hashtable<String, Unmarshaller>();

    /**
     * Provides a reusable way to serialize objects.  This method
     * currently uses the Castor XML binding product to handle the
     * serialization.  As such, the serialized data will be
     * well-formed XML, and the input object must meet the conditions
     * expected by Castor, *OR* the input object class must have been
     * generated by Castor from an XML Schema Definition file.
     *
     * If non-null, the map name specifies a file to be used as the
     * Castor map file.  It is treated as a Java resource relative to
     * the inputObject class.  So, if the file exists somewhere in the
     * classpath independent of the input object use "/<filename>".  If
     * it exists in the classpath relative to the input object, use
     * "<relativePath>/<filename>".
     *
     * @param inputObject the object that will be marshalled (serialized)
     *        into an XML string
     * @param mapName the name of the Castor mapping file to use, or null
     *        if no special mapping is required
     * @return the string containing the XML representation of the input object
     * @throws novadaq.Exception if a problem is encountered during the 
     *         marshalling process (the original exception will be
     *         specified as the cause of the novadaq.Exception)
     */
    public static String marshal(Object inputObject, String mapName)
        throws novadaq.Exception {

        // local variables
        StringWriter localWriter = null;
        Marshaller worker = null;

        // translate empty or null map file name
        if (mapName == null || mapName.trim().equals("")) {
            mapName = _NO_MAP_SPECIFIED;
        }

        // check if we already have a marshaller team for this map name
        MarshalTeam team = _teamTable.get(mapName);
        if (team != null) {

            // fetch the team members
            localWriter = team.writer;
            worker = team.marshaller;
        }

        // otherwise, create the team members
        else {

            // create the string writer (convenient string storage)
            localWriter = new StringWriter();

            // create the mapping object for Castor, if needed
            Mapping mapper = null;
            if (! mapName.equals(_NO_MAP_SPECIFIED)) {
                mapper = new Mapping();
                InputStream src =
                    inputObject.getClass().getResourceAsStream(mapName);
                //System.out.println("SOURCE = " + src + ", mapName = " +
                //mapName);
                mapper.loadMapping(new org.xml.sax.InputSource(src));
            }

            // create the Marshal class instance
            try {
                worker = new Marshaller(localWriter);
            }
            catch (IOException ioExcpt) {
                String msg = "Error creating a new Castor Marshaller instance.";
                throw new novadaq.Exception(msg, ioExcpt);
            }
            worker.setSuppressXSIType(true);
            if (mapper != null) {
                try {
                    worker.setMapping(mapper);
                }
                catch (MappingException mapExcpt) {
                    String msg = "Error assigning a map file to a Castor " +
                        "Marshaller instance (mapName = \"" + mapName + "\").";
                    throw new novadaq.Exception(msg, mapExcpt);
                }
            }

            // add this team to the team table
            team = new MarshalTeam();
            team.marshaller = worker;
            team.writer = localWriter;
            _teamTable.put(mapName, team);
        }

        // ensure that only one thread accesses a marshaller at one time
        synchronized (worker) {

            // clear out any stale data in the string writer
            localWriter.getBuffer().setLength(0);

            // do the work
            try {
                worker.marshal(inputObject);
            }
            catch (MarshalException mExcpt) {
                String msg = "Error marshalling object of type " +
                    inputObject.getClass().getName() + " using map \"" +
                    mapName + "\".";
                throw new novadaq.Exception(msg, mExcpt);
            }
            catch (ValidationException vExcpt) {
                String msg = "Error marshalling object of type " +
                    inputObject.getClass().getName() + " using map \"" +
                    mapName + "\".";
                throw new novadaq.Exception(msg, vExcpt);
            }

            // return the string contained in the string writer
            return localWriter.toString();
        }
    }

    /**
     * Provides a reusable way to deserialize objects.  This method
     * currently uses the Castor XML binding product to handle the
     * deserialization.  As such, the serialized data is expected to be
     * well-formed XML, and the target object must meet the conditions
     * expected by Castor *OR* the target class must have been generated
     * by Castor from an XSD file.
     *
     * These conditions can be somewhat overruled by specifying a Castor
     * map file.  If the mapName argument to this method is non-null, it
     * specifies a file to be used as the map.  When it is read in,
     * it is treated as a Java resource relative to
     * the inputObject class.  So, if the file exists somewhere in the
     * classpath independent of the input object use "/<filename>".  If
     * it exists in the classpath relative to the input object, use
     * "<relativePath>/<filename>".
     *
     * @param targetObject the object that the specified XML will be
     *        unmarshalled (deserialized) into
     * @param xmlData the string containing the XML that will be unmarshalled
     * @param mapName the name of the Castor mapping file to use, or null
     *        if no special mapping is required
     * @throws novadaq.Exception if a problem is encountered during the 
     *         unmarshalling process (the original exception will be
     *         specified as the cause of the novadaq.Exception)
     */
    public static void unmarshal(Object targetObject, String xmlData,
                                 String mapName) throws novadaq.Exception {

        // check for invalid inputs
        if (targetObject == null || xmlData == null) {return;}

        // translate empty or null map file name
        if (mapName == null || mapName.trim().equals("")) {
            mapName = _NO_MAP_SPECIFIED;
        }

        // build up the hashtable key from the target class name and the
        // map file name
        Class targetClass = targetObject.getClass();
        String className = targetClass.getName();
        String tableKey = className + mapName;

        // check if we already have an unmarshaller for this class and map
        Unmarshaller worker = _unmarshallerTable.get(tableKey);
        if (worker == null) {

            // create the mapping object for Castor, if needed
            Mapping mapper = null;
            if (! mapName.equals(_NO_MAP_SPECIFIED)) {
                mapper = new Mapping();
                InputStream src =
                    targetObject.getClass().getResourceAsStream(mapName);
                mapper.loadMapping(new org.xml.sax.InputSource(src));
            }

            // create the Unmarshal instance
            worker = new Unmarshaller(targetClass);
            if (mapper != null) {
                try {
                    worker.setMapping(mapper);
                }
                catch (MappingException mapExcpt) {
                    String msg = "Error assigning a map file to a Castor " +
                        "Unmarshaller instance (mapName = \"" + mapName + "\".";
                    throw new novadaq.Exception(msg, mapExcpt);
                }
            }

            // save this unmarshaller for later re-use
            _unmarshallerTable.put(tableKey, worker);
        }

        // create a tempory string reader with the serialize data (XML)
        StringReader localReader = new StringReader(xmlData);

        // ensure that only one thread accesses an unmarshaller at one time
        synchronized (worker) {

            // create a silly object factory to get the unmarshaller
            // to fill in the contents of the existing target object
            // rather than creating a new one
            final Object fixedTargetObject = targetObject;
            ObjectFactory sillyFactory = new ObjectFactory() {
                public Object createInstance(Class type) {
                    return fixedTargetObject;
                }
                public Object createInstance(Class type, Object[] args) {
                    return fixedTargetObject;
                }
                public Object createInstance(Class type, Class[] argTypes,
                                             Object[] args) {
                    return fixedTargetObject;
                }
            };
            worker.setObjectFactory(sillyFactory);

            // actually do the unmarshalling
            try {
                Object resObj = worker.unmarshal(localReader);
            }
            catch (MarshalException mExcpt) {
                String msg = "Error unmarshalling object of type " +
                    targetObject.getClass().getName() + " using map \"" +
                    mapName + "\".";
                throw new novadaq.Exception(msg, mExcpt);
            }
            catch (ValidationException vExcpt) {
                String msg = "Error unmarshalling object of type " +
                    targetObject.getClass().getName() + " using map \"" +
                    mapName + "\".";
                throw new novadaq.Exception(msg, vExcpt);
            }
        }
    }

    /**
     * Reads in an XML document, parses it, and returns an instance of
     * the specified class.  As with every other method in this class,
     * the Castor XML binding product is used to handle the fetch and
     * parse the document.
     *
     * @param targetClassName the fully qualified class name of the object
     *        that is created and returned
     * @param urlString the location of the XML document (e.g.
     *        file:///test1.xml or http://host1/getXML)
     * @returns the object that is created as a result of the unmarshalling
     *          (i.e. deserializing) of the XML
     * @throws novadaq.Exception if a problem is encountered during the 
     *         unmarshalling process (the original exception will be
     *         specified as the cause of the novadaq.Exception)
     */
    public static Object unmarshal(String targetClassName, String urlString)
        throws novadaq.Exception {

        try {
            // create a URL instance from the URL string
            URL xmlURL = null;
            try {
                xmlURL = new URL(urlString);
            }
            catch (MalformedURLException mfuExcpt) {
                // allow for a missing "file" protocol
                if (urlString.indexOf(":") < 0) {
                    try {
                        xmlURL = new URL("file://" + urlString);
                    }
                    catch (java.lang.Exception excpt ) {
                        throw mfuExcpt;
                    }
                }
                else {
                    throw mfuExcpt;
                }
            }

            // run the unmarshalling and return the result
            return Unmarshaller.unmarshal(Class.forName(targetClassName),
                                new InputStreamReader(xmlURL.openStream()));
        }
        catch (java.lang.Exception excpt) {
            String msg = "Error unmarshalling object of type " +
                targetClassName + " from URL \"" + urlString + "\".";
            throw new novadaq.Exception(msg, excpt);
        }
    }
}
