package novadaq.demo;

/**
 * Instances of this class represent the various states of a system
 * as part of a finite state machine.
 *
 * @author Kurt Biery
 * @version $Revision: 1.1 $ $Date: 2009/12/28 17:49:10 $
 */

import java.util.Vector;

/*
 * State.java
 * June 28, 2006
 * Fermi National Accelerator Lab
 */

public class State {

    /**
     * The name of the State.
     */
    private String _name;

    /**
     * The list of Transitions that are possible from this State.
     */
    private Vector<Transition> _transitionList;

    /**
     * Constructs a State instance with the specified name.
     *
     * @param name the string that names the State
     *
     * @throws java.lang.Exception if a null or empty name is specified
     */
    public State(String name) throws Exception {

        // validate input parameters
        if (name == null || name.length() == 0) {
            throw new Exception("State names must be non-null");
        }

		// initialization
        _name = name;
        _transitionList = new Vector<Transition>(2, 2);
    }

    /**
     * Returns the name of the State.
     *
     * @returns the name of the State as a string
     */
    public String getName() {
        return _name;
    }

    /**
     * Adds the specified Transition to the list of ones that
     * are possible from this State.
     *
     * <b>Note</b> that the Transition object that is passed to this
     * method is stored without making a copy.  As long as Transitions
     * are immutable, this should not cause any problems.
     *
     * @param newTransition the Transition to add
     */
    public void addTransition(Transition newTransition) {
        // for now, we'll presume that Transition objects are immutable,
        // so we can store the one passed in rather than creating a copy
        _transitionList.addElement(newTransition);
    }

    /**
     * Returns the list of possible Transitions from this State.
     *
     * @returns a Vector containing the allowed Transitions
     */
    public Vector<Transition> getTransitions() {
        // return a copy of the list so the internal version can not
        // be accidentally corrupted
        return new Vector<Transition>(_transitionList);
    }

    /**
     * Tests if the specified object is equal to this one.
     * Currently, equality is solely based on the State name.
     * That is, two distinct State objects that have the same name
     * will be declared equal.
     *
     * @param testObject the object to compare this this State
     * @returns true if the objects are equal, false otherwise
     */
    public boolean equals(Object testObject) {
        if (testObject == null) return false;
        if (testObject == this) return true;
        if (! (testObject instanceof State)) return false;
        if (((State) testObject).getName().equals(_name)) return true;
        return false;
    }
}
