package novadaq.demo;

/**
 * Instances of this class hold the states and transitions that make
 * up a particulare finite state machine.
 * In addition, this class provides convenience
 * methods for creating the states and the transitions that link them.
 *
 * @author Kurt Biery
 * @version $Revision: 1.1 $ $Date: 2009/12/28 17:49:10 $
 */

import java.util.Vector;

/*
 * StateDiagram.java
 * June 28, 2006
 * Fermi National Accelerator Lab
 */

public class StateDiagram {

    /**
     * The list of states.
     */
    private Vector<State> _stateList;

    /**
     * The list of transitions.
     */
    private Vector<Transition> _transitionList;

    /**
     * The current state of the system.
     * This should be non-null even during transitions (so that if the
     * transition fails, we know what state to go back to).
     */
    private State _currentState = null;

    /**
     * The current transition that is in progress, if any.  If no
     * transition is in progress, this should be null.
     */
    private Transition _currentTransition = null;

    /**
     * Constructs a new StateDiagram with the specified initial state.
     *
     * @param name the string name for the initial state
     * @throws java.lang.Exception if the state name is invalid
     */
    public StateDiagram(String initialStateName) throws Exception {

        // initialization
        _stateList = new Vector<State>();
        _transitionList = new Vector<Transition>();

        // stateName must be non-null
        if (initialStateName == null || initialStateName.length() == 0) {
            throw new Exception("StateDiagram state names must be non-null");
        }

        // define the initial state
        _currentState = new State(initialStateName);
        _stateList.addElement(_currentState);
    }

    /**
     * Returns the current State of the system.
     *
     * @returns the current State
     */
    public State getCurrentState() {
        return _currentState;
    }

    /**
     * Returns the Transition that is in progess, if any.  If no
     * Transition is in progess, null will be returned.
     *
     * @returns the current Transition or null
     */
    public Transition getCurrentTransition() {
        return _currentTransition;
    }

    /**
     * Tests whether a Transition is in progress.
     *
     * @returns boolean flag indicating if a Transition is in progress
     */
    public boolean transitionInProgress() {
        return (_currentTransition != null);
    }

    /**
     * Returns the list of States that exist in this diagram.
     *
     * @returns a Vector containing the States
     */
    public Vector getStates() {
        // return a copy of the list so that the internal version can not
        // be accidentally corrupted
        return new Vector<State>(_stateList);
    }

    /**
     * Returns the list of Transitions that exist in this diagram.
     *
     * @returns a Vector containing the Transitions
     */
    public Vector<Transition> getTransitions() {
        // return a copy of the list so that the internal version can not
        // be accidentally corrupted
        return new Vector<Transition>(_transitionList);
    }

    /**
     * Adds a state to the diagram with the specified name.
     *
     * @param name the string name for the new state
     * @returns the newly created State object
     * @throws java.lang.Exception if the state name is invalid or already
     *        exists in the diagram
     */
    public State addState(String stateName) throws Exception {

        // sanity check: stateName must be non-null
        if (stateName == null || stateName.length() == 0) {
            throw new Exception("StateDiagram state names must be non-null");
        }

        // require states to be unique
        State newState = new State(stateName);
        if (_stateList.contains(newState)) {
            throw new Exception("The state diagram already contains a " +
                                "state named \"" + stateName + "\"");
        }

        // add the state to the list and return it
        _stateList.addElement(newState);
        return newState;
    }

    /**
     * Adds the specified transition to this diagram and link it to the
     * specified initial and final states.
     *
     * @param transitionAction action string associated with the transition
     * @param transitionDescription string description of the transition
     * @param continueOn boolean to indicate whether this
     *        transition should continue or rollback when an error is
     *        encountered
     * @param initialStateName string name of the state from which this
     *        transition is allowed
     * @param finalStateName string name of the state the system is in
     *        after this transition successfully completes
     * @returns the new created Transition object
     * @throws java.lang.Exception if any of the input arguments are invalid,
     *        if the specified states cannot be found, or if a transition
     *        with the specified action already exists
     */
    public Transition addTransition(String transitionAction,
                                    String transitionDescription,
                                    boolean continueOnError,
                                    String initialStateName,
                                    String finalStateName)
        throws Exception {

        // validate inputs
        if (transitionAction == null || transitionAction.length() == 0) {
            throw new Exception("Transition action strings must be non-null");
        }
        if (transitionDescription == null ||
            transitionDescription.length() == 0) {
            throw new Exception("Transition descriptions must be non-null");
        }
        if (initialStateName == null || initialStateName.length() == 0) {
            throw new Exception("StateDiagram state names must be non-null");
        }
        if (finalStateName == null || finalStateName.length() == 0) {
            throw new Exception("StateDiagram state names must be non-null");
        }

        // look up the initial and final states
        State initialState = null;
        State finalState = null;
        for (int idx = 0; idx < _stateList.size(); idx++) {
            State workingState = (State) _stateList.elementAt(idx);
            if (initialStateName.equals(workingState.getName())) {
                initialState = workingState;
            }
            if (finalStateName.equals(workingState.getName())) {
                finalState = workingState;
            }
        }
        if (initialState == null) {
            throw new Exception("Unable to find the \"" + initialStateName +
                                "\" state in the StateDiagram");
        }
        if (finalState == null) {
            throw new Exception("Unable to find the \"" + finalStateName +
                                "\" state in the StateDiagram");
        }

        // create the transition and verify that it is unique
        Transition newTransition =
            new Transition(transitionAction, transitionDescription,
                           finalState, continueOnError);
        if (_transitionList.contains(newTransition)) {
            throw new Exception("A transition with action \"" +
                                transitionAction +
                                "\" already exists in the StateDiagram");

        }

        // add the transition to the initial state and save it in our list
        initialState.addTransition(newTransition);
        _transitionList.addElement(newTransition);
        return newTransition;
    }

    /**
     * Links the specified transition to the specified initial state.
     *
     * @param transitionAction action string associated with the transition
     * @param initialStateName string name of the state from which this
     *        transition is allowed
     * @returns the associated Transition object
     * @throws java.lang.Exception if any of the input arguments are invalid
     *        or if the specified state or transition cannot be found
     */
    public Transition linkTransition(String transitionAction,
                                     String initialStateName)
        throws Exception {

        // validate inputs
        if (transitionAction == null || transitionAction.length() == 0) {
            throw new Exception("Transition action strings must be non-null");
        }
        if (initialStateName == null || initialStateName.length() == 0) {
            throw new Exception("StateDiagram state names must be non-null");
        }

        // look up the initial states
        State initialState = null;
        for (int idx = 0; idx < _stateList.size(); idx++) {
            State workingState = (State) _stateList.elementAt(idx);
            if (initialStateName.equals(workingState.getName())) {
                initialState = workingState;
            }
        }
        if (initialState == null) {
            throw new Exception("Unable to find the \"" + initialStateName +
                                "\" state in the StateDiagram");
        }

        // look up the transition
        Transition requestedTransition = null;
        for (int idx = 0; idx < _transitionList.size(); idx++) {
            Transition transition =
                (Transition) _transitionList.elementAt(idx);
            if (transitionAction.equals(transition.getAction())) {
                requestedTransition = transition;
            }
        }
        if (requestedTransition == null) {
            throw new Exception("Unable to find the \"" + transitionAction +
                                "\" transition in the StateDiagram");
        }

        // add the transition to the initial state
        initialState.addTransition(requestedTransition);
        return requestedTransition;
    }

    /**
     * Starts the transition associated with the specified action.
     *
     * @param requestedAction the action string for the requested Transition
     * @returns the Transition that corresponds to the requested action
     * @throws java.lang.Exception if the requested action is invalid or
     *        the requested transition is unavailable
     */
    public Transition startTransition(String requestedAction)
        throws Exception {

        // sanity checks
        if (requestedAction == null || requestedAction.length() == 0) {
            throw new Exception("State diagram action requests must " +
                                "be non-null");
        }

        // look up the transition that was requested
        Transition requestedTransition = null;
        Vector<Transition> allowedTransitionList =
	    _currentState.getTransitions();
        for (int idx = 0; idx < allowedTransitionList.size(); idx++) {
            Transition allowedTransition =
                (Transition) allowedTransitionList.elementAt(idx);
            if (requestedAction.equals(allowedTransition.getAction())) {
                requestedTransition = allowedTransition;
                break;
            }
        }
        if (requestedTransition == null) {
            throw new Exception("No allowed transition found " +
                                "for action \"" + requestedAction + "\"");
        }

        // update the current transition
        _currentTransition = requestedTransition;
        return requestedTransition;
    }

    /* TODO - handle errors */
    /**
     * Completes the transition that is in progress and moves the system
     * to the resulting state.
     *
     * @returns the Transition that had been in progress
     * @throws java.lang.Exception if no transition was in progress
     */
    public Transition completeTransition(boolean errorOccurred)
        throws Exception {

        // verify that a transition is in progress
        if (_currentTransition == null) {
            throw new Exception("No transition in progress - completion " +
                                "request ignored");
        }

        // update the current state, etc.
        _currentState = _currentTransition.getResultingState();
        Transition transInProgress = _currentTransition;
        _currentTransition = null;
        return transInProgress;
    }

	/**
	 * Returns text that describes the current state of the StateDiagram.
	 */
	public String currentStateToString() {
		StringBuffer buff = new StringBuffer();
		buff.append(super.toString());
		buff.append(" ");
		buff.append("Current state is " + _currentState.getName());

		buff.append(" ");
		buff.append("Allowed transitions are ");
        Vector<Transition> allowedTransitionList =
	    _currentState.getTransitions();
        for (int idx = 0; idx < allowedTransitionList.size(); idx++) {
            Transition allowedTransition =
                (Transition) allowedTransitionList.elementAt(idx);
			buff.append(" ");
			buff.append(allowedTransition.getAction());
		}

		return buff.toString();
	}
}
