package novadaq.demo;

/**
 * Instances of this class represent transitions that are possible in
 * a finite state machine.  No interactions between this class and external
 * processes are currently supported - it simply represents the logical 
 * transition from one state to another.
 *
 * @author Kurt Biery
 * @version $Revision: 1.1 $ $Date: 2009/12/28 17:49:10 $
 */

import java.util.Vector;

/*
 * Transition.java
 * June 28, 2006
 * Fermi National Accelerator Lab
 */

public class Transition {

    /**
     * The action (or command) that initiates this Transition.
     */
    private String _action;

    /**
     * The description of the Transition.
     */
    private String _description;

    /**
     * The State that results from the successful completion of this
     * Transition.
     */
    private State _resultingState;

    /**
     * The flag that indicates whether the transition should continue
     * to the resulting State even if an error occurs.  If this is false,
     * then the system should return to the original state when an
     * error occurs.
     */
    private boolean _continueOnError;

    /**
     * Constructs a Transition instance with the specified parameters.
     *
     * @param action the string that defines the action that initiates
     *        this Transition
     * @param description the string description (This can be used to display
     *        the transitional "state" of the system. For example,
     *        a "Configure" transition might have a description of
     *        "Configuring".
     * @param resultingState the State that results from the successful
     *        execution of this Transition
     * @param continueOnError the boolean flag that indicates whether the
     *        transition should continue to the resultingState even
     *        if an error occurs
     *
     * @throws java.lang.Exception if a null or empty action, description,
     *        or resultingState are specified
     */
    public Transition(String action, String description, State resultingState,
                      boolean continueOnError) throws Exception {

        // validate input parameters
        if (action == null || action.length() == 0) {
            throw new Exception("Transition actions must be non-null");
        }
        if (description == null || description.length() == 0) {
            throw new Exception("Transition descriptions must be non-null");
        }
        if (resultingState == null) {
            throw new Exception("Transition resulting states must be non-null");
        }

        // assign parameters
        _action = action;
        _description = description;
        _resultingState = resultingState;
        _continueOnError = continueOnError;
    }

    /**
     * Returns the action that initiates this Transition.
     *
     * @returns the action string
     */
    public String getAction() {
        return _action;
    }

    /**
     * Returns the description of this Transition.
     *
     * @returns the description string
     */
    public String getDescription() {
        return _description;
    }

    /**
     * Returns the state that results from a successful completion of
     * this Transition.
     *
     * @returns resulting State
     */
    public State getResultingState() {
        return _resultingState;
    }

    /**
     * Returns true if the transition should continue even if an error
     * is encountered.
     *
     * @returns boolean flag indicating whether to continue after an error
     */
    public boolean continueOnError() {
        return _continueOnError;
    }

    /**
     * Tests if the specified object is equal to this one.
     * Currently, equality is solely based on the Transition action.
     * That is, two distinct Transition objects that have the same action
     * will be declared equal.
     *
     * @param testObject the object to compare this this Transition
     * @returns true if the objects are equal, false otherwise
     */
    public boolean equals(Object testObject) {
        if (testObject == null) return false;
        if (testObject == this) return true;
        if (! (testObject instanceof Transition)) return false;
        if (((Transition) testObject).getAction().equals(_action)) {
            return true;
        }
        return false;
    }
}
