#include <cppunit/extensions/HelperMacros.h>
#include <NovaDatabase/Table.h>
#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/shared_ptr.hpp>

namespace BPT = boost::posix_time;
using namespace nova::database;

const std::string TABLE1_DEFINITION_FILE =
  "NovaDatabase/cxx/unittest/basicTestTable1.xml";

const std::string TABLE2_DEFINITION_FILE =
  "NovaDatabase/cxx/unittest/basicTestTable2.xml";

const std::string TABLE3_DEFINITION_FILE =
  "NovaDatabase/cxx/unittest/basicTestTable3.xml";

class BasicTests : public CppUnit::TestFixture
{
  CPPUNIT_TEST_SUITE(BasicTests);
  CPPUNIT_TEST(populateDBTable);
  CPPUNIT_TEST(testRowInsertWithAutoIncrColumn);
  CPPUNIT_TEST(testColumnLookup);
  CPPUNIT_TEST(testDistinctAndOrder);
  CPPUNIT_TEST(testTableCreationAndDestruction);
  CPPUNIT_TEST(testTableCreateFromDB);  
  CPPUNIT_TEST(testEmptyStringVsNULL);
  CPPUNIT_TEST(testUpdateTwice);
  CPPUNIT_TEST(testBooleanValidity);
  CPPUNIT_TEST(cleanupDBTable);
  CPPUNIT_TEST_SUITE_END();

public:
  void setUp();
  void tearDown();

  void populateDBTable();
  void testRowInsertWithAutoIncrColumn();
  void testColumnLookup();
  void testDistinctAndOrder();
  void testTableCreationAndDestruction();
  void testTableCreateFromDB();
  void testEmptyStringVsNULL();
  void testUpdateTwice();
  void testBooleanValidity();
  void cleanupDBTable();

private:

  static boost::shared_ptr<Table> _dbt1;
  static boost::shared_ptr<Table> _dbt2;
  static boost::shared_ptr<Table> _dbt3;
};

boost::shared_ptr<Table> BasicTests::_dbt1;
boost::shared_ptr<Table> BasicTests::_dbt2;
boost::shared_ptr<Table> BasicTests::_dbt3;

void
BasicTests::setUp()
{
  // flush stdout to get progress markers to display in a timely way
  std::cout.flush();
}

void
BasicTests::tearDown()
{
  // flush stdout to get progress markers to display in a timely way
  std::cout.flush();
}

void
BasicTests::testRowInsertWithAutoIncrColumn()
{
  CPPUNIT_ASSERT(_dbt2->ExistsInDB());
  _dbt2->Clear();
  CPPUNIT_ASSERT(_dbt2->LoadFromDB());

  // some random, but known, values
  int initialRowCount = _dbt2->NRow();
  int channel = 3 * (initialRowCount + 1);
  time_t now = time(0);
  int tmpVal = now % 10000;
  float value1 = (float) (((double) tmpVal) / 100.0);
  float value2 = (float) (((double) tmpVal) / 357.0);
  std::string string1 = (((initialRowCount % 5) != 0) ? "three" : "one");

  // create the test row and store it in the database
  boost::shared_ptr<Row> rowPtr(_dbt2->NewRow());
  CPPUNIT_ASSERT(rowPtr->Set("channel", channel));
  CPPUNIT_ASSERT(rowPtr->Set("value1", value1));
  CPPUNIT_ASSERT(rowPtr->Set("value2", value2));
  CPPUNIT_ASSERT(rowPtr->Set("string1", string1));
  CPPUNIT_ASSERT(rowPtr->Set("string2", "two"));
  _dbt2->AddRow(rowPtr.get());
  CPPUNIT_ASSERT(_dbt2->WriteToDB());

  // verify that we can fetch the ID of the new row
  int64_t idOfInsertedRow;
  CPPUNIT_ASSERT(_dbt2->GetCurrSeqVal("id", idOfInsertedRow));

  // reload the table from the database
  _dbt2->Clear();
  CPPUNIT_ASSERT(_dbt2->LoadFromDB());
  int newRowCount = _dbt2->NRow();

  // verify that the number of rows has increased by one
  CPPUNIT_ASSERT(newRowCount == (initialRowCount + 1));

  // loop over the result set until we find the new row
  // (testing of validity ranges will be done in a later test)
  bool found = false;
  for (int idx = 0; idx < newRowCount; ++idx) {
    Row* row = _dbt2->GetRow(idx);
    Column* col = row->Col("id");
    int64_t rowId;
    CPPUNIT_ASSERT(col->Get(rowId));
    if (rowId == idOfInsertedRow) {
      found = true;
      int testChannelValue;
      float testValue;

      col = row->Col("channel");
      CPPUNIT_ASSERT(col->Get(testChannelValue));
      CPPUNIT_ASSERT(testChannelValue == channel);

      col = row->Col("value1");
      CPPUNIT_ASSERT(col->Get(testValue));
      CPPUNIT_ASSERT(testValue > (0.99 * value1) &&
                     testValue < (1.01 * value1));

      col = row->Col("value2");
      CPPUNIT_ASSERT(col->Get(testValue));
      CPPUNIT_ASSERT(testValue > (0.99 * value2) &&
                     testValue < (1.01 * value2));

      break;
    }
  }
  CPPUNIT_ASSERT(found);

  // test the lookup of a row by ID
  _dbt2->Clear();
  _dbt2->SetValidityRange("id", idOfInsertedRow);
  CPPUNIT_ASSERT(_dbt2->LoadFromDB());
  CPPUNIT_ASSERT(_dbt2->NRow() == 1);
}

void
BasicTests::testColumnLookup()
{
  CPPUNIT_ASSERT(_dbt2->ExistsInDB());
  _dbt2->Clear();
  CPPUNIT_ASSERT(_dbt2->LoadFromDB());

  int nRow = _dbt2->NRow();
  CPPUNIT_ASSERT(nRow > 0);

  // test individual columns
  std::string columnValue;
  Row* row = _dbt2->GetRow(0);
  Column* col = row->Col("id");
  CPPUNIT_ASSERT(col != 0);
  CPPUNIT_ASSERT(col->Get(columnValue));
  col = row->Col("channel");
  CPPUNIT_ASSERT(col != 0);
  CPPUNIT_ASSERT(col->Get(columnValue));
  col = row->Col("value1");
  CPPUNIT_ASSERT(col != 0);
  CPPUNIT_ASSERT(col->Get(columnValue));
  col = row->Col("value2");
  CPPUNIT_ASSERT(col != 0);
  CPPUNIT_ASSERT(col->Get(columnValue));
  col = row->Col("inserttime");
  CPPUNIT_ASSERT(col != 0);
  CPPUNIT_ASSERT(col->Get(columnValue));
  col = row->Col("insertuser");
  CPPUNIT_ASSERT(col != 0);
  CPPUNIT_ASSERT(col->Get(columnValue));

  // test the full set of columns
  bool foundChannelColumn = false;
  CPPUNIT_ASSERT(_dbt2->NCol() == 11);
  for (int idx = 0; idx < _dbt2->NCol(); ++idx) {
    Column& colRef = row->Col(idx);
    std::string cname = colRef.Name();
    if (cname != "nameid" && cname != "updateuser" &&
        cname != "updatetime" && cname != "insertuser" &&
        cname != "inserttime") {
      CPPUNIT_ASSERT(colRef.Get(columnValue));
      if (colRef.Name() == "channel") {
        foundChannelColumn = true;
      }
    }
  }
  CPPUNIT_ASSERT(foundChannelColumn);
}

void
BasicTests::testDistinctAndOrder()
{
  CPPUNIT_ASSERT(_dbt2->ExistsInDB());

  _dbt2->Clear();
  CPPUNIT_ASSERT(_dbt2->AddDistinctColumn("string1"));
  CPPUNIT_ASSERT(_dbt2->LoadFromDB());
  CPPUNIT_ASSERT(_dbt2->NRow() == 2);

  _dbt2->Clear();
  CPPUNIT_ASSERT(_dbt2->AddDistinctColumn("string2"));
  CPPUNIT_ASSERT(_dbt2->LoadFromDB());
  CPPUNIT_ASSERT(_dbt2->NRow() == 2);

  _dbt2->Clear();
  CPPUNIT_ASSERT(_dbt2->AddDistinctColumn("string1"));
  CPPUNIT_ASSERT(_dbt2->AddDistinctColumn("string2"));
  CPPUNIT_ASSERT(_dbt2->LoadFromDB());
  CPPUNIT_ASSERT(_dbt2->NRow() == 3);


  _dbt2->Clear();
  CPPUNIT_ASSERT(_dbt2->AddOrderColumn("channel"));
  CPPUNIT_ASSERT(_dbt2->LoadFromDB());

  int previousChannel = 999999999;
  int channel;
  for (int i=0; i<_dbt2->NRow(); ++i) {
    CPPUNIT_ASSERT(_dbt2->GetRow(i)->Col("channel")->Get(channel));
    CPPUNIT_ASSERT(channel <= previousChannel);
    previousChannel = channel;
  }

  _dbt2->Clear();
  CPPUNIT_ASSERT(_dbt2->AddOrderColumn("channel"));
  _dbt2->SetOrderAsc();
  CPPUNIT_ASSERT(_dbt2->LoadFromDB());

  previousChannel = -1;
  for (int i=0; i<_dbt2->NRow(); ++i) {
    CPPUNIT_ASSERT(_dbt2->GetRow(i)->Col("channel")->Get(channel));
    CPPUNIT_ASSERT(channel >= previousChannel);
    previousChannel = channel;
  }

  _dbt2->Clear();
  CPPUNIT_ASSERT(_dbt2->AddOrderColumn("string2"));
  CPPUNIT_ASSERT(_dbt2->AddOrderColumn("value2"));
  _dbt2->SetOrderAsc();
  CPPUNIT_ASSERT(_dbt2->LoadFromDB());

  double previousValue2 = -1.0;
  double value2;
  std::string previousString2 = "aaaaaa";
  std::string string2;
  for (int i=0; i<_dbt2->NRow(); ++i) {
    CPPUNIT_ASSERT(_dbt2->GetRow(i)->Col("string2")->Get(string2));
    CPPUNIT_ASSERT(string2 >= previousString2);
    if (string2 == previousString2) {
      CPPUNIT_ASSERT(_dbt2->GetRow(i)->Col("value2")->Get(value2));
      CPPUNIT_ASSERT(value2 >= previousValue2);
      previousValue2 = value2;
    }
    else {
      previousValue2 = -1.0;
    }
    previousString2 = string2;
  }
}

void
BasicTests::testTableCreationAndDestruction()
{
  boost::shared_ptr<Table> localDBTablePtr;

  for (int idx = 0; idx < 20; ++idx) {
      //std::cout << "Table creation test, version 1, index "
      //          << idx << std::endl;
      localDBTablePtr.reset(new Table(TABLE2_DEFINITION_FILE.c_str()));
      CPPUNIT_ASSERT(localDBTablePtr->ExistsInDB());
  }

  for (int idx = 0; idx < 20; ++idx) {
      //std::cout << "Table creation test, version 1, index "
      //          << idx << std::endl;
      localDBTablePtr.reset(new Table(TABLE2_DEFINITION_FILE.c_str()));
      localDBTablePtr->GetConnection();
      CPPUNIT_ASSERT(localDBTablePtr->ExistsInDB());
  }
}

void
BasicTests::testTableCreateFromDB()
{
  boost::shared_ptr<Table> localDBTablePtr;

  localDBTablePtr.reset(new Table());
  CPPUNIT_ASSERT(localDBTablePtr->CreateFromDB("public","hardware_position",
					       "ifdbprod.fnal.gov","nova_hardware",
					       "5432","nova_reader"));
}

void
BasicTests::testEmptyStringVsNULL()
{
  _dbt3->Clear();
  Row* rowPtr = _dbt3->NewRow();

  CPPUNIT_ASSERT(rowPtr->Set("channel", 1024));
  CPPUNIT_ASSERT(rowPtr->Set("value1", 3.1415));
  CPPUNIT_ASSERT(rowPtr->Set("value2", 0.01));
  CPPUNIT_ASSERT(! rowPtr->Set("string1", ""));
  CPPUNIT_ASSERT(rowPtr->Set("string2", "NULL"));
  _dbt3->AddRow(rowPtr);
  CPPUNIT_ASSERT(_dbt3->WriteToDB());
  delete rowPtr;

  _dbt3->Clear();
  CPPUNIT_ASSERT(_dbt3->LoadFromDB());
  int nRows = _dbt3->NRow();
  rowPtr = _dbt3->GetRow(nRows-1);
  CPPUNIT_ASSERT(rowPtr->Col("string1")->IsNull());
  CPPUNIT_ASSERT(rowPtr->Col("string2")->IsNull());
  _dbt3->Clear();
}

void
BasicTests::testUpdateTwice()
{
  _dbt3->Clear();
  _dbt3->SetUser(getenv("USER"));
  Row* rowPtr = _dbt3->NewRow();

  CPPUNIT_ASSERT(rowPtr->Set("channel", 2048));
  CPPUNIT_ASSERT(rowPtr->Set("value1", 929.929));
  CPPUNIT_ASSERT(rowPtr->Set("value2", 0.929));
  CPPUNIT_ASSERT(rowPtr->Set("string1", "929"));
  CPPUNIT_ASSERT(rowPtr->Set("string2", "92929"));
  _dbt3->AddRow(rowPtr);
  CPPUNIT_ASSERT(_dbt3->WriteToDB());
  delete rowPtr;

  _dbt3->Clear();
  CPPUNIT_ASSERT(_dbt3->LoadFromDB());
  int nRows = _dbt3->NRow();
  CPPUNIT_ASSERT(_dbt3->GetRow(nRows-1)->Update("channel",2049));
  CPPUNIT_ASSERT(_dbt3->WriteToDB());
  _dbt3->Clear();
  CPPUNIT_ASSERT(_dbt3->LoadFromDB());
  int myChan;
  CPPUNIT_ASSERT(_dbt3->GetRow(nRows-1)->Col("channel")->Get(myChan));
  CPPUNIT_ASSERT(myChan == 2049);
  std::string updateUser, updateTime;
  CPPUNIT_ASSERT(_dbt3->GetRow(nRows-1)->Col("updateuser")->Get(updateUser));
  CPPUNIT_ASSERT(_dbt3->GetRow(nRows-1)->Col("updatetime")->Get(updateTime));
  sleep(2);
  CPPUNIT_ASSERT(_dbt3->GetRow(nRows-1)->Update("channel",2050));
  CPPUNIT_ASSERT(_dbt3->WriteToDB());
  _dbt3->Clear();
  CPPUNIT_ASSERT(_dbt3->LoadFromDB());
  CPPUNIT_ASSERT(_dbt3->GetRow(nRows-1)->Col("channel")->Get(myChan));
  CPPUNIT_ASSERT(myChan == 2050);
  std::string updateUser_save = updateUser;
  std::string updateTime_save = updateTime;
  CPPUNIT_ASSERT(_dbt3->GetRow(nRows-1)->Col("updateuser")->Get(updateUser));
  CPPUNIT_ASSERT(_dbt3->GetRow(nRows-1)->Col("updatetime")->Get(updateTime));
  CPPUNIT_ASSERT(updateUser_save == updateUser);

  BPT::ptime firstTime = BPT::time_from_string(updateTime_save);
  BPT::ptime secondTime = BPT::time_from_string(updateTime);
  BPT::time_duration timeDiff = secondTime - firstTime;
  CPPUNIT_ASSERT(timeDiff.seconds() == 2 || timeDiff.seconds() == 3);
}

void
BasicTests::testBooleanValidity()
{
  CPPUNIT_ASSERT(_dbt1->ExistsInDB());

  _dbt1->Clear();
  _dbt1->SetValidityRange("isenabled", true);
  CPPUNIT_ASSERT(_dbt1->LoadFromDB());
  CPPUNIT_ASSERT(_dbt1->NRow() == 2);

  _dbt1->Clear();
  _dbt1->SetValidityRange("isenabled", false);
  CPPUNIT_ASSERT(_dbt1->LoadFromDB());
  CPPUNIT_ASSERT(_dbt1->NRow() == 1);
}

void
BasicTests::populateDBTable()
{
  // create the tables that are used in these tests
  _dbt1.reset(new Table(TABLE1_DEFINITION_FILE.c_str()));
  CPPUNIT_ASSERT(_dbt1->ExistsInDB());
  _dbt1->Clear();
  CPPUNIT_ASSERT(_dbt1->LoadFromDB());

  _dbt2.reset(new Table(TABLE2_DEFINITION_FILE.c_str()));
  CPPUNIT_ASSERT(_dbt2->ExistsInDB());
  _dbt2->Clear();
  CPPUNIT_ASSERT(_dbt2->LoadFromDB());

  _dbt3.reset(new Table(TABLE3_DEFINITION_FILE.c_str()));
  CPPUNIT_ASSERT(_dbt3->ExistsInDB());
  _dbt3->Clear();
  CPPUNIT_ASSERT(_dbt3->LoadFromDB());

  // create some filler rows, if needed
  int nRow = _dbt2->NRow();
  if (nRow == 0) {
    boost::shared_ptr<Row> rowPtr(_dbt2->NewRow());
    CPPUNIT_ASSERT(rowPtr->Set("channel", 101));
    CPPUNIT_ASSERT(rowPtr->Set("value1", 5.0));
    CPPUNIT_ASSERT(rowPtr->Set("value2", 42.6));
    CPPUNIT_ASSERT(rowPtr->Set("string1", "one"));
    CPPUNIT_ASSERT(rowPtr->Set("string2", "two"));
    _dbt2->AddRow(rowPtr.get());
    CPPUNIT_ASSERT(_dbt2->WriteToDB());
    rowPtr.reset(_dbt2->NewRow());
    CPPUNIT_ASSERT(rowPtr->Set("channel", 102));
    CPPUNIT_ASSERT(rowPtr->Set("value1", 5.0));
    CPPUNIT_ASSERT(rowPtr->Set("value2", 42.7));
    CPPUNIT_ASSERT(rowPtr->Set("string1", "one"));
    CPPUNIT_ASSERT(rowPtr->Set("string2", "four"));
    _dbt2->AddRow(rowPtr.get());
    CPPUNIT_ASSERT(_dbt2->WriteToDB());
    rowPtr.reset(_dbt2->NewRow());
    CPPUNIT_ASSERT(rowPtr->Set("channel", 103));
    CPPUNIT_ASSERT(rowPtr->Set("value1", 5.0));
    CPPUNIT_ASSERT(rowPtr->Set("value2", 42.8));
    CPPUNIT_ASSERT(rowPtr->Set("string1", "one"));
    CPPUNIT_ASSERT(rowPtr->Set("string2", "two"));
    _dbt2->AddRow(rowPtr.get());
    CPPUNIT_ASSERT(_dbt2->WriteToDB());
  }

  nRow = _dbt1->NRow();
  if (nRow == 0) {
    boost::shared_ptr<Row> rowPtr(_dbt1->NewRow());
    CPPUNIT_ASSERT(rowPtr->Set("Name", "AAA"));
    CPPUNIT_ASSERT(rowPtr->Set("isEnabled", true));
    _dbt1->AddRow(rowPtr.get());
    CPPUNIT_ASSERT(_dbt1->WriteToDB());
    rowPtr.reset(_dbt1->NewRow());
    CPPUNIT_ASSERT(rowPtr->Set("Name", "BBB"));
    CPPUNIT_ASSERT(rowPtr->Set("isEnabled", false));
    _dbt1->AddRow(rowPtr.get());
    CPPUNIT_ASSERT(_dbt1->WriteToDB());
    rowPtr.reset(_dbt1->NewRow());
    CPPUNIT_ASSERT(rowPtr->Set("Name", "CCC"));
    CPPUNIT_ASSERT(rowPtr->Set("isEnabled", true));
    _dbt1->AddRow(rowPtr.get());
    CPPUNIT_ASSERT(_dbt1->WriteToDB());
  }

  // clear the Table objects as a nicety
  _dbt1->Clear();
  _dbt2->Clear();
  _dbt3->Clear();
}

void
BasicTests::cleanupDBTable()
{
  _dbt1.reset();
  _dbt2.reset();
  _dbt3.reset();
}

CPPUNIT_TEST_SUITE_NAMED_REGISTRATION(BasicTests, "NovaDatabase");
CPPUNIT_REGISTRY_ADD_TO_DEFAULT("NovaDatabase");
