#include <cppunit/extensions/HelperMacros.h>
#include <NovaDatabase/Table.h>
#include <boost/shared_ptr.hpp>

#include <NovaDAQConventions/DAQConventions.h>

// dummy comment

using namespace nova::database;

const std::string TABLE3_DEFINITION_FILE =
  "NovaDatabase/cxx/unittest/basicTestTable3.xml";

class DevelopmentTests : public CppUnit::TestFixture
{
  CPPUNIT_TEST_SUITE(DevelopmentTests);
  CPPUNIT_TEST(populateDBTable);
  CPPUNIT_TEST(testUpdateTwice);
  //  CPPUNIT_TEST(testEmptyStringVsNULL);
  CPPUNIT_TEST(testColumnLookup);
  CPPUNIT_TEST(testDistinctAndOrder);
  //  CPPUNIT_TEST(testTableCreationAndDestruction);
  //  CPPUNIT_TEST(testDBCommandCache);
  CPPUNIT_TEST(testUserName);
  CPPUNIT_TEST(cleanupDBTable);
  CPPUNIT_TEST_SUITE_END();

public:
  void setUp();
  void tearDown();

  void populateDBTable();
  void testColumnLookup();
  void testDistinctAndOrder();
  void testTableCreationAndDestruction();
  void testDBCommandCache();
  void testEmptyStringVsNULL();
  void testUpdateTwice();
  void cleanupDBTable();
  void testUserName();

private:

  static boost::shared_ptr<Table> _dbt3;
};

boost::shared_ptr<Table> DevelopmentTests::_dbt3;

void
DevelopmentTests::setUp()
{
  // flush stdout to get progress markers to display in a timely way
  std::cout.flush();
}

void
DevelopmentTests::tearDown()
{
  // flush stdout to get progress markers to display in a timely way
  std::cout.flush();
}

void
DevelopmentTests::testColumnLookup()
{
  std::cout << "Column lookup test:" << std::endl;
  CPPUNIT_ASSERT(_dbt3->ExistsInDB());
  _dbt3->Clear();
  CPPUNIT_ASSERT(_dbt3->LoadFromDB());

  int nRow = _dbt3->NRow();
  CPPUNIT_ASSERT(nRow > 0);

  // test individual columns
  std::string columnValue;
  Row* row = _dbt3->GetRow(0);
  Column* col = row->Col("id");
  CPPUNIT_ASSERT(col != 0);
  CPPUNIT_ASSERT(col->Get(columnValue));
  col = row->Col("channel");
  CPPUNIT_ASSERT(col != 0);
  CPPUNIT_ASSERT(col->Get(columnValue));
  col = row->Col("value1");
  CPPUNIT_ASSERT(col != 0);
  CPPUNIT_ASSERT(col->Get(columnValue));
  col = row->Col("value2");
  CPPUNIT_ASSERT(col != 0);
  CPPUNIT_ASSERT(col->Get(columnValue));
  col = row->Col("inserttime");
  CPPUNIT_ASSERT(col != 0);
  CPPUNIT_ASSERT(col->Get(columnValue));
  col = row->Col("insertuser");
  CPPUNIT_ASSERT(col != 0);
  CPPUNIT_ASSERT(col->Get(columnValue));

  // test the full set of columns
  bool foundChannelColumn = false;
  CPPUNIT_ASSERT(_dbt3->NCol() == 11);
  for (int idx = 0; idx < _dbt3->NCol(); ++idx) {
    Column& colRef = row->Col(idx);
    std::string cname = colRef.Name();
    if (! colRef.IsNull()) {
      if (cname != "nameid" && cname != "updateuser" &&
	  cname != "updatetime" && cname != "insertuser" &&
	  cname != "inserttime") {
	CPPUNIT_ASSERT(colRef.Get(columnValue));
      }
      if (colRef.Name() == "channel") {
        foundChannelColumn = true;
      }
    }
  }
  CPPUNIT_ASSERT(foundChannelColumn);
}

void
DevelopmentTests::testDistinctAndOrder()
{
  CPPUNIT_ASSERT(_dbt3->ExistsInDB());

  _dbt3->Clear();
  CPPUNIT_ASSERT(_dbt3->AddDistinctColumn("string1"));
  CPPUNIT_ASSERT(_dbt3->LoadFromDB());
  CPPUNIT_ASSERT(_dbt3->NRow() == 2);

  _dbt3->Clear();
  CPPUNIT_ASSERT(_dbt3->AddDistinctColumn("string2"));
  CPPUNIT_ASSERT(_dbt3->LoadFromDB());
  CPPUNIT_ASSERT(_dbt3->NRow() == 3);

  _dbt3->Clear();
  CPPUNIT_ASSERT(_dbt3->AddDistinctColumn("string1"));
  CPPUNIT_ASSERT(_dbt3->AddDistinctColumn("string2"));
  CPPUNIT_ASSERT(_dbt3->LoadFromDB());
  CPPUNIT_ASSERT(_dbt3->NRow() == 3);

  _dbt3->Clear();
  CPPUNIT_ASSERT(_dbt3->AddOrderColumn("channel"));
  CPPUNIT_ASSERT(_dbt3->AddOrderColumn("value2"));
  CPPUNIT_ASSERT(_dbt3->LoadFromDB());

  std::vector<int> id;
  int tid;
  for (int i=0; i<_dbt3->NRow(); ++i) {
    CPPUNIT_ASSERT(_dbt3->GetRow(i)->Col(0).Get(tid));    
    id.push_back(tid);
  }
  unsigned int sz = id.size();
  CPPUNIT_ASSERT(id[sz-3]==2 && id[sz-2]==3 && id[sz-1]==4);
  
}

void
DevelopmentTests::testTableCreationAndDestruction()
{
  boost::shared_ptr<Table> localDBTablePtr;

  std::cout << "Table creation test:" << std::endl;
  for (int idx = 0; idx < 200; ++idx) {
      //std::cout << "Table creation test, version 1, index "
      //          << idx << std::endl;
      localDBTablePtr.reset(new Table(TABLE3_DEFINITION_FILE.c_str()));
      localDBTablePtr->SetDetector(novadaq::cnv::kNDOS);
      CPPUNIT_ASSERT(localDBTablePtr->ExistsInDB());
  }

  std::cout << "Table connection test:" << std::endl;
  for (int idx = 0; idx < 200; ++idx) {
      //std::cout << "Table creation test, version 1, index "
      //          << idx << std::endl;
      localDBTablePtr.reset(new Table(TABLE3_DEFINITION_FILE.c_str()));
      localDBTablePtr->SetDetector(novadaq::cnv::kNDOS);
      localDBTablePtr->GetConnection();
      CPPUNIT_ASSERT(localDBTablePtr->ExistsInDB());
  }
}

void
DevelopmentTests::populateDBTable()
{
  std::cout << "Populate dB table test:" << std::endl;
  // create the table that is used in these tests
  _dbt3.reset(new Table(TABLE3_DEFINITION_FILE.c_str()));
  _dbt3->SetDetector(novadaq::cnv::kNDOS);
  CPPUNIT_ASSERT(_dbt3->ExistsInDB());
  _dbt3->Clear();
  CPPUNIT_ASSERT(_dbt3->LoadFromDB());

  // create some filler rows, if needed
  int nRow = _dbt3->NRow();
  if (nRow == 0) {
    boost::shared_ptr<Row> rowPtr(_dbt3->NewRow());
    CPPUNIT_ASSERT(rowPtr->Set("channel", 101));
    CPPUNIT_ASSERT(rowPtr->Set("value1", 5.0));
    CPPUNIT_ASSERT(rowPtr->Set("value2", 42.6));
    CPPUNIT_ASSERT(rowPtr->Set("string1", "one"));
    CPPUNIT_ASSERT(rowPtr->Set("string2", "two"));
    _dbt3->AddRow(rowPtr.get());
    CPPUNIT_ASSERT(_dbt3->WriteToDB());
    rowPtr.reset(_dbt3->NewRow());
    CPPUNIT_ASSERT(rowPtr->Set("channel", 102));
    CPPUNIT_ASSERT(rowPtr->Set("value1", 5.0));
    CPPUNIT_ASSERT(rowPtr->Set("value2", 42.7));
    CPPUNIT_ASSERT(rowPtr->Set("string1", "one"));
    CPPUNIT_ASSERT(rowPtr->Set("string2", "four"));
    _dbt3->AddRow(rowPtr.get());
    CPPUNIT_ASSERT(_dbt3->WriteToDB());
    rowPtr.reset(_dbt3->NewRow());
    CPPUNIT_ASSERT(rowPtr->Set("channel", 103));
    CPPUNIT_ASSERT(rowPtr->Set("value1", 5.0));
    CPPUNIT_ASSERT(rowPtr->Set("value2", 42.8));
    CPPUNIT_ASSERT(rowPtr->Set("string1", "one"));
    CPPUNIT_ASSERT(rowPtr->Set("string2", "two"));
    _dbt3->AddRow(rowPtr.get());
    CPPUNIT_ASSERT(_dbt3->WriteToDB());
  }

  // clear the Table object as a nicety
  _dbt3->Clear();
}

void
DevelopmentTests::testDBCommandCache()
{
  std::cout << "dB command cache test:" << std::endl;
  // create the table that is used in these tests
  _dbt3.reset(new Table(TABLE3_DEFINITION_FILE.c_str()));
  _dbt3->SetDetector(novadaq::cnv::kNDOS);
  CPPUNIT_ASSERT(_dbt3->ExistsInDB());
  _dbt3->Clear();
  CPPUNIT_ASSERT(_dbt3->LoadFromDB());

  _dbt3->LoadFromDB();
  int initialRowCount = _dbt3->NRow();
  int channel = 3 * (initialRowCount + 1);
  time_t now = time(0);
  int tmpVal = now % 10000;
  float value1 = (float) (((double) tmpVal) / 100.0);
  float value2 = (float) (((double) tmpVal) / 357.0);

  // create the test row and store it in the database
  boost::shared_ptr<Row> rowPtr(_dbt3->NewRow());
  CPPUNIT_ASSERT(rowPtr->Set("channel", channel));
  CPPUNIT_ASSERT(rowPtr->Set("value1", value1));
  CPPUNIT_ASSERT(rowPtr->Set("value2", value2));
  CPPUNIT_ASSERT(rowPtr->Set("string1", "three"));
  CPPUNIT_ASSERT(rowPtr->Set("string2", "two"));
  _dbt3->AddRow(rowPtr.get());

  _dbt3->SetUser("dummy");
  CPPUNIT_ASSERT(_dbt3->WriteToDB());
  _dbt3->SetUser(getenv("USER"));

  // JMP NOTE: need to add some code to check that cache file was created/appended to
  // for now, I'm doing this by hand...

  _dbt3->Clear();
}

void
DevelopmentTests::testUserName()
{
  std::cout << "\ntestUserName\n" << std::endl;
  sleep(2);
  // create the table that is used in these tests
  _dbt3.reset(new Table(TABLE3_DEFINITION_FILE.c_str()));
  _dbt3->SetDetector(novadaq::cnv::kNDOS);
  CPPUNIT_ASSERT(_dbt3->ExistsInDB());
  _dbt3->Clear();
  CPPUNIT_ASSERT(_dbt3->LoadFromDB());

  _dbt3->LoadFromDB();
  int initialRowCount = _dbt3->NRow();
  int channel = 3 * (initialRowCount + 1);
  time_t now = time(0);
  int tmpVal = now % 10000;
  float value1 = (float) (((double) tmpVal) / 100.0);
  float value2 = (float) (((double) tmpVal) / 357.0);

  // create the test row and store it in the database
  boost::shared_ptr<Row> rowPtr(_dbt3->NewRow());
  CPPUNIT_ASSERT(rowPtr->Set("channel", channel));
  CPPUNIT_ASSERT(rowPtr->Set("value1", value1));
  CPPUNIT_ASSERT(rowPtr->Set("value2", value2));
  CPPUNIT_ASSERT(rowPtr->Set("string1", "three"));
  CPPUNIT_ASSERT(rowPtr->Set("string2", "two"));
  _dbt3->AddRow(rowPtr.get());

  _dbt3->SetIgnoreEnvVar(true);
  _dbt3->SetUser("nova_reader");
  _dbt3->SetPasswordFile("./unittest/nova_reader.dat");
  bool didIt = (! _dbt3->WriteToDB());
  if (didIt) std::cout << "[INSERT should have failed]" << std::endl;
  CPPUNIT_ASSERT(didIt);
  _dbt3->SetUser(getenv("USER"));
  _dbt3->DisablePasswordAccess();
  CPPUNIT_ASSERT(_dbt3->WriteToDB());
  CPPUNIT_ASSERT(_dbt3->GetConnection());
  _dbt3->SetUser("nova_reader");
  rowPtr->Set("channel", channel+1);
  rowPtr->Set("value1", value1+10.);
  rowPtr->Set("value2", value2+10.);
  rowPtr->Set("string1", "three+");
  rowPtr->Set("string2", "two++");
  _dbt3->AddRow(rowPtr.get());
  CPPUNIT_ASSERT(_dbt3->WriteToDB());
  CPPUNIT_ASSERT(_dbt3->CloseConnection());

  _dbt3->Clear();
}

void
DevelopmentTests::testEmptyStringVsNULL()
{
  _dbt3->Clear();
  std::cout << "Empty String vs. NULL Test:" << std::endl;
  Row* rowPtr = _dbt3->NewRow();

  CPPUNIT_ASSERT(rowPtr->Set("channel", 1024));
  CPPUNIT_ASSERT(rowPtr->Set("value1", 3.1415));
  CPPUNIT_ASSERT(rowPtr->Set("value2", 0.01));
  CPPUNIT_ASSERT(rowPtr->Set("string1", ""));
  CPPUNIT_ASSERT(rowPtr->Set("string2", "NULL"));
  _dbt3->AddRow(rowPtr);
  CPPUNIT_ASSERT(_dbt3->WriteToDB());
  delete rowPtr;

  _dbt3->Clear();
  CPPUNIT_ASSERT(_dbt3->LoadFromDB());
  int nRows = _dbt3->NRow();
  rowPtr = _dbt3->GetRow(nRows-1);
  std::string testStr;
  CPPUNIT_ASSERT(rowPtr->Col("string1")->Get(testStr));
  CPPUNIT_ASSERT(testStr == "");
  std::cout << "\"string1\" is correctly read back as an empty string." << std::endl;
  CPPUNIT_ASSERT(rowPtr->Col("string2")->IsNull());
  std::cout << "\"string2\" is correctly read back as NULL." << std::endl;
  _dbt3->Clear();
}

void
DevelopmentTests::testUpdateTwice()
{
  _dbt3->Clear();
  _dbt3->SetUser(getenv("USER"));
  std::cout << "Update-Twice Test:" << std::endl;
  Row* rowPtr = _dbt3->NewRow();

  CPPUNIT_ASSERT(rowPtr->Set("channel", 2048));
  CPPUNIT_ASSERT(rowPtr->Set("value1", 929.929));
  CPPUNIT_ASSERT(rowPtr->Set("value2", 0.929));
  CPPUNIT_ASSERT(rowPtr->Set("string1", "929"));
  CPPUNIT_ASSERT(rowPtr->Set("string2", "92929"));
  _dbt3->AddRow(rowPtr);
  CPPUNIT_ASSERT(_dbt3->WriteToDB());
  delete rowPtr;

  _dbt3->Clear();
  CPPUNIT_ASSERT(_dbt3->LoadFromDB());
  int nRows = _dbt3->NRow();
  CPPUNIT_ASSERT(_dbt3->GetRow(nRows-1)->Update("channel",2049));
  CPPUNIT_ASSERT(_dbt3->WriteToDB());
  _dbt3->Clear();
  CPPUNIT_ASSERT(_dbt3->LoadFromDB());
  int myChan;
  CPPUNIT_ASSERT(_dbt3->GetRow(nRows-1)->Col("channel")->Get(myChan));
  CPPUNIT_ASSERT(myChan == 2049);
  std::string updateUser, updateTime;
  CPPUNIT_ASSERT(_dbt3->GetRow(nRows-1)->Col("updateuser")->Get(updateUser));
  CPPUNIT_ASSERT(_dbt3->GetRow(nRows-1)->Col("updatetime")->Get(updateTime));
  sleep(2);
  CPPUNIT_ASSERT(_dbt3->GetRow(nRows-1)->Update("channel",2050));
  CPPUNIT_ASSERT(_dbt3->WriteToDB());
  _dbt3->Clear();
  CPPUNIT_ASSERT(_dbt3->LoadFromDB());
  CPPUNIT_ASSERT(_dbt3->GetRow(nRows-1)->Col("channel")->Get(myChan));
  CPPUNIT_ASSERT(myChan == 2050);
  std::string updateUser_save = updateUser;
  std::string updateTime_save = updateTime;
  CPPUNIT_ASSERT(_dbt3->GetRow(nRows-1)->Col("updateuser")->Get(updateUser));
  CPPUNIT_ASSERT(_dbt3->GetRow(nRows-1)->Col("updatetime")->Get(updateTime));
  CPPUNIT_ASSERT(updateUser_save == updateUser);
  std::cout << "Two updatetimes: " << updateTime_save << ", " 
	    << updateTime << std::endl;
  
}

void
DevelopmentTests::cleanupDBTable()
{
  _dbt3.reset();
}

CPPUNIT_TEST_SUITE_NAMED_REGISTRATION(DevelopmentTests, "NovaDatabase");
CPPUNIT_REGISTRY_ADD_TO_DEFAULT("NovaDatabase");
