#include <cppunit/extensions/HelperMacros.h>
#include <NovaTimingUtilities/TimingUtilities.h>

namespace NDTU = novadaq::timeutils;

class TimingUtilitiesTests : public CppUnit::TestFixture
{
    CPPUNIT_TEST_SUITE(TimingUtilitiesTests);
    CPPUNIT_TEST(testBasicFunctions);
    CPPUNIT_TEST(testSecMinHourDayMonYear);
    CPPUNIT_TEST(testTMStruct);
    CPPUNIT_TEST(testLeapSeconds);
    CPPUNIT_TEST_SUITE_END();

public:
    void setUp();
    void tearDown();

    void testBasicFunctions();
    void testSecMinHourDayMonYear();
    void testTMStruct();
    void testLeapSeconds();

private:

};

void
TimingUtilitiesTests::setUp()
{ 
    // flush stdout to get progress markers to display in a timely way
    std::cout.flush();
}

void
TimingUtilitiesTests::tearDown()
{ 
    // flush stdout to get progress markers to display in a timely way
    std::cout.flush();
}

void
TimingUtilitiesTests::testBasicFunctions()
{
    struct timeval unixTime;
    uint64_t novaTime;

    // test the conversion of NOvA time to UNIX time at boundary conditions
    novaTime = 0;
    CPPUNIT_ASSERT(NDTU::convertNovaTimeToUnixTime(novaTime, unixTime));
    CPPUNIT_ASSERT(unixTime.tv_sec == 1262304000);
    CPPUNIT_ASSERT(unixTime.tv_usec == 0);
    novaTime = 64000000;
    CPPUNIT_ASSERT(NDTU::convertNovaTimeToUnixTime(novaTime, unixTime));
    CPPUNIT_ASSERT(unixTime.tv_sec == (1 + 1262304000));
    CPPUNIT_ASSERT(unixTime.tv_usec == 0);

    // test a round trip conversion, novaTime -> unixTime -> novaTime
    // (Note that the starting time needs to be something that can be
    // accurately represented in both units)
    uint64_t sampleNovaTime = (uint64_t)((double) 64000000 * 123456789.5);
    CPPUNIT_ASSERT(NDTU::convertNovaTimeToUnixTime(sampleNovaTime, unixTime));
    CPPUNIT_ASSERT(NDTU::convertUnixTimeToNovaTime(unixTime, novaTime));
    CPPUNIT_ASSERT(novaTime == sampleNovaTime);

    // test the fetching of the current NOvA time
    gettimeofday(&unixTime, 0);
    uint64_t beforeTime;
    CPPUNIT_ASSERT(NDTU::convertUnixTimeToNovaTime(unixTime, beforeTime));
    uint64_t currentNovaTime;
    CPPUNIT_ASSERT(NDTU::getCurrentNovaTime(currentNovaTime));
    gettimeofday(&unixTime, 0);
    uint64_t afterTime;
    CPPUNIT_ASSERT(NDTU::convertUnixTimeToNovaTime(unixTime, afterTime));
    CPPUNIT_ASSERT(beforeTime <= currentNovaTime);
    CPPUNIT_ASSERT(currentNovaTime <= afterTime);
}

void
TimingUtilitiesTests::testSecMinHourDayMonYear()
{
  uint16_t sec = 0;
  uint16_t min = 0;
  uint16_t hour = 0;
  uint16_t mday = 1;
  uint16_t mon = 0;
  uint32_t year = 110;
  uint64_t novaTime;
  CPPUNIT_ASSERT(NDTU::convertUnixTimeToNovaTime(sec,min,hour,mday,mon,year, novaTime));
  CPPUNIT_ASSERT(novaTime == 0);

  uint16_t osec, omin, ohour, omday, omon;
  uint32_t oyear;
  CPPUNIT_ASSERT(NDTU::convertNovaTimeToUnixTime(novaTime,osec,omin,ohour,omday,omon,oyear));
  CPPUNIT_ASSERT(sec == osec);
  CPPUNIT_ASSERT(min == omin);
  CPPUNIT_ASSERT(hour == ohour);
  CPPUNIT_ASSERT(mday == omday);
  CPPUNIT_ASSERT(mon == omon);
  CPPUNIT_ASSERT(year == oyear);

  // A time earlier than nova epoch should fail to convert
  sec = 0;
  min = 0;
  hour = 0;
  mday = 1;
  mon = 0;
  year = 109;
  CPPUNIT_ASSERT(NDTU::convertUnixTimeToNovaTime(sec,min,hour,mday,mon,year,novaTime) == false);

  // try Feb 4, 2010, 10:11:12 (UTC) 
  sec = 12;
  min = 11;
  hour = 10;
  mday = 4;
  mon = 1;
  year = 110;
  CPPUNIT_ASSERT(NDTU::convertUnixTimeToNovaTime(sec,min,hour,mday,mon,year,novaTime));
  uint64_t novaTimeCompare = 2974272llu*64000000llu;
  CPPUNIT_ASSERT(novaTime == novaTimeCompare);
  CPPUNIT_ASSERT(NDTU::convertNovaTimeToUnixTime(novaTime,osec,omin,ohour,omday,omon,oyear));
  CPPUNIT_ASSERT(sec == osec);
  CPPUNIT_ASSERT(min == omin);
  CPPUNIT_ASSERT(hour == ohour);
  CPPUNIT_ASSERT(mday == omday);
  CPPUNIT_ASSERT(mon == omon);
  CPPUNIT_ASSERT(year == oyear);
 
}

void
TimingUtilitiesTests::testTMStruct()
{
 
  tm timeTM; 
  timeTM.tm_sec = 0;
  timeTM.tm_min = 0;
  timeTM.tm_hour = 0;
  timeTM.tm_mday = 1;
  timeTM.tm_mon = 0;
  timeTM.tm_year = 110;
  uint64_t novaTime;
  CPPUNIT_ASSERT(NDTU::convertUnixTimeToNovaTime(timeTM, novaTime));
  CPPUNIT_ASSERT(novaTime == 0);

  tm outputTimeTM;
  CPPUNIT_ASSERT(NDTU::convertNovaTimeToUnixTime(novaTime,outputTimeTM));

  CPPUNIT_ASSERT(timeTM.tm_sec == outputTimeTM.tm_sec);
  CPPUNIT_ASSERT(timeTM.tm_min == outputTimeTM.tm_min);
  CPPUNIT_ASSERT(timeTM.tm_hour == outputTimeTM.tm_hour);
  CPPUNIT_ASSERT(timeTM.tm_mday == outputTimeTM.tm_mday);
  CPPUNIT_ASSERT(timeTM.tm_mon == outputTimeTM.tm_mon);
  CPPUNIT_ASSERT(timeTM.tm_year == outputTimeTM.tm_year);

  // A time earlier than nova epoch should fail to convert
  timeTM.tm_sec = 0;
  timeTM.tm_min = 0;
  timeTM.tm_hour = 0;
  timeTM.tm_mday = 1;
  timeTM.tm_mon = 0;
  timeTM.tm_year = 109;
  CPPUNIT_ASSERT(NDTU::convertUnixTimeToNovaTime(timeTM,novaTime) == false);

  // try Feb 4, 2010, 10:11:12 (UTC) 
  timeTM.tm_sec = 12;
  timeTM.tm_min = 11;
  timeTM.tm_hour = 10;
  timeTM.tm_mday = 4;
  timeTM.tm_mon = 1;
  timeTM.tm_year = 110;
  CPPUNIT_ASSERT(NDTU::convertUnixTimeToNovaTime(timeTM,novaTime));
  uint64_t novaTimeCompare = 2974272llu*64000000llu;
  CPPUNIT_ASSERT(novaTime == novaTimeCompare);
  CPPUNIT_ASSERT(NDTU::convertNovaTimeToUnixTime(novaTime,outputTimeTM));

  CPPUNIT_ASSERT(timeTM.tm_sec == outputTimeTM.tm_sec);
  CPPUNIT_ASSERT(timeTM.tm_min == outputTimeTM.tm_min);
  CPPUNIT_ASSERT(timeTM.tm_hour == outputTimeTM.tm_hour);
  CPPUNIT_ASSERT(timeTM.tm_mday == outputTimeTM.tm_mday);
  CPPUNIT_ASSERT(timeTM.tm_mon == outputTimeTM.tm_mon);
  CPPUNIT_ASSERT(timeTM.tm_year == outputTimeTM.tm_year);
 
}

void
TimingUtilitiesTests::testLeapSeconds()
{
    struct timespec unixTimeSpec;
    struct timeval unixTimeVal;
    struct tm unixTMTime;
    uint64_t novaTime;
    uint64_t expectedNovaTime;
    struct timespec reconstructedUnixTimeSpec;
    struct timeval reconstructedUnixTimeVal;
    struct tm reconstructedUnixTMTime;

    // ****************************************
    // *** One second before the time of the first leap second
    // *** 30-Jun-2012 23:59:59.0
    // ****************************************
    expectedNovaTime = 5042995136000000;

    unixTMTime.tm_sec = 59;
    unixTMTime.tm_min = 59;
    unixTMTime.tm_hour = 23;
    unixTMTime.tm_mday = 30;
    unixTMTime.tm_mon = 5;
    unixTMTime.tm_year = 112;
    CPPUNIT_ASSERT(NDTU::convertUnixTimeToNovaTime(unixTMTime, novaTime));
    CPPUNIT_ASSERT(novaTime == expectedNovaTime);

    CPPUNIT_ASSERT(NDTU::convertNovaTimeToUnixTime(novaTime,
                                                   reconstructedUnixTMTime));
    CPPUNIT_ASSERT(reconstructedUnixTMTime.tm_sec == unixTMTime.tm_sec);
    CPPUNIT_ASSERT(reconstructedUnixTMTime.tm_min == unixTMTime.tm_min);
    CPPUNIT_ASSERT(reconstructedUnixTMTime.tm_hour == unixTMTime.tm_hour);
    CPPUNIT_ASSERT(reconstructedUnixTMTime.tm_mday == unixTMTime.tm_mday);
    CPPUNIT_ASSERT(reconstructedUnixTMTime.tm_mon == unixTMTime.tm_mon);
    CPPUNIT_ASSERT(reconstructedUnixTMTime.tm_year == unixTMTime.tm_year);

    CPPUNIT_ASSERT(NDTU::convertUnixTimeToNovaTime(59, 59, 23, 30, 5, 112,
                                                   novaTime));
    CPPUNIT_ASSERT(novaTime == expectedNovaTime);

    // ****************************************
    // *** Half a second before the time of the first leap second
    // *** 30-Jun-2012 23:59:59.5
    // ****************************************
    expectedNovaTime += 32000000;

    unixTimeSpec.tv_sec = 1341100799;
    unixTimeSpec.tv_nsec = 500000000;
    CPPUNIT_ASSERT(NDTU::convertUnixTimeToNovaTime(unixTimeSpec, novaTime));
    CPPUNIT_ASSERT(novaTime == expectedNovaTime);

    CPPUNIT_ASSERT(NDTU::convertNovaTimeToUnixTime(novaTime,
                                                   reconstructedUnixTimeSpec));
    CPPUNIT_ASSERT(reconstructedUnixTimeSpec.tv_sec == unixTimeSpec.tv_sec);
    CPPUNIT_ASSERT(reconstructedUnixTimeSpec.tv_nsec == unixTimeSpec.tv_nsec);

    unixTimeVal.tv_sec = 1341100799;
    unixTimeVal.tv_usec = 500000;
    CPPUNIT_ASSERT(NDTU::convertUnixTimeToNovaTime(unixTimeVal, novaTime));
    CPPUNIT_ASSERT(novaTime == expectedNovaTime);

    CPPUNIT_ASSERT(NDTU::convertNovaTimeToUnixTime(novaTime,
                                                   reconstructedUnixTimeVal));
    CPPUNIT_ASSERT(reconstructedUnixTimeVal.tv_sec == unixTimeVal.tv_sec);
    CPPUNIT_ASSERT(reconstructedUnixTimeVal.tv_usec == unixTimeVal.tv_usec);

    // ****************************************
    // *** Exactly at the time of the first leap second
    // *** 01-Ju1-2012 00:00:00.0
    // ****************************************
    expectedNovaTime += 32000000 + 64000000;

    unixTimeSpec.tv_sec = 1341100800;
    unixTimeSpec.tv_nsec = 0;
    CPPUNIT_ASSERT(NDTU::convertUnixTimeToNovaTime(unixTimeSpec, novaTime));
    CPPUNIT_ASSERT(novaTime == expectedNovaTime);

    CPPUNIT_ASSERT(NDTU::convertNovaTimeToUnixTime(novaTime,
                                                   reconstructedUnixTimeSpec));
    CPPUNIT_ASSERT(reconstructedUnixTimeSpec.tv_sec == unixTimeSpec.tv_sec);
    CPPUNIT_ASSERT(reconstructedUnixTimeSpec.tv_nsec == unixTimeSpec.tv_nsec);

    unixTimeVal.tv_sec = 1341100800;
    unixTimeVal.tv_usec = 0;
    CPPUNIT_ASSERT(NDTU::convertUnixTimeToNovaTime(unixTimeVal, novaTime));
    CPPUNIT_ASSERT(novaTime == expectedNovaTime);

    CPPUNIT_ASSERT(NDTU::convertNovaTimeToUnixTime(novaTime,
                                                   reconstructedUnixTimeVal));
    CPPUNIT_ASSERT(reconstructedUnixTimeVal.tv_sec == unixTimeVal.tv_sec);
    CPPUNIT_ASSERT(reconstructedUnixTimeVal.tv_usec == unixTimeVal.tv_usec);

    unixTMTime.tm_sec = 0;
    unixTMTime.tm_min = 0;
    unixTMTime.tm_hour = 0;
    unixTMTime.tm_mday = 1;
    unixTMTime.tm_mon = 6;
    unixTMTime.tm_year = 112;
    CPPUNIT_ASSERT(NDTU::convertUnixTimeToNovaTime(unixTMTime, novaTime));
    CPPUNIT_ASSERT(novaTime == expectedNovaTime);

    CPPUNIT_ASSERT(NDTU::convertNovaTimeToUnixTime(novaTime,
                                                   reconstructedUnixTMTime));
    CPPUNIT_ASSERT(reconstructedUnixTMTime.tm_sec == unixTMTime.tm_sec);
    CPPUNIT_ASSERT(reconstructedUnixTMTime.tm_min == unixTMTime.tm_min);
    CPPUNIT_ASSERT(reconstructedUnixTMTime.tm_hour == unixTMTime.tm_hour);
    CPPUNIT_ASSERT(reconstructedUnixTMTime.tm_mday == unixTMTime.tm_mday);
    CPPUNIT_ASSERT(reconstructedUnixTMTime.tm_mon == unixTMTime.tm_mon);
    CPPUNIT_ASSERT(reconstructedUnixTMTime.tm_year == unixTMTime.tm_year);

    CPPUNIT_ASSERT(NDTU::convertUnixTimeToNovaTime(0, 0, 0, 1, 6, 112,
                                                   novaTime));
    CPPUNIT_ASSERT(novaTime == expectedNovaTime);

    // ****************************************
    // *** Half a second after the time of the first leap second
    // *** 01-Ju1-2012 00:00:00.5
    // ****************************************
    expectedNovaTime += 32000000;

    unixTimeSpec.tv_sec = 1341100800;
    unixTimeSpec.tv_nsec = 500000000;
    CPPUNIT_ASSERT(NDTU::convertUnixTimeToNovaTime(unixTimeSpec, novaTime));
    CPPUNIT_ASSERT(novaTime == expectedNovaTime);

    CPPUNIT_ASSERT(NDTU::convertNovaTimeToUnixTime(novaTime,
                                                   reconstructedUnixTimeSpec));
    CPPUNIT_ASSERT(reconstructedUnixTimeSpec.tv_sec == unixTimeSpec.tv_sec);
    CPPUNIT_ASSERT(reconstructedUnixTimeSpec.tv_nsec == unixTimeSpec.tv_nsec);

    unixTimeVal.tv_sec = 1341100800;
    unixTimeVal.tv_usec = 500000;
    CPPUNIT_ASSERT(NDTU::convertUnixTimeToNovaTime(unixTimeVal, novaTime));
    CPPUNIT_ASSERT(novaTime == expectedNovaTime);

    CPPUNIT_ASSERT(NDTU::convertNovaTimeToUnixTime(novaTime,
                                                   reconstructedUnixTimeVal));
    CPPUNIT_ASSERT(reconstructedUnixTimeVal.tv_sec == unixTimeVal.tv_sec);
    CPPUNIT_ASSERT(reconstructedUnixTimeVal.tv_usec == unixTimeVal.tv_usec);

    // ****************************************
    // *** One second after the time of the first leap second
    // *** 01-Ju1-2012 00:00:01.0
    // ****************************************
    expectedNovaTime += 32000000;

    unixTMTime.tm_sec = 1;
    unixTMTime.tm_min = 0;
    unixTMTime.tm_hour = 0;
    unixTMTime.tm_mday = 1;
    unixTMTime.tm_mon = 6;
    unixTMTime.tm_year = 112;
    CPPUNIT_ASSERT(NDTU::convertUnixTimeToNovaTime(unixTMTime, novaTime));
    CPPUNIT_ASSERT(novaTime == expectedNovaTime);

    CPPUNIT_ASSERT(NDTU::convertNovaTimeToUnixTime(novaTime,
                                                   reconstructedUnixTMTime));
    CPPUNIT_ASSERT(reconstructedUnixTMTime.tm_sec == unixTMTime.tm_sec);
    CPPUNIT_ASSERT(reconstructedUnixTMTime.tm_min == unixTMTime.tm_min);
    CPPUNIT_ASSERT(reconstructedUnixTMTime.tm_hour == unixTMTime.tm_hour);
    CPPUNIT_ASSERT(reconstructedUnixTMTime.tm_mday == unixTMTime.tm_mday);
    CPPUNIT_ASSERT(reconstructedUnixTMTime.tm_mon == unixTMTime.tm_mon);
    CPPUNIT_ASSERT(reconstructedUnixTMTime.tm_year == unixTMTime.tm_year);

    CPPUNIT_ASSERT(NDTU::convertUnixTimeToNovaTime(1, 0, 0, 1, 6, 112,
                                                   novaTime));
    CPPUNIT_ASSERT(novaTime == expectedNovaTime);
}

CPPUNIT_TEST_SUITE_NAMED_REGISTRATION(TimingUtilitiesTests, "NovaTimingUtilities");
CPPUNIT_REGISTRY_ADD_TO_DEFAULT("NovaTimingUtilities");
