# this function processes the cstxsd files
# it will fill the variable name CSTXSD_LIB_SOURCES with the 
# processed .cpp files

# Rules for generating C++ classes from XML schema definitions using
# the CodeSynthesis XSD package.
#
# The variables that are used in this file, but need to be defined outside
# of this file, are listed here.
#
# Variables that *should* be defined in arch_spec_cstxsd.mk (or its children)
# (of course, these variables can be overridden elsewhere):
# - CSTXSD_COMPILE - the application that generates the classes from the XSD files
# - CSTXSD_COMPILE_OPTIONS - the options that should be passed to the "compiler"
# - CSTXSD_TAILOR - the script that tailors the generated files for local use
#
# Variables that *should* be defined in the makefile that includes this file:
# - CSTXSD_XSD_DIR - the directory that contains XML schema definition files
# - CSTXSD_XSD_FILES - the list of XML Schema Definition files that should be
#                      used to create message classes (including paths)
# - CSTXSD_INC_PREFIX - the prefix that should be used to include the
#                       generated header file in the generated source file
#
# Variables that *can* be defined in the makefile that includes this file:
# - CSTXSD_NAMESPACE - the namespace that should be used for the generated classes
# - CSTXSD_EXTRA_OPTIONS - any options that need to be passed to the XSD
#                          compiler but are not covered by other variables.
# - CSTXSD_TAILOR_INC_OPTIONS - the options to be passed to the "tailor" script
#                               for the header files, if any.  If this variable is
#                               not set and the CSTXSD_RUN_TAILOR_SCRIPT described
#                               below is not set, then no tailoring of header
#                               files will take place.
# - CSTXSD_TAILOR_SRC_OPTIONS - the options to be passed to the "tailor" script
#                               for the source files, if any.  If this variable is
#                               not set and the CSTXSD_RUN_TAILOR_SCRIPT described
#                               below is not set, then no tailoring of source
#                               files will take place.
# - CSTXSD_RUN_TAILOR_SCRIPT - a flag to indicate that the tailor script should
#                              be run on header and source files even though no
#                              header- or source-specific options are needed.  
#                              Setting this variable to any non-empty value
#                              enables the tailoring.
# Just a short note on tailoring:  as of cstXSD 3.2.0, the local "tailoring"
# that we support for the C++ code that CodeSynthesis XSD generates is the
# addition of some accessor methods, addition of inheritance from a common base
# class, addition of convenience methods for serialization and deserialization,
# and addition of warning messages when class instances are serialized but some
# of the non-optional elements have not been specified.
#

function(process_cstxsd CSTXSD_LIB_SOURCES CSTXSD_XSD_DIR CSTXSD_NAMESPACE CSTXSD_INC_PREFIX)

  # get the collection of .xsd files
  FILE( GLOB xsd_files ${CSTXSD_XSD_DIR}/*.xsd )
  MESSAGE("XSD files are ${xsd_files} from ${CSTXSD_XSD_DIR}")

  # set the location fo the tailoring script
  SET(CSTXSD_TAILOR "${TOPLEVEL_NOVADAQ_SRC_DIR}/NovaDAQUtilities/tools/tailorXsdGeneratedClassesForNova.pl" )

  # loop over all .xsd files 
  foreach(source ${xsd_files})
    #define some local variables for the produced .h and .cpp files
    get_filename_component(XSD_BASENAME     ${source} NAME_WE)
    get_filename_component(XSD_BASENAME_EXT ${source} NAME)
    string(REPLACE ".xsd" ".cpp" XSD_CPP_OUTPUT ${XSD_BASENAME_EXT})
    string(REPLACE ".xsd" ".h"   XSD_H_OUTPUT   ${XSD_BASENAME_EXT})

    MESSAGE("generating source for ${XSD_BASENAME}")
    
    set(LOG_NAME ${XSD_BASENAME}_log.txt)

    # add the .cpp file to the list of sources
    list(APPEND TMP_SOURCES ${XSD_CPP_OUTPUT})

    # call xsd and generate the source and header files
    # then tailor the files.  Move the .h file into the 
    # top level directory for use in compiling
    add_custom_command(OUTPUT ${CMAKE_CURRENT_BINARY_DIR}/${XSD_CPP_OUTPUT} 
    			      ${CMAKE_CURRENT_BINARY_DIR}/../../${XSD_H_OUTPUT}
      COMMAND $ENV{CSTXSD_FQ_DIR}/bin/xsd cxx-tree --generate-serialization --generate-default-ctor --hxx-suffix .h --cxx-suffix .cpp --include-with-brackets --root-element-all --output-dir ${CMAKE_CURRENT_BINARY_DIR} --include-prefix ${CSTXSD_INC_PREFIX} --namespace-map =${CSTXSD_NAMESPACE} ${source} 
      COMMAND echo "Tailoring ${XSD_CPP_OUTPUT}" 
      COMMAND ${CSTXSD_TAILOR} ${CMAKE_CURRENT_BINARY_DIR}/${XSD_CPP_OUTPUT} && rm ${CMAKE_CURRENT_BINARY_DIR}/${XSD_CPP_OUTPUT}.orig 
      COMMAND echo "Tailoring ${XSD_H_OUTPUT}"
      COMMAND ${CSTXSD_TAILOR} ${CMAKE_CURRENT_BINARY_DIR}/${XSD_H_OUTPUT} && rm ${CMAKE_CURRENT_BINARY_DIR}/${XSD_H_OUTPUT}.orig
      COMMAND ${CMAKE_COMMAND} -E copy ${CMAKE_CURRENT_BINARY_DIR}/${XSD_H_OUTPUT} ${CMAKE_CURRENT_BINARY_DIR}/../../${XSD_H_OUTPUT}
      COMMAND ${CMAKE_COMMAND} -E remove ${CMAKE_CURRENT_BINARY_DIR}/${XSD_H_OUTPUT}
    )
    #install the header and source files
    install_source(LIST ${CMAKE_CURRENT_BINARY_DIR}/${XSD_CPP_OUTPUT})
    install(FILES ${CMAKE_CURRENT_BINARY_DIR}/../../${XSD_H_OUTPUT}
            DESTINATION ${${product}_inc_dir}/${CSTXSD_INC_PREFIX})
  endforeach()

  # set the output variable
  set(${CSTXSD_LIB_SOURCES} ${TMP_SOURCES} PARENT_SCOPE)
endfunction()
