//  This file (ShmMilliBlockTester.cc) was created by Ron Rechenmacher <ron@fnal.gov> on
//  Nov  7, 2011. "TERMS AND CONDITIONS" governing this file are in the README
//  or COPYING file. If you do not have such a file, one can be obtained by
//  contacting Ron or Fermi Lab in Batavia IL, 60510, phone: 630-840-3000.
//  $RCSfile: ShmMilliBlockTester.cc,v $
//  rev="$Revision: 1.11 $$Date: 2013/12/19 05:32:25 $";

#include <stdio.h>		         // printf
#include <getopt.h>		// getopt_long
#include <libgen.h>             // basename
#include <unistd.h>		// getpid
#include "ShmMilliBlock/ShmMilliBlock.h"
#include "DAQDataFormats/RawMilliBlock.h"    // RawMilliBlock

#define DFLT_TMO_US	1000000

#define USAGE "\
   Usage: %s [opts]\n\
Examples: %s\n\
          %s --count=10 --dump\n\
 Options: --count=N\n\
          --dump\n\
          -i        - include a header\n\
          -t<tmo_us>  (dflt:%d)\n\
          -? -h --help  help/usage\n\
The output is:\n\
milliIdx microIdx diblock_id dcm_id feb_id num_nano micro_time delta_from_prv\n\
",      basename(argv[0]), basename(argv[0]), basename(argv[0]),\
	DFLT_TMO_US


int
main(  int	argc
     , char	*argv[] )
{
    ShmMilliBlock	shm_milliBlock;
    RawMilliBlock	mb;
    uint32_t		got_cnt=0, ungot_cnt=0;
    uint32_t		slice_cnt=0;
    uint32_t		tot_crc_errors=0;
    int			opt_dump=0, opt_count=-1, opt_header=0;
    int			opt_tmo_us=DFLT_TMO_US;
    pid_t		mypid=getpid();	// for when multiple programs are forked in one window

    while (1)
    {   int                  c, option_index=0;
        static struct option long_options[]=
            {   // name,    has_arg,flag,val
                {"count",    1,     0,  'c'},
                {"dump",     0,     0,  'd'},
		{"help",     0,     0,  'h'},
                {0, 0, 0, 0}
            };
        //                            +=stop parsing at first non-option
        c = getopt_long( argc, argv, "+?hc:dit:", long_options, &option_index );
        if (c == -1)  break;
        switch (c)
        {
        case '?': case 'h': printf(USAGE); exit(0);  break;
        case 'c': opt_count=strtoul(optarg,NULL,0);  break;
        case 't': opt_tmo_us=strtoul(optarg,NULL,0); break;
        case 'd': ++opt_dump;                        break;
	case 'i': opt_header=1;                      break;
        default:printf("?? getopt returned character code %d (0x%x) ??\n",c,c);
        }
    }

    char const* fmt;
    char const* hdr =      "#mil mic dib dcm feb nn time             delta_prv\n";
    if (opt_header==0) fmt= "%2d %2d %2d %2d %2d %2d 0x%014lx %ld\n";
    else               fmt=" %3d %3d %3d %3d %3d %2d 0x%014lx %ld\n";

    uint64_t prev=0;
    while (opt_count--)
    {
	if (shm_milliBlock.getMilliBlock( opt_tmo_us, mb ))
	{   int sz=mb.slices.size();
	    if (opt_dump==1)
	    {   uint64_t tThis=mb.getTimeStart();
		if (prev==0) prev=tThis;
		printf( "seqNum:%6d  mbTime:0x%014lx  dT:%09ld\n"
		       , mb.getSeqNum(), tThis, tThis-prev );
		prev=tThis;
	    }
	    slice_cnt += sz;
	    bool crc_chkd=false, crc_err=false;
	    for (int mm=0; mm<sz; ++mm)
	    {
		if (mb.slices[mm].isCRCCalculationUsed())
		{   crc_chkd = true;
		    uint32_t crc = mb.slices[mm].calculateCheckSum(0);
		    if (crc != 0) { crc_err=true; ++tot_crc_errors; }
		}
		if (opt_dump > 1)
		{   int64_t                        prv_micro_time=0;
		    daqdataformats::RawMicroSlice* uslcp;
		    if (opt_header) printf(hdr);
		    for (int uu=0; mb.slices[mm].setFloatingMicroSlice(uu); ++uu)
		    {   uslcp =    mb.slices[mm].getFloatingMicroSlice();
			int64_t micro_time=uslcp->getTime();
			daqdataformats::RawMilliSliceHeader* mslchp=mb.slices[mm].getHeader();
			uint32_t                  diblock_id=mslchp->getDiblockId();
			uint32_t                      dcm_id=mslchp->getDCMId();
			uint32_t    numnanos=uslcp->getNumNanoSlices();
			if (numnanos == 0)
			{
			    if (prv_micro_time==0)
				printf( fmt, mm, uu, diblock_id, dcm_id, -1, 0, micro_time, (long)0 );
			    else
				printf( fmt, mm, uu, diblock_id, dcm_id, -1, 0, micro_time, micro_time-prv_micro_time );
			}
			else
			{   int feb_id, prv_feb_id=-1;
			    int nn_each_feb=0;
			    for (uint32_t nn=0; nn<numnanos; ++nn)
			    {   uslcp->setFloatingNanoSlice(nn);
				daqdataformats::RawNanoSliceHeader* nslchp=uslcp->getFloatingNanoSlice()->getHeader();
				feb_id = nslchp->getFEBId();
				if ((prv_feb_id!=-1) && (feb_id!=prv_feb_id))
				{
				    if (prv_micro_time==0)
					printf( fmt, mm, uu, diblock_id, dcm_id, prv_feb_id, nn_each_feb, micro_time, (long)0 );
				    else
					printf( fmt, mm, uu, diblock_id, dcm_id, prv_feb_id, nn_each_feb, micro_time, micro_time-prv_micro_time );
				    nn_each_feb=0;
				}
				++nn_each_feb;
				prv_feb_id = feb_id;
			    }
			    if (prv_micro_time==0)
				printf( fmt, mm, uu, diblock_id, dcm_id, prv_feb_id, nn_each_feb, micro_time, (long)0 );
			    else
				printf( fmt, mm, uu, diblock_id, dcm_id, prv_feb_id, nn_each_feb, micro_time, micro_time-prv_micro_time );
			}
			prv_micro_time = micro_time;
		    }
		}
	    }
	    if (crc_chkd)
	    {
		if (crc_err) { printf("crc checked -- ERROR found\n"); }
		else         { printf("crc checked -- OK\n"); }
	    }

	    if (++got_cnt%100 == 0)
	    {   printf( "%6d got %u  crc_chkd=%d  ave slice_cnt=%f tot_crc_errors=%u\n"
		       , mypid,got_cnt, crc_chkd, (float)slice_cnt/100.0, tot_crc_errors );
		slice_cnt=0;
	    }
	}
	else
	{
	    if (++ungot_cnt%500 == 0) printf( "ungot %u tot_crc_errors=%u\n", ungot_cnt, tot_crc_errors );
	}
    }
    return (0);
}   // main
