#ifndef _RMS_SENDER_H
#define _RMS_SENDER_H

#include <rms/base/RmsCloseable.h>
#include <rms/base/RmsDestination.h>
#include <rms/base/RmsRuntimeException.h>

#include <boost/shared_ptr.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/interprocess/smart_ptr/unique_ptr.hpp>

#include <string>
#include <iostream>

namespace gov {

namespace fnal {

namespace cd {

namespace rms {

/**
 * Populates the correlation fields in the target message using the
 * appropriate fields in the source message.
 */
template <class S, class T>
void fillCorrelationFields(S const& sourceMessage, T& targetMessage) {
    targetMessage.header.correlationId = sourceMessage.header.id;
    targetMessage.header.correlationTime = sourceMessage.header.sentTime;
}

/**
 * This class contains functionality for sending messages of specified
 * types to a specified destination.
 *
 * @author Kurt Biery
 * @version $Revision: 1.9 $ $Date: 2011/07/28 21:02:06 $
 */

template <class C, class M>
class RmsSender : public base::RmsCloseable {

public:
    RmsSender(boost::shared_ptr<C> parentConnection, base::RmsDestination dest);
    ~RmsSender();

    void sendMessage(M& message);

    void close() throw();
    bool hasBeenClosed() throw(){return _hasBeenClosed;};
private:
    /**
     * Handle to the RMS connection
     */
    boost::shared_ptr<C> _myConnection;

    /**
     * RmsDestination that we are listening to
     */
    base::RmsDestination _myDestination;

    unsigned int _senderHandle;

    bool _hasBeenClosed;

};

/**
 * Constructs an RmsSender instance.
 *
 * @param parentConn Connection object which is used for
 * sending messages.
 * @param dest The destination that we are sending messages to.
 * @throws RmsRuntimeException If the destination is invalid.
 */
template <class C, class M> RmsSender<C, M>::
RmsSender(boost::shared_ptr<C> parentConn, base::RmsDestination dest) :
        _myConnection(parentConn),
        _myDestination(dest),
        _hasBeenClosed(false)
{
    _myDestination.validate();
    if (! _myConnection->supportsDestination(_myDestination)) {
        std::string msg;
        msg.append("The current connection does not support ");
        msg.append("the requested destination: ");
        msg.append(_myDestination.toString());
        GENERATE_RMS_RUNTIME_EXCEPTION(msg);
    }

    M dummyObj;
    _senderHandle = _myConnection->registerSender(dummyObj, _myDestination);
}

/**
 * Destructs an RmsSender instance.
 */
template <class C, class M> RmsSender<C, M>::~RmsSender()
{
    try {
        this->close();
    }
    catch (...) {
        std::cerr << "Exception in RmsSender destructor." << std::endl;
    }
}

/**
 * Sends the specified message.  The message argument is modified by this
 * method - "standard" message fields are filled in as a convenience.
 *
 * @param message The message to be sent.
 * @throws RmsCommunicationException If an error occurs when trying
 * to send the message to the destination(s).
 */
template <class C, class M> void RmsSender<C, M>::sendMessage(M& message)
{
    try
    {
        if (_hasBeenClosed)
            GENERATE_RMS_NOTCONNECTED_EXCEPTION("RMS connection was closed.");

        _myConnection->sendMessage(_senderHandle, message);
    }
    catch ( gov::fnal::cd::rms::base::RmsNotConnectedException& ex  ) {
        LOG_EXCEPTION("Caught RmsNotConnectedException in RmsSender::sendMessage(); message=", ex );
    }
    catch ( ... ) {
        this->close();

        LOG_EXCEPTION("Caught exception in RmsSender::sendMessage(). Closing RMS connection.","");
    }
}

/**
 * Closes the RmsSender.
 */
template <class C, class M> void RmsSender<C, M>::close() throw()
{
    if (_hasBeenClosed) {
        return;
    }

    _hasBeenClosed = true;

    try {
        _myConnection->unRegisterSender(_senderHandle);
    } catch (...) {}

    _myConnection.reset();
    
#if CETDDS_DEBUG1
    std::cout << "Closed RmsSender connection" << std::endl;
#endif
}

};  // end of namespace rms

};  // end of namespace cd

};  // end of namespace fnal

};  // end of namespace gov

#endif
// kate: indent-mode cstyle; space-indent on; indent-width 0; 
