#ifndef _RMS_RMSSTATUS_H
#define _RMS_RMSSTATUS_H

#include <string>

namespace gov {

namespace fnal {

namespace cd {

namespace rms {

namespace base {

/**
 * Class to hold information about errors that occur
 * in the process of receiving and decoding RmsMessages.
 *
 * @author Kurt Biery
 * @author Steve Foulkes
 * @version $Revision: 1.1.1.1 $ $Date: 2009/10/09 20:03:00 $
 */

class RmsStatus {
 public:
    /**
     * Codes for the three classes of problems that we handle:
     * success, exception and error.
     */
    static const int SUCCESS_CODE = 0;
    static const int EXCEPTION_CODE = 1;
    static const int ERROR_CODE = 2;

    /**
     * Human readable strings for the three classes of 
     * problems that we handle: success, exception and error.
     */
    static const std::string SUCCESS_MESSAGE;
    static const std::string EXCPETION_MESSAGE;
    static const std::string ERROR_MESSAGE;

    /**
     * Default constructor.  Sets up everything to be success.
     */
    RmsStatus() {
        _code = SUCCESS_CODE;
        _message = SUCCESS_MESSAGE;
    }

    /**
     * Constructs an RmsStatus object with the specified status code
     * and message.
     *
     * @param code integer value that indicates success or failure
     * @param message text string that describes the status
     */
    RmsStatus(int code, std::string message) {
        _code = code;
        _message = message;
    }
    
    /**
     * Constructs an RmsStatus object from the specified exception.
     *
     * @param excp the exception object that was generated by some
     *             exceptional condition
     */
    RmsStatus(int excp) {
        _excp = excp;
    }

    /**
     * Returns the status code.
     *
     * @return The status code
     */
    int getCode() const {
        return _code;
    }

    /**
     * Returns the status message.
     *
     * @return The status message
     */
    std::string getMessage() const {
        return _message;
    }

    /**
     * Returns the exception that was encountered when the status code is
     * EXCEPTION and null otherwise.
     *
     * @return The status exception, or null if the problem code
     * is not EXCPETION.
     */
    int getException() const {
        return _excp;
    }

 private:
    /**
     * This indicates whether we're reporting success, an error,
     * or on exception.
     */
    int _code; 

    /**
     * If we're reporting an exception, this will indicate
     * what the exception was.
     */
    int _excp;

    /**
     * The human readable text for of the error code.
     */
    std::string _message;
};

};  // end of namespace base

};  // end of namespace rms

};  // end of namespace cd

};  // end of namespace fnal

};  // end of namespace gov

#endif
 
