#ifndef _RMS_DESTINATION_H
#define _RMS_DESTINATION_H

#include <string>
#include <map>
#include <iostream>

#include <rms/RmsMessageSchema.h>
#include <rms/util/ReentrantGetEnv.h>

#include <boost/algorithm/string/trim.hpp>

namespace gov {

namespace fnal {

namespace cd {

namespace rms {

namespace base {

/**
 * Instances of this class represent message destinations in the
 * Responsive Messaging System, where a "destination" is an abstract
 * location to which producers send messages and from which consumers
 * receive messages.
 *
 * In RMS, a destination is composed of an arbitrary number of properties
 * that describe various characteristics of the abstract location such as
 * target (hostname or process name) and channel (type of functionality
 * that the destination is used for).  Standard property names and typical
 * property values are defined in this class and experiment-specific child
 * classes can provide additional ones based on specific needs.
 *
 * @author Kurt Biery
 * @author Steve Foulkes
 * @version $Revision: 1.17 $ $Date: 2014/08/01 06:40:16 $
 */

class RmsDestination : public gov::fnal::cd::rms::xsd::destination {
 public:

    /**
     * The following string constants define common broadcast targets.
     */
    static const std::string ALL_DCM_TARGET;
    static const std::string ALL_BNEVB_TARGET;
    static const std::string ALL_ELEMENT_TARGET;

    /**
     * The following string constants define expected application names.
     */
    static const std::string SPILL_SERVER_APPNAME;
    static const std::string GLOBAL_TRIGGER_APPNAME;
    static const std::string DATA_LOGGER_APPNAME;
    static const std::string DAQ_MONITOR_APPNAME;
    static const std::string CONFIG_MANAGER_APPNAME;
    static const std::string RUN_CONTROL_APPNAME;
    static const std::string SIMULATION_MANAGER_APPNAME;
    static const std::string RESOURCE_MANAGER_APPNAME;
    static const std::string DCM_APPNAME;
    static const std::string DAQDCS_APPNAME;
    static const std::string TDU_MANAGER_APPNAME;
    static const std::string BNEVB_APPNAME;
    static const std::string MESSAGE_ANALYZER_APPNAME;
    static const std::string DAQAPP_MANAGER_APPNAME;
    static const std::string DDT_MANAGER_APPNAME;

    /**
     * The following string constants are used to set and retrieve
     * common destination properties.
     */
    static const std::string TARGET_PROPERTY_NAME;
    static const std::string CHANNEL_PROPERTY_NAME;
    static const std::string MESSAGE_TYPE_PROPERTY_NAME;
    static const std::string SOURCE_PROPERTY_NAME;
    static const std::string PARTITION_PROPERTY_NAME;
    static const std::string ENVIRONMENT_PROPERTY_NAME;
    
    /**
     * The following string constants are common targets
     * that are used with the target property.
     */
    static const std::string SERVICE_MANAGER_TARGET;
    
    /**
     * The following string constants are common channels
     * that are used with the channel property.
     */
    static const std::string HEARTBEAT_CHANNEL;
    static const std::string CONTROL_CHANNEL;
    static const std::string TRIGGER_CHANNEL;
    static const std::string STATUS_REPORT_CHANNEL;
    static const std::string MONITOR_CHANNEL;
    static const std::string ALERT_CHANNEL;
    static const std::string CONFIGURATION_CHANNEL;
    static const std::string DAQDCS_CHANNEL;
    static const std::string EH_CHANNEL;    

    /**
     * The following string constants are common message types
     * that are used with the message type property.
     */
    static const std::string NOTIFICATION_MESSAGE_TYPE;
    static const std::string REQUEST_MESSAGE_TYPE;
    static const std::string REPLY_MESSAGE_TYPE;
    static const std::string LOOPBACK_MESSAGE_TYPE;

    /**
     * Used as an identifier for the null partition.
     */
    static const int NULL_PARTITION;
    
    /**
     * Default constructor for the destination object.  Will create an 
     * empty destination.
     */
    RmsDestination() {
        setDefaults();
    };
    
    /**
     * Copy constructor for the destination object.  Will create a destination
     * from an existing destination object.
     *
     * @param originalDestination The object we want to copy
     */
    RmsDestination(const RmsDestination &originalDestination) :
        xsd::destination(originalDestination) {}
    
    /**
     * Creates an instance from an instance of the XSD base class.
     *
     * @param baseDest The object we want to copy
     */
    RmsDestination(const xsd::destination &baseDest) : 
        xsd::destination(baseDest) {}
    
    RmsDestination(std::string target, std::string channel);
    RmsDestination(std::string target, std::string channel, 
                   std::string messageType);
    RmsDestination(std::string target, std::string channel, 
                   std::string messageType, int partitionNumber);

    std::string getProperty(std::string name) const;
    std::string getProperty(std::string name, std::string defaultValue) const;
    void setProperty(std::string name, std::string value);
    void setPartitionProperty(int partitionNumber);
    
    bool equals(const RmsDestination& testObject) const;
    bool matches(const RmsDestination& testObject) const;

    std::string toString() const;

    void validate() const;

 private:
    
    /**
     * If a custom environment has been specified, this field is used
     * to store its value so that we can automatically set the
     * environment property in destinations.
     */
    static std::auto_ptr<std::string> _defaultEnvironment;

    /**
     * Indicates whether static initialization has been done.
     */
    static bool _staticInitializationDone;
    
    void setDefaults();
};

};  // end of namespace base

};  // end of namespace rms

};  // end of namespace cd

};  // end of namespace fnal

};  // end of namespace gov

#endif  
