#ifndef _RMS_MESSAGE_H
#define _RMS_MESSAGE_H

/**
 * @package gov.fnal.cd.rms.base
 *
 * @brief A group of data structures that are
 * used with RMS.
 *
 * The classes inside base consist mostly of data structures
 * that are passed between user programs and RMS itself.  These
 * include the destination class that is used to define where a
 * message is headed, as well as the message class itself.
 */

#include <rms/RmsMessageSchema.h>
#include <rms/base/RmsDestination.h>
#include <rms/base/RmsMessageSource.h>
#include <rms/base/RmsMessageBody.h>
#include <rms/base/RmsStatus.h>

#include <boost/shared_ptr.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>

#include <string>

namespace gov {

namespace fnal {

namespace cd {

namespace rms {

namespace base {

/**
 * A class for the message and replies that are transmitted
 * between applications.
 *
 * @author Kurt Biery
 * @author Steve Foulkes
 * @version $Revision: 1.3 $ $Date: 2009/11/11 16:26:52 $
 */

class RmsMessage : public gov::fnal::cd::rms::xsd::rmsMessage {
 public:
    /**
     * Creates an instance from an instance of the XSD base class.
     *
     * @param baseMsg The object we want to copy
     */
    RmsMessage(const xsd::rmsMessage &baseMsg) : xsd::rmsMessage(baseMsg) {}
    
    RmsMessage();
    RmsMessage(const RmsMessageBody& body);
    RmsMessage(std::string serializedData);
    RmsMessage(const novadaq::XMLSerializable& bodyObject);
    
    const std::string getId() const;
    const std::string getCorrelationId() const;
  
    RmsDestination getDestination() const;
    RmsDestination getReplyDestination() const;
    
    boost::posix_time::ptime getSentTime() const;
    boost::posix_time::ptime getCorrelationTime() const;

    long getTimeToLive() const;
    int getPriority() const;

    RmsMessageSource getSource() const;
    RmsMessageBody getBody() const;
    boost::shared_ptr<novadaq::XMLSerializable> getBodyObject() const;

    void setId(const std::string uuid);
    void setCorrelationId(const std::string uuid);
    void setDestination(const RmsDestination& destination);
    void setReplyDestination(const RmsDestination& destination);

    void setSentTime();
    void setSentTime(boost::posix_time::ptime timestamp);
    void setCorrelationTime(boost::posix_time::ptime timestamp);
    void setTimeToLive(long seconds);

    void setPriority(int priority);

    void setSource(const RmsMessageSource& source);

    void setBody(const RmsMessageBody& body);
    void setBody(const novadaq::XMLSerializable& bodyObject);

    void setMessageStatus(const RmsStatus& messageStatus);
    RmsStatus getMessageStatus() const;

    boost::shared_ptr<RmsMessage> createReply() const;

    /**
     * Returns the fully-qualified class name of the RmsMessage.
     */
    std::string getClassName() const {
        return "gov::fnal::cd::rms::base::RmsMessage";
    }

    /**
     * Returns a copy of the RmsMessage.
     */
    boost::shared_ptr<novadaq::XMLSerializable> clone() const {
        RmsMessage *copy = new RmsMessage(*this);
        boost::shared_ptr<novadaq::XMLSerializable> copyPtr(copy);

        return copyPtr;
    }

 private:
    /**
     * The status of the message 
     */
    RmsStatus _messageStatus; 
    
};

};  // end of namespace base

};  // end of namespace rms

};  // end of namespace cd

};  // end of namespace fnal

};  // end of namespace gov

#endif
