#ifndef _PROCESSSIGNAL_HANDLER_H
#define _PROCESSSIGNAL_HANDLER_H

#include <list>
#include <utility>
#include <boost/interprocess/smart_ptr/unique_ptr.hpp>
#include <boost/function.hpp>
#include <boost/thread/shared_mutex.hpp>
#include <boost/thread/pthread/recursive_mutex.hpp>
#include "rms/base/RmsCloseable.h"
#include "rms/base/RmsRuntimeException.h"

namespace gov {
namespace fnal {
namespace cd {
namespace rms {

namespace base
{

class Notifiable : virtual public RmsCloseable 
{
public:
    Notifiable();
    virtual ~Notifiable();

    void doNotify();
    bool hasBeenNotified() const;

private:
    virtual void doNotify_impl();
    virtual bool hasBeenNotified_impl() const;

private:
    bool _beenNotified;
    mutable boost::mutex _lock;
};
}; // end of namespace base

namespace provider {

//Provides a functionality that replaces the default user signal handling
//mechanism with the customSignalHandler function that frees up DDS server
//resources used by this client.
class ProcessSignalHandler
{
    typedef std::list<base::Notifiable* > NotifiableList;
    typedef NotifiableList::iterator NotifiableListIterator;
    typedef boost::function < void( int ) > SignalHandler;
    typedef std::pair<int, SignalHandler > SignalHandlerPair;
    typedef std::list< SignalHandlerPair > SignalHandlerPairList;
    typedef SignalHandlerPairList::iterator SignalHandlerPairIterator;

    friend class ProcessSignalHandlerDeleter;

public:
    ProcessSignalHandler();

    //adds Notifiable to the process exit notification list
    void addToNotificationList( base::Notifiable* notifiable );

    //removes Notifiable to the process exit notification list
    void removeFromNotificationList( base::Notifiable * notifiable );

    //replaces the default process signal handler by customSignalHandler
    //for a specified signal number
    bool setupSignalHandler( int signalNumber );

    //Frees up resources on the DDS server upon receiving a process signal
    static void customSignalHandler( int signalNumber );

private:
    ProcessSignalHandler( ProcessSignalHandler const& );
    ProcessSignalHandler& operator=( const ProcessSignalHandler& );
    ~ProcessSignalHandler() {};

    //Gracefully closes DDS connections and frees up resources used by this client
    void closeDDSConnections();
    
    //Terminates DDS connections prior exiting this process
    void terminateDDSConnections();
    
    //Acquires a unique lock for RmsLockable::sharedLock and calls closeDDSConnections & terminateDDSConnections
    void processSignal( int signalNumber, ProcessSignalHandler* handler );

private:
    //fatalErrorInProgress is set to true, once the cleanup has begun
    volatile sig_atomic_t _cleanupRequired;

    //A list of client connections that should be closed upon process exit
    NotifiableList _notificationList;
    boost::mutex _listLock; //provides locking for _listLock
    boost::recursive_timed_mutex _selfLock; //provides locking for _selfLock

    //Maintains a list of signal-number--signal-handler pairs;
    //signalhHandlerList is used by customSignalHandler
    static SignalHandlerPairList _signalhHandlerList;
    struct sigaction _previousSignalHandler;
};

class ProcessSignalHandlerDeleter
{
public:
    void operator()( ProcessSignalHandler *handler );
};

typedef boost::interprocess::unique_ptr<ProcessSignalHandler, ProcessSignalHandlerDeleter>   ProcessSignalHandler_t;


class SignalInhibitor
{
  public:
    SignalInhibitor();
    ~SignalInhibitor();
};

};  // end of namespace provider


namespace base
{
  //Provides ProcessSignalHandler with a unique lock on DDS communication channels,
  //so ProcessSignalHandler can free up the DDS server resources used by this process.
  class RmsLockable
  {
    friend  class provider::ProcessSignalHandler;
    
  protected:
    //The purpose of the sharedLock is to prevent DDS communication while processing
    //the exit signal sent by user. Otherwise this process may die prior to unlocking
    //a mutex shared with DDS server.
    static boost::shared_mutex sharedLock;
  };
}; // end of namespace base

#define SCOPED_SHARED_LOCK(a) boost::shared_lock<boost::shared_mutex>  \
lock( RmsLockable::sharedLock, boost::try_to_lock_t() ); \
if(!lock.owns_lock()) {return a;}

};  // end of namespace rms
};  // end of namespace cd
};  // end of namespace fnal
};  // end of namespace gov
#endif //_PROCESSSIGNAL_HANDLER_H
// kate: indent-mode cstyle; space-indent on; indent-width 0; 
