#include <rms/base/RmsDestination.h>
#include <rms/base/RmsRuntimeException.h>

namespace gov {

namespace fnal {

namespace cd {

namespace rms {

namespace base {

/**
 * The following string constants define common broadcast targets.
 */
const std::string RmsDestination::ALL_DCM_TARGET("allDCMs");
const std::string RmsDestination::ALL_BNEVB_TARGET("allBufferNodeEVBs");
const std::string RmsDestination::ALL_ELEMENT_TARGET("allElements");

/**
 * The following string constants define expected application names.
 */
const std::string RmsDestination::SPILL_SERVER_APPNAME("SpillServer");
const std::string RmsDestination::GLOBAL_TRIGGER_APPNAME("GlobalTrigger");
const std::string RmsDestination::DATA_LOGGER_APPNAME("DataLogger");
const std::string RmsDestination::DAQ_MONITOR_APPNAME("DaqMonitor");
const std::string RmsDestination::CONFIG_MANAGER_APPNAME("ConfigurationManager");
const std::string RmsDestination::RUN_CONTROL_APPNAME("RunControl");
const std::string RmsDestination::SIMULATION_MANAGER_APPNAME("SimulationManager");
const std::string RmsDestination::RESOURCE_MANAGER_APPNAME("ResourceManager");
const std::string RmsDestination::DCM_APPNAME("DCMApplication");
const std::string RmsDestination::DAQDCS_APPNAME("DAQDCSMonitor");
const std::string RmsDestination::TDU_MANAGER_APPNAME("TDUManager");
const std::string RmsDestination::BNEVB_APPNAME("BufferNodeEVB");
const std::string RmsDestination::MESSAGE_ANALYZER_APPNAME("MessageAnalyzer");
const std::string RmsDestination::DAQAPP_MANAGER_APPNAME("DAQApplicationManager");
const std::string RmsDestination::DDT_MANAGER_APPNAME("DDTManager");

/**
 * The following string constants are used to set and retrieve
 * common destination properties.
 */
const std::string RmsDestination::TARGET_PROPERTY_NAME("target");
const std::string RmsDestination::CHANNEL_PROPERTY_NAME("channel");
const std::string RmsDestination::MESSAGE_TYPE_PROPERTY_NAME("messageType");
const std::string RmsDestination::SOURCE_PROPERTY_NAME("source");
const std::string RmsDestination::PARTITION_PROPERTY_NAME("partition");
const std::string RmsDestination::ENVIRONMENT_PROPERTY_NAME("environment");

/**
 * The following string constants are common targets
 * that are used with the target property.
 */
const std::string RmsDestination::SERVICE_MANAGER_TARGET("serviceManager");

/**
 * The following string constants are common channels
 * that are used with the channel property.
 */
const std::string RmsDestination::HEARTBEAT_CHANNEL("heartbeat");
const std::string RmsDestination::CONTROL_CHANNEL("control");
const std::string RmsDestination::TRIGGER_CHANNEL("trigger");
const std::string RmsDestination::STATUS_REPORT_CHANNEL("status");
const std::string RmsDestination::MONITOR_CHANNEL("monitor");
const std::string RmsDestination::ALERT_CHANNEL("alert");
const std::string RmsDestination::CONFIGURATION_CHANNEL("config");
const std::string RmsDestination::DAQDCS_CHANNEL("daqdcs");
const std::string RmsDestination::EH_CHANNEL("eh");

/**
 * The following string constants are common message types
 * that are used with the message type property.
 */
const std::string RmsDestination::NOTIFICATION_MESSAGE_TYPE("notification");
const std::string RmsDestination::REQUEST_MESSAGE_TYPE("request");
const std::string RmsDestination::REPLY_MESSAGE_TYPE("reply");
const std::string RmsDestination::LOOPBACK_MESSAGE_TYPE("loopback");

/**
 * Used as an identifier for the null partition.
 */
const int RmsDestination::NULL_PARTITION(-1);

/**
 * Creates an RmsDestination instance for the specified target and channel.
 *
 * @param target The target name
 * @param channel The channel name
 */
RmsDestination::RmsDestination(std::string target, std::string channel) {
    setDefaults();
    setProperty(TARGET_PROPERTY_NAME, target);
    setProperty(CHANNEL_PROPERTY_NAME, channel);
}

/**
 * Creates an RmsDestination instance for the specified target, channel,
 * and message type.
 *
 * @param target The target name
 * @param channel The channel name
 * @param messageType The message type
 */
RmsDestination::RmsDestination(std::string target, std::string channel, 
                               std::string messageType) {
    setDefaults();
    setProperty(TARGET_PROPERTY_NAME, target);
    setProperty(CHANNEL_PROPERTY_NAME, channel);
    setProperty(MESSAGE_TYPE_PROPERTY_NAME, messageType);
}

/**
 * Creates an RmsDestination instance for the specified target,
 * channel, message type, and partition.
 *
 * @param target The target name
 * @param channel The channel name
 * @param messageType The message type
 * @param partitionNumber The partition number
 */
RmsDestination::RmsDestination(std::string target, std::string channel,
                               std::string messageType, int partitionNumber) {
    setDefaults();
    setProperty(TARGET_PROPERTY_NAME, target);
    setProperty(CHANNEL_PROPERTY_NAME, channel);
    setProperty(MESSAGE_TYPE_PROPERTY_NAME, messageType);
    setPartitionProperty(partitionNumber);
}
  
/**
 * Returns the requested property value or null if the specified
 * property name is not found.
 *
 * @param name The property name to look up
 *
 * @return The value that corresponds to the property name passed in,
 * or an empty string if the property is not in the table.
 */
std::string RmsDestination::getProperty(std::string name) const {
    property_sequence propertyList = this->property();
    std::string value = "";

    for (int idx = 0; idx < (int) propertyList.size(); idx++) {
        xsd::property propertyPair = propertyList[idx];
        std::string propertyName = propertyPair.getName();
        if (name == propertyName) {
            value = propertyPair.getValue();
            break;
        }
    }

    return value;
}

/**
 * Returns the requested property value or the specified default value
 * null if the property name is not found.
 *
 * @param name The property name to look up
 * @param defaultValue The value that will be returned if the property
 * doesn't exist in the table.
 * 
 * @return The value that corresponds to the property name passed in,
 * if that property exists.  This will return defaultValue otherwise.
 */
std::string RmsDestination::getProperty(std::string name, 
                                        std::string defaultValue) const {
    std::string value = getProperty(name);

    return (value != "" ? value : defaultValue);
}

/**
 * Sets the named property to the specified value.
 *
 * @param name The property name that we're setting
 * @param value The value that the property is being set to.
 *
 * @throws RmsRuntimeException If either the name or the value is
 * an empty string.
 */
void RmsDestination::setProperty(std::string name, std::string value) {
    bool found = false;

    // 02-Nov-2007, KAB - added conditions that inputs must be non-empty
    if (name == "") {
        std::string msg("Destination property names can not be empty.");
        GENERATE_RMS_RUNTIME_EXCEPTION(msg);
    }
    if (value == "") {
        std::string msg("Destination property values can not be empty.");
        GENERATE_RMS_RUNTIME_EXCEPTION(msg);
    }

    for (int idx = 0; idx < (int) this->getPropertyCount(); idx++) {
        xsd::property& propertyPair = this->getPropertyReference(idx);
        std::string propertyName = propertyPair.getName();
        if (name == propertyName) {
            propertyPair.setValue(value);
            found = true;
            break;
        }
    }

    if (found != true) {
        xsd::property propertyPair;
        propertyPair.setName(name);
        propertyPair.setValue(value);
        this->addProperty(propertyPair);
    }
}

/**
 * Sets the partition property for this destination.
 *
 * @param partitionNumber The partition number
 */
void RmsDestination::setPartitionProperty(int partitionNumber) {    
    char tempString[20];
    snprintf(tempString, 20, "%d", partitionNumber);
    setProperty(PARTITION_PROPERTY_NAME, std::string(tempString));
}

/**
 * Test to see if another destination object has the
 * same properties as this one.
 *
 * @param testObject The object we are testing to see
 * if it matches this one.
 *
 * @return True if both destinations objects have the
 * same properties and values, false otherwise.
 */
bool RmsDestination::equals(const RmsDestination& testObject) const {
    property_sequence thisPropertyList = this->property();
    property_sequence thatPropertyList = testObject.property();

    // require the property list sizes to be the same
    if (thatPropertyList.size() != thisPropertyList.size()) {
        return false;
    }
    
    // check that each property in this object matches one in the test object
    for (int idx = 0; idx < (int) thisPropertyList.size(); idx++) {
        xsd::property propertyPair = thisPropertyList[idx];
        std::string thisName = propertyPair.getName();
        std::string thisValue = this->getProperty(thisName);
        std::string thatValue = testObject.getProperty(thisName);
        if (thatValue == "" || thatValue != thisValue) {
            return false;
        }
    }

    return true;
}

/**
 * Tests if the specified object matches this destination.
 * In this case, "matching" means that the values for the
 * properties that both destinations have in common are equal.
 * Properties that are specified in only one destination are
 * not considered in determining whether the destinations match.
 * (So, if destination A has a "messageType" property and destination
 * B does not, the this method acts as though destination B
 * had a wildcard for the messageType property.)
 * 
 * @param testObject The object that we are comparing this one with.
 *
 * @return True if the objects match, false otherwise.
 */
bool RmsDestination::matches(const RmsDestination& testObject) const {
    property_sequence thisPropertyList = this->property();
    property_sequence thatPropertyList = testObject.property();

    // check that each property in this object matches one in the test object
    for (int idx = 0; idx < (int) thisPropertyList.size(); idx++) {
        xsd::property propertyPair = thisPropertyList[idx];
        std::string thisName = propertyPair.getName();
        std::string thisValue = this->getProperty(thisName);
        std::string thatValue = testObject.getProperty(thisName);
        if (thatValue != "" && thatValue != thisValue) {
            return false;
        }
    }

    return true;
}

/**
 * Generate a string representation of this destination.
 *
 * @return A string representation of this destination.
 */
std::string RmsDestination::toString() const {
    std::string stringRep = "";

    property_sequence propertyList = this->property();
    for (int idx = 0; idx < (int) propertyList.size(); idx++) {
        xsd::property propertyPair = propertyList[idx];
        if (idx > 0) {
            stringRep.append(";");
        }
        stringRep.append(propertyPair.getName());
        stringRep.append("=");
        stringRep.append(propertyPair.getValue());
    }

    return stringRep;
}

/**
 * Validates this destination.  This consists of testing if
 * required properties are set and verifying that mutually exclusive
 * properties are not set.
 *
 * @throws RmsRuntimeException If the destination is invalid.
 */
void RmsDestination::validate() const {
    std::string exceptionText;
    std::string targetPropertyValue = getProperty(TARGET_PROPERTY_NAME);
    std::string sourcePropertyValue = getProperty(SOURCE_PROPERTY_NAME);
    std::string channelPropertyValue = getProperty(CHANNEL_PROPERTY_NAME);
    std::string partitionPropertyValue = getProperty(PARTITION_PROPERTY_NAME);
    
    // check for required properties
    // --> as of 11-Oct-2007, these are
    // A) target or source
    // B) channel
    // C) partition
    if (targetPropertyValue == "" && sourcePropertyValue == "") {
        if (exceptionText.size() > 0) {exceptionText.append("; ");}
        exceptionText.append("either the " + TARGET_PROPERTY_NAME +
                             " property or the " + SOURCE_PROPERTY_NAME +
                             " property must be set");
    }

    if (channelPropertyValue == "") {
        if (exceptionText.size() > 0) {exceptionText.append("; ");}
        exceptionText.append("the " + CHANNEL_PROPERTY_NAME +
                             " property must be set");
    }

    if (partitionPropertyValue == "") {
        if (exceptionText.size() > 0) {exceptionText.append("; ");}
        exceptionText.append("the " + PARTITION_PROPERTY_NAME +
                             " property must be set");
    }

    // check for mutually exclusive properties
    // --> as of 26-Mar-2007, these are
    // A) target and source
    if (targetPropertyValue != "" && sourcePropertyValue != "") {
        if (exceptionText.size() > 0) {exceptionText.append("; ");}
        exceptionText.append("the " + TARGET_PROPERTY_NAME +
                             " property and the " + SOURCE_PROPERTY_NAME +
                             " property must not both be set");
    }

    // throw an exception if one or more problems were found
    if (exceptionText.size() != 0) {
        exceptionText.insert(0, "Invalid RmsDestination " +
                             this->toString() + ": ");
        exceptionText.append(".");
        GENERATE_RMS_RUNTIME_EXCEPTION(exceptionText);
    }
}

/**
 * Initializes the defaultEnvironment static attribute.
 */
std::auto_ptr<std::string> RmsDestination::_defaultEnvironment;

/**
 * Sets the initial value for the static init done flag.
 */
bool RmsDestination::_staticInitializationDone = false;

/**
 * Sets default values at construction time.
 */
void RmsDestination::setDefaults() {
    if (!_staticInitializationDone) {
        _staticInitializationDone = true;
        
        std::string *envString = 
            new std::string(util::ReentrantGetEnv("RMS_CUSTOM_ENVIRONMENT"));

        if (*envString != "") {
            boost::algorithm::trim(*envString);
            if (envString->length() != 0) {
                _defaultEnvironment.reset(envString);
            }
        }
        else {
            delete envString;
        }
    }
    
    if (_defaultEnvironment.get() != NULL) {
        this->setProperty(ENVIRONMENT_PROPERTY_NAME, *_defaultEnvironment);
    }

    setProperty(MESSAGE_TYPE_PROPERTY_NAME, REQUEST_MESSAGE_TYPE);
    this->setPartitionProperty(NULL_PARTITION);
}

};  // end of namespace base

};  // end of namespace rms

};  // end of namespace cd

};  // end of namespace fnal

};  // end of namespace gov
