#include <rms/provider/RmsConnectionFactory.h>
#include <rms/provider/LocalhostConnection.h>
#include <rms/util/ReentrantGetEnv.h>

#include <boost/algorithm/string/trim.hpp>

using namespace gov::fnal::cd::rms::provider;

/**
 * Constant string used for a EPICS provider connection.
 */
const std::string RmsConnectionFactory::EPICS_PROVIDER("epics");

/**
 * Constant string used for a localhost provider connection.
 */
const std::string RmsConnectionFactory::LOCALHOST_PROVIDER("localhost");

/**
 * Creates a connection to the messaging system  using the 
 * default provider.  To determine the default provider,
 * this method first checks for a choice specified in
 * an environmental variabled named RMS_PROVIDER.  If
 * that env. var. is defined and contains an unempty string,
 * it is used.  If not, then a hard-coded default is used.
 *
 * @param applicationName The name of the application generating/receiving
 * messages.
 * @param applicationPartition The partition of the application that is
 * generating/receiving messages.
 *
 * @return an instance of the connection class specified in
 *        the RMS_PROVIDER environmental variable or the
 *        default provider type if the env. var. is not defined.
 */
boost::shared_ptr<RmsConnection> RmsConnectionFactory::getConnection(const std::string applicationName,
                                                                     const int applicationPartition) {
    std::string selectedProvider(util::ReentrantGetEnv("RMS_PROVIDER"));    

    if (selectedProvider == "") {
        return getConnection(applicationName, applicationPartition, EPICS_PROVIDER);
    }

    boost::algorithm::trim(selectedProvider);

    if (selectedProvider.length() != 0) {
        return getConnection(applicationName, applicationPartition, selectedProvider);
    }

    return getConnection(applicationName, applicationPartition, EPICS_PROVIDER);
}

/**
 * Creates a connection to a particular region of the
 * messaging system using a specified messaging provider.
 *
 * <b>Note</b> that this method not intended for general use!  The
 * methods in this class that automatically determine which
 * provider to use should be used by most, if not all, applications.
 * It has only been declared public to allow unit testing of specific
 * provider implementations.
 *
 * @param applicationName The name of the application generating/receiving
 * messages.
 * @param applicationPartition The partition of the application that is
 * generating/receiving messages.
 * @param providerType string containing the requested provider type
 *
 * @return an instance of the requested connection class
 */
boost::shared_ptr<RmsConnection> RmsConnectionFactory::getConnection(const std::string applicationName,
                                                                     const int applicationPartition,
                                                                     std::string providerType) {
    boost::shared_ptr<RmsConnection> newConnection;

    if (providerType.length() == 0) {
        return newConnection;
    }

    if (providerType == EPICS_PROVIDER) {
        //newConnection.reset(new EpicsConnection(applicationName, applicationPartition));
        return newConnection;
    }

    if (providerType == LOCALHOST_PROVIDER) {
        newConnection.reset(new LocalhostConnection(applicationName, applicationPartition));
        return newConnection;
    }

    return newConnection;
}
