#include <rms/provider/CETDDS.h>
#include <rms/provider/DDSConnection.h>
#include <rms/base/RmsDestination.h>
#include <rms/RmsReceiver.h>
#include <rms/RmsSender.h>

#include <rmsexample/ccpp_RmsExampleMessages.h>

#include <boost/shared_ptr.hpp>
#include <boost/lexical_cast.hpp>

#include <iostream>
#include <string>
#include <csignal>
#include <cstdlib>
#include <boost/thread.hpp>

#include <signal.h>
#include <sys/types.h>

using namespace gov::fnal::cd::rms;

/***********************************************************************************************************************
 *
 *                 AbstractListener
 *
 **********************************************************************************************************************/
class AbstractListener {
public:
    AbstractListener(boost::shared_ptr<provider::DDSConnection> ddsConnection, std::string const& appName)
            :m_ddsConnection(ddsConnection), m_msgCount(0), m_appName(appName) {}

    virtual ~AbstractListener() {}

    virtual void processMessage(boost::shared_ptr<rmsexample::RmsExampleMessage> message)=0;

    int getMessageCount() {
        return m_msgCount;
    }

protected:
    boost::shared_ptr<provider::DDSConnection> m_ddsConnection;
    int m_msgCount;
    std::string m_appName;
};



/***********************************************************************************************************************
 *
 *                 AbstractThreadFunctor
 *
 **********************************************************************************************************************/
class AbstractThreadFunctor
{
public:
    AbstractThreadFunctor(std::sig_atomic_t & exitNow):
            m_exitCode(0),m_exitNow(exitNow) {}

    int getExitCode() const {
        return m_exitCode;
    }

    virtual ~AbstractThreadFunctor() {}

    virtual void operator()()=0;

protected:
    bool exitNow() {
        return m_exitNow!=0;
    }
protected:
    int m_exitCode;
private:
    std::sig_atomic_t & m_exitNow;
};


/***********************************************************************************************************************
 *
 *                 BroadcastResponderListener
 *
 **********************************************************************************************************************/
class BroadcastResponderListener : public AbstractListener
{
public:
    BroadcastResponderListener(boost::shared_ptr<provider::DDSConnection> ddsConnection, std::string const& appName)
            :AbstractListener(ddsConnection,appName) {}

    virtual ~BroadcastResponderListener() {}

    virtual void processMessage(boost::shared_ptr<rmsexample::RmsExampleMessage> message);
};

/***********************************************************************************************************************
 *
 *                 BroadcastRecieverListener
 *
 **********************************************************************************************************************/
class BroadcastRecieverListener : public AbstractListener
{
public:
    BroadcastRecieverListener(boost::shared_ptr<provider::DDSConnection> ddsConnection, std::string const& appName)
            :AbstractListener(ddsConnection,appName) {}

    virtual ~BroadcastRecieverListener() {}

    virtual void processMessage(boost::shared_ptr<rmsexample::RmsExampleMessage> message);
};

/***********************************************************************************************************************
 *
 *                 RuntimeConfigOptions
 *
 **********************************************************************************************************************/
struct RuntimeConfigOptions
{
    RuntimeConfigOptions():
            m_verboseFlag(false),m_showIdleStatus(true),m_appName("RmsTestApplication"),
            m_loopCount(10),m_sleepMilliseconds(50) {}

    void printOptions() const {
        std::cout << "RuntimeConfigOptions:" <<  std::endl;
        std::cout << "\tverboseFlag =" << (m_verboseFlag ? "True":"False") << std::endl;
        std::cout << "\tshowIdleStatus =" << (m_showIdleStatus ? "True":"False") << std::endl;
        std::cout << "\tappName =" <<m_appName << std::endl;
        std::cout << "\tloopCount =" <<m_loopCount << std::endl;
        std::cout << "\tsleepTime =" <<m_sleepMilliseconds << " milisecond(s)" << std::endl<<std::flush;
    }

    bool m_verboseFlag;
    bool m_showIdleStatus;
    std::string m_appName;
    int m_loopCount;
    int m_sleepMilliseconds;
};


/***********************************************************************************************************************
 *
 *                 RmsBroadcastSender
 *
 **********************************************************************************************************************/
class RmsBroadcastSender : public AbstractThreadFunctor
{
public:
    RmsBroadcastSender(std::sig_atomic_t & exitNow, RuntimeConfigOptions const&  opts) :
            AbstractThreadFunctor(exitNow), m_opts(opts) {}

    virtual ~RmsBroadcastSender() {}

    virtual void operator()();

private:
    RuntimeConfigOptions m_opts;
};



/***********************************************************************************************************************
 *
 *                 RmsBroadcastReceiver
 *
 **********************************************************************************************************************/
template <class RR > class RmsBroadcastReceiver : public AbstractThreadFunctor
{
public:
    RmsBroadcastReceiver(std::sig_atomic_t & exitNow, RuntimeConfigOptions const&  opts) :
            AbstractThreadFunctor(exitNow), m_opts(opts),m_isResponder(false) { }
    RmsBroadcastReceiver(std::sig_atomic_t & exitNow, RuntimeConfigOptions const&  opts, bool isResponder) :
            AbstractThreadFunctor(exitNow), m_opts(opts),m_isResponder(isResponder) { }
    virtual ~RmsBroadcastReceiver() {}

    virtual void operator()();

private:
    RuntimeConfigOptions m_opts;
    bool m_isResponder;
};


/***********************************************************************************************************************
 *
 *                 RmsBroadcastSenderAndReceiverApplication
 *
 **********************************************************************************************************************/

class RmsBroadcastSenderAndReceiverApplication
{
private:
    RmsBroadcastSenderAndReceiverApplication(): m_isSenderMode(false) {}

    ~RmsBroadcastSenderAndReceiverApplication() {}

public:
    RuntimeConfigOptions const& opts() const {
        return m_opts;
    }

    static void showUsage();
    static int main(int argc, char **argv);

private:
    int parseCommandLineOptions(int argc, char **argv);

private:
    RuntimeConfigOptions m_opts;
    bool m_isSenderMode;
};



/**
 * This is a simple example showing how to send messages
 * and receive replies using RMS.
 */
int main(int argc, char **argv) {
    return RmsBroadcastSenderAndReceiverApplication::main(argc,argv);
};



//*******************************************************************************************
//***                                        BroadcastResponderListener                   ***
//*******************************************************************************************

void BroadcastResponderListener::
processMessage(boost::shared_ptr<rmsexample::RmsExampleMessage> message) {
    // Create a destination to represent where replies are to be sent.
    std::string replyTarget(message->header.replyTarget);
    base::RmsDestination sendDest(replyTarget,
                                  base::RmsDestination::HEARTBEAT_CHANNEL,
                                  base::RmsDestination::REQUEST_MESSAGE_TYPE);

    // Create the sender with the connection and the destination.
    RmsSender<provider::DDSConnection, rmsexample::RmsExampleMessage>
    replySender(m_ddsConnection, sendDest);

    // Create and send the reply.
    rmsexample::RmsExampleMessage replyMessage;

    std::string msgText("Hello World Reply! Message Number=");
    msgText.append(boost::lexical_cast<std::string>(m_msgCount));
    replyMessage.message = msgText.c_str();

    std::string msgId(m_appName);
    msgId.append(" ID=").append(m_ddsConnection->getUUID());
    replyMessage.header.id = DDS::string_dup(msgId.c_str());

    replySender.sendMessage(replyMessage);

    // Clean up
    replySender.close();

    ++m_msgCount;
}


//*******************************************************************************************
//***                                        BroadcastRecieverListener                    ***
//*******************************************************************************************

void BroadcastRecieverListener::
processMessage(boost::shared_ptr<rmsexample::RmsExampleMessage> message) {
    std::cout << "Received a reply message: ";
    std::cout << message->message << std::endl;
    // ++m_msgCount;
}


//*******************************************************************************************
//***               RmsBroadcastSenderAndReceiverApplication                              ***
//*******************************************************************************************
int RmsBroadcastSenderAndReceiverApplication::parseCommandLineOptions(int argc, char **argv)
{
    int optionChar;

    while ((optionChar = getopt (argc, argv, "xvia:n:s:")) != -1) {
        switch (optionChar) {
        case 'a':
            m_opts.m_appName = optarg;
            break;
        case 'n':
            try {
                m_opts.m_loopCount = boost::lexical_cast<int>(optarg);
            }
            catch (boost::bad_lexical_cast &excpt) {
                std::cerr << "Error: loopCount must be numeric."
                          << std::endl;
                return 3;
            }
            break;
        case 's':
            try {
                m_opts.m_sleepMilliseconds = boost::lexical_cast<int>(optarg);
            }
            catch (boost::bad_lexical_cast &excpt) {
                std::cerr << "Error: sleepTime must be numeric."
                          << std::endl;
                return 5;
            }
            break;
        case 'i':
            m_opts.m_showIdleStatus = true;
            break;
        case 'v':
            m_opts.m_verboseFlag = true;
            break;
        case 'x':
            m_isSenderMode = true;
            break;
        case '?':
        default:
            return 1;
        }
    }
    return 0;
}



int RmsBroadcastSenderAndReceiverApplication::main(int argc, char **argv) {
    RmsBroadcastSenderAndReceiverApplication theApp;

    int retCode = theApp.parseCommandLineOptions(argc,argv);

    if (retCode)
    {
        RmsBroadcastSenderAndReceiverApplication::showUsage();
        return retCode;
    }
    else
        theApp.m_opts.printOptions();

    std::cout << "Starting sender app with name \"" << theApp.m_opts.m_appName
              << "\"." << std::endl;

    std::sig_atomic_t exitNow = 0;

    boost::thread_group threadGroup;
  
    if (theApp.m_isSenderMode)
    {
        std::cout <<"Sender mode" <<std::endl;
        RmsBroadcastSender sender(exitNow,theApp.opts());
        RmsBroadcastReceiver<BroadcastRecieverListener> receiver(exitNow,theApp.opts());
        RmsBroadcastReceiver<BroadcastResponderListener> responder(exitNow,theApp.opts(),true);
        threadGroup.create_thread(receiver);
        threadGroup.create_thread(responder);
        threadGroup.create_thread(sender);
    }
    else
    {
        std::cout <<"Responder mode" <<std::endl;
        RmsBroadcastReceiver<BroadcastResponderListener> responder(exitNow,theApp.opts(),true);
        threadGroup.create_thread(responder);
    }

    sleep(7);

    std::cout <<"Hit any key to stop." <<std::endl;
    system("read aa");

 //   kill(getpid() ,2);

    sleep(600);
    
    exitNow = 1;

    std::cout << "Stopping..." << std::endl;

    threadGroup.join_all();

    std::cout << "Done" << std::endl;
return retCode;
    
}

void RmsBroadcastSenderAndReceiverApplication::showUsage()
{
    std::cout << "Usage: " << "RmsBroadcastSenderAndReceiverApplication"
              << " [-v] [-i] [-n loopCount] [-s sleepTime] [-x]"
              << "[-a appName]" << std::endl;
    std::cout << "  where -v enables all diagnostic messages,"
              << std::endl;
    std::cout << "        -i enables just the idle status printout,"
              << std::endl;
    std::cout << "        -s specifies a non-zero sleep time "
              << "(in milliseconds) between messages,"
              << std::endl;
    std::cout << "        -x enables a sender mode "
              << "(default operation mode is a receiver mode),"
              << std::endl;
    std::cout << "    and the application name is used in the routing "
              << "of the messages." << std::endl <<std::flush;
}

void RmsBroadcastSender::operator()() {
    /**
     * Create an RmsConnection object using the connection factory.
     * We want replies routed to the destination with a target equal
     * to "RmsSenderExample" so we set that as the applicationName
     * and we use the NULL partition since we don't care about partitions.
     */
    boost::shared_ptr<provider::DDSConnection> rmsConnection;
    rmsConnection.reset(
        new provider::DDSConnection(m_opts.m_appName,
                                    base::RmsDestination::NULL_PARTITION)
    );

    /**
     * Create a destination to represent where the messages that are
     * generated here will go to.  For the example, we're sending messages
     * to the "RmsReceiverExample" target.  The messages will use the control
     * channel and be request messages.
     */
    base::RmsDestination sendDest(m_opts.m_appName,
                                  base::RmsDestination::ALL_ELEMENT_TARGET,
                                  base::RmsDestination::REQUEST_MESSAGE_TYPE);

    /**
     * Create a sender with the connection and destination.
     */
    RmsSender<provider::DDSConnection, rmsexample::RmsExampleMessage>
    requestSender(rmsConnection, sendDest);
    int idx = 0;
    for (; !exitNow() ; ++idx) {

        if (m_opts.m_sleepMilliseconds > 0) {
            std::cout << "Sleeping for " << m_opts.m_sleepMilliseconds
                      << " miliseconds [" << idx << "]." << std::endl;
            usleep(m_opts.m_sleepMilliseconds*1000);
        }

        /**
         * Create the message and send it.
         */
        rmsexample::RmsExampleMessage exampleMessage;
        std::string msgText("Hello World! Message Number=");
        msgText.append(boost::lexical_cast<std::string>(idx));
        exampleMessage.message =msgText.c_str();
        requestSender.sendMessage(exampleMessage);
    }

    std::cout << idx << " messages sent." << std::endl;

    /**
     * Cleanup, and exit.
     */
    requestSender.close();
    rmsConnection->close();

    std::cout << "Exiting RmsSender thread; application name=\"" << m_opts.m_appName
              << "\"." << std::endl;
}

template < class RR > void RmsBroadcastReceiver< RR >::operator()() {
    boost::shared_ptr<provider::DDSConnection> rmsConnection;

    // Create an RmsConnection object using the connection factory.
    // We want replies routed to the destination with a target equal
    // to "RmsReceiverExample" so we set that as the applicationName
    // and we use the NULL partition since we don't care about partitions.
    rmsConnection.reset(
        new provider::DDSConnection(m_opts.m_appName,
                                    base::RmsDestination::NULL_PARTITION)
    );

    // Create a destination to represent where we want to receive messages
    // from.  For the example, we're receiving messages to the
    // "RmsReceiverExample" target.  The messages will use the control channel
    // and be request messages.
    base::RmsDestination receiveDest(m_opts.m_appName,m_isResponder? base::RmsDestination::ALL_ELEMENT_TARGET:
                                     base::RmsDestination::HEARTBEAT_CHANNEL,
                                     base::RmsDestination::REQUEST_MESSAGE_TYPE);

    // Create a receiver with the connection object and destination.
    RmsReceiver<provider::DDSConnection,
    rmsexample::RmsExampleMessage,RR>
    requestReceiver(rmsConnection, receiveDest);
    boost::shared_ptr<RR>
    theListener(new RR(rmsConnection,m_opts.m_appName));
    requestReceiver.setListener(theListener);

    while (!exitNow()) {
        usleep(250000);
    }

    /**
     * Cleanup, and exit.
     */
    requestReceiver.close();
    rmsConnection->close();
    std::cout << "Exiting RmsReceiver thread; application name=\"" << m_opts.m_appName
              << "\"." << std::endl;
}

