#include <rms/provider/RmsConnection.h>
#include <rms/provider/RmsConnectionFactory.h>
#include <rms/base/RmsDestination.h>
#include <rms/base/RmsMessage.h>
#include <rms/base/RmsMessageBody.h>
#include <rms/RmsConsumer.h>

#include <NovaDAQUtilities/XMLSerializable.h>
#include <RmsExampleMessages.h>

#include <boost/shared_ptr.hpp>

#include <rms/RmsMessageListener.h>
#include <rms/util/TimeUtils.h>
#include <boost/thread/mutex.hpp>

using namespace gov::fnal::cd::rms;

// NOTE that the model of sharing data shown in this example is just
// a sample.  In a non-example application, we may structure things
// differently so that the shared data (like the message counter) and
// the consumer to use for sending replies is made available in a
// different way.  (And, we probably wouldn't pass a mutex around...)
class AsyncMessageHandler : public RmsMessageListener {

  public:

    /**
     * Constructor.
     */
    AsyncMessageHandler(boost::mutex& countLock, int& messageCounter,
                        RmsConsumer& replyConsumer) :
        _countLock(countLock),
        _messageCounter(messageCounter),
        _replyConsumer(replyConsumer)
    {}

    /**
     * Handles all messages that are received.
     */
    void messageReceived(boost::shared_ptr<base::RmsMessage> incomingMessage) {

        // ignore stale messages
        // (21-Sep-2009, KAB - I need to look at the time classes that became
        // available in later versions of Boost to see if there are better
        // ways to manage times.)
        boost::posix_time::ptime sentTime = incomingMessage->getSentTime();
        boost::posix_time::ptime comparisonTime = sentTime +
            boost::posix_time::milliseconds(incomingMessage->getTimeToLive());
        boost::posix_time::ptime now = util::TimeUtils::getUniversalPtime();
        boost::posix_time::time_duration timeDiff = comparisonTime - now;
        if (timeDiff.is_negative()) {
            //std::cout << "Ignoring stale message" << std::endl;
            return;
        }

        // update the number of messages received
        {
            boost::mutex::scoped_lock sl(_countLock);
            ++_messageCounter;
        }

        // initialize the status that we'll send in the reply to sucess
        example::SampleRequestStatus replyStatus;
        replyStatus.setCode(0);
        replyStatus.setMessage("Success");

        // parse the request
        std::string requestedAction = "unknown";
        int partitionNumber = -1;

        // fetch the body object of the incoming message
        boost::shared_ptr<novadaq::XMLSerializable> bodyObjPtr =
            incomingMessage->getBodyObject();
        if (replyStatus.getCode() == 0 && bodyObjPtr.get() == NULL) {
            replyStatus.setCode(-103);
            replyStatus.setMessage("Unable to parse request message: " +
                                   incomingMessage->serialize());
        }

        // verify that the class of the message body is what we expect
        example::SampleTransitionRequest *pTransRequest =
            dynamic_cast<example::SampleTransitionRequest *>(bodyObjPtr.get());
        if (replyStatus.getCode() == 0 && pTransRequest == NULL) {
            replyStatus.setCode(-101);
            replyStatus.setMessage("Invalid message type received: " +
                                   incomingMessage->getBody().getClassName());
        }

        // pull out the requested action and the partition number
        if (replyStatus.getCode() == 0) {
            requestedAction = pTransRequest->getAction();
            partitionNumber = pTransRequest->getPartitionNumber();
        }

        // print out a simple message to the screen (an application that is
        // more than an example would probably act on the message contents)
        std::cout << "Received an asynchronous message with requested action = "
                  << "\"" << requestedAction << "\" and partition number = "
                  << "\"" << partitionNumber << "\"." << std::endl;
        

        // send the reply
        if (replyStatus.getCode() >= -100) {
            example::SampleTransitionResponse transResponse;
            transResponse.setSampleRequestStatus(replyStatus);
            // should get the application name in non-harded-coded way...
            transResponse.setApplicationName("RMS Consumer Example");

            boost::shared_ptr<base::RmsMessage> reply = incomingMessage->createReply();
            reply->setBody(transResponse);

            // it would be nice to move the assignment of the source information
            // into the consumer...
            base::RmsMessageSource msgSource;
            msgSource.setApplicationName("RMS Consumer Example");
            // should get the hostname once and re-use it
            char *envVarPtr = getenv("HOSTNAME");
            if (envVarPtr != NULL) {
                std::string tmpString(envVarPtr);
                msgSource.setHostName(tmpString);
            }
            else {
                msgSource.setHostName("Unknown");
            }
            reply->setSource(msgSource);

            _replyConsumer.sendReply(reply);
        }
    }

  private:

    boost::mutex& _countLock;
    int& _messageCounter;
    RmsConsumer& _replyConsumer;
};

/**
 * This is simple example showing how to receive messages asynchronously
 * and send replies using RMS.
 *
 * Using the EPICS provider, the following commands need
 * to be run to provide the necessary PVs:
 *   [sfoulkes@cd-90837 rms]$ export RMS_CUSTOM_ENVIRONMENT="<yourEnv>"
 *   [sfoulkes@cd-90837 rms]$ autoCas
 */
int main() {
    provider::RmsConnectionFactory connectionFactory;
    boost::shared_ptr<provider::RmsConnection> rmsConnection;

    /**
     * Register the needed message types with the deserialization registry
     * so that automatic message serialization/deserialization works.
     */
    gov::fnal::cd::rms::example::RmsExampleMessage::registerForDeserialization();
    gov::fnal::cd::rms::example::SampleTransitionRequest::registerForDeserialization();
    gov::fnal::cd::rms::example::SampleTransitionResponse::registerForDeserialization();

    /**
     * Create an RmsConnection object using the connection factory.
     * We want replies routed to the destination with a target equal
     * to "RmsConsumerExample" so we set that as the applicationName
     * and we use the NULL partition since we don't care about partitions.
     */
    rmsConnection = connectionFactory.getConnection("RmsConsumerExample",
                                                    base::RmsDestination::NULL_PARTITION);

    /**
     * Create a destination to represent where we want to receive messages
     * from.  For the example, we're receiving messages to the "RmsConsumerExample" 
     * target.  The messages will use the control channel and be request messages.
     */
    base::RmsDestination exampleDest("RmsConsumerExample",
                                     base::RmsDestination::CONTROL_CHANNEL,
                                     base::RmsDestination::REQUEST_MESSAGE_TYPE);

    /**
     * Create a consumer with the connection object and destination.
     */
    RmsConsumer exampleConsumer(rmsConnection, exampleDest);


    // Create a destination and consumer for asynchronous messages.
    base::RmsDestination asyncDest("asyncMessageTarget",
                                   base::RmsDestination::CONTROL_CHANNEL,
                                   base::RmsDestination::REQUEST_MESSAGE_TYPE);
    RmsConsumer asyncConsumer(rmsConnection, asyncDest);
    boost::mutex countLock;
    int asyncMessageCount = 0;
    AsyncMessageHandler backgroundHandler(countLock, asyncMessageCount,
                                          asyncConsumer);
    asyncConsumer.setMessageListener(&backgroundHandler);


    /**
     * Wait for a message to show up.
     */
    std::cout << "Waiting for the primary message to arrive..." << std::endl;
    boost::shared_ptr<base::RmsMessage> incomingMessage;
    incomingMessage = exampleConsumer.receiveMessage();

    /**
     * Check the message to make sure it isn't empty.
     */
    boost::shared_ptr<novadaq::XMLSerializable> bodyObjPtr =
        incomingMessage->getBodyObject();

    if (bodyObjPtr.get() == NULL) {
        std::cout << "Unable to parse messages:" << std::endl;
        std::cout << incomingMessage->serialize() << std::endl;
        exampleConsumer.close();
        rmsConnection->close();
        return -1;
    }

    /**
     * Verify that the message we received is of the type that we expect.
     */
    gov::fnal::cd::rms::example::RmsExampleMessage *messageContent =
        dynamic_cast<gov::fnal::cd::rms::example::RmsExampleMessage *> (bodyObjPtr.get());
    if (messageContent == NULL) {
        std::cout << "Invalid message type: " << incomingMessage->getBody().getClassName() << std::endl;
        exampleConsumer.close();
        rmsConnection->close();
        return -1;
    }

    /**
     * Print the message content to the screen.
     */
    std::cout << "Received a message: ";
    std::cout << messageContent->getMessage() << std::endl;
    {
        boost::mutex::scoped_lock sl(countLock);
        std::cout << "Received " << asyncMessageCount << " messages in the background "
                  << "while waiting for the primary message to arrive." << std::endl;
    }

    /**
     * Create a reply and stuff some content into it.
     * This data type was created from the RmsExampleMessages xsd file.
     */
    boost::shared_ptr<base::RmsMessage> exampleReply = 
        incomingMessage->createReply();

    gov::fnal::cd::rms::example::RmsExampleMessage replyBody;
    replyBody.setMessage("Hello World Reply!");
    exampleReply->setBody(replyBody);

    /**
     * Send the reply.
     */
    std::cout << "Sending reply..." << std::endl;
    exampleConsumer.sendReply(exampleReply);

    /**
     * Cleanup, and exit.
     */
    asyncConsumer.close();
    exampleConsumer.close();
    rmsConnection->close();

    return 0;
}
