#include <rms/provider/RmsConnection.h>
#include <rms/provider/RmsConnectionFactory.h>
#include <rms/base/RmsDestination.h>
#include <rms/base/RmsMessage.h>
#include <rms/base/RmsMessageBody.h>
#include <rms/RmsConsumer.h>

#include <NovaDAQUtilities/XMLSerializable.h>
#include <RmsExampleMessages.h>

#include <boost/shared_ptr.hpp>

using namespace gov::fnal::cd::rms;

/**
 * This is simple example showing how to receive messages
 * and send replies using RMS.  It assumes that two
 * destinations exist in the message passing system:
 *   target=RmsConsumerExample, channel=control, message type=request
 *   target=RmsProducerExample, channel=control, message type=request
 * (These will be automatically created when using the EPICS autoCas.)
 *
 * Using the EPICS provider, the following commands need
 * to be run to provide the necessary PVs:
 *   [sfoulkes@cd-90837 rms]$ export RMS_CUSTOM_ENVIRONMENT="<yourEnv>"
 *   [sfoulkes@cd-90837 rms]$ autoCas
 *
 * This program should produce the following output:
 *   [sfoulkes@cd-90837 rms]$ RmsConsumerExample
 *   Received a message: Hello World!
 *   Sending reply...
 */
int main() {
    provider::RmsConnectionFactory connectionFactory;
    boost::shared_ptr<provider::RmsConnection> rmsConnection;

    /**
     * Register the example message type with the deserialization registry
     * so that automatic message serialization/deserialization works.
     * (For production applications, these operations will be handled
     * in a more automatic way.)
     */
    gov::fnal::cd::rms::example::RmsExampleMessage::registerForDeserialization();

    /**
     * Create an RmsConnection object using the connection factory.
     * We want replies routed to the destination with a target equal
     * to "RmsConsumerExample" so we set that as the applicationName
     * and we use the NULL partition since we don't care about partitions.
     */
    rmsConnection = connectionFactory.getConnection("RmsConsumerExample",
                                                    base::RmsDestination::NULL_PARTITION);

    /**
     * Create a destination to represent where we want to receive messages
     * from.  For the example, we're receiving messages to the "RmsConsumerExample" 
     * target.  The messages will use the control channel and be request messages.
     */
    base::RmsDestination exampleDest("RmsConsumerExample",
                                     base::RmsDestination::CONTROL_CHANNEL,
                                     base::RmsDestination::REQUEST_MESSAGE_TYPE);

    /**
     * Create a consumer with the connection object and destination.
     */
    RmsConsumer exampleConsumer(rmsConnection, exampleDest);

    /**
     * Wait for a message to show up.
     */
    boost::shared_ptr<base::RmsMessage> incomingMessage;
    incomingMessage = exampleConsumer.receiveMessage();

    /**
     * Check the message to make sure it isn't empty.
     */
    boost::shared_ptr<novadaq::XMLSerializable> bodyObjPtr =
        incomingMessage->getBodyObject();

    if (bodyObjPtr.get() == NULL) {
        std::cout << "Unable to parse messages:" << std::endl;
        std::cout << incomingMessage->serialize() << std::endl;
        exampleConsumer.close();
        rmsConnection->close();
        return -1;
    }

    /**
     * Verify that the message we received is of the type that we expect.
     */
    gov::fnal::cd::rms::example::RmsExampleMessage *messageContent =
        dynamic_cast<gov::fnal::cd::rms::example::RmsExampleMessage *> (bodyObjPtr.get());
    if (messageContent == NULL) {
        std::cout << "Invalid message type: " << incomingMessage->getBody().getClassName() << std::endl;
        exampleConsumer.close();
        rmsConnection->close();
        return -1;
    }

    /**
     * Print the message content to the screen.
     */
    std::cout << "Received a message: ";
    std::cout << messageContent->getMessage() << std::endl;
    std::cout << "The message was sent at "
              << incomingMessage->getSentTime() << " (GMT)." << std::endl;

    /**
     * Create a reply and stuff some content into it.
     * This data type was created from the RmsExampleMessages xsd file.
     */
    boost::shared_ptr<base::RmsMessage> exampleReply = incomingMessage->createReply();

    gov::fnal::cd::rms::example::RmsExampleMessage replyBody;
    replyBody.setMessage("Hello World Reply!");
    exampleReply->setBody(replyBody);

    /**
     * Send the reply.
     */
    std::cout << "Sending reply..." << std::endl;
    exampleConsumer.sendReply(exampleReply);

    /**
     * Cleanup, and exit.
     */
    exampleConsumer.close();
    rmsConnection->close();

    return 0;
}
