#include <rms/provider/RmsConnection.h>
#include <rms/provider/RmsConnectionFactory.h>
#include <rms/base/RmsDestination.h>
#include <rms/base/RmsMessage.h>
#include <rms/base/RmsMessageBody.h>
#include <rms/RmsProducer.h>

#include <NovaDAQUtilities/XMLSerializable.h>
#include <RmsExampleMessages.h>

#include <boost/shared_ptr.hpp>

#include <rms/RmsMessageListener.h>
#include <rms/util/TimeUtils.h>
#include <boost/thread/mutex.hpp>

using namespace gov::fnal::cd::rms;

// NOTE that the model of sharing data shown in this example is just
// a sample.  In a non-example application, we may structure things
// differently so that the shared data (like the reply counter) is
// made available in a different way.  (And, we probably wouldn't
// pass a mutex around...)
class AsyncMessageHandler : public RmsMessageListener {

  public:

    /**
     * Constructor.
     */
    AsyncMessageHandler(boost::mutex& countLock, int& replyCounter) :
        _countLock(countLock),
        _replyCounter(replyCounter)
    {}

    /**
     * Handles all messages that are received.
     */
    void messageReceived(boost::shared_ptr<base::RmsMessage> incomingMessage) {

        // update the number of messages received
        {
            boost::mutex::scoped_lock sl(_countLock);
            ++_replyCounter;
        }

        // parse the request
        std::string statusMessage = "unknown";

        // fetch the body object of the incoming message
        boost::shared_ptr<novadaq::XMLSerializable> bodyObjPtr =
            incomingMessage->getBodyObject();
        if (bodyObjPtr.get() != NULL) {
            example::SampleTransitionResponse *pTransResponse =
                dynamic_cast<example::SampleTransitionResponse *>(bodyObjPtr.get());
            if (pTransResponse != NULL) {
                statusMessage = pTransResponse->getSampleRequestStatus().getMessage();
            }
        }

        // print out a simple message to the screen (an application that is
        // more than an example would probably act on the message contents)
        std::cout << "Received an asynchronous reply with status message = "
                  << "\"" << statusMessage << "\"." << std::endl;
    }

  private:

    boost::mutex& _countLock;
    int& _replyCounter;
};

/**
 * This is simple example showing how to receive messages asynchronously
 * and send replies using RMS.
 *
 * Using the EPICS provider, the following commands need
 * to be run to provide the necessary PVs:
 *   [sfoulkes@cd-90837 rms]$ export RMS_CUSTOM_ENVIRONMENT="<yourEnv>"
 *   [sfoulkes@cd-90837 rms]$ autoCas
 */
int main(void) {
    provider::RmsConnectionFactory connectionFactory;
    boost::shared_ptr<provider::RmsConnection> rmsConnection;

    /**
     * Register the needed message types with the deserialization registry
     * so that automatic message serialization/deserialization works.
     */
    gov::fnal::cd::rms::example::RmsExampleMessage::registerForDeserialization();
    gov::fnal::cd::rms::example::SampleTransitionRequest::registerForDeserialization();
    gov::fnal::cd::rms::example::SampleTransitionResponse::registerForDeserialization();
    
    /**
     * Create an RmsConnection object using the connection factory.
     * We want replies routed to the destination with a target equal
     * to "RmsProducerExample" so we set that as the applicationName
     * and we use the NULL partition since we don't care about partitions.
     */
    rmsConnection = connectionFactory.getConnection("RmsProducerExample",
                                                    base::RmsDestination::NULL_PARTITION);

    /**
     * Create a destination to represent where the messages that are
     * generated here will go to.  For the example, we're sending messages
     * to the "RmsConsumerExample" target.  The messages will use the control
     * channel and be request messages.
     */
    base::RmsDestination exampleDest("RmsConsumerExample",
                                     base::RmsDestination::CONTROL_CHANNEL,
                                     base::RmsDestination::REQUEST_MESSAGE_TYPE);

    /**
     * Create a producer with the connection, destination and reply destination.
     */
    RmsProducer exampleProducer(rmsConnection, exampleDest);


    // Create a destination and producer for asynchronous messages.
    base::RmsDestination asyncDest("asyncMessageTarget",
                                   base::RmsDestination::CONTROL_CHANNEL,
                                   base::RmsDestination::REQUEST_MESSAGE_TYPE);
    RmsProducer asyncProducer(rmsConnection, asyncDest);
    boost::mutex countLock;
    int asyncReplyCount = 0;
    AsyncMessageHandler replyHandler(countLock, asyncReplyCount);
    asyncProducer.setReplyListener(&replyHandler);

    // Send a number of transition request messages
    gov::fnal::cd::rms::example::SampleTransitionRequest transitionRequest;
    for (int idx = 0; idx < 5; ++idx) {
        transitionRequest.setAction("Take Data!");
        transitionRequest.setPartitionNumber(idx);
        base::RmsMessage asyncMessage(transitionRequest);
        asyncProducer.sendMessage(asyncMessage);
        ::sleep(1);
    }

    /**
     * Create an RmsExampleMessage and stuff some content into
     * it.  This data type was created from the RmsExampleMessages xsd file.
     */
    gov::fnal::cd::rms::example::RmsExampleMessage exampleMessageBody;
    exampleMessageBody.setMessage("Hello World!");
    base::RmsMessage exampleMessage(exampleMessageBody);

    /**
     * Send the message.
     */
    std::cout << "Sending the primary message..." << std::endl;
    exampleProducer.sendMessage(exampleMessage);

    /**
     * Wait for a reply.
     */
    boost::shared_ptr<base::RmsMessage> incomingReply;
    incomingReply = exampleProducer.receiveReply();

    /**
     * Check the reply to make sure that it isn't empty.
     */
    boost::shared_ptr<novadaq::XMLSerializable> bodyObjPtr =
        incomingReply->getBodyObject();

    if (bodyObjPtr.get() == NULL) {
        std::cout << "Unable to parse messages:" << std::endl;
        std::cout << incomingReply->serialize() << std::endl;
        exampleProducer.close();
        rmsConnection->close();
        return -1;
    }

    /**
     * Verify that the message we received is of the type 
     * that we are expecting.
     */
    std::string messageClassName = incomingReply->getBody().getClassName();
    if (messageClassName != "gov.fnal.cd.rms.example.RmsExampleMessage") {
        std::cout << "Invalid message type: " << messageClassName << std::endl;
        exampleProducer.close();
        rmsConnection->close();
        return -1;
    }

    /**
     * Retrieve the message content.
     */
    gov::fnal::cd::rms::example::RmsExampleMessage *messageContent =
        static_cast<gov::fnal::cd::rms::example::RmsExampleMessage *> (bodyObjPtr.get());

    std::cout << "Received a reply: ";
    std::cout << messageContent->getMessage() << std::endl;

    /**
     * Cleanup, and exit.
     */
    asyncProducer.close();
    exampleProducer.close();
    rmsConnection->close();

    return 0;
}

