#include <rms/provider/RmsConnection.h>
#include <rms/provider/RmsConnectionFactory.h>
#include <rms/base/RmsDestination.h>
#include <rms/base/RmsMessage.h>
#include <rms/base/RmsMessageBody.h>
#include <rms/RmsProducer.h>


#include <NovaDAQUtilities/XMLSerializable.h>
#include <RmsExampleMessages.h>

#include <boost/shared_ptr.hpp>

using namespace gov::fnal::cd::rms;

/**
 * This is simple example showing how to send messages
 * and receive replies using RMS.  It assumes that two
 * destinations exist in the message passing system:
 *   target=RmsConsumerExample, channel=control, message type=request
 *   target=RmsProducerExample, channel=control, message type=request
 * (These will be automatically created when using the EPICS autoCas.)
 *
 * Using the EPICS provider, the following commands need
 * to be run to provide the necessary PVs:
 *   [sfoulkes@cd-90837 rms]$ export RMS_CUSTOM_ENVIRONMENT="<yourEnv>"
 *   [sfoulkes@cd-90837 rms]$ autoCas
 *
 * This program should produce the following output:
 *   [sfoulkes@cd-90837 rms]$ RmsProducerExample
 *   Sending message...
 *   Received a reply: Hello World Reply!
 */
int main(void) {
    provider::RmsConnectionFactory connectionFactory;
    boost::shared_ptr<provider::RmsConnection> rmsConnection;

    /**
     * Register the example message type with the deserialization registry
     * so that automatic message serialization/deserialization works.
     */
    gov::fnal::cd::rms::example::RmsExampleMessage::registerForDeserialization();

    /**
     * Create an RmsConnection object using the connection factory.
     * We want replies routed to the destination with a target equal
     * to "RmsProducerExample" so we set that as the applicationName
     * and we use the NULL partition since we don't care about partitions.
     */
    rmsConnection = connectionFactory.getConnection("RmsProducerExample",
                                                    base::RmsDestination::NULL_PARTITION);

    /**
     * Create a destination to represent where the messages that are
     * generated here will go to.  For the example, we're sending messages
     * to the "RmsConsumerExample" target.  The messages will use the control
     * channel and be request messages.
     */
    base::RmsDestination exampleDest("RmsConsumerExample",
                                     base::RmsDestination::CONTROL_CHANNEL,
                                     base::RmsDestination::REQUEST_MESSAGE_TYPE);

    /**
     * Create a producer with the connection, destination and reply destination.
     */
    RmsProducer exampleProducer(rmsConnection, exampleDest);

    /**
     * Create an RmsExampleMessage and stuff some content into
     * it.  This data type was created from the RmsExampleMessages xsd file.
     */
    gov::fnal::cd::rms::example::RmsExampleMessage exampleMessageBody;
    exampleMessageBody.setMessage("Hello World!");
    base::RmsMessage exampleMessage(exampleMessageBody);

    /**
     * Send the message.
     */
    std::cout << "Sending message..." << std::endl;
    exampleProducer.sendMessage(exampleMessage);

    /**
     * Wait for a reply.
     */
    boost::shared_ptr<base::RmsMessage> incomingReply;
    incomingReply = exampleProducer.receiveReply();

    /**
     * Check the reply to make sure that it isn't empty.
     */
    boost::shared_ptr<novadaq::XMLSerializable> bodyObjPtr =
        incomingReply->getBodyObject();

    if (bodyObjPtr.get() == NULL) {
        std::cout << "Unable to parse messages:" << std::endl;
        std::cout << incomingReply->serialize() << std::endl;
        exampleProducer.close();
        rmsConnection->close();
        return -1;
    }

    /**
     * Verify that the message we received is of the type 
     * that we are expecting.
     */
    std::string messageClassName = incomingReply->getBody().getClassName();

    if (messageClassName != "gov.fnal.cd.rms.example.RmsExampleMessage") {
        std::cout << "Invalid message type: " << messageClassName << std::endl;
        exampleProducer.close();
        rmsConnection->close();
        return -1;
    }

    /**
     * Retrieve the message content.
     */
    gov::fnal::cd::rms::example::RmsExampleMessage *messageContent =
        static_cast<gov::fnal::cd::rms::example::RmsExampleMessage *> (bodyObjPtr.get());

    std::cout << "Received a reply: ";
    std::cout << messageContent->getMessage() << std::endl;

    /**
     * Cleanup, and exit.
     */
    exampleProducer.close();
    rmsConnection->close();

    return 0;
}

