#include <rms/provider/CETDDS.h>
#include <rms/provider/DDSConnection.h>
#include <rms/base/RmsDestination.h>
#include <rms/RmsReceiver.h>
#include <rms/RmsSender.h>

#include <rmsexample/ccpp_RmsExampleMessages.h>

#include <boost/shared_ptr.hpp>

using namespace gov::fnal::cd::rms;

/**
 * This is simple example showing how to receive messages
 * and send replies using RMS.  It uses two
 * destinations in the message passing system:
 *   target=RmsReceiverExample, channel=control, message type=request
 *   target=RmsSenderExample, channel=control, message type=request
 *
 * Include info on the DDS server setup?
 *
 * This program should produce the following output:
 *   [sfoulkes@cd-90837 rms]$ RmsReceiverExample
 *   Received a message: Hello World!
 *   Sending reply...
 */

class ExampleListener {
 public:
    ExampleListener(boost::shared_ptr<provider::DDSConnection> ddsConnection);
    void processMessage(boost::shared_ptr<rmsexample::RmsExampleMessage> message);
    int getMessageCount() {return _msgCount;}
 private:
    boost::shared_ptr<provider::DDSConnection> _ddsConnection;
    int _msgCount;
};

ExampleListener::
ExampleListener(boost::shared_ptr<provider::DDSConnection> ddsConnection) {
    _ddsConnection = ddsConnection;
    _msgCount = 0;
}

void ExampleListener::
processMessage(boost::shared_ptr<rmsexample::RmsExampleMessage> message) {
    //std::cout << "Received a message: ";
    //std::cout << message->message << std::endl;

    // Create a destination to represent where replies are to be sent.
    std::string replyTarget(message->header.replyTarget);
    base::RmsDestination sendDest(replyTarget,
                                  base::RmsDestination::CONTROL_CHANNEL,
                                  base::RmsDestination::REQUEST_MESSAGE_TYPE);

    // Create the sender with the connection and the destination.
    RmsSender<provider::DDSConnection, rmsexample::RmsExampleMessage>
        replySender(_ddsConnection, sendDest);

    // Create and send the reply.
    rmsexample::RmsExampleMessage replyMessage;
    replyMessage.message = "Hello World Reply!";
    std::string id = "RmsReceiverExample " + _ddsConnection->getUUID();
    replyMessage.header.id = DDS::string_dup(id.c_str());
    //std::cout << "Sending the reply." << std::endl;
    replySender.sendMessage(replyMessage);

    // Clean up
    replySender.close();

    ++_msgCount;
}

int main() {
    boost::shared_ptr<provider::DDSConnection> rmsConnection;

    // Create an RmsConnection object using the connection factory.
    // We want replies routed to the destination with a target equal
    // to "RmsReceiverExample" so we set that as the applicationName
    // and we use the NULL partition since we don't care about partitions.
    rmsConnection.reset(
        new provider::DDSConnection("RmsReceiverExample",
                                    base::RmsDestination::NULL_PARTITION)
        );

    // Create a destination to represent where we want to receive messages
    // from.  For the example, we're receiving messages to the
    // "RmsReceiverExample" target.  The messages will use the control channel
    // and be request messages.
    base::RmsDestination receiveDest("RmsReceiverExample",
                                     base::RmsDestination::CONTROL_CHANNEL,
                                     base::RmsDestination::REQUEST_MESSAGE_TYPE);

    // Create a receiver with the connection object and destination.
    RmsReceiver<provider::DDSConnection, rmsexample::RmsExampleMessage,
                ExampleListener>
        requestReceiver(rmsConnection, receiveDest);
    boost::shared_ptr<ExampleListener>
        theListener(new ExampleListener(rmsConnection));
    requestReceiver.setListener(theListener);

    // Wait for at least one message to be received.
    //std::cout << "Waiting for a message..." << std::endl;
    while (theListener->getMessageCount() < 20000) {
        usleep(250000);
    }

    // Cleanup, and exit.
    requestReceiver.close();
    rmsConnection->close();

    return 0;
}
