#include <rms/provider/CETDDS.h>
#include <rms/provider/DDSConnection.h>
#include <rms/base/RmsDestination.h>
#include <rms/RmsReceiver.h>
#include <rms/RmsSender.h>

#include <rmsexample/ccpp_RmsExampleMessages.h>

#include <boost/shared_ptr.hpp>
#include <boost/lexical_cast.hpp>
#include <unistd.h>

using namespace gov::fnal::cd::rms;

/**
 * This is simple example showing how to receive messages
 * and send replies using RMS.
 */

int main(int argc, char **argv) {

    bool verboseFlag = false;
    bool showIdleStatus = false;
    bool skipReply = false;
    std::string thisAppName = "RmsReceiverExample";
    int loopCount = 10;
    int optionChar;
    while ((optionChar = getopt (argc, argv, "a:vin:R")) != -1) {
        switch (optionChar) {
        case 'a':
            thisAppName = optarg;
            break;
        case 'n':
            try {
                loopCount = boost::lexical_cast<int>(optarg);
            }
            catch (boost::bad_lexical_cast &excpt) {
                std::cerr << "Error: loopCount must be numeric."
                          << std::endl;
                return 3;
            }
            break;
        case 'i':
            showIdleStatus = true;
            break;
        case 'v':
            verboseFlag = true;
            break;
        case 'R':
            skipReply = true;
            break;
        case '?':
            std::cout << "Usage: " << argv[0]
                      << " [-v] [-i] [-R] [-n loopCount] [-a appName]"
                      << std::endl;
            std::cout << "  where -v enables all diagnostic messages,"
                      << std::endl;
            std::cout << "        -i enables just the idle status printout."
                      << std::endl;
            std::cout << "        -R skips the sending of replies."
                      << std::endl;
            std::cout << "    and the application name is "
                      << "used in the routing of the messages." << std::endl;
            return 1;
        default:
            abort ();
        }
    }

    /**
     * Create an RmsConnection object using the connection factory.
     * We want replies routed to the destination with a target equal
     * to "RmsReceiverExample" so we set that as the applicationName
     * and we use the NULL partition since we don't care about partitions.
     */
    boost::shared_ptr<provider::DDSConnection> rmsConnection;
    rmsConnection.reset(
        new provider::DDSConnection(thisAppName,
                                    base::RmsDestination::NULL_PARTITION)
        );

    /**
     * Create a destination to represent where we want to receive messages
     * from.  For the example, we're receiving messages to the
     * "RmsReceiverExample" target.  The messages will use the control channel
     * and be request messages.
     */
    base::RmsDestination receiveDest(thisAppName,
                                     base::RmsDestination::CONTROL_CHANNEL,
                                     base::RmsDestination::REQUEST_MESSAGE_TYPE);

    /**
     * Create a receiver with the connection object and destination.
     */
    RmsReceiver<provider::DDSConnection, rmsexample::RmsExampleMessage>
        requestReceiver(rmsConnection, receiveDest);

    std::map<std::string, boost::shared_ptr< RmsSender<provider::DDSConnection,
        rmsexample::RmsExampleMessage> > > senderTable;

    for (int idx = 0; idx < loopCount; ++idx) {

        /**
         * Wait for a message to show up.
         */
        if (verboseFlag) {
            std::cout << "Waiting for a message..." << std::endl;
        }
        rmsexample::RmsExampleMessage incomingMessage;
        while (! requestReceiver.receiveMessage(incomingMessage, 2000000)) {
            if (verboseFlag || showIdleStatus) {
                std::cout << idx << " messages received so far." << std::endl;
            }
        }
        if (verboseFlag) {
            std::cout << "Received a message: ";
            std::cout << idx << " " << incomingMessage.message << std::endl;
        }

        if (skipReply) {continue;}

        std::string replyTarget(incomingMessage.header.replyTarget);
        boost::shared_ptr< RmsSender<provider::DDSConnection,
            rmsexample::RmsExampleMessage> > replySender;
        std::map<std::string,
            boost::shared_ptr< RmsSender<provider::DDSConnection,
            rmsexample::RmsExampleMessage> > >::const_iterator senderIter;
        senderIter = senderTable.find(replyTarget);
        if (senderIter == senderTable.end()) {

            /**
             * Create a destination to represent where replies are to be sent.
             */
            base::RmsDestination sendDest(replyTarget,
                                          base::RmsDestination::CONTROL_CHANNEL,
                                          base::RmsDestination::REQUEST_MESSAGE_TYPE);

            /**
             * Create the sender with the connection and the destination.
             */
            replySender.reset(new RmsSender<provider::DDSConnection, rmsexample::RmsExampleMessage>(rmsConnection, sendDest));

            senderTable[replyTarget] = replySender;
        }
        else {
            replySender = senderIter->second;
        }

        /**
         * Create and send the reply.
         */
        rmsexample::RmsExampleMessage replyMessage;
        replyMessage.message = "Hello World Reply!";
        std::string id = thisAppName + " " + rmsConnection->getUUID();
        fillCorrelationFields<rmsexample::RmsExampleMessage,
            rmsexample::RmsExampleMessage>(incomingMessage, replyMessage);
        if (verboseFlag) {
            std::cout << "Sending the reply." << std::endl;
        }
        replySender->sendMessage(replyMessage);
    }
    std::cout << loopCount << " messages received." << std::endl;

    /**
     * Cleanup, and exit.
     */
    std::map<std::string, boost::shared_ptr< RmsSender<provider::DDSConnection,
        rmsexample::RmsExampleMessage> > >::iterator sendIter;
    for (sendIter = senderTable.begin(); sendIter != senderTable.end(); sendIter++) {
        sendIter->second->close();
    }
    requestReceiver.close();
    rmsConnection->close();

    return 0;
}
