#include <rms/RmsConsumer.h>
#include <rms/RmsProducer.h>
#include <rms/base/RmsDestination.h>
#include <rms/base/RmsMessage.h>
#include <rms/base/RmsMessageBody.h>
#include <rms/provider/LocalhostConnection.h>
#include <rms/provider/RmsConnection.h>
#include <rms/provider/RmsConnectionFactory.h>

#include <boost/shared_ptr.hpp>

#include <iostream>
#include <string>

using namespace std;
using namespace gov::fnal::cd::rms;

/**
 * A class to send and receive messages using the 
 * local host provider.
 *
 * @author Kurt Biery
 * @author Steve Foulkes
 * @version $Revision: 1.1.1.1 $ $Date: 2009/10/09 20:03:00 $
 */
class RmsSendAndReceive {
private:
  boost::shared_ptr<provider::RmsConnection> _myProvider; /**< Handle to the local host provider */
  RmsProducer *_myProducer;            /**< Handle to the producer */
  RmsConsumer *_myConsumer;            /**< Handle to the consumer */

public:
  /**
   * Create the destinations, localhost provider, consumer
   * and producer.
   */
  RmsSendAndReceive();

  /**
   * Call the destructor on the provider, consumer and producer
   * so that we can exit in an orderly fashion.
   */
  ~RmsSendAndReceive();

  /**
   * Send a message through the producer.
   */
  void sendMessage();

  /**
   * Send a string through the producer 
   */
  void sendString();

  /**
   * Receive a message through the consumer, and
   * send a reply.
   */
  void receiveMessage();

  /**
   * Receive a string through the consumer.
   */
  void receiveString();

  /**
   * Receive a reply through the producer.
   */
  void receiveReply();
};

/**
 * Create the destinations, localhost provider, consumer
 * and producer.
 */
RmsSendAndReceive::RmsSendAndReceive() {
  base::RmsDestination myDestination("configManager", "control", "request");

  _myConsumer = new RmsConsumer(_myProvider, myDestination);
  _myProducer = new RmsProducer(_myProvider, myDestination);
}

/**
 * Call the destructor on the provider, consumer and producer
 * so that we can exit in an orderly fashion.
 */
RmsSendAndReceive::~RmsSendAndReceive() {
  delete _myConsumer;
  delete _myProducer;
}

/**
 * Send a message through the producer.
 */
void RmsSendAndReceive::sendMessage() {
  base::RmsMessage myMessage;
  base::RmsMessageBody myMessageBody;

  myMessageBody.setContent("Hello World!");
  myMessage.setBody(myMessageBody);

  cout << "Sending the following:" << endl;
  cout << myMessageBody.getContent() << endl;
  cout << endl;

  _myProducer->sendMessage(myMessage);
}

/**
 * Send a string through the producer.
 */
void RmsSendAndReceive::sendString() {
  string myString("Test String");

  cout << "Sending the following string:" << endl;
  cout << "Test String" << endl;
  cout << endl;

  _myProducer->sendString(myString);
}

/**
 * Receive message from the consumer, and send back
 * a reply.
 */
void RmsSendAndReceive::receiveMessage() {
  boost::shared_ptr <base::RmsMessage> incomingMessage;
  boost::shared_ptr <base::RmsMessage> outgoingReply;

  incomingMessage = _myConsumer->receiveMessage();

  cout << "Received the following:" << endl;
  cout << incomingMessage->getBody().getContent() << endl;
  cout << endl;

  outgoingReply = incomingMessage->createReply();

  base::RmsMessageBody replyMessageBody;
  replyMessageBody.setContent("Hello World! REPLY!!");
  outgoingReply->setBody(replyMessageBody);

  _myConsumer->sendReply(outgoingReply);
}

/**
 * Receive a string from the consumer.
 */
void RmsSendAndReceive::receiveString() {
  boost::shared_ptr <string> incomingString;

  incomingString = _myConsumer->receiveString();

  cout << "Received the following string:" << endl;
  cout << *incomingString << endl;
  cout << endl;
}

/**
 * Send a message through the producer.
 */
void RmsSendAndReceive::receiveReply() {
  boost::shared_ptr <base::RmsMessage> incomingReply;

  incomingReply = _myProducer->receiveReply();

  cout << "Received the following reply:" << endl;
  cout << incomingReply->getBody().getContent() << endl;
  cout << endl;
}  

/**
 * Intstantiate a RmsSendAndReceive object and use it to
 * send a series of messages between a consumer and a
 * producer.
 */  
int main(int argc, char *argv[]) {
  RmsSendAndReceive testObject;

  if(argc > 2) {
    cout << "Usage:" << endl;
    cout << "./RmsSendAndReceive [ -strings ]" << endl;
    return 0;
  }

  if(argc == 2) {
    if(strcmp(argv[1], "-strings") == 0) {
      cout << "Sending strings instead of messages" << endl;

      for(int i = 0; i < 10; i++) {
	testObject.sendString();
	testObject.receiveString();
      }

      return 0;
    }
  }
  
  for(int i = 0; i < 10; i++) {
    testObject.sendMessage();
    testObject.receiveMessage();
    testObject.receiveReply();
  }

  // close
  
  return 1;
}

