#include <rms/provider/CETDDS.h>
#include <rms/provider/DDSConnection.h>
#include <rms/base/RmsDestination.h>
#include <rms/RmsReceiver.h>
#include <rms/RmsSender.h>

#include <rmsexample/ccpp_RmsExampleMessages.h>

#include <boost/shared_ptr.hpp>
#include <boost/lexical_cast.hpp>
#include <unistd.h>

using namespace gov::fnal::cd::rms;

/**
 * This is a simple example showing how to send messages
 * and receive replies using RMS.
 */

int main(int argc, char **argv) {

    bool verboseFlag = false;
    bool showIdleStatus = false;
    bool skipReply = false;
    std::string thisAppName = "RmsSenderExample";
    std::string targetName = "RmsReceiverExample";
    int loopCount = 10;
    int sleepTime = 0;
    int optionChar;
    while ((optionChar = getopt (argc, argv, "via:n:s:t:R")) != -1) {
        switch (optionChar) {
        case 'a':
            thisAppName = optarg;
            break;
        case 'n':
            try {
                loopCount = boost::lexical_cast<int>(optarg);
            }
            catch (boost::bad_lexical_cast &excpt) {
                std::cerr << "Error: loopCount must be numeric."
                          << std::endl;
                return 3;
            }
            break;
        case 's':
            try {
                sleepTime = boost::lexical_cast<int>(optarg);
            }
            catch (boost::bad_lexical_cast &excpt) {
                std::cerr << "Error: sleepTime must be numeric."
                          << std::endl;
                return 5;
            }
            break;
        case 't':
            targetName = optarg;
            break;
        case 'i':
            showIdleStatus = true;
            break;
        case 'v':
            verboseFlag = true;
            break;
        case 'R':
            skipReply = true;
            break;
        case '?':
            std::cout << "Usage: " << argv[0]
                      << " [-v] [-i] [-R] [-n loopCount] [-s sleepTime] "
                      << "[-a appName] [-t targetName] " << std::endl;
            std::cout << "  where -v enables all diagnostic messages,"
                      << std::endl;
            std::cout << "        -i enables just the idle status printout,"
                      << std::endl;
            std::cout << "        -R skips waiting for replies."
                      << std::endl;
            std::cout << "        -s specifies a non-zero sleep time "
                      << "(in microseconds) between messages,"
                      << std::endl;
            std::cout << "    and the application and target names are "
                      << "used in the routing of the messages." << std::endl;
            return 1;
        default:
            abort ();
        }
    }
    std::cout << "Starting sender app with name \"" << thisAppName
              << "\"." << std::endl;

    /**
     * Create an RmsConnection object using the connection factory.
     * We want replies routed to the destination with a target equal
     * to "RmsSenderExample" so we set that as the applicationName
     * and we use the NULL partition since we don't care about partitions.
     */
    boost::shared_ptr<provider::DDSConnection> rmsConnection;
    rmsConnection.reset(
        new provider::DDSConnection(thisAppName,
                                    base::RmsDestination::NULL_PARTITION)
        );

    /**
     * Create a destination to represent where replies are received from.
     */
    base::RmsDestination receiveDest(thisAppName,
                                     base::RmsDestination::CONTROL_CHANNEL,
                                     base::RmsDestination::REQUEST_MESSAGE_TYPE);

    /**
     * Create a receiver with the connection and destination.
     * (Create the receiver before the sender so that it is ready to
     * receive replies.)
     */
    RmsReceiver<provider::DDSConnection, rmsexample::RmsExampleMessage>
        replyReceiver(rmsConnection, receiveDest);

    /**
     * Create a destination to represent where the messages that are
     * generated here will go to.  For the example, we're sending messages
     * to the "RmsReceiverExample" target.  The messages will use the control
     * channel and be request messages.
     */
    base::RmsDestination sendDest(targetName,
                                  base::RmsDestination::CONTROL_CHANNEL,
                                  base::RmsDestination::REQUEST_MESSAGE_TYPE);

    /**
     * Create a sender with the connection and destination.
     */
    RmsSender<provider::DDSConnection, rmsexample::RmsExampleMessage>
        requestSender(rmsConnection, sendDest);

    int loopsPerSleep = 1;
    int trueSleepTime = sleepTime;
    if (sleepTime > 0 && sleepTime <= 1000) {
        trueSleepTime = 1000;
        loopsPerSleep = (int) (1000 / sleepTime);
    }

    for (int idx = 0; idx < loopCount; ++idx) {

        if (trueSleepTime > 0 && ((idx+1) % loopsPerSleep) == 0) {
            if (verboseFlag) {
                std::cout << "Sleeping for " << trueSleepTime
                          << " microseconds [" << idx << "]." << std::endl;
            }
            usleep(trueSleepTime);
        }

        /**
         * Create the message and send it.
         */
        rmsexample::RmsExampleMessage exampleMessage;
        exampleMessage.message = "Hello World!";
        requestSender.sendMessage(exampleMessage);

        if (! skipReply) {
            /**
             * Wait for a reply.
             */
            rmsexample::RmsExampleMessage incomingReply;
            while (! replyReceiver.receiveMessage(incomingReply, 2000000)) {
                if (verboseFlag || showIdleStatus) {
                    std::cout << idx << " replies received so far." << std::endl;
                }
            }

            if (verboseFlag) {
                std::cout << idx << " " << incomingReply.message << std::endl;
            }
        }
    }
    std::cout << loopCount << " messages sent." << std::endl;

    /**
     * Cleanup, and exit.
     */
    replyReceiver.close();
    requestSender.close();
    rmsConnection->close();

    std::cout << "Exiting sender app with name \"" << thisAppName
              << "\"." << std::endl;

    return 0;
}
