#include <NovaDAQUtilities/CppUnitTestDriver.h>
#include <cppunit/extensions/HelperMacros.h>

#include <rms/RmsMessageSchema.h>
#include <rms/util/TimeUtils.h>

#include <boost/regex.hpp>

class BaseRmsmMessageTest : public CppUnit::TestFixture
{
    CPPUNIT_TEST_SUITE(BaseRmsmMessageTest);
    CPPUNIT_TEST(serialization);
    CPPUNIT_TEST(deserialization);
    CPPUNIT_TEST(reserialization);
    CPPUNIT_TEST_SUITE_END();

public:
    void setUp();
    void tearDown();

    void serialization();
    void deserialization();
    void reserialization();

private:

};

void
BaseRmsmMessageTest::setUp()
{ 
}

void
BaseRmsmMessageTest::tearDown()
{ 
}

void 
BaseRmsmMessageTest::serialization()
{
    // build up a sample message using the underlying XSD 
    // message classes
    gov::fnal::cd::rms::xsd::messageSource msource;
    msource.hostName("localhost");
    msource.applicationName("MyApplication");

    gov::fnal::cd::rms::xsd::messageBody mbody;
    mbody.content("Hello, World");
    mbody.className("a.b.c.d");

    gov::fnal::cd::rms::xsd::property prop1;
    prop1.name("target");
    prop1.value("dcm005");
    gov::fnal::cd::rms::xsd::property prop2;
    prop2.name("service");
    prop2.value("control");

    gov::fnal::cd::rms::xsd::destination dest1;
    dest1.property().push_back(prop1);
    dest1.property().push_back(prop2);

    gov::fnal::cd::rms::xsd::property prop3;
    prop3.name("target");
    prop3.value("RunControl");
    gov::fnal::cd::rms::xsd::property prop4;
    prop4.name("service");
    prop4.value("control");

    gov::fnal::cd::rms::xsd::destination dest2;
    dest2.property().push_back(prop3);
    dest2.property().push_back(prop4);

    gov::fnal::cd::rms::xsd::rmsMessage rmsMsg;
    rmsMsg.body(mbody);
    rmsMsg.destination(dest1);
    rmsMsg.replyDestination(dest2);
    rmsMsg.id("IDIDIDID");
    rmsMsg.timeToLive(5000);
    rmsMsg.sentTime(
        gov::fnal::cd::rms::util::TimeUtils::convertPtimeToString(
        gov::fnal::cd::rms::util::TimeUtils::getUniversalPtime()));
    rmsMsg.source(msource);

    // serialize the message to a string
    std::string xmlString = rmsMsg.serialize();
    //std::cout << xmlString << std::endl;

    // it would be a little difficult to match the full XML string
    // to a standard string, so we check for expected substrings
    std::string patString;

    // check the message destination
    patString.clear();
    patString.append(".*?<destination>");
    patString.append("\\s*<property\\s+name=\"target\">");
    patString.append("\\s*<value>dcm005</value>");
    patString.append("\\s*</property>");
    patString.append("\\s*<property\\s+name=\"service\">");
    patString.append("\\s*<value>control</value>");
    patString.append("\\s*</property>");
    patString.append("\\s*</destination>");
    patString.append(".*");
    boost::regex pat1(patString); //, boost::regex_constants::icase);
    CPPUNIT_ASSERT(boost::regex_match(xmlString, pat1));

    // check the reply destination
    patString.clear();
    patString.append(".*?<replyDestination>");
    patString.append("\\s*<property\\s+name=\"target\">");
    patString.append("\\s*<value>RunControl</value>");
    patString.append("\\s*</property>");
    patString.append("\\s*<property\\s+name=\"service\">");
    patString.append("\\s*<value>control</value>");
    patString.append("\\s*</property>");
    patString.append("\\s*</replyDestination>");
    patString.append(".*");
    boost::regex pat2(patString); //, boost::regex_constants::icase);
    CPPUNIT_ASSERT(boost::regex_match(xmlString, pat2));

    // check the source
    patString.clear();
    patString.append(".*?<source\\s+applicationName=\"MyApplication\"");
    patString.append("\\s+hostName=\"localhost\"/>");
    patString.append(".*");
    boost::regex pat3(patString); //, boost::regex_constants::icase);
    CPPUNIT_ASSERT(boost::regex_match(xmlString, pat3));

    // check the message body
    patString.clear();
    patString.append(".*?<body className=\"a.b.c.d\">");
    patString.append("\\s*<content>Hello, World</content>");
    patString.append("\\s*</body>");
    patString.append(".*");
    boost::regex pat4(patString); //, boost::regex_constants::icase);
    CPPUNIT_ASSERT(boost::regex_match(xmlString, pat4));

    // check the outermost tags
    patString.clear();
    patString.append("\\s*<\\?xml.*?\\?>");
    patString.append("\\s*<rmsMessage\\s+id=\"IDIDIDID\"");
    patString.append("\\s+sentTime=\".*?\"\\s+timeToLive=\"5000\"");
    patString.append("\\s*>.*?");
    patString.append("\\s*</rmsMessage>\\s*");
    boost::regex pat5(patString); //, boost::regex_constants::icase);
    CPPUNIT_ASSERT(boost::regex_match(xmlString, pat5));
}

void 
BaseRmsmMessageTest::deserialization()
{
    // build up a sample message using the underlying XSD 
    // message classes
    gov::fnal::cd::rms::xsd::messageSource msource;
    msource.hostName("localhost");
    msource.applicationName("MyApplication");

    gov::fnal::cd::rms::xsd::messageBody mbody;
    mbody.content("Hello, World");
    mbody.className("a.b.c.d");

    gov::fnal::cd::rms::xsd::property prop1;
    prop1.name("target");
    prop1.value("dcm005");
    gov::fnal::cd::rms::xsd::property prop2;
    prop2.name("service");
    prop2.value("control");

    gov::fnal::cd::rms::xsd::destination dest1;
    dest1.property().push_back(prop1);
    dest1.property().push_back(prop2);

    gov::fnal::cd::rms::xsd::property prop3;
    prop3.name("target");
    prop3.value("RunControl");
    gov::fnal::cd::rms::xsd::property prop4;
    prop4.name("service");
    prop4.value("control");

    gov::fnal::cd::rms::xsd::destination dest2;
    dest2.property().push_back(prop3);
    dest2.property().push_back(prop4);

    gov::fnal::cd::rms::xsd::rmsMessage rmsMsg;
    rmsMsg.body(mbody);
    rmsMsg.destination(dest1);
    rmsMsg.replyDestination(dest2);
    rmsMsg.id("IDIDIDID");
    rmsMsg.timeToLive(5000);
    rmsMsg.sentTime(
        gov::fnal::cd::rms::util::TimeUtils::convertPtimeToString(
        gov::fnal::cd::rms::util::TimeUtils::getUniversalPtime()));
    rmsMsg.source(msource);

    // serialize the message to a string
    std::string xmlString = rmsMsg.serialize();

    // deserialize the string into a second message object
    gov::fnal::cd::rms::xsd::rmsMessage msgCopy;
    msgCopy.deserialize(xmlString);

    // check that the IDs match
    CPPUNIT_ASSERT(msgCopy.hasId() && rmsMsg.hasId() &&
                   msgCopy.id() == rmsMsg.id());

    // need to add more tests...
}

void 
BaseRmsmMessageTest::reserialization()
{
    // build up a sample message using the underlying XSD 
    // message classes
    gov::fnal::cd::rms::xsd::messageSource msource;
    msource.hostName("localhost");
    msource.applicationName("MyApplication");

    gov::fnal::cd::rms::xsd::messageBody mbody;
    mbody.content("Hello, World");
    mbody.className("a.b.c.d");

    gov::fnal::cd::rms::xsd::property prop1;
    prop1.name("target");
    prop1.value("dcm005");
    gov::fnal::cd::rms::xsd::property prop2;
    prop2.name("service");
    prop2.value("control");

    gov::fnal::cd::rms::xsd::destination dest1;
    dest1.property().push_back(prop1);
    dest1.property().push_back(prop2);

    gov::fnal::cd::rms::xsd::property prop3;
    prop3.name("target");
    prop3.value("RunControl");
    gov::fnal::cd::rms::xsd::property prop4;
    prop4.name("service");
    prop4.value("control");

    gov::fnal::cd::rms::xsd::destination dest2;
    dest2.property().push_back(prop3);
    dest2.property().push_back(prop4);

    gov::fnal::cd::rms::xsd::rmsMessage rmsMsg;
    rmsMsg.body(mbody);
    rmsMsg.destination(dest1);
    rmsMsg.replyDestination(dest2);
    rmsMsg.id("IDIDIDID");
    rmsMsg.timeToLive(5000);
    rmsMsg.sentTime(
        gov::fnal::cd::rms::util::TimeUtils::convertPtimeToString(
        gov::fnal::cd::rms::util::TimeUtils::getUniversalPtime()));
    rmsMsg.source(msource);

    // serialize the message to a string
    std::string xmlString = rmsMsg.serialize();
    //std::cout << xmlString << std::endl;

    // deserialize the string into a second message object
    gov::fnal::cd::rms::xsd::rmsMessage msgCopy;
    msgCopy.deserialize(xmlString);

    // serialize the second message object to a second string
    std::string xmlString2 = msgCopy.serialize();
    //std::cout << xmlString2 << std::endl;

    // ensure that the two strings match
    CPPUNIT_ASSERT(xmlString2 == xmlString);
}

// This macro writes the 'main' for this test.
CPPUNIT_TEST_SUITE_REGISTRATION(BaseRmsmMessageTest);
