package gov.fnal.cd.rms;

import gov.fnal.cd.rms.base.RmsMessage;
import gov.fnal.cd.rms.base.RmsStatus;

import java.util.concurrent.LinkedBlockingQueue;

/**
 * A class for handling asynchronous message notifications.
 * ClientListenerLoop objects wait for messages to appear on a
 * specified input queue and notify a listener when they do.
 * In addition, a message filter can be specified to select
 * which notifications are forwarded to the listener based on
 * the message contents or other criteria.  
 *
 * Users should expect that this class will create one or
 * more background threads to manage the asynchronous notifications.
 *
 * @author Kurt Biery
 * @author Steve Foulkes
 * @version $Revision: 1.2 $ $Date: 2009/11/25 14:52:19 $
 * @see RmsConsumer.java
 * @see RmsProducer.java
 */
public class ClientListenerLoop {
    /**
     * A queue that messages are placed onto before they are
     * posted to any listeners.
     */
    private LinkedBlockingQueue <String> _inputQueue;

    /**
     * The handle to the client listener thread that waits
     * for messages to arrive.
     */
    private Thread _clientListenerThread;

    /**
     * The handle to the message listener that messages are
     * posted to as they arrive.
     */
    private RmsMessageListener _clientListener;

    /**
     * The handle to the messageFilter that messages are passed
     * through to make sure that they are relevant to the listener.
     */
    private MessageFilter _messageFilter;

    private class ProcessingThread implements Runnable {
        /**
         * Retrieve replies off of the input queue and
         * pass them to the client call back.
         */
        public void run() {
            String messageText;

            while (!Thread.currentThread().isInterrupted()) {
                try {
                    messageText = (String)_inputQueue.take();
                }
                catch (InterruptedException excp) {
                    return;
                }

                RmsMessage receivedMessage = new RmsMessage();
                try {
                    receivedMessage.deserialize(messageText);
                }
                catch (novadaq.DataException excp) {
                    RmsStatus replyStatus = new RmsStatus(excp);
                    receivedMessage.setMessageStatus(replyStatus);
                }

                if (_messageFilter == null ||
                    _messageFilter.verify(receivedMessage)) {
                    if (_clientListener != null) {
                        _clientListener.messageReceived(receivedMessage);
                    }
                }
            }
        }
    }  

    /**
     * Create a new ClientListenerLoop with a specific input
     * queue and message listener.
     *
     * @param inputQueue The queue to pull messages from
     * @param listener The listener to send messages to
     * @param messageFilter The filter that should be used when deciding
     *        whether messages should be added to the queue, or null
     *        if no filter is desired
     */
    public ClientListenerLoop (LinkedBlockingQueue <String> inputQueue,
                               RmsMessageListener listener, 
                               MessageFilter messageFilter) {
        _inputQueue = inputQueue;
        _clientListener = listener;
        _messageFilter = messageFilter;
        
        _clientListenerThread = new Thread(new ProcessingThread());
        _clientListenerThread.start();
    }

    /**
     * Change the message listener to be the one that is 
     * passed into this function.
     *
     * @param listener The new RmsMessageListener.
     */
    public void setMessageListener(RmsMessageListener listener) {
        _clientListener = listener;
    }

    /**
     * Change the message filter to be the one that is
     * passed into this function.
     *
     * @param filter The new MessageFilter.
     */
    public void setMessageFilter(MessageFilter filter) {
        _messageFilter = filter;
    }

    /**
     * Stops the listener loop.  After this method is called, the
     * listener loop should no longer be used for anything since
     * it will no longer be capable of responding to new messages.
     */
    public void stop() {
        if (_clientListenerThread != null) {

            // interrupt the listener thread
            _clientListenerThread.interrupt();

            // if the listener thread is different than the current thread, join
            // it to ensure that when we exit this method, it has really died
            Thread currentThread = Thread.currentThread();
            if (_clientListenerThread != null && currentThread != null &&
                ! _clientListenerThread.equals(currentThread)) {
                try {
                    _clientListenerThread.join();
                }
                catch (InterruptedException iExcpt) {
                    // it's really hard to imagine how this operation would be
                    // interrupted and why we would care, so it seems reasonable
                    // to simply print the exception to the console
                    iExcpt.printStackTrace(System.err);
                }
            }
        }
        _clientListenerThread = null;
    }

    /**
     * Return the RmsMessageListener.
     *
     * @return The current RmsMessageListener
     */
    public RmsMessageListener getReplyListener() {
        return _clientListener;
    }

    /**
     * Cleans up during garbage collection.
     */
    protected void finalize() {
        stop();
    }
}
