package gov.fnal.cd.rms;

import gov.fnal.cd.rms.provider.RmsConnection;
import gov.fnal.cd.rms.provider.RmsConnectionFactory;

import gov.fnal.cd.rms.base.RmsDestination;
import gov.fnal.cd.rms.base.RmsMessage;
import gov.fnal.cd.rms.RmsConsumer;

import java.util.LinkedList;

/**
 * Poll for heartbeat messages from a particular destination.
 *
 * @author Kurt Biery
 * @author Steve Foulkes
 * @version $Revision: 1.1.1.1 $ $Date: 2009/10/09 20:03:00 $
 */
public class HeartbeatConsumer {
    /**
     * Handle to the rmsConnection, needed to create producers
     * and consumers.
     */
    private RmsConnection _myConnection;

    /**
     * RMS Consumer to receive the heartbeat messages.
     */
    private RmsConsumer _myConsumer;

    /**
     * List of heartbeat responses that have been received.
     */
    private LinkedList <Heartbeat> _heartbeatResponses;

    /**
     * Parse the command line parameters, create a connection to RMS
     * and then start polling for heartbeat messages.
     *
     * @param args The command line arguments wheres args[0] is the target
     * name and args[1] is the heartbeat interval.
     */
    public static void main(String[] args) {
	HeartbeatConsumer testObject = new HeartbeatConsumer();
	Heartbeat receivedHeartbeat = null;

	if (args.length != 2) {
	    System.out.println("Usage: HeartbeatConsumer [target name] " + 
                               "[interval]");
	    
	    System.exit(0);
	}

	int interval = 0;
	try {
	    interval = Integer.parseInt(args[1], 10);
	} catch (Exception excp) {}

	testObject.initializeRms(args);

	testObject._heartbeatResponses = new LinkedList<Heartbeat>();

	testObject.pollHeartbeats(interval);
    }

    /**
     * Go through the list of _hearbeatResponses and check to
     * see if we are missing updates from any of the applications
     * we have received heartbeats from so far.
     *
     * @param interval The maximum amount of time we should expect
     * between two heartbeat messages.
     */
    private void checkForDead(int interval) {
	Heartbeat loggedHeartbeat;
	long currentTime = System.currentTimeMillis();
	long deltaTime;

	if (_heartbeatResponses.size() == 0) {
	    System.out.println("Have not received any heartbeats.");
	    return;
	}

	for (int i = 0; i < _heartbeatResponses.size(); i++) {
	    loggedHeartbeat = _heartbeatResponses.get(i);

	    deltaTime = currentTime - loggedHeartbeat.getTimestamp();

	    if (deltaTime / 1000 > interval) {
		System.out.println("");
		System.out.println(loggedHeartbeat.getApplicationName() + 
                                   " failed to report within " +
				   interval + " seconds.");
	    }
	    else {
		System.out.print(".");
	    }
	}
    }

    /**
     * Check to see if we have received a heartbeat from a particular
     * application before.  If this is the first time we've heard from
     * this application, add it to the list otherwise find it on the 
     * list and update its timestamp.
     *
     * @param incomingHeartbeat The heartbeat we just received.
     */
    private void addApplication(Heartbeat incomingHeartbeat) {
	Heartbeat loggedHeartbeat;

	for (int i = 0; i < _heartbeatResponses.size(); i++) {
	    loggedHeartbeat = _heartbeatResponses.get(i);

	    if (loggedHeartbeat.getApplicationName().equals(incomingHeartbeat.getApplicationName()) &&
                loggedHeartbeat.getApplicationVersion().equals(incomingHeartbeat.getApplicationVersion())) {
		_heartbeatResponses.set(i, incomingHeartbeat);
		return;
	    }
	}

	_heartbeatResponses.add(incomingHeartbeat);

	System.out.println("Received a heartbeat: " + 
                           incomingHeartbeat.getApplicationName() +
			   ", version " + 
                           incomingHeartbeat.getApplicationVersion());
    }

    /**
     * Receive messages from the consumer, waiting at most interval
     * seconds.  Deserialize the incoming message and add it to
     * the list of received heartbeats.  We check the list for dead
     * heartbeat producers on each loop iteration.
     *
     * @param interval The maximum amount of time we should expect
     * between two heartbeat messages.
     */
    private void pollHeartbeats(int interval) {
	RmsMessage incomingMessage;
	Heartbeat incomingHeartbeat;

	while (true) {
	    incomingMessage = _myConsumer.receiveMessage(1000 * interval);

	    if (incomingMessage != null) {
		try {
		    incomingHeartbeat = (Heartbeat)incomingMessage.getBodyObject(); 
		} 
		catch (Exception excp) {
		    System.out.println("There was a problem de-serializing " +
                                       "the incoming heartbeat:");
		    System.out.println(excp);
		    continue;
		}

		addApplication(incomingHeartbeat);
	    }

	    checkForDead(interval);
	}	
    }

    /**
     * Create provider, connection and consumer objects based
     * upon what was passed in from the command line.
     *
     * @param args The command line arguments where args[0] is the 
     * target name.
     */ 
    private void initializeRms(String[] args) {
	RmsDestination myDestination;

	try {
	    _myConnection = RmsConnectionFactory.getConnection("rmsHeartbeat", -1);
	}
	catch (Exception excp) {
	    System.out.println("Had some problems creating the connection:");
	    System.out.println(excp);
	
	    return;
	}

	myDestination = new RmsDestination(args[0], 
                                           RmsDestination.HEARTBEAT_CHANNEL,
					   RmsDestination.NOTIFICATION_MESSAGE_TYPE);

	try {
	    _myConsumer = new RmsConsumer(_myConnection, myDestination);
	}
	catch (Exception excp) {
	    System.out.println("Failed to create a consumer:");
	    System.out.println(excp);
	}

	return;
    }
}
