package gov.fnal.cd.rms;

import gov.fnal.cd.rms.Heartbeat;
import gov.fnal.cd.rms.provider.RmsConnection;

import gov.fnal.cd.rms.base.RmsDestination;
import gov.fnal.cd.rms.base.RmsMessage;
import gov.fnal.cd.rms.RmsProducer;

/**
 * This class creates a thread that periodically sends out
 * heartbeat messages based on the parameters that the
 * instantiating applications passes in.
 *
 * @author Kurt Biery
 * @author Steve Foulkes
 * @version $Revision: 1.1.1.1 $ $Date: 2009/10/09 20:03:00 $
 */
public class HeartbeatProducer {
    /**
     * Handle to an RmsConnection that is used to create
     * producers and consumers.
     */
    private RmsConnection _myConnection;

    /**
     * An RmsProducer used to send out heartbeats.
     */
    private RmsProducer _myProducer;

    /**
     * The destination that heartbeats are supposed
     * to be sent to.
     */
    private RmsDestination _myDestination;

    /**
     * The amount of time, in seconds, between heartbeat
     * messages.
     */
    private int _heartbeatInterval;

    /**
     * Handle to the thread that actually sends out the 
     * messages.
     */
    private Thread _heartbeatThread;

    /**
     * Instance of the heartbeat class that manages
     * everything.
     */
    private Heartbeat _heartbeat;
    
    /**
     * The message to send out.
     */
    private RmsMessage _heartbeatMessage;

    /**
     * Update the timestamp in the heartbeat message and
     * broadcast the heartbeat.
     */
    private void sendHeartbeat() {
	_heartbeat.setTimestamp(System.currentTimeMillis());

	try {
	    _heartbeatMessage.setBody(_heartbeat);
	    _myProducer.sendMessage(_heartbeatMessage);
	}
	catch (Exception excp) {
	    System.out.println("Sending of the heartbeat failed:");
	    System.out.println(excp);
	}

	return;
    }

    /**
     * The actual heartbeat is quite simple: sleep for a certain
     * period of time and then send out a message.
     */
    private class HeartbeatThread implements Runnable {
	public void run() {
	    while (true) {
		sendHeartbeat();
		try {
		    Thread.sleep(_heartbeatInterval * 1000);
		}
		catch (Exception excp) {
		}		
	    }
	}
    }

    /**
     * Create a heartbeat producer and begin sending out heartbeats.
     *
     * @param applicationName The name of the application that wants to send
     * out heartbeat messages.
     * @param applicationVersion The version of the application that wants to
     * send out heartbeat messages.
     * @param interval The amount of time, in seconds to wait between 
     * broadcasting two heartbeat messages.
     * @param rmsConnection An initialized connection to RMS.
     */
    public HeartbeatProducer(String applicationName, String applicationVersion,
                             int interval, RmsConnection rmsConnection) {
	_myDestination = new RmsDestination(applicationName, "heartbeat", "notification");

	try {
	    _myProducer = new RmsProducer(rmsConnection, _myDestination);
	}
	catch (Exception excp) {
	    System.out.println("Exception when trying to create a producer:");
	    System.out.println(excp);
	    return;
	}

	_heartbeatInterval = interval;

	_heartbeat = new Heartbeat();
	_heartbeat.setApplicationName(applicationName);
	_heartbeat.setApplicationVersion(applicationVersion);

	try {
	    _heartbeatMessage = new RmsMessage(_heartbeat);
	}
	catch (Exception excp) {
	}

	_heartbeatThread = new Thread(new HeartbeatThread());
	_heartbeatThread.start();
    }
}
