package gov.fnal.cd.rms;

/**
 * @package gov.fnal.cd.rms
 *
 * @brief Classes that allow applications to send and receive
 * messages utilizing the Responsive Messaging System.
 *
 */

import gov.fnal.cd.rms.base.RmsCommunicationException;
import gov.fnal.cd.rms.base.RmsDestination;
import gov.fnal.cd.rms.base.RmsMessage;
import gov.fnal.cd.rms.base.RmsMessageBody;
import gov.fnal.cd.rms.base.RmsRuntimeException;
import gov.fnal.cd.rms.base.RmsStatus;
import gov.fnal.cd.rms.provider.BufferedProviderListener;
import gov.fnal.cd.rms.provider.RmsConnection;
import gov.fnal.cd.rms.provider.RmsProviderUtil;

import java.util.UUID;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;
import java.util.Hashtable;

/**
 * This class contains functionality for receiving messages and sending
 * replies.
 *
 * @author Kurt Biery
 * @author Steve Foulkes
 * @version $Revision: 1.2 $ $Date: 2009/11/25 14:52:19 $
 */
public class RmsConsumer implements MessageFilter {
    /**
     * Handle to an RmsConnection that is used to send and
     * receive messages.
     */
    private RmsConnection _myConnection = null;

    /**
     * The destination that this consumer is listening to.
     */
    private RmsDestination[] _myDestinations;

    /**
     * The ClientListenerLoop receives messages and puts them
     * onto the input queue.
     */
    private ClientListenerLoop _listenerLoop = null;

    /**
     * The queue that messages are placed onto before they are
     * sent out to listeners.
     */
    private LinkedBlockingQueue <String> _inputQueue;

    /**
     * The provider listener that is responsible for receiving
     * messages from the provider and storing them on our queue.
     */
    private BufferedProviderListener _providerListener = null;

    /**
     * A map from source destinations to corresponding target
     * destinations.  This is used A) to indicate that a particular
     * destination is a "source destination" (only source destinations
     * will have entries in the map) and B) to save us the trouble of
     * creating a corresponding "target destination" every time we need
     * to verify if a particular message matches a source destination.
     * (For reference, a "source destination" is an RmsDestination in
     * which the "source" property is set.  A "target destination" is
     * one in which the "target" property is set.  Normally, these two
     * properties are mutually exclusive in a single RmsDestination
     * [see the RmsDestination.validate() method], but we allow the
     * target destinations in this internal map to have both.)
     */
    private Hashtable<RmsDestination, RmsDestination> sourceToTargetMap =
        new Hashtable<RmsDestination, RmsDestination>();

    /**
     * Create a new consumer given a connection and a destination.
     *
     * @param parentConnection RmsConnection object which is used for receiving
     * and sending messages.
     * @param dest The destination that we are receiving messages from.
     * @throws RmsCommunicationException If an error occurs while
     * connecting to the destination.
     * @throws RmsRuntimeException If the destination is invalid.
     */
    public RmsConsumer(RmsConnection parentConnection, RmsDestination dest)
        throws RmsCommunicationException {
        // Here, we create the input queue and register a listener with
        // the provider for the messages that we are interested in.

        _myConnection = parentConnection;

        dest.validate();
        if (! _myConnection.supportsDestination(dest)) {
            throw new RmsRuntimeException("Connection " + _myConnection + 
                                          " does not support " +
                                          "destination " + dest);
        }

        _myDestinations = new RmsDestination[1];
        _myDestinations[0] = dest;

        // create an entry in the "source destination" to "target
        // destination" map if this destination is a source destination
        // (see the comments on the map declaration for more details)
        String targetProperty = _myDestinations[0].
            getProperty(RmsDestination.TARGET_PROPERTY_NAME);
        String sourceProperty = _myDestinations[0].
            getProperty(RmsDestination.SOURCE_PROPERTY_NAME);
        if (targetProperty == null && sourceProperty != null) {
            RmsDestination mappedDestination =
                new RmsDestination(_myDestinations[0]);
            mappedDestination.setProperty(RmsDestination.TARGET_PROPERTY_NAME,
                                          sourceProperty);
            sourceToTargetMap.put(_myDestinations[0], mappedDestination);
        }

        _inputQueue = new LinkedBlockingQueue<String>();

        _providerListener = new BufferedProviderListener(_inputQueue);
        _myConnection.addListener(_myDestinations[0], _providerListener);
    }

    /**
     * Create a new consumer given a connection and an array of 
     * destinations.
     *
     * @param parentConnection RmsConnection object which is used for receiving
     * and sending messages.
     * @param dest The array of destinations that we are receiving messages
     * from.
     * @throws RmsCommunicationException If an error occurs while
     * connecting to one of the destinations.
     * @throws RmsRuntimeException If any of the destinations are invalid.
     */
    public RmsConsumer(RmsConnection parentConnection, RmsDestination[] dest)
        throws RmsCommunicationException {
        // Here we create the input queue and register  a listener with the
        // provider for the messages that we are interested in.

        _myConnection = parentConnection;

        for (int i = 0; i < dest.length; i++) {
            dest[i].validate();
            if (! _myConnection.supportsDestination(dest[i])) {
                throw new RmsRuntimeException("Connection " + _myConnection + 
                                              " does not support " +
                                              "destination " + dest[i]);
            }
        }

        _myDestinations = dest;

        _inputQueue = new LinkedBlockingQueue<String>();

        _providerListener = new BufferedProviderListener(_inputQueue);

        for (int i = 0; i < _myDestinations.length; i++) {
            _myConnection.addListener(_myDestinations[i], _providerListener);

            // create an entry in the "source destination" to "target
            // destination" map if this destination is a source destination
            // (see the comments on the map for more details)
            String targetProperty = _myDestinations[i].
                getProperty(RmsDestination.TARGET_PROPERTY_NAME);
            String sourceProperty = _myDestinations[i].
                getProperty(RmsDestination.SOURCE_PROPERTY_NAME);
            if (targetProperty == null && sourceProperty != null) {
                RmsDestination mappedDestination =
                    new RmsDestination(_myDestinations[i]);
                mappedDestination.setProperty(RmsDestination.TARGET_PROPERTY_NAME,
                                              sourceProperty);
                sourceToTargetMap.put(_myDestinations[i], mappedDestination);
            }
        }
    }

    /**
     * Pull strings off the incoming message queue.  The messages
     * will not be deserialized, nor will they be verified to see
     * if they are relevant to this consumer.  This function will
     * block until it is able to return a string.
     *
     * @return A string that was received by this consumer.
     * @throws java.lang.InterruptedException If the blocking wait
     * for a message string is interrupted.
     */
    public String receiveString() throws InterruptedException {
        return (String)_inputQueue.take();
    }

    /**
     * Pull strings off the incoming message queue, waiting at most
     * timeout milliseconds.  The messages will not be deserialized, nor
     * will they be verified to see if they are relevant to this
     * consumer.  This function will block until it is able to return a
     * string, or the timeout expires, whichever comes first.
     *
     * @param timeout The maximum amount of time to wait for a message,
     * in milliseconds.
     * @return A string that was received by this consumer.
     * @throws java.lang.InterruptedException If the blocking wait
     * for a message string is interrupted.
     */
    public String receiveString(long timeout) throws InterruptedException {
        return (String)_inputQueue.poll(timeout, TimeUnit.MILLISECONDS);
    }

    /**
     * Pull strings off the incoming message queue.  This function
     * won't block.  The messages will not be deserialized, nor
     * will they be verified to see if they are relevant to this
     * consumer.  
     *
     * @return A string that was received by this consumer, or null
     * if the queue was empty.
     */
    public String receiveStringNoWait() {
        return (String)_inputQueue.poll();
    }

    /**
     * Pull messages off the incoming message queue.  This function will
     * block until a message is received.
     *
     * @return The received RmsMessage on success.  If an error occurs,
     * a non-null RmsMessage will be returned, and the RmsStatus
     * object inside the message will contain information on the error
     * (and the message body will be empty).
     */
    public RmsMessage receiveMessage() {
        RmsMessage receivedMessage = null;
        RmsStatus messageStatus = null;

        do {
            receivedMessage = new RmsMessage();
            try {
                String messageText = (String)_inputQueue.take();
                receivedMessage.deserialize(messageText);
            }
            catch (novadaq.DataException excp) {
                messageStatus = new RmsStatus(excp);
            }
            catch (InterruptedException excp) {
                messageStatus = new RmsStatus(excp);
            }
            if (messageStatus != null) {
                receivedMessage.setMessageStatus(messageStatus);
            }
        } while (!verify(receivedMessage));

        return receivedMessage;
    }

    /**
     * Pull messages off the incoming message queue.  This function will
     * block until a message is received, or the time expires; whichever
     * happens first.
     *
     * @param timeout The maximum amount of time in milliseconds to wait if
     * the message queue is empty before giving up and returning.
     * @return The received RmsMessage on success or null if the timeout
     * expires before a message is received.  If an error occurs,
     * a non-null RmsMessage will be returned, and the RmsStatus
     * object inside the message will contain information on the error
     * (and the message body will be empty).
     */
    public RmsMessage receiveMessage(long timeout) {
        RmsMessage receivedMessage = null;
        RmsStatus messageStatus = null;

        do {
            receivedMessage = new RmsMessage();
            try {
                String messageText = 
                    (String)_inputQueue.poll(timeout, TimeUnit.MILLISECONDS);

                if (messageText == null) {
                    return null;
                }

                receivedMessage.deserialize(messageText);
            }
            catch (novadaq.DataException excp) {
                messageStatus = new RmsStatus(excp);
            }
            catch (InterruptedException excp) {
                messageStatus = new RmsStatus(excp);
            }
            if (messageStatus != null) {
                receivedMessage.setMessageStatus(messageStatus);
            }
        } while (!verify(receivedMessage));

        return receivedMessage;
    }

    /**
     * Pull messages off the incoming message queue.  This function will
     * not block. 
     *
     * @return The received RmsMessage on success or null if no messages
     * are available.  If an error occurs,
     * a non-null RmsMessage will be returned, and the RmsStatus
     * object inside the message will contain information on the error
     * (and the message body will be empty).
     */
    public RmsMessage receiveMessageNoWait() {
        RmsMessage receivedMessage = null;
        RmsStatus messageStatus = null;

        do {
            receivedMessage = new RmsMessage();
            try {
                String messageText = (String)_inputQueue.poll();

                if (messageText == null) {
                    return null;
                }

                receivedMessage.deserialize(messageText);
            }
            catch (novadaq.DataException excp) {
                messageStatus = new RmsStatus(excp);
            }
            if (messageStatus != null) {
                receivedMessage.setMessageStatus(messageStatus);
            }
        } while (!verify(receivedMessage));

        return receivedMessage;
    }

    /**
     * Register a callback with the consumer.  The specified
     * listener will be notified of new messages as they arrive.
     *
     * @param listener The listener that will receive messages
     */
    public void setMessageListener(RmsMessageListener listener) {
        if (_listenerLoop == null) {
            _listenerLoop = new ClientListenerLoop(_inputQueue, 
                                                   listener, this);
        }
        else {
            _listenerLoop.setMessageListener(listener);
        }

        return;
    }

    /**
     * Return the RmsMessageListener, if there is one.
     *
     * @return The RmsMessageListener if one has been set,
     * otherwise null.
     */
    public RmsMessageListener getMessageListener() {
        if (_listenerLoop != null) {
            return _listenerLoop.getReplyListener();
        }
        else {
            return null;
        }
    }

    /**
     * Sends a reply message in response to a request that was received.
     * All reply messages must be created by calling the
     * createReply() method on the request message.
     * The only attribute of reply messages that needs to be set
     * by application code is the message body.
     *
     * @param message The reply that is being sent
     * @throws RmsCommunicationException If an error occurs when trying
     *         to send the message to the destination.
     * @throws novadaq.DataException If an error occurs when trying
     *         to prepare the message for sending.
     * @throws RmsRuntimeException If either of the destinations in the
     *         message are invalid.
     * @see RmsMessage.createReply()
     */
    public void sendReply(RmsMessage message)
        throws RmsCommunicationException, novadaq.DataException {
        RmsDestination replyDestination = _myConnection.getMessageSource();
        replyDestination.setProperty(RmsDestination.CHANNEL_PROPERTY_NAME,
                                     message.getDestination().getProperty(RmsDestination.CHANNEL_PROPERTY_NAME));
        replyDestination.validate();
        message.setReplyDestination(replyDestination);

        RmsProviderUtil.setStandardFields(message, _myConnection);

        message.getDestination().validate();
        _myConnection.sendMessage(message.getDestination(), message);
    }

    /**
     * Closes down the consumer.  After this method is called, the
     * consumer should no longer be used since it will no
     * longer be connected to the messaging system.
     *
     * @throws RmsCommunicationException If a problem occurs when
     * disconnecting from the messaging system.
     */
    public void close() throws RmsCommunicationException {
        RmsCommunicationException firstException = null;

        // remove the provider listener from all destinations
        if (_myConnection != null && _myDestinations != null &&
            _providerListener != null) {
            for (int i = 0; i < _myDestinations.length; i++) {
                // we want to remove as much as possible, so we keep going
                // if exceptions are encountered, but save the first one
                // for later reporting
                try {
                    _myConnection.removeListener(_myDestinations[i],
                                                 _providerListener);
                }
                catch (RmsCommunicationException rcExcpt) {
                    if (firstException == null) {
                        firstException = rcExcpt;
                    }
                }
            }
        }
        _providerListener = null;
        _myDestinations = null;
        _myConnection = null;

        // clear out the input queue
        if (_inputQueue != null) {
            _inputQueue.clear();
        }
        _inputQueue = null;

        // stop the listener
        if (_listenerLoop != null) {
            _listenerLoop.stop();
        }
        _listenerLoop = null;

        // report the first exception, if any
        if (firstException != null) {
            throw firstException;
        }
    }

    /**
     * Verify that any messages received are relevant to this
     * consumer.  This includes checking to see if the message
     * is a "loopback" message.  If this is the case, we process it
     * and send it back to its source.  
     *
     * If the consumer is consuming messages from any destination 
     * with a "source" parameter set we check the replyDestination
     * of the message as opposed to its actual destination.
     *
     * @param message The message to be checked
     * @return True if the message matches the destination that was
     * specified when the consumer was created, false otherwise.
     */
    public boolean verify (RmsMessage message) {
        if (message.getMessageStatus() == null ||
            message.getMessageStatus().getCode() != RmsStatus.SUCCESS_CODE) {
            return true;
        }

        String messageType = 
            message.getDestination().getProperty(RmsDestination.MESSAGE_TYPE_PROPERTY_NAME);
        if (messageType != null &&
            messageType.equals(RmsDestination.LOOPBACK_MESSAGE_TYPE)) {
            handleLoopback(message);
            return false;
        }

        for (int i = 0; i < _myDestinations.length; i++) {

            // check if the _myDestinations entry is a "source destination"
            // (this is indicated by the presence of an entry in the
            // source-to-target map).  If the entry is a "source destination",
            // we compare to the reply destination in the message (because
            // we don't care where the message is going *to*, we care where
            // it came *from*).
            RmsDestination mappedDestination =
                sourceToTargetMap.get(_myDestinations[i]);
            if (mappedDestination != null) {
                if (message.getReplyDestination().matches(mappedDestination)) {
                    return true;
                }
            }
            else {
                if (message.getDestination().matches(_myDestinations[i])) {
                    return true;
                }
            }
        }

        return false;
    }

    /**
     * Handles cleanup when garbage collected.
     */
    protected void finalize() {
        try {
            this.close();
        }
        catch (RmsCommunicationException rcExcpt) {
            rcExcpt.printStackTrace(System.err);
        }
    }

    /**
     * Handle the sending of loopback requests.  Generate a reply,
     * and send it out with the same body that was inside the request.
     *
     * @param loopbackMessage The message we received with the loopback
     * request.
     */
    private void handleLoopback(RmsMessage loopbackMessage) {
        RmsMessage loopbackReply;
        RmsMessageBody loopbackBody = new RmsMessageBody();

        loopbackBody.setContent(loopbackMessage.getBody().getContent());
        loopbackReply = loopbackMessage.createReply();
        loopbackReply.setBody(loopbackBody);

        try {
            sendReply(loopbackReply);
        }
        // TODO - improve this error handling, if needed, when we
        // fully implement the loopback functionality
        catch (Exception excp) {
            System.out.println("Received a loopback request, but " + 
                               "something happend:");
            System.out.println(excp);
        }
    }

    // Static initialization block to run an initial RmsMessage serialization
    // and deserialization.  This is useful because these operations can
    // have longer runtimes the first time they are called, and it
    // is good to incur that overhead *before* the first real message
    // is sent.
    static {
        try {
            RmsMessage dummyMessage = new RmsMessage();
            String dummyString = dummyMessage.serialize();
            dummyMessage = new RmsMessage(dummyString);
        }
        catch (Exception excpt) {}
    }
}
