package gov.fnal.cd.rms;

import gov.fnal.cd.rms.base.RmsCommunicationException;
import gov.fnal.cd.rms.base.RmsDestination;
import gov.fnal.cd.rms.base.RmsMessage;
import gov.fnal.cd.rms.base.RmsRuntimeException;
import gov.fnal.cd.rms.base.RmsStatus;
import gov.fnal.cd.rms.provider.BufferedProviderListener;
import gov.fnal.cd.rms.provider.RmsConnection;
import gov.fnal.cd.rms.provider.RmsProviderUtil;

import java.util.Date;
import java.util.UUID;
import java.util.LinkedHashMap;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;

/**
 * This class contains functionality for sending messages and receiving
 * replies.
 *
 * @author Kurt Biery
 * @author Steve Foulkes
 * @version $Revision: 1.2 $ $Date: 2009/11/25 14:52:19 $
 */
public class RmsProducer implements MessageFilter {
    /**
     * The RmsConnection that is used to send and receive messages.
     */
    private RmsConnection _myConnection;

    /**
     * An array of destinations that this producer sends messages to.
     */
    private RmsDestination[] _myDestinations;

    /**
     * Keep an array of all the unique channels we're sending
     * messages to so we can reconstruct all the reply destinations.
     */
    private HashSet <String> _myReplyChannels;

    /**
     * A handle to the client listener loop that listens
     * for messages and places them onto the input queue.
     */
    private ClientListenerLoop _listenerLoop = null;

    /**
     * The input queue holds messages until they are sent
     * out to listeners.
     */
    private LinkedBlockingQueue <String> _inputQueue;

    /**
     * The provider listener that is responsible for receiving
     * messages from the provider and storing them on our queue.
     */
    private BufferedProviderListener _providerListener = null;

    /**
     * The reply ID table holds the IDs of all messages
     * sent from this producer in order to determine
     * if replies received are relevant.
     */
    private LinkedHashMap <String, Date> _replyIdTable;

    /**
     * The reply timeout is hard coded to be 5 minutes.
     */
    private int _replyTimeout = 300; /* 5 Minutes */

    /**
     * Create a new producer given a connection and a destination.  
     *
     * @param parentConnection RmsConnection object which is used for receiving
     * and sending messages.
     * @param dest The destination that we are sending messages to.
     * @throws RmsCommunicationException If an error occurs while
     * connecting to the destination.
     * @throws RmsRuntimeException If either of the destinations are invalid.
     */
    public RmsProducer(RmsConnection parentConnection, RmsDestination dest)
        throws RmsCommunicationException {

        _myConnection = parentConnection;
        _myDestinations = new RmsDestination[1];
        _myDestinations[0] = dest;

        initialize();
    }

    /**
     * Create a new producer given a connection, a destination, and
     * a reply time window.
     *
     * @param parentConnection RmsConnection object which is used for receiving
     * and sending messages.
     * @param dest The destination that we are sending messages to.
     * @param replyTimeWindow The interval after a message
     * is sent, in seconds, in which a reply must be received to be
     * considered valid..
     * @throws RmsCommunicationException If an error occurs while
     * connecting to the destination.
     * @throws RmsRuntimeException If either of the destinations are invalid.
     */
    public RmsProducer(RmsConnection parentConnection, RmsDestination dest, int replyTimeWindow)
        throws RmsCommunicationException {

        _myConnection = parentConnection;
        _myDestinations = new RmsDestination[1];
        _myDestinations[0] = dest;
        _replyTimeout = replyTimeWindow;

        initialize();
    }

    /**
     * Create a new producer given a connection and an array of destinations.  
     *
     * @param parentConnection RmsConnection object which is used for receiving
     * and sending messages.
     * @param dest The destination that we are sending messages to.
     * @throws RmsCommunicationException If an error occurs while
     * connecting to one of the destinations.
     * @throws RmsRuntimeException If any of the destinations are invalid.
     */
    public RmsProducer(RmsConnection parentConnection, RmsDestination[] dest)
        throws RmsCommunicationException {

        _myConnection = parentConnection;
        _myDestinations = dest;

        initialize();
    }

    /**
     * Create a new producer given a connection, an array of destinations,
     * and a reply time window.
     *
     * @param parentConnection RmsConnection object which is used for receiving
     * and sending messages.
     * @param dest The destination that we are sending messages to.
     * @param replyTimeWindow The interval after a message
     * is sent, in seconds, in which a reply must be received to be
     * considered valid..
     * @throws RmsCommunicationException If an error occurs while
     * connecting to one of the destinations.
     * @throws RmsRuntimeException If any of the destinations are invalid.
     */
    public RmsProducer(RmsConnection parentConnection, RmsDestination[] dest, 
                       int replyTimeWindow)
        throws RmsCommunicationException {

        _myConnection = parentConnection;
        _myDestinations = dest;
        _replyTimeout = replyTimeWindow;

        initialize();
    }

    /**
     * Send a string to the destinations that are associated with
     * this producer.  Replies to string messages are not supported.
     *
     * @param messageText The string to send out.
     * @throws RmsCommunicationException If an error occurs when trying
     * to send the string to the destination(s).
     * @throws novadaq.DataException If an error occurs when trying
     * to prepare the string for sending.
     */
    public void sendString(String messageText)
        throws RmsCommunicationException, novadaq.DataException {

        for (int i = 0; i < _myDestinations.length; i++) {
            _myConnection.sendString(_myDestinations[i], messageText);
        }
    }

    /**
     * Send a message.  Application code should set the message body
     * inside the RmsMessage - all other fields will be filled as needed.
     * 
     * @param message The message to be sent.
     * @throws RmsCommunicationException If an error occurs when trying
     * to send the message to the destination(s).
     * @throws novadaq.DataException If an error occurs when trying
     * to prepare the message for sending.
     */
    public void sendMessage(RmsMessage message)
        throws RmsCommunicationException, novadaq.DataException {
        RmsDestination replyDestination;
        // This is responsible for setting the replyDestination
        // of the message, the sentTime and the destination.  This will also
        // add the message's ID to the reply list.

        RmsProviderUtil.setStandardFields(message, _myConnection);
        message.setTimeToLive(_myConnection.getDefaultTimeToLive());

        replyDestination = _myConnection.getMessageSource();

        synchronized (_replyIdTable) {
            _replyIdTable.put(message.getId(), message.getSentTime());
        }

        for (int i = 0; i < _myDestinations.length; i++) {
            message.setDestination(_myDestinations[i]);
            replyDestination.setProperty(RmsDestination.CHANNEL_PROPERTY_NAME,
                                         _myDestinations[i].getProperty(RmsDestination.CHANNEL_PROPERTY_NAME));
            replyDestination.validate();
            message.setReplyDestination(replyDestination);

            _myConnection.sendMessage(_myDestinations[i], message);
        }
    }

    /**
     * Pull replies off the incoming message queue.  This function will
     * block until a message is received.
     *
     * @return The received RmsMessage on success.  If an error occurs,
     * a non-null RmsMessage will be returned, and the RmsStatus
     * object inside the message will contain information on the error
     * (and the message body will be empty).
     */
    public RmsMessage receiveReply() {
        RmsMessage receivedReply = null;
        RmsStatus messageStatus = null;

        do {
            receivedReply = new RmsMessage();
            try {
                String messageText = (String)_inputQueue.take();
                receivedReply.deserialize(messageText);
            }
            catch (novadaq.DataException excp) {
                messageStatus = new RmsStatus(excp);
            }
            catch (InterruptedException excp) {
                messageStatus = new RmsStatus(excp);
            }
            if (messageStatus != null) {
                receivedReply.setMessageStatus(messageStatus);
            }
        } while (!verify(receivedReply));

        return receivedReply;
    }

    /**
     * Pull replies off the incoming message queue.  This function will
     * block until a message is received, or the time expires; whichever
     * happens first.
     *
     * @param timeout The maximum amount of time in milliseconds to wait if
     * the message queue is empty before giving up and returning.
     * @return The received RmsMessage on success or null if the timeout
     * expires before a message is received.  If an error occurs,
     * a non-null RmsMessage will be returned, and the RmsStatus
     * object inside the message will contain information on the error
     * (and the message body will be empty).
     */
    public RmsMessage receiveReply(long timeout) {
        RmsMessage receivedReply = null;
        RmsStatus messageStatus = null;

        do {
            receivedReply = new RmsMessage();
            try {
                String messageText = 
                    (String)_inputQueue.poll(timeout, TimeUnit.MILLISECONDS);

                if (messageText == null) {
                    return null;
                }

                receivedReply.deserialize(messageText);
            }
            catch (novadaq.DataException excp) {
                messageStatus = new RmsStatus(excp);
            }
            catch (InterruptedException excp) {
                messageStatus = new RmsStatus(excp);
            }
            if (messageStatus != null) {
                receivedReply.setMessageStatus(messageStatus);
            }
        } while (!verify(receivedReply));

        return receivedReply;
    }

    /**
     * Pull replies off the incoming message queue.  This function will
     * not block.
     *
     * @return The received RmsMessage on success or null if no messages
     * are available.  If an error occurs,
     * a non-null RmsMessage will be returned, and the RmsStatus
     * object inside the message will contain information on the error
     * (and the message body will be empty).
     */
    public RmsMessage receiveReplyNoWait() {
        RmsMessage receivedReply = null;
        RmsStatus messageStatus = null;

        do {
            receivedReply = new RmsMessage();
            try {
                String messageText = (String)_inputQueue.poll();

                if (messageText == null) {
                    return null;
                }

                receivedReply.deserialize(messageText);
            }
            catch (novadaq.DataException excp) {
                messageStatus = new RmsStatus(excp);
            }
            if (messageStatus != null) {
                receivedReply.setMessageStatus(messageStatus);
            }
        } while (!verify(receivedReply));

        return receivedReply;
    }

    /**
     * Register a callback with the producer.  This will start a
     * thread that pulls replies off the input queue and sends them
     * to the listener.
     *
     * @param listener The listener that will receive messages
     *
     * @return NONE
     */
    public void setReplyListener(RmsMessageListener listener) {
        if (_listenerLoop == null) {
            _listenerLoop = new ClientListenerLoop(_inputQueue, 
                                                   listener, this);
        }
        else {
            _listenerLoop.setMessageListener(listener);
        }

        return;
    }

    /**
     * Return the RmsMessageListener, if there is one.
     *
     * @return The RmsMessageListener if one has been set,
     * otherwise null.
     */
    public RmsMessageListener getReplyListener() {
        if (_listenerLoop != null) {
            return _listenerLoop.getReplyListener();
        }
        else {
            return null;
        }
    }

    /**
     * Closes down the producer.  After this method is called, the
     * producer should no longer be used since it will no
     * longer be connected to the messaging system.
     *
     * @throws RmsCommunicationException If a problem occurs when
     * disconnecting from the messaging system.
     */
    public void close() throws RmsCommunicationException {
        RmsCommunicationException firstException = null;

        // remove the provider listener from all reply destinations
        if (_myConnection != null && _providerListener != null) {
            RmsDestination replyDestination = _myConnection.getMessageSource();
            for (Iterator<String> channelIterator = _myReplyChannels.iterator();
                 channelIterator.hasNext();) {
                replyDestination.setProperty(RmsDestination.CHANNEL_PROPERTY_NAME,
                                             channelIterator.next());
                try {
                    _myConnection.removeListener(replyDestination,
                                                 _providerListener);
                }
                catch (RmsCommunicationException rcExcpt) {
                    if (firstException == null) {
                        firstException = rcExcpt;
                    }
                }
            }
        }
        _providerListener = null;
        _myDestinations = null;
        _myConnection = null;

        // clear out the input queue
        if (_inputQueue != null) {
            _inputQueue.clear();
        }
        _inputQueue = null;

        // stop the listener
        if (_listenerLoop != null) {
            _listenerLoop.stop();
        }
        _listenerLoop = null;

        // clear the reply table
        if (_replyIdTable != null) {
            _replyIdTable.clear();
        }
        _replyIdTable = null;

        // report the first exception, if any
        if (firstException != null) {
            throw firstException;
        }
    }

    /**
     * Verify that a reply received is one that the producer
     * is actually interested in.  We maintain a list of all
     * the message IDs that are sent out, and compare replies
     * received to this list.
     *
     * @param message The message that we are checking.
     * @return True is the message is supposed to come to
     * this producer, false otherwise.
     */
    public boolean verify(RmsMessage message) {
        if (message.getMessageStatus() == null ||
            message.getMessageStatus().getCode() != RmsStatus.SUCCESS_CODE) {
            /* Let all messages with errors go back to the user */
            return true;
        }

        if (_replyIdTable.size() == 0) {
            return false;
        }

        Date now = new Date();

        /**
         * The list of message IDs is ordered, with the oldest ID
         * being the first entry.  Here, we repeatedly pull the first
         * entry off the list until we hit a message ID that has not
         * expired.
         */
        synchronized (_replyIdTable) {
            Iterator<String> idIter = _replyIdTable.keySet().iterator();
            while (idIter.hasNext()) {
                String replyId = idIter.next();
                Date sentTime = _replyIdTable.get(replyId);

                if ((now.getTime() - sentTime.getTime()) >
                    (1000 * _replyTimeout)) {
                    idIter.remove();
                }
                else {
                    break;
                }
            }
        }

        /**
         * Go through the list of message IDs and see if
         * we can find one that matches the message that was
         * passed in.
         */
        synchronized (_replyIdTable) {
            Date sentTime = _replyIdTable.get(message.getCorrelationId());
            if (sentTime != null) {
                // we definitely don't want to remove correlation IDs
                // from the table here in case there are multiple replies
                // to a given request
                return true;
            }
        }

        return false;
    }

    /**
     * Handles cleanup when garbage collected.
     */
    protected void finalize() {
        try {
            this.close();
        }
        catch (RmsCommunicationException rcExcpt) {
            rcExcpt.printStackTrace(System.err);
        }
    }

    /**
     * Called by the constructor, this will verify that our provider
     * can talk to the supplied destination, as well as creating the
     * input queue, replyId list, the provider listener.  This will
     * also determine the reply destination.
     *
     * @throws RmsCommunicationException If an error occurs while
     * connecting to one of the destinations.
     */
    private void initialize() throws RmsCommunicationException {
        _myReplyChannels = new HashSet<String>();

        for (int i = 0; i < _myDestinations.length; i++) {
            _myDestinations[i].validate();
            _myReplyChannels.add(_myDestinations[i].getProperty(RmsDestination.CHANNEL_PROPERTY_NAME));

            if (! _myConnection.supportsDestination(_myDestinations[i])) {
                throw new RmsRuntimeException("Connection " + _myConnection + 
                                              " does not support " +
                                              "destination " +
                                              _myDestinations[i]);
            }
        }

        _inputQueue = new LinkedBlockingQueue<String>();
        _replyIdTable = new LinkedHashMap<String, Date>();

        if (_providerListener == null) {
            _providerListener =
                new BufferedProviderListener(_inputQueue);
        }

        RmsDestination replyDestination = _myConnection.getMessageSource();

        for (Iterator<String> channelIterator = _myReplyChannels.iterator();
             channelIterator.hasNext();) {
            replyDestination.setProperty(RmsDestination.CHANNEL_PROPERTY_NAME,
                                         channelIterator.next());
            replyDestination.validate();

            _myConnection.addListener(replyDestination,
                                      _providerListener);
        }
    }

    // Static initialization block to run an initial RmsMessage serialization
    // and deserialization.  This is useful because these operations can
    // have longer runtimes the first time they are called, and it
    // is good to incur that overhead *before* the first real message
    // is sent.
    static {
        try {
            RmsMessage dummyMessage = new RmsMessage();
            String dummyString = dummyMessage.serialize();
            dummyMessage = new RmsMessage(dummyString);
        }
        catch (Exception excpt) {}
    }
}
