package gov.fnal.cd.rms;

/**
 * @package gov.fnal.cd.rms
 *
 * @brief Classes that allow applications to send and receive
 * messages utilizing the Responsive Messaging System.
 *
 */

import gov.fnal.cd.rms.base.RmsCommunicationException;
import gov.fnal.cd.rms.base.RmsDestination;
import gov.fnal.cd.rms.base.RmsRuntimeException;
import gov.fnal.cd.rms.provider.RmsConnection;
import gov.fnal.cd.rms.provider.DDSConnection;

import java.util.*;

/**
 * This class contains functionality for receiving messages of a
 * specified type from a specific destination.
 *
 * It is generic on type T so that we can return message objects of
 * the correct type from the receive() method.  At the moment, the
 * generic type doesn't relieve the need to pass the data type of
 * the messages to be received to the constructor.  
 *
 * @author Kurt Biery
 * @version $Revision: 1.3 $ $Date: 2010/03/15 20:52:04 $
 */

public class RmsReceiver<T> {

    /**
     * Handle to an RmsConnection that is used to receive messages.
     */
    private RmsConnection _myConnection = null;

    /**
     * The type (Class) of the messages that this Receiver is listening for.
     */
    private Class<T> _messageDataType;

    /**
     * The destination that this Receiver is listening to.
     */
    private RmsDestination _myDestination;

    /**
     * Create a new Receiver given a connection and a destination.
     *
     * @param parentConn RmsConnection object which is used for
     * receiving and sending messages.
     * @param messageDataType The java.lang.Class of the datatype
     * (message type) that will be received with this Receiver.
     * @param dest The destination that we are receiving messages from.
     * @throws RmsCommunicationException If an error occurs while
     * connecting to the destination.
     * @throws RmsRuntimeException If the destination is invalid.
     */
    public RmsReceiver(RmsConnection parentConn, Class<T> messageDataType,
                       RmsDestination dest)
        throws RmsCommunicationException {

        _myConnection = parentConn;
        _messageDataType = messageDataType;
        dest.validate();
        if (! _myConnection.supportsDestination(dest)) {
            throw new RmsRuntimeException("Connection " + _myConnection + 
                                          " does not support " +
                                          "destination " + dest);
        }
        _myDestination = dest;

        if (_myConnection instanceof DDSConnection) {
            ((DDSConnection)_myConnection).registerReceiver(_messageDataType,
                                                            _myDestination);
        }
    }

    /**
     * Receives the next message.  This method will block until a
     * message is received.
     *
     * @param messageDataType the java.lang.Class of the message that will
     *        be received
     * @param dest the destination that the message was sent to
     * @return The received message.
     * @throws RmsCommunicationException if there is a problem receiving
     *         the next message.
     */
    public Vector<T> receiveMessage() throws RmsCommunicationException {
        return _myConnection.receiveMessage(_messageDataType, _myDestination);
    }

    /**
     * Receives the next message.  If no message is received within
     * the specified timeout, a null message is returned.
     *
     * @param messageDataType the java.lang.Class of the message that will
     *        be received
     * @param dest the destination that the message was sent to
     * @param timeout The maximum amount of time to wait for a message,
     * in milliseconds.
     * @return The received message or null if none.
     * @throws RmsCommunicationException if there is a problem receiving
     *         the next message.
     */
    public Vector<T> receiveMessage(long timeout) throws RmsCommunicationException {
        return _myConnection.receiveMessage(_messageDataType, _myDestination,
                                            timeout);
    }

    /**
     * Closes down the Receiver.  After this method is called, the
     * Receiver should no longer be used since it will no
     * longer be connected to the messaging system.
     *
     * @throws RmsCommunicationException If a problem occurs when
     * disconnecting from the messaging system.
     */
    public void close() throws RmsCommunicationException {
        RmsCommunicationException firstException = null;

        if (_myConnection == null) {return;}

        if (_myConnection instanceof DDSConnection) {
            ((DDSConnection)_myConnection).unRegisterReceiver(_messageDataType,
                                                              _myDestination);
        }
        _myConnection = null;
        _myDestination = null;

        // report the first exception, if any
        if (firstException != null) {
            throw firstException;
        }
    }

    /**
     * Handles cleanup when an instance of this class is garbage collected.
     */
    protected void finalize() {
        try {
            this.close();
        }
        catch (RmsCommunicationException rcExcpt) {
            rcExcpt.printStackTrace(System.err);
        }
    }
}
