package gov.fnal.cd.rms;

import gov.fnal.cd.rms.base.RmsCommunicationException;
import gov.fnal.cd.rms.base.RmsDestination;
import gov.fnal.cd.rms.base.RmsRuntimeException;
import gov.fnal.cd.rms.provider.RmsConnection;
import gov.fnal.cd.rms.provider.DDSConnection;

/**
 * This class contains functionality for sending messages of a
 * specified type to a specific destination.
 *
 * It is generic on type T so that we get the benefit of type checking
 * on the sendMessage method.  It is not clear whether we also need to
 * pass the data type of the messages to the constructor, but we'll include
 * that, for now.
 *
 * @author Kurt Biery
 * @version $Revision: 1.3 $ $Date: 2010/03/03 22:30:16 $
 */

public class RmsSender<T> {

    /**
     * The RmsConnection that is used to send messages.
     */
    private RmsConnection _myConnection;

    /**
     * The type (Class) of the messages that this Sender will send.
     */
    private Class<T> _messageDataType;

    /**
     * The destination that this Sender is sending messages to.
     */
    private RmsDestination _myDestination;

    /**
     * The reply timeout is hard coded to be 5 minutes.
     */
    private int _replyTimeout = 300; /* 5 Minutes */

    /**
     * Create a new Sender given a connection and a destination.  
     *
     * @param parentConnection RmsConnection object which is used for receiving
     * and sending messages.
     * @param dest The destination that we are sending messages to.
     * @throws RmsCommunicationException If an error occurs while
     * connecting to the destination.
     * @throws RmsRuntimeException If either of the destinations are invalid.
     */
    public RmsSender(RmsConnection parentConn, Class<T> messageDataType,
                     RmsDestination dest)
        throws RmsCommunicationException {

        _myConnection = parentConn;
        _messageDataType = messageDataType;
        dest.validate();
        if (! _myConnection.supportsDestination(dest)) {
            throw new RmsRuntimeException("Connection " + _myConnection + 
                                          " does not support " +
                                          "destination " + dest);
        }
        _myDestination = dest;

        if (_myConnection instanceof DDSConnection) {
            ((DDSConnection)_myConnection).registerSender(_messageDataType,
                                                          _myDestination);
        }
    }

    /**
     * Sends the specified message.
     * 
     * @param message The message to be sent.
     * @throws RmsCommunicationException If an error occurs when trying
     * to send the message to the destination(s).
     */
    public void sendMessage(T message) throws RmsCommunicationException {
        _myConnection.sendMessage(_messageDataType, _myDestination, message);
    }

    /**
     * Closes down the Sender.  After this method is called, the
     * Sender should no longer be used since it will no
     * longer be connected to the messaging system.
     *
     * @throws RmsCommunicationException If a problem occurs when
     * disconnecting from the messaging system.
     */
    public void close() throws RmsCommunicationException {
        RmsCommunicationException firstException = null;

        if (_myConnection == null) {return;}

        if (_myConnection instanceof DDSConnection) {
            ((DDSConnection)_myConnection).unRegisterSender(_messageDataType,
                                                            _myDestination);
        }
        _myConnection = null;
        _myDestination = null;

        // report the first exception, if any
        if (firstException != null) {
            throw firstException;
        }
    }

    /**
     * Handles cleanup when garbage collected.
     */
    protected void finalize() {
        try {
            this.close();
        }
        catch (RmsCommunicationException rcExcpt) {
            rcExcpt.printStackTrace(System.err);
        }
    }
}
