package gov.fnal.cd.rms.base;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Map;

/**
 * Instances of this class represent message destinations in the
 * Responsive Messaging System, where a "destination" is an abstract
 * location to which producers send messages and from which consumers
 * receive messages.
 *
 * In RMS, a destination is composed of an arbitrary number of properties
 * that describe various characteristics of the abstract location such as
 * target (hostname or process name) and channel (type of functionality
 * that the destination is used for).  Standard property names and typical
 * property values are defined in this class and experiment-specific child
 * classes can provide additional ones based on specific needs.
 *
 * @author Kurt Biery
 * @author Steve Foulkes
 * @version $Revision: 1.3 $ $Date: 2010/04/12 20:18:52 $
 */
public class RmsDestination extends novadaq.CastorSerializableBase {

    /**
     * The destination target for all DCMs.
     */
    public static final String ALL_DCM_TARGET = "allDCMs";

    /**
     * The destination target for all buffer node event builders.
     */
    public static final String ALL_BNEVB_TARGET = "allBufferNodeEVBs";

    /**
     * The destination target for all DAQ elements.
     */
    public static final String ALL_ELEMENT_TARGET = "allElements";

    /**
     * The name used to identify the commonly used "target" property.
     */
    public static final String TARGET_PROPERTY_NAME = "target";

    /**
     * The name used to identify the commonly used "channel" property.
     */
    public static final String CHANNEL_PROPERTY_NAME = "channel";

    /**
     * The name used to identify the commonly used "messageType" property.
     */
    public static final String MESSAGE_TYPE_PROPERTY_NAME = "messageType";

    /**
     * The name used to identify the occasionally used "source" property.
     */
    public static final String SOURCE_PROPERTY_NAME = "source";

    /**
     * The name used to identify the "partition" property.
     */
    public static final String PARTITION_PROPERTY_NAME = "partition";

    /**
     * The name used to identify the "environment" property.
     */
    public static final String ENVIRONMENT_PROPERTY_NAME = "environment";

    /**
     * The property value corresponding to the service manager target.
     */
    public static final String SERVICE_MANAGER_TARGET = "serviceManager";

    /**
     * The property value corresponding to the heartbeat channel.
     */
    public static final String HEARTBEAT_CHANNEL = "heartbeat";

    /**
     * The property value corresponding to the control message channel.
     */
    public static final String CONTROL_CHANNEL = "control";

    /**
     * The property value corresponding to the status report channel.
     */
    public static final String STATUS_REPORT_CHANNEL = "status";

    /**
     * The property value corresponding to the monitoring channel.
     */
    public static final String MONITOR_CHANNEL = "monitor";

    /**
     * The property value corresponding to the alert channel.
     */
    public static final String ALERT_CHANNEL = "alert";

    /**
     * The property value corresponding to the notification message type.
     */
    public static final String NOTIFICATION_MESSAGE_TYPE = "notification";

    /**
     * The property value corresponding to the request message type.
     */
    public static final String REQUEST_MESSAGE_TYPE = "request";

    /**
     * The property value corresponding to the reply message type.
     */
    public static final String REPLY_MESSAGE_TYPE = "reply";

    /**
     * The property value corresponding to the loopback message type.
     */
    public static final String LOOPBACK_MESSAGE_TYPE = "loopback";

    /**
     * The property value corresponding to the configuration channel.
     */
    public static final String CONFIGURATION_CHANNEL = "config";

    /**
     * Used as an identifier for the null partition.
     */
    public static final int NULL_PARTITION = -1;

    /**
     * The properties that describe the destination.
     */
    private Map<String,String> _propertyTable = new Hashtable<String,String>();

    /**
     * Creates an empty RmsDestination (one with no destination properties).
     */
    public RmsDestination() {
        this.setDefaults();
    }

    /**
     * Creates an RmsDestination instance by copying the properties of the
     * input RmsDestination.
     *
     * @param originalDest The destination object that we are
     * copying.
     */
    public RmsDestination(RmsDestination originalDest) {
        this._castorMapName = originalDest._castorMapName;

        this.addProperties(originalDest.getPropertyTable());
    }

    /**
     * Creates an RmsDestination instance for the specified target, channel,
     * and message type.
     *
     * @param target The target name
     * @param channel The channel name
     * @param messageType The message type
     */
    public RmsDestination(String target, String channel, String messageType) {
        this.setDefaults();

        this.setProperty(TARGET_PROPERTY_NAME, target);
        this.setProperty(CHANNEL_PROPERTY_NAME, channel);
        this.setProperty(MESSAGE_TYPE_PROPERTY_NAME, messageType);
    }

    /**
     * Creates an RmsDestination instance for the specified target,
     * channel, message type, and partition.
     *
     * @param target The target name
     * @param channel The channel name
     * @param messageType The message type
     * @param partitionNumber The partition number
     */
    public RmsDestination(String target, String channel,
                          String messageType, int partitionNumber) {
        this.setDefaults();

        this.setProperty(TARGET_PROPERTY_NAME, target);
        this.setProperty(CHANNEL_PROPERTY_NAME, channel);
        this.setProperty(MESSAGE_TYPE_PROPERTY_NAME, messageType);
        this.setPartitionProperty(partitionNumber);
    }

    /**
     * Returns the requested property value or null if the specified
     * property name is not found.
     *
     * @param name The property name to retrieve.
     *
     * @return The value of the property requested.
     */
    public String getProperty(String name) {
        return (String) _propertyTable.get(name);
    }

    /**
     * Returns the requested property value or the specified default value
     * null if the property name is not found.
     *
     * @param name The property name
     * @param defaultValue The value to return if the property is
     * not found.
     *
     * @return The value of the property if it is in the table,
     * the defaultValue otherwise.
     */
    public String getProperty(String name, String defaultValue) {
        String value = this.getProperty(name);
        return (value != null ? value : defaultValue);
    }

    /**
     * Sets the named property to the specified value.
     *
     * @param name The name of the property
     * @param value The value to set the property to
     */
    public void setProperty(String name, String value) {
        _propertyTable.put(name, value);
    }

    /**
     * Sets the partition property for this destination.
     *
     * @param partitionNumber The value to set the partition number
     * property to.
     */
    public void setPartitionProperty(int partitionNumber) {
        this.setProperty(PARTITION_PROPERTY_NAME,
                         Integer.toString(partitionNumber));
    }

    /**
     * Returns the properties contained in this destination.  This method
     * should be used with care - it returns a reference the to internal
     * property table.  This is needed for Castor serialization, but should
     * not be used for much else.
     *
     * @return The property table.
     */
    public Map getPropertyTable() {
        // 27-Nov-2006, KAB: we need to return the bare property table (not
        // a copy) since Castor uses this map to store the properties when
        // deserializing from XML.
        return _propertyTable;
    }

    /**
     * Adds the specified properties to this destination.
     *
     * @param newProps The new properties to setup in the table.
     */
    public void addProperties(java.util.Map newProps) {
        java.util.Set theKeySet = newProps.keySet();
        java.util.Iterator keyIter = theKeySet.iterator();
        while (keyIter.hasNext()) {
            String name = (String) keyIter.next();
            String value = (String) newProps.get(name);
            this.setProperty(name, value);
        }
    }

    /**
     * Sets the properties for this destination.
     *
     * @param newProps The new property table.
     */
    public void setPropertyTable(java.util.Map<String,String> newProps) {
        this._propertyTable = new Hashtable<String,String>(newProps);
    }

    /**
     * Validates this destination.  This consists of testing if
     * required properties are set and verifying that mutually exclusive
     * properties are not set.  A runtime exception with text describing
     * the problem(s) is thrown if the destination is not valid.
     *
     * Throwing a runtime exception (and subsequently halting program
     * execution) may seem like a rather severe reaction
     * to an invalid destination, but given that an invalid destination
     * would be the result of a coding error (trying to use a destination
     * with insufficient information), it seems like the appropriate response.
     *
     * @throws RmsRuntimeException if the destination is not valid
     */
    public void validate() {
        StringBuffer exceptionText = new StringBuffer();
        String targetPropertyValue = 
            _propertyTable.get(TARGET_PROPERTY_NAME);
        String sourcePropertyValue = 
            _propertyTable.get(SOURCE_PROPERTY_NAME);
        String channelPropertyValue = 
            _propertyTable.get(CHANNEL_PROPERTY_NAME);
        String partitionPropertyValue =
            _propertyTable.get(PARTITION_PROPERTY_NAME);

        // check for required properties
        // --> as of 12-Oct-2007, these are
        // A) target or source
        // B) channel
        // C) partition
        if (targetPropertyValue == null &&
            sourcePropertyValue == null) {
            if (exceptionText.length() > 0) {exceptionText.append("; ");}
            exceptionText.append("either the " + TARGET_PROPERTY_NAME +
                                 " property or the " + SOURCE_PROPERTY_NAME +
                                 " property must be set");
        }
        if (channelPropertyValue == null) {
            if (exceptionText.length() > 0) {exceptionText.append("; ");}
            exceptionText.append("the " + CHANNEL_PROPERTY_NAME +
                                 " property must be set");
        }
        if (partitionPropertyValue == null) {
            if (exceptionText.length() > 0) {exceptionText.append("; ");}
            exceptionText.append("the " + PARTITION_PROPERTY_NAME +
                                 " property must be set");
        }

        // check for mutually exclusive properties
        // --> as of 26-Mar-2007, these are
        // A) target and source
        if (targetPropertyValue != null &&
            sourcePropertyValue != null) {
            if (exceptionText.length() > 0) {exceptionText.append("; ");}
            exceptionText.append("the " + TARGET_PROPERTY_NAME +
                                 " property and the " + SOURCE_PROPERTY_NAME +
                                 " property must not both be set");
        }

        // throw an exception if one or more problems were found
        if (exceptionText.length() != 0) {
            exceptionText.insert(0, "Invalid RmsDestination " +
                                 this.toString() + ": ");
            exceptionText.append(".");
            throw new RmsRuntimeException(exceptionText.toString());
        }
    }

    /**
     * Tests if the specified object matches this destination.
     * In this case, "matching" means that the values for the
     * properties that both destinations have in common are equal.
     * Properties that are specified in only one destination are
     * not considered in determining whether the destinations match.
     * (So, if destination A has a "messageType" property and destination
     * B does not, the this method acts as though destination B
     * had a wildcard for the messageType property.)
     * 
     * @param testObject The object that we are comparing this one with.
     *
     * @return True if the object match, false otherwise.
     */
    public boolean matches(Object testObject) {
        if (testObject == null ||
            ! (testObject instanceof RmsDestination)) return false;
        RmsDestination testDestination = (RmsDestination) testObject;
        boolean objectsMatch = true;

        // loop through each of the properties in this destination
        java.util.Set<String> theKeySet = _propertyTable.keySet();
        java.util.Iterator<String> keyIter = theKeySet.iterator();
        while (keyIter.hasNext() && objectsMatch) {
            String thisName = keyIter.next();
            String thisValue = _propertyTable.get(thisName);
            String thatValue = testDestination.getProperty(thisName);
            if (thisValue != null && thatValue != null) {
                objectsMatch = thisValue.equals(thatValue);
            }
        }

        return objectsMatch;
    }

    /**
     * Tests if the specified object is equal to this destination.
     *
     * @param testObject The object that we are comparing this one to.
     *
     * @return True if the objects are equal, false otherwise.
     */
    public boolean equals(Object testObject) {
        if (testObject == null ||
            ! (testObject instanceof RmsDestination)) return false;
        RmsDestination testDestination = (RmsDestination) testObject;
        return _propertyTable.equals(testDestination.getPropertyTable());
    }

    /**
     * Returns a hash code for this destination.
     *
     * @return A hash code representing the property table.
     */
    public int hashCode() {
        return _propertyTable.hashCode();
    }

    /**
     * Returns a String representation of this RmsDestination.
     *
     * @return A string representation of this destination.
     */
    public String toString() {
        return _propertyTable.toString();
    }

    /**
     * Runs unit tests for the RmsDestination class.
     */
    public static void main(String[] args) {
        int testCount = 0;

        // create a sample destination
        testCount++;
        System.out.println("\n>>>>> Test " + testCount + " <<<<<");
        System.out.println("Creating a sample destination with " +
                           "three properties...");
        RmsDestination destOrig = new RmsDestination();
        destOrig.setProperty(TARGET_PROPERTY_NAME, "sampleTarget");
        destOrig.setProperty(CHANNEL_PROPERTY_NAME, CONTROL_CHANNEL);
        destOrig.setProperty(MESSAGE_TYPE_PROPERTY_NAME,
                             NOTIFICATION_MESSAGE_TYPE);

        // check the behavior of toString();
        System.out.println("\nResult of toString() call on original " +
                           "destination:");
        System.out.println("----------------------------------------");
        System.out.println(destOrig.toString());
        System.out.println("----------------------------------------");

        // create a copy of the original destination
        testCount++;
        System.out.println("\n>>>>> Test " + testCount + " <<<<<");
        System.out.println("Creating a copy of the original destination...");
        RmsDestination destCopy = new RmsDestination(destOrig);

        // check the behavior of toString();
        System.out.println("\nResult of toString() call on " +
                           "destination copy:");
        System.out.println("----------------------------------------");
        System.out.println(destCopy.toString());
        System.out.println("----------------------------------------");

        // check behavior of equals(), matches(), and hashcode()
        testCount++;
        System.out.println("\n>>>>> Test " + testCount + " <<<<<");
        System.out.println("Does the copy equal the original? " +
                           (destCopy.equals(destOrig) ? "yes" : "no") +
                           " (it should)");
        System.out.println("Does the copy match the original? " +
                           (destCopy.matches(destOrig) ? "yes" : "no") +
                           " (it should)");
        System.out.println("Original and copy hashcodes: " +
                           destOrig.hashCode() + ", " + destCopy.hashCode());

        // modify the copy
        testCount++;
        System.out.println("\n>>>>> Test " + testCount + " <<<<<");
        System.out.println("Modifying the copy...");
        destCopy.setProperty(CHANNEL_PROPERTY_NAME, ALERT_CHANNEL);

        // check the behavior of toString();
        System.out.println("\nResult of toString() call on " +
                           "modified copy:");
        System.out.println("----------------------------------------");
        System.out.println(destCopy.toString());
        System.out.println("----------------------------------------");

        // check behavior of equals(), matches(), and hashcode()
        testCount++;
        System.out.println("\n>>>>> Test " + testCount + " <<<<<");
        System.out.println("Does the copy now equal the original? " +
                           (destCopy.equals(destOrig) ? "yes" : "no") +
                           " (it shouldn't)");
        System.out.println("Does the copy now match the original? " +
                           (destCopy.matches(destOrig) ? "yes" : "no") +
                           " (it shouldn't)");
        System.out.println("Original and copy hashcodes: " +
                           destOrig.hashCode() + ", " + destCopy.hashCode());

        // further check of the behavior of matches()
        testCount++;
        System.out.println("\n>>>>> Test " + testCount + " <<<<<");
        System.out.println("Creating a destination to test for matches...");
        RmsDestination matchDest = new RmsDestination();
        matchDest.setProperty(TARGET_PROPERTY_NAME, "sampleTarget");
        matchDest.setProperty(CHANNEL_PROPERTY_NAME, CONTROL_CHANNEL);
        System.out.println("\nResult of toString() call on " +
                           "matchTest destination:");
        System.out.println("----------------------------------------");
        System.out.println(matchDest.toString());
        System.out.println("----------------------------------------");
        System.out.println("Does the matchTest match the original? " +
                           (matchDest.matches(destOrig) ? "yes" : "no") +
                           " (it should)");
        System.out.println("Does the matchTest match the copy? " +
                           (matchDest.matches(destCopy) ? "yes" : "no") +
                           " (it shouldn't)");

        testCount++;
        System.out.println("\n>>>>> Test " + testCount + " <<<<<");
        System.out.println("Modifying the matchTest destination...");
        matchDest.setProperty(SOURCE_PROPERTY_NAME, "RunControl0");
        System.out.println("\nResult of toString() call on " +
                           "modified matchTest destination:");
        System.out.println("----------------------------------------");
        System.out.println(matchDest.toString());
        System.out.println("----------------------------------------");
        System.out.println("Does the matchTest match the original? " +
                           (matchDest.matches(destOrig) ? "yes" : "no") +
                           " (it should)");
        System.out.println("Does the matchTest match the copy? " +
                           (matchDest.matches(destCopy) ? "yes" : "no") +
                           " (it shouldn't)");

        testCount++;
        System.out.println("\n>>>>> Test " + testCount + " <<<<<");
        System.out.println("Modifying the matchTest destination...");
        matchDest.setProperty(MESSAGE_TYPE_PROPERTY_NAME,
                              NOTIFICATION_MESSAGE_TYPE);
        System.out.println("\nResult of toString() call on " +
                           "modified matchTest destination:");
        System.out.println("----------------------------------------");
        System.out.println(matchDest.toString());
        System.out.println("----------------------------------------");
        System.out.println("Does the matchTest match the original? " +
                           (matchDest.matches(destOrig) ? "yes" : "no") +
                           " (it should)");
        System.out.println("Does the matchTest match the copy? " +
                           (matchDest.matches(destCopy) ? "yes" : "no") +
                           " (it shouldn't)");

        // check that validation is working
        testCount++;
        System.out.println("\n>>>>> Test " + testCount + " <<<<<");
        System.out.println("Creating a destination to test validation...");
        RmsDestination vdnDest = new RmsDestination();
        vdnDest.setProperty(PARTITION_PROPERTY_NAME, "0");
        System.out.println("\nResult of toString() call on " +
                           "validation destination:");
        System.out.println("----------------------------------------");
        System.out.println(vdnDest.toString());
        System.out.println("----------------------------------------");
        System.out.println("Is the destination valid? (it shouldn't be)");
        try {
            vdnDest.validate();
            System.out.println("  ERROR: the destination was incorrectly " +
                               "determined to be valid!");
        }
        catch (Exception excpt) {
            System.out.println("  The destination was correctly " +
                               "determined to be invalid:");
            System.out.println("    " + excpt);
        }

        testCount++;
        System.out.println("\n>>>>> Test " + testCount + " <<<<<");
        System.out.println("Modifying the validation destination...");
        vdnDest.setProperty(CHANNEL_PROPERTY_NAME, MONITOR_CHANNEL);
        System.out.println("\nResult of toString() call on " +
                           "validation destination:");
        System.out.println("----------------------------------------");
        System.out.println(vdnDest.toString());
        System.out.println("----------------------------------------");
        System.out.println("Is the destination valid? (it shouldn't be)");
        try {
            vdnDest.validate();
            System.out.println("  ERROR: the destination was incorrectly " +
                               "determined to be valid!");
        }
        catch (Exception excpt) {
            System.out.println("  The destination was correctly " +
                               "determined to be invalid:");
            System.out.println("    " + excpt);
        }

        testCount++;
        System.out.println("\n>>>>> Test " + testCount + " <<<<<");
        System.out.println("Modifying the validation destination...");
        vdnDest.setProperty(TARGET_PROPERTY_NAME, "dcm005");
        System.out.println("\nResult of toString() call on " +
                           "validation destination:");
        System.out.println("----------------------------------------");
        System.out.println(vdnDest.toString());
        System.out.println("----------------------------------------");
        System.out.println("Is the destination valid? (it should be)");
        try {
            vdnDest.validate();
            System.out.println("  The destination was correctly " +
                               "determined to be valid.");
        }
        catch (Exception excpt) {
            System.out.println("  ERROR: The destination was incorrectly " +
                               "determined to be invalid:");
            System.out.println("    " + excpt);
        }

        testCount++;
        System.out.println("\n>>>>> Test " + testCount + " <<<<<");
        System.out.println("Modifying the validation destination...");
        vdnDest.setProperty(SOURCE_PROPERTY_NAME, "RunControl2");
        System.out.println("\nResult of toString() call on " +
                           "validation destination:");
        System.out.println("----------------------------------------");
        System.out.println(vdnDest.toString());
        System.out.println("----------------------------------------");
        System.out.println("Is the destination valid? (it shouldn't be)");
        try {
            vdnDest.validate();
            System.out.println("  ERROR: the destination was incorrectly " +
                               "determined to be valid!");
        }
        catch (Exception excpt) {
            System.out.println("  The destination was correctly " +
                               "determined to be invalid:");
            System.out.println("    " + excpt);
        }

        // check how null inputs are handled
        testCount++;
        System.out.println("\n>>>>> Test " + testCount + " <<<<<");
        System.out.println("Trying NULL property name and value...");
        try {
            destCopy.setProperty(null, "Test Value");
            System.out.println("WARNING: a null property name did NOT " +
                               "generate an exception!  This could lead to " +
                               "undefined behavior!");
        }
        catch (Exception excpt) {
            System.out.println("setProperty() with a null property name " +
                               "generated an exception, as expected:\n    " +
                               excpt);
        }
        try {
            destCopy.setProperty("Test Name", null);
            System.out.println("WARNING: a null property value did NOT " +
                               "generate an exception!  This could lead to " +
                               "undefined behavior!");
        }
        catch (Exception excpt) {
            System.out.println("setProperty() with a null property value " +
                               "generated an exception, as expected:\n    " +
                               excpt);
        }

        // test lookups
        String name, value;
        testCount++;
        System.out.println("\n>>>>> Test " + testCount + " <<<<<");
        System.out.println("For the original destination:");
        name = TARGET_PROPERTY_NAME;
        value = destOrig.getProperty(name);
        System.out.println("    The value for the \"" + name + "\"" +
                           " property is \"" + value + "\".");
        name = CHANNEL_PROPERTY_NAME;
        value = destOrig.getProperty(name);
        System.out.println("    The value for the \"" + name + "\"" +
                           " property is \"" + value + "\".");
        name = "nonsense";
        value = destOrig.getProperty(name);
        System.out.println("    The value for the \"" + name + "\"" +
                           " property is \"" + value + "\" (should be null).");

        // test constructor with 3 arguments
        testCount++;
        System.out.println("\n>>>>> Test " + testCount + " <<<<<");
        System.out.println("Creating a destination with " +
                           "the three-argument constructor...");
        RmsDestination dest3Arg =
            new RmsDestination("configManager", CONFIGURATION_CHANNEL,
                               REQUEST_MESSAGE_TYPE);
        System.out.println("\nResult of toString() call on 3-arg " +
                           "destination:");
        System.out.println("----------------------------------------");
        System.out.println(dest3Arg.toString());
        System.out.println("----------------------------------------");

        // test constructor with 4 arguments
        testCount++;
        System.out.println("\n>>>>> Test " + testCount + " <<<<<");
        System.out.println("Creating a destination with " +
                           "the four-argument constructor...");
        RmsDestination dest4Arg =
            new RmsDestination("allElements", HEARTBEAT_CHANNEL,
                               REQUEST_MESSAGE_TYPE, 15);
        System.out.println("\nResult of toString() call on 4-arg " +
                           "destination:");
        System.out.println("----------------------------------------");
        System.out.println(dest4Arg.toString());
        System.out.println("----------------------------------------");

        // test serialization of destinations
        testCount++;
        System.out.println("\n>>>>> Test " + testCount + " <<<<<");
        System.out.println("Serialized original destination...");
        System.out.println("----------------------------------------");
        try {
            System.out.println(destOrig.serialize());
        }
        catch (Exception excpt) {
            System.err.println("Exception caught while serializing");
            excpt.printStackTrace(System.err);
        }
        System.out.println("----------------------------------------");
        testCount++;
        System.out.println("\n>>>>> Test " + testCount + " <<<<<");
        System.out.println("Serialized copied destination...");
        System.out.println("----------------------------------------");
        try {
            System.out.println(destCopy.serialize());
        }
        catch (Exception excpt) {
            System.err.println("Exception caught while serializing");
            excpt.printStackTrace(System.err);
        }
        System.out.println("----------------------------------------");
        testCount++;
        System.out.println("\n>>>>> Test " + testCount + " <<<<<");
        System.out.println("Serialized 4-arg destination...");
        System.out.println("----------------------------------------");
        String xmlString = "";
        try {
            xmlString = dest4Arg.serialize();
            System.out.println(xmlString);
        }
        catch (Exception excpt) {
            System.err.println("Exception caught while serializing");
            excpt.printStackTrace(System.err);
        }
        System.out.println("----------------------------------------");

        // test deserialization
        testCount++;
        System.out.println("\n>>>>> Test " + testCount + " <<<<<");
        System.out.println("Testing deserialization on 4-arg destination...");
        System.out.println("Result of deserialization and " +
                           "re-serialization...");
        System.out.println("----------------------------------------");
        try {
            RmsDestination destTester = new RmsDestination();
            destTester.deserialize(xmlString);
            System.out.println(destTester.serialize());
        }
        catch (Exception excpt) {
            System.err.println("Exception caught while deserializing");
            excpt.printStackTrace(System.err);
        }
        System.out.println("----------------------------------------");

        System.out.println("");
    }

    /**
     * If a custom environment has been specified, this field is used
     * to store its value so that we can automatically set the
     * environment property in destinations.  If this field is null, the
     * environment property in destination objects is not automatically set.
     */
    private static String _defaultEnvironment = null;
    static {
        // check for a custom environment
        _defaultEnvironment = System.getenv("RMS_CUSTOM_ENVIRONMENT");
        if (_defaultEnvironment != null &&
            _defaultEnvironment.trim().equals("")) {
            _defaultEnvironment = null;
        }
    }

    /**
     * Sets default values of needed attributes.
     */
    private void setDefaults() {
        setPartitionProperty(NULL_PARTITION);

        // set the name of the map we'll use for Castor XML binding
        _castorMapName = "/RmsCastorMap.xml";

        // set the environment property if needed
        if (_defaultEnvironment != null) {
            this.setProperty(ENVIRONMENT_PROPERTY_NAME, _defaultEnvironment);
        }
    }
}
