package gov.fnal.cd.rms.base;

import novadaq.XMLSerializable;

/**
 * @package gov.fnal.cd.rms.base
 *
 * @brief Basic classes that encapsulate the data that is 
 * transferred by RMS.
 *
 * The classes inside base consist mostly of data structures
 * that are passed between user programs and RMS itself.  These
 * include the destination class that is used to define where a 
 * message is headed, as well as the message class itself.
 */

/**
 * Class for the messages and replies that are transmitted between
 * applications.
 *
 * @author Kurt Biery
 * @author Steve Foulkes
 * @version $Revision: 1.3 $ $Date: 2009/11/25 14:52:19 $
 */
public class RmsMessage extends novadaq.CastorSerializableBase {
    /**
     * synchronization token for creating UUIDs
     */
    private static Integer _syncToken = new Integer(42);

    /**
     * message ID
     */
    private String _id = null;

    /**
     * correlation ID
     */
    private String _correlationId = null;

    /**
     * the message destination
     */
    private RmsDestination _destination = null;

    /**
     * the destination to which replies should be sent
     */
    private RmsDestination _replyDestination = null;

    /**
     * the time at which the message was sent
     */
    private java.util.Date _sentTime = null;

    /**
     * the amount of time that the message is valid
     */
    private long _timeToLive = 5000;

    /**
     * for replies, the time at which the original message was sent
     */
    private java.util.Date _correlationTime = null;

    /**
     * the message source
     */
    private RmsMessageSource _source = null;

    /**
     * the message priority
     */
    private int _priority = 5;  // need to define high/med/low, if needed

    /**
     * the message body
     */
    private RmsMessageBody _body = null;

    /**
     * Container for the message status
     */
    private RmsStatus _messageStatus = RmsStatus.getSuccessStatus();

    /**
     * Creates an RMS message with an empty body.
     */
    public RmsMessage() {
        // set the name of the map we'll use for Castor XML binding
        _castorMapName = "/RmsCastorMap.xml";
    }

    /**
     * Creates an RMS message and populates the internal data of the
     * object using the specified serialized data string.
     *
     * @param serializedData The serialized message
     * @throws novadaq.DataException if an error occurs deserializing the
     *         string representation of the object into its internal data
     */
    public RmsMessage(String serializedData) throws novadaq.DataException {
        // set the name of the map we'll use for Castor XML binding
        _castorMapName = "/RmsCastorMap.xml";

        this.deserialize(serializedData);
    }

    /**
     * Creates an RMS message with the specified body.
     *
     * @param body The message body to associate with this
     * message.
     */
    public RmsMessage(RmsMessageBody body) {
        // set the name of the map we'll use for Castor XML binding
        _castorMapName = "/RmsCastorMap.xml";

        this._body = body;  // make copy??
    }

    /**
     * Creates an RMS message with the specified Tranmittable object
     * as the message body.
     *
     * @param bodyObject The serializable to associate with this message.
     * @throws novadaq.DataException if an error occurs serializing the
     *         specified XMLSerializable object into the message body
     */
    public RmsMessage(XMLSerializable bodyObject) throws novadaq.DataException {
        // set the name of the map we'll use for Castor XML binding
        _castorMapName = "/RmsCastorMap.xml";

        this.setBody(bodyObject);
    }

    /**
     * Returns the ID for the message.
     *
     * @return The message ID.
     */
    public String getId() {
        return _id;
    }

    /**
     * Returns the correlation ID for the message.  This ID is only valid
     * for replies, and it corresponds to the ID of the original message.
     *
     * @return The correlation ID for the message.
     */
    public String getCorrelationId() {
        return _correlationId;
    }

    /**
     * Returns the destination for the message.
     *
     * @return The destination for this message.
     */
    public RmsDestination getDestination() {
        // 27-Nov-2006, KAB: returning the bare destination (not a copy) is a
        // little dangerous in that the caller could accidentally modify what
        // we have stored internally, but it saves time when (de)serializing.
        //(copy) return (_destination == null) ? null :
        //(copy)     new RmsDestination(_destination);
        return _destination;
    }

    /**
     * Returns the destination to which replies to the message should be sent.
     *
     * @return The destination that replies to this message should
     * go to.
     */
    public RmsDestination getReplyDestination() {
        // 27-Nov-2006, KAB: returning the bare destination (not a copy) is a
        // little dangerous in that the caller could accidentally modify what
        // we have stored internally, but it saves time when (de)serializing.
        //(copy) return (_replyDestination == null) ? null :
        //(copy)     new RmsDestination(_replyDestination);
        return _replyDestination;
    }

    /**
     * Returns the time at which the message was sent.
     *
     * @return The time the message was sent.
     */
    public java.util.Date getSentTime() {
        return (_sentTime == null) ? null :
            new java.util.Date(_sentTime.getTime());
    }

    /**
     * Returns the amount of time for which the message is valid.
     *
     * @return The amount of times, in seconds the message is
     * valid for after it is sent.
     */
    public long getTimeToLive() {
        return _timeToLive;
    }

    /**
     * Returns the correlation time for the message.  This time is only
     * valid for reply messages, and it corresponds to the sent time of
     * the original message.
     *
     * @return The correlation time.
     */
    public java.util.Date getCorrelationTime() {
        return (_correlationTime == null) ? null :
            new java.util.Date(_correlationTime.getTime());
    }

    /**
     * Returns the source of the message.
     *
     * @return The source of the message.
     */
    public RmsMessageSource getSource() {
        return _source;
    }

    /**
     * Returns the priority for the message.
     *
     * @return The priority of the message.
     */
    public int getPriority() {
        return _priority;
    }

    /**
     * Returns the body of the message.
     *
     * @return The body of the message.
     */
    public RmsMessageBody getBody() {
        return _body;
    }

    /**
     * Returns the body of the message restored to an XMLSerializable object
     * or null if the body is empty.
     *
     * @return The body of the message in the form of an
     *         XMLSerializable object.
     * @throws novadaq.DataException if an error occurs deserializing the
     *         message body into the appropriate XMLSerializable object
     */
    public XMLSerializable getBodyObject() throws novadaq.DataException {
        XMLSerializable resultObject = null;

        // fetch body quantities
        RmsMessageBody ourBody = this.getBody();
        if (ourBody != null) {
            String className = ourBody.getClassName();
            if (className != null) {

                // create an instance of the class that is in the body
                Object testObj = null;

                // At some level, catch bare exceptions is bad form, but the
                // alternative in this case is catching three separate
                // exceptions and handling them in exactly the same way.  So,
                // we go ahead and catch bare exceptions, but keep the scope
                // to a minimum so that, hopefully, we don't accidentally
                // catch any Runtime exceptions.
                try {
                    Class targetClass = Class.forName(className);
                    testObj = targetClass.newInstance();
                }
                catch (Exception excpt) {
                    String msg = "Error deserializing object of type " +
                        className;
                    throw new novadaq.DataException(msg, excpt,
                                                    ourBody.getContent());
                }

                if (testObj != null &&
                    testObj instanceof novadaq.XMLSerializable &&
                    ourBody.getContent() != null) {

                    // deserialize the body contents into the object
                    resultObject = (XMLSerializable) testObj;
                    resultObject.deserialize(ourBody.getContent());
                }
            }
        }

        // return the resulting object
        return resultObject;
    }

    /**
     * Sets the ID for the message.  NOTE that this method is primarily
     * to be used by deserialization routines.  It is not intended for use
     * in application-level code.
     *
     * @param uuid The new message ID.
     */
    public void setId(String uuid) {
        _id = uuid;
    }

    /**
     * Sets the correlation ID for the message.
     *
     * @param uuid The new correlation ID.
     */
    public void setCorrelationId(String uuid) {
        _correlationId = uuid;
    }

    /**
     * Sets the destination for the message.
     *
     * @param destination The new destination for the message.
     */
    public void setDestination(RmsDestination destination) {
        // 27-Nov-2006, KAB: storing the bare destination that is passed in
        // (not a copy) is a little dangerous in that the caller could
        // accidentally modify what we have stored internally, but it saves
        // time when (de)serializing.
        //(copy) _destination = (destination == null) ? null :
        //(copy)     new RmsDestination(destination);
        _destination = destination;
    }

    /**
     * Sets the destination to which replies to the message should be sent.
     *
     * @param destination The reply destination for the message.
     */
    public void setReplyDestination(RmsDestination destination) {
        // 27-Nov-2006, KAB: storing the bare destination that is passed in
        // (not a copy) is a little dangerous in that the caller could
        // accidentally modify what we have stored internally, but it saves
        // time when (de)serializing.
        //(copy) _replyDestination = (destination == null) ? null :
        //(copy)     new RmsDestination(destination);
        _replyDestination = destination;
    }

    /**
     * Sets the time when the message is sent.
     *
     * @param timestamp The time to set the sent time
     * to.  If this is null, the sent time will be set
     * to the current time.
     */
    public void setSentTime(java.util.Date timestamp) {
        _sentTime = (timestamp == null) ? null :
            new java.util.Date(timestamp.getTime());
    }

    /**
     * Sets the amount time (in milliseconds) that the message will be valid.
     *
     * @param milliSeconds The TTL for the message.
     */
    public void setTimeToLive(long milliSeconds) {
        _timeToLive = milliSeconds;
    }

    /**
     * Sets the correlation time for the message.
     *
     * @param timestamp The correlation time for the 
     * message.  If this is null, the correlation time
     * will be set to the current time.
     */
    public void setCorrelationTime(java.util.Date timestamp) {
        _correlationTime = (timestamp == null) ? null :
            new java.util.Date(timestamp.getTime());
    }

    /**
     * Sets the source of the message.
     *
     * @param source The message source.
     */
    public void setSource(RmsMessageSource source) {
        _source = source;
    }

    /**
     * Sets the message priority.
     *
     * @param value The message priority.
     */
    public void setPriority(int value) {
        _priority = value;
    }

    /**
     * Sets the body of the message to the specified RmsMessageBody.
     *
     * @param body The message body.
     */
    public void setBody(RmsMessageBody body) {
        _body = body;
    }

    /**
     * Sets the body of the message to the specified XMLSerializable.
     *
     * @param bodyObject The message body in the form of an XMLSerializable.
     * @throws novadaq.DataException if an error occurs serializing the
     *         specified XMLSerializable object into the message body
     */
    public void setBody(XMLSerializable bodyObject) throws novadaq.DataException {
        if (bodyObject != null) {
            _body = new RmsMessageBody();
            _body.setClassName(bodyObject.getClass().getName());
            _body.setContent(bodyObject.serialize());
        }
        else {
            _body = null;
        }
    }

    /**
     * Creates a new message to be sent as a reply to the message.
     *
     * @return An RmsMessage that has been setup as a reply to
     * this message.
     */
    public RmsMessage createReply() {
        RmsMessage replyMsg = new RmsMessage();
        replyMsg.setCorrelationId(this.getId());

        RmsDestination newDest =
            new RmsDestination(this.getReplyDestination());
        newDest.setProperty(RmsDestination.MESSAGE_TYPE_PROPERTY_NAME,
                            RmsDestination.REPLY_MESSAGE_TYPE);
        replyMsg.setDestination(newDest);

        replyMsg.setTimeToLive(this.getTimeToLive());
        return replyMsg;
    }

    /**
     * Attach an RmsStatus object to this message
     *
     * @param messageStatus The RmsStatus object to attach.
     */
    public void setMessageStatus(RmsStatus messageStatus) {
        _messageStatus = messageStatus;
    }

    /**
     * Get the RmsStatus object from this message, if there 
     * is one.
     *
     * @return The message status.
     */
    public RmsStatus getMessageStatus() {
        return _messageStatus;
    }

    /**
     * Runs unit tests for the RmsMessage class.
     */
    public static void main(String[] args) {
        int testCount = 0;

        // create a sample message
        testCount++;
        System.out.println("\n>>>>> Test " + testCount + " <<<<<");
        System.out.println("Creating a sample message with " +
                           "no body...");
        RmsMessage message = new RmsMessage();
        System.out.println("Result of serialization...");
        System.out.println("----------------------------------------");
        try {
            System.out.println(message.serialize());
        }
        catch (Exception excpt) {
            System.err.println("Exception caught while serializing");
            excpt.printStackTrace(System.err);
        }
        System.out.println("----------------------------------------");

        // add time and body
        testCount++;
        System.out.println("\n>>>>> Test " + testCount + " <<<<<");
        System.out.println("Adding a sent time and a body " +
                           "to the message...");
        message.setSentTime(new java.util.Date());
        RmsMessageBody body = new RmsMessageBody();
        body.setClassName("gov.fnal.cd.rms.test1");
        body.setContent("This is a test");
        message.setBody(body);
        System.out.println("Result of serialization...");
        System.out.println("----------------------------------------");
        try {
            System.out.println(message.serialize());
        }
        catch (Exception excpt) {
            System.err.println("Exception caught while serializing");
            excpt.printStackTrace(System.err);
        }
        System.out.println("----------------------------------------");

        // add destinations
        testCount++;
        System.out.println("\n>>>>> Test " + testCount + " <<<<<");
        System.out.println("Adding destinations " +
                           "to the message...");
        RmsDestination targetDest =
            new RmsDestination("dcm005", RmsDestination.CONTROL_CHANNEL,
                               RmsDestination.REQUEST_MESSAGE_TYPE);
        message.setDestination(targetDest);
        RmsDestination replyDest =
            new RmsDestination("RunControl_2", RmsDestination.CONTROL_CHANNEL,
                               RmsDestination.REPLY_MESSAGE_TYPE);
        message.setReplyDestination(replyDest);
        String xmlString = "";
        System.out.println("Result of serialization...");
        System.out.println("----------------------------------------");
        try {
            xmlString = message.serialize();
            System.out.println(xmlString);
        }
        catch (Exception excpt) {
            System.err.println("Exception caught while serializing");
            excpt.printStackTrace(System.err);
        }
        System.out.println("----------------------------------------");

        // test deserialization
        testCount++;
        System.out.println("\n>>>>> Test " + testCount + " <<<<<");
        System.out.println("Testing deserialization...");
        System.out.println("Result of deserialization and " +
                           "re-serialization...");
        System.out.println("----------------------------------------");
        try {
            RmsMessage msgTester = new RmsMessage();
            msgTester.deserialize(xmlString);
            System.out.println(msgTester.serialize());
        }
        catch (Exception excpt) {
            System.err.println("Exception caught while deserializing");
            excpt.printStackTrace(System.err);
        }
        System.out.println("----------------------------------------");
    }
}
