package gov.fnal.cd.rms.base;

/**
 * Class to hold information about errors that occur
 * in the process of receiving and decoding RmsMessages.
 * 
 * @author Kurt Biery
 * @author Steve Foulkes
 * @version $Revision: 1.2 $ $Date: 2009/11/25 14:52:20 $
 */

public class RmsStatus {

    /**
     * Constant value indicating successful completion.
     */
    public static final int SUCCESS_CODE = 0;

    /**
     * Constant value indicating that an exception occurred.
     */
    public static final int EXCEPTION_CODE = 1;

    /**
     * Constant value indicating that an error occurred.
     */
    public static final int ERROR_CODE = 2;

    /**
     * Constant string indicating successful completion.
     */
    public static final String SUCCESS_MESSAGE = "Success";

    /**
     * Constant string indicating that an exception occurred.
     */
    public static final String EXCEPTION_MESSAGE = "Exception";

    /**
     * The status code.
     */
    private int _code = SUCCESS_CODE;

    /**
     * The message associated with an ERROR status code.
     */
    private String _message = SUCCESS_MESSAGE;

    /**
     * The exception object associated with an EXCEPTION status code.
     */
    private Exception _excp = null;

    /**
     * A reusable success status object.
     */
    private static RmsStatus _successStatus = null;

    /**
     * Constructs an RmsStatus object with the specified status code
     * and message.
     * @param code integer value that indicates success or failure
     * @param message text string that describes the status
     */
    public RmsStatus(int code, String message) {
        _code = code;
        _message = message;
    }

    /**
     * Constructs an RmsStatus object from the specified exception.
     * @param excp the exception object that was generated by some
     *             exceptional condition
     */
    public RmsStatus(Exception excp) {
        _code = EXCEPTION_CODE;
        _message = EXCEPTION_MESSAGE;
        _excp = excp;
    }

    /**
     * Returns the status code.
     *
     * @return the status code (int)
     */
    public int getCode() {
        return _code;
    }

    /**
     * Returns the status message.
     *
     * @return the status message (String)
     */
    public String getMessage() {
        return _message;
    }

    /**
     * Returns the exception that was encountered when the status code is
     * EXCEPTION and null otherwise.
     *
     * @return the status exception (or null if not EXCEPTION status type)
     */
    public Exception getException() {
        return _excp;
    }

    /**
     * Returns a string representation of the status object.
     *
     * @return a string describing the status
     */
    public String toString() {
        StringBuffer buff = new StringBuffer();
        buff.append(super.toString());
        buff.append("\n  code = " + _code);
        buff.append("\n  message = " + _message);
        buff.append("\n  exception = ");
        if (_excp == null) {
            buff.append("null");
        }
        else {
            _excp.printStackTrace(System.out);
            buff.append(_excp.toString());
            if (_excp instanceof novadaq.DataException) {
                buff.append("\n    Data = " +
                            ((novadaq.DataException) _excp).getData());
            }
            StackTraceElement[] traceElements = _excp.getStackTrace();
            for (int idx = 0; idx < traceElements.length; idx++) {
                buff.append("\n" + traceElements[idx].toString());
            }
        }
        return buff.toString();
    }

    /**
     * Returns a status object initialized to indicate success.
     *
     * @return An RmsStatus object initialized to success.
     */
    public static RmsStatus getSuccessStatus() {
        // if and when we allow RmsStatus to be modified, we'll
        // need to change this to create new instances each time.
        if (_successStatus == null) {
            _successStatus =
                new RmsStatus(RmsStatus.SUCCESS_CODE,
                              RmsStatus.SUCCESS_MESSAGE);
        }
        return _successStatus;
    }
}
