package gov.fnal.cd.rms.provider;

import gov.fnal.cd.rms.base.RmsCommunicationException;
import gov.fnal.cd.rms.base.RmsDestination;
import gov.fnal.cd.rms.base.RmsMessage;
import gov.fnal.cd.rms.util.RmsException;
import gov.fnal.cd.rms.util.UUIDGenerator;
import java.util.*;
import java.lang.reflect.*;

/**
 * This class takes care of setting up and managing connections
 * to DDS for RMS.
 *
 * @author Kurt Biery
 * @version $Revision: 1.6 $ $Date: 2010/04/06 19:37:42 $
 */

public class DDSConnection implements RmsConnection {

    /**
     * The UUID generator instance for this connection.
     */
    private UUIDGenerator _uuidGenerator;

    /**
     * Destination to represent where messages originate from.
     */
    private RmsDestination _messageSource;

    private String _myHostName;

    /**
     * The default time-to-live for DDS messages.
     */
    private long _defaultTimeToLive = 5000;

    private Hashtable<String, DDSWriter> writerTable =
        new Hashtable<String, DDSWriter>();
    private Hashtable<String, DDSReader> readerTable =
        new Hashtable<String, DDSReader>();

    /**
     * Creates a new connection to the messaging system.
     *
     * @param applicationName The name of the application generating/receiving
     * messages.
     * @param applicationPartition The partition of the application that is
     * generating/receiving messages.  
     *
     * @throws RmsCommunicationException If there is an error
     *         making the connection to DDS.
     */
    public DDSConnection(String applicationName, int applicationPartition)
        throws RmsCommunicationException {

        // other initialization
        _myHostName = "Unknown";
        try {
            _myHostName = java.net.InetAddress.getLocalHost().getHostName();
            _uuidGenerator = new UUIDGenerator(_myHostName);
        }
        catch (java.net.UnknownHostException uhExcpt) {
            _uuidGenerator = new UUIDGenerator("JavaDDSConnection");
        }

        _messageSource = new RmsDestination();
        _messageSource.setProperty(RmsDestination.TARGET_PROPERTY_NAME, 
                                   applicationName);
        _messageSource.setPartitionProperty(applicationPartition);        
    }

    /**
     * Sends a message of the specified type to the specified
     * destination.
     *
     * @param messageDataType the java.lang.Class of the message to be sent
     * @param dest the destination that the message should be sent to
     * @param theMessage the message to be sent
     * @throws RmsCommunicationException if an error occurs when trying
     *         to send the message to the destination
     */
    public <T> void sendMessage(Class<T> messageDataType, RmsDestination dest,
                                T theMessage)
        throws RmsCommunicationException {

        String targetProperty =
            dest.getProperty(RmsDestination.TARGET_PROPERTY_NAME);
        String channelProperty =
            dest.getProperty(RmsDestination.CHANNEL_PROPERTY_NAME);
        String key = messageDataType.getName() + targetProperty +
            channelProperty;

        DDSWriter myWriter = writerTable.get(key);
        if (myWriter != null) {

            // sloppy, sloppy
            Object hdrObject = null;
            try {
                Field hdrField = messageDataType.getDeclaredField("header");
                hdrObject = hdrField.get(theMessage);
            }
            catch (Exception excpt) {}

            if (hdrObject != null &&
                hdrObject instanceof rmscore.RmsMessageHeader) {
                rmscore.RmsMessageHeader theHeader =
                    (rmscore.RmsMessageHeader) hdrObject;

                theHeader.id = this.getUUIDNoThrow();
                theHeader.target = targetProperty;
                theHeader.channel = channelProperty;

                String appName =
                    _messageSource.getProperty(RmsDestination.
                                               TARGET_PROPERTY_NAME);
                theHeader.replyTarget = appName;
                theHeader.applicationName = appName;

                long now = System.currentTimeMillis();
                long seconds = (long) (now / 1000);
                theHeader.sentTime.sec = (int) seconds;
                theHeader.sentTime.usec = (int) ((now - (seconds * 1000)) * 1000);

                String partitionString =
                    _messageSource.getProperty(RmsDestination.
                                               PARTITION_PROPERTY_NAME);
                int partitionNumber = -99;
                try {
                    partitionNumber = Integer.parseInt(partitionString);
                }
                catch (Exception excpt) {}
                theHeader.partitionNumber = partitionNumber;

                theHeader.hostName = _myHostName;

                //System.out.println("^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^");
                //System.out.println("ID = " + theHeader.id);
                //System.out.println("Target = " + theHeader.target);
                //System.out.println("Channel = " + theHeader.channel);
                //System.out.println("Reply Target = " + theHeader.replyTarget);
                //System.out.println("Corrrelation ID = " + theHeader.correlationId);
                //System.out.println("Sent Time = " + theHeader.sentTime.sec + " " + theHeader.sentTime.usec);
                //System.out.println("Correlation Time = " + theHeader.correlationTime.sec + " " + theHeader.correlationTime.usec);
                //System.out.println("Priority = " + theHeader.priority);
                //System.out.println("Partition Number = " + theHeader.partitionNumber);
                //System.out.println("Host Name = " + theHeader.hostName);
                //System.out.println("Application Name = " + theHeader.applicationName);
            }

            myWriter.send(theMessage);
        }
    }

    /**
     * Receives a message of the specified type from the specified
     * destination.  This is a blocking call, so it does not return
     * until a message becomes available.
     *
     * @param messageDataType the java.lang.Class of the message that will
     *        be received
     * @param dest the destination that the message was sent to
     * @throws RmsCommunicationException if an error occurs when trying
     *         to receive the message from the destination
c     */
    public <T> Vector<T> receiveMessage(Class<T> messageDataType,
                                        RmsDestination dest)
        throws RmsCommunicationException {
        Vector<T> tList = new Vector<T>();

        String targetProperty =
            dest.getProperty(RmsDestination.TARGET_PROPERTY_NAME);
        String channelProperty =
            dest.getProperty(RmsDestination.CHANNEL_PROPERTY_NAME);
        String key = messageDataType.getName() + targetProperty +
            channelProperty;
        DDSReader myReader = readerTable.get(key);
        if (myReader != null) {
            Vector<Object> objList = myReader.receive();
            if (objList != null && objList.size() > 0) {
                for (int idx = 0; idx < objList.size(); idx++) {
                    T receivedMessage =
                        messageDataType.cast(objList.elementAt(idx));
                    tList.addElement(receivedMessage);
                }
            }
        }
        return tList;
    }

    /**
     * Receives a message of the specified type from the specified
     * destination.  If no message is received within the specified timeout,
     * a null message is returned.
     *
     * @param messageDataType the java.lang.Class of the message that will
     *        be received
     * @param dest the destination that the message was sent to
     * @param timeout The maximum amount of time to wait for a message,
     * in milliseconds.
     * @throws RmsCommunicationException if an error occurs when trying
     *         to receive the message from the destination
     */
    public <T> Vector<T> receiveMessage(Class<T> messageDataType,
                                        RmsDestination dest,
                                        long timeout)
        throws RmsCommunicationException {
        Vector<T> tList = new Vector<T>();

        String targetProperty =
            dest.getProperty(RmsDestination.TARGET_PROPERTY_NAME);
        String channelProperty =
            dest.getProperty(RmsDestination.CHANNEL_PROPERTY_NAME);
        String key = messageDataType.getName() + targetProperty +
            channelProperty;
        DDSReader myReader = readerTable.get(key);
        if (myReader != null) {
            Vector<Object> objList = myReader.receive(timeout);
            if (objList != null && objList.size() > 0) {
                for (int idx = 0; idx < objList.size(); idx++) {
                    T receivedMessage =
                        messageDataType.cast(objList.elementAt(idx));
                    tList.addElement(receivedMessage);
                }
            }
        }
        return tList;
    }

    public <T> void registerSender(Class<T> messageDataType,
                                   RmsDestination dest) {
        String targetProperty =
            dest.getProperty(RmsDestination.TARGET_PROPERTY_NAME);
        String channelProperty =
            dest.getProperty(RmsDestination.CHANNEL_PROPERTY_NAME);
        String topicName = targetProperty + "_" + channelProperty +
            "_" + messageDataType.getName().replaceAll("\\.", "_");
        String partitionName = "Partition" +
            _messageSource.getProperty(RmsDestination.
                                       PARTITION_PROPERTY_NAME);
        DDSWriter myWriter =
            new DDSWriter(messageDataType.getName(), null,
                          topicName, partitionName);

        String key = messageDataType.getName() + targetProperty +
            channelProperty;
        writerTable.put(key, myWriter);
    }

    public <T> void registerReceiver(Class<T> messageDataType,
                                     RmsDestination dest) {
        String targetProperty =
            dest.getProperty(RmsDestination.TARGET_PROPERTY_NAME);
        String channelProperty =
            dest.getProperty(RmsDestination.CHANNEL_PROPERTY_NAME);
        String topicName = targetProperty + "_" + channelProperty +
            "_" + messageDataType.getName().replaceAll("\\.", "_");
        String partitionName = "Partition" +
            _messageSource.getProperty(RmsDestination.
                                       PARTITION_PROPERTY_NAME);
        DDSReader myReader =
            new DDSReader(messageDataType.getName(), null,
                          topicName, partitionName);

        String key = messageDataType.getName() + targetProperty +
            channelProperty;
        readerTable.put(key, myReader);
    }

    public <T> void unRegisterSender(Class<T> messageDataType,
                                     RmsDestination dest) {
        String targetProperty =
            dest.getProperty(RmsDestination.TARGET_PROPERTY_NAME);
        String channelProperty =
            dest.getProperty(RmsDestination.CHANNEL_PROPERTY_NAME);
        String key = messageDataType.getName() + targetProperty +
            channelProperty;
        DDSWriter myWriter = writerTable.get(key);
        if (myWriter != null) {
            myWriter.close();
            writerTable.remove(key);
        }
    }

    public <T> void unRegisterReceiver(Class<T> messageDataType,
                                       RmsDestination dest) {
        String targetProperty =
            dest.getProperty(RmsDestination.TARGET_PROPERTY_NAME);
        String channelProperty =
            dest.getProperty(RmsDestination.CHANNEL_PROPERTY_NAME);
        String key = messageDataType.getName() + targetProperty +
            channelProperty;
        DDSReader myReader = readerTable.get(key);
        if (myReader != null) {
            myReader.close();
            readerTable.remove(key);
        }
    }

    /**
     * This operation is not supported by a DDSConnection.  It is only
     * provided to fulfill the RmsConnection interface, and it should
     * be removed if/when we modify that interface.
     *
     * @throws RmsCommunicationException Always.
     */
    public void sendString(RmsDestination dest, String messageString)
        throws RmsCommunicationException {

        String msg = "The sendString() method is not currently " +
            "supported in the DDSConnection class";
        throw new RmsCommunicationException(msg, null);
    }

    /**
     * This operation is not supported by a DDSConnection.  It is only
     * provided to fulfill the RmsConnection interface, and it should
     * be removed if/when we modify that interface.
     *
     * @throws RmsCommunicationException Always.
     */
    public void sendMessage(RmsDestination dest, RmsMessage message)
        throws RmsCommunicationException {

        String msg = "The sendMessage() method is not currently " +
            "supported in the DDSConnection class";
        throw new RmsCommunicationException(msg, null);
    }

    /**
     * This operation is not supported by a DDSConnection.  It is only
     * provided to fulfill the RmsConnection interface, and it should
     * be removed if/when we modify that interface.
     *
     * @throws RmsCommunicationException Always.
     */
    public void addListener(RmsDestination dest, ProviderListener listener)
        throws RmsCommunicationException {

        String msg = "The addListener() method is not currently " +
            "supported in the DDSConnection class";
        throw new RmsCommunicationException(msg, null);
    }

    /**
     * This operation is not supported by a DDSConnection.  It is only
     * provided to fulfill the RmsConnection interface, and it should
     * be removed if/when we modify that interface.
     *
     * @throws RmsCommunicationException Always.
     */
    public void removeListener(RmsDestination dest, ProviderListener listener)
        throws RmsCommunicationException {

        String msg = "The removeListener() method is not currently " +
            "supported in the DDSConnection class";
        throw new RmsCommunicationException(msg, null);
    }

    /**
     * This operation is not supported by a DDSConnection.  It is only
     * provided to fulfill the RmsConnection interface, and it should
     * be removed if/when we modify that interface.
     *
     * @param pingDestination The destination to be tested.
     * @return False always.
     */
    public boolean ping(RmsDestination pingDestination) {
        return false;
    }

    /**
     * Closes the connection.  After calling this method, the
     * connection should no longer be used since it will no
     * longer be connected to the underlying provider system.
     *
     * @throws RmsCommunicationException If a problem occurs when trying
     *         to disconnect from the provider.
     */
    public void close() throws RmsCommunicationException {
        Exception firstException = null;

        // if we've already been closed, simply return
        //if (_jcaContext == null) {}

        // re-package the first exception and report it
        if (firstException != null) {
            String msg = "Error closing DDSConnection.";
            throw new RmsCommunicationException(msg, firstException);
        }
        return;
    }

    /**
     * Handles cleanup when garbage collected.
     */
    protected void finalize() {
        try {
            this.close();
        }
        catch (RmsCommunicationException rcExcpt) {
            rcExcpt.printStackTrace(System.err);
        }
    }

    /**
     * Get a reference to the message source.
     *
     * @return A destination object that contains the source of messages
     * for this connection.
     */    
    public RmsDestination getMessageSource() {
        return new RmsDestination(_messageSource);
    }

    /**
     * Returns a universally unique identifier.  This can be used to uniquely
     * identify a message.
     *
     * @return The unique ID string.
     * @throws novadaq.DataException If an error occurs while creating the UUID.
     */
    public String getUUID() throws novadaq.DataException {
        try {
            return _uuidGenerator.getUUID().toString();
        }
        catch (RmsException rmsExcpt) {
            String msg = "Unable to create unique identifier.";
            throw new novadaq.DataException(msg, rmsExcpt);
        }
    }

    /**
     * Returns a universally unique identifier.  This can be used to uniquely
     * identify a message.  This version of the method will not throw a checked
     * exception.  Instead it will make a best-effort attempt to create a
     * unique ID.
     *
     * @return The unique ID string.
     */
    public String getUUIDNoThrow() {
        try {
            return _uuidGenerator.getUUID().toString();
        }
        catch (RmsException rmsExcpt) {
            return _messageSource.getProperty(RmsDestination.TARGET_PROPERTY_NAME) +
                System.currentTimeMillis();
        }
    }

    /**
     * Returns the default message expiration interval, in milliseconds
     * (also known as the time-to-live).
     *
     * @return The default time-to-live for messages.
     */
    public long getDefaultTimeToLive() {
        return _defaultTimeToLive;
    }

    /**
     * Tests whether the specified destination is supported by this
     * connection type.
     *
     * @param candidateDestination The destination that should be tested.
     * @return True if the destination is supported, false otherwise.
     */
    public boolean supportsDestination(RmsDestination candidateDestination) {
        return true;
    }
}
