package gov.fnal.cd.rms.provider;

import gov.fnal.cd.rms.base.RmsCommunicationException;
import gov.fnal.cd.rms.base.RmsDestination;
import gov.fnal.cd.rms.base.RmsMessage;
import gov.fnal.cd.rms.base.RmsRuntimeException;
import gov.fnal.cd.rms.util.RmsException;
import gov.fnal.cd.rms.util.UUIDGenerator;

import java.io.FileReader;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.InputStream;
import java.io.IOException;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;

import gov.aps.jca.CAException;
import gov.aps.jca.Context;
import gov.aps.jca.Channel;
import gov.aps.jca.JCALibrary;
import gov.aps.jca.jni.SingleThreadedContext;

/**
 * This class handles the details of sending messages using EPICS
 * and JCA.
 *
 * @author Kurt Biery
 * @author Steve Foulkes
 * @version $Revision: 1.7 $ $Date: 2010/03/15 20:38:46 $
 */
public class EpicsConnection implements RmsConnection {
    /**
     * The delimiter used when constructing process variable names.
     */
    private static final String PV_DELIMITER = "/";

    /**
     * The JCA context used by this connection.
     */
    private Context _jcaContext;

    /**
     * The UUID generator instance for this connection.
     */
    private UUIDGenerator _uuidGenerator;

    /**
     * Destination to represent where messages originate from.
     */
    private RmsDestination _messageSource;

    /**
     * The default time-to-live for EPICS messages.
     */
    private long _defaultTimeToLive = 5000;

    /**
     * A cache of messenger objects that are used to communicate
     * with EPICS process variables.
     */
    private Hashtable<String, EpicsMessenger> _cachedPVs;

    /**
     * Creates a new connection to the messaging system.
     *
     * @param applicationName The name of the application generating/receiving
     * messages.
     * @param applicationPartition The partition of the application that is
     * generating/receiving messages.  
     *
     * @throws NumberFormatException If the RMS_EPICS_TTL environmental
     *         variable is set but contains a value that can not be
     *         interpreted as a number.
     * @throws RmsCommunicationException If there is an error
     *         making the connection to EPICS.
     */
    public EpicsConnection(String applicationName, int applicationPartition)
        throws RmsCommunicationException {

        // check for a default time-to-live
        String valueString = System.getenv("RMS_EPICS_TTL");
        if (valueString != null && valueString.trim().length() > 0) {
            _defaultTimeToLive = Long.parseLong(valueString, 10);
        }

        // other initialization
        try {
            _uuidGenerator = new UUIDGenerator(java.net.InetAddress.getLocalHost().getHostName());
        }
        catch (java.net.UnknownHostException uhExcpt) {
            _uuidGenerator = new UUIDGenerator("JavaEpicsConnection");
        }

        _cachedPVs = new Hashtable<String, EpicsMessenger>();

        // a single-threaded context seems to be the right choice
        // from a reliability standpoint
        JCALibrary jcaLib = JCALibrary.getInstance();
        try {
            _jcaContext = jcaLib.createContext(JCALibrary.JNI_SINGLE_THREADED);
        }
        catch (CAException caExcpt) {
            String msg = "Unable to create an EPICS JCA context.";
            throw new RmsCommunicationException(msg, caExcpt);
        }

        _messageSource = new RmsDestination();
        _messageSource.setProperty(RmsDestination.TARGET_PROPERTY_NAME, 
                                   applicationName);
        _messageSource.setPartitionProperty(applicationPartition);        
    }

    /**
     * Sends a string to a particular destination.
     *
     * @param dest The destination that the message is going to.
     * @param messageString The string that is being sent.
     * @throws RmsCommunicationException If an error occurs when trying
     *         to send the message to the destination.
     * @throws novadaq.DataException If an error occurs when trying
     *         to prepare the message for sending.
     */
    public void sendString(RmsDestination dest, String messageString)
        throws RmsCommunicationException, novadaq.DataException {

        String destinationPV;
        EpicsMessenger pvHandle;

        // translate the abstract RmsDestination object into concrete PV names
        destinationPV = getDestinationPVFromDest(dest);

        // catch exceptions and re-package them
        try {
            // fetch a handle to the messenger that we'll use to talk to the PV
            pvHandle = getPVHandle(destinationPV);

            // since we created the context as single-threaded, we need to
            // prevent multiple threads from accessing it at the same time
            synchronized(_jcaContext) {

                // tell EPICS/JCA that it is OK to use access the 
                // context from the current thread
                if (_jcaContext instanceof SingleThreadedContext) {
                    ((SingleThreadedContext) _jcaContext).attachCurrentThread();
                }

                // send the message
                pvHandle.sendMessage(messageString);
            }
        }
        catch (CAException caExcpt) {
            String msg = "Error sending a string to EPICS " +
                "(process variable " + destinationPV + ").";
            throw new RmsCommunicationException(msg, caExcpt, dest);
        }
        catch (IOException ioExcpt) {
            String msg = "Error preparing a message to send to EPICS.";
            throw new novadaq.DataException(msg, ioExcpt, messageString);
        }

        return;
    }

    /**
     * Sends an RmsMessage to a particular destination.
     *
     * @param dest The destination that the message is going to.
     * @param message The RmsMessage that is being sent.
     * @throws RmsCommunicationException If an error occurs when trying
     *         to send the message to the destination.
     * @throws novadaq.DataException If an error occurs when trying
     *         to prepare the message for sending.
     */
    public void sendMessage(RmsDestination dest, RmsMessage message)
        throws RmsCommunicationException, novadaq.DataException {

        String destinationPV;
        String monitoringOutboxPV;
        EpicsMessenger pvHandle;
        EpicsMessenger mopvHandle;

        // translate the abstract RmsDestination object into concrete PV names
        destinationPV = getDestinationPVFromDest(dest);
        monitoringOutboxPV = getMonitoringPVFromDest(message.getReplyDestination());

        // catch exceptions and re-package them
        try {
            // fetch a handle to the messenger that we'll use to talk to the PV
            pvHandle = getPVHandle(destinationPV);
            mopvHandle = getPVHandle(monitoringOutboxPV);

            // since we created the context as single-threaded, we need to
            // prevent multiple threads from accessing it at the same time
            synchronized(_jcaContext) {

                // tell EPICS/JCA that it is OK to use access the 
                // context from the current thread
                if (_jcaContext instanceof SingleThreadedContext) {
                    ((SingleThreadedContext) _jcaContext).attachCurrentThread();
                }

                // send the message
                String serializedMessage = message.serialize();
                pvHandle.sendMessage(serializedMessage);

                // if monitoring is supported for this PV, fetch the handle
                // to the monitor outbox and send a copy of the message to it
                if (monitoringOutboxPV != null) {
                    mopvHandle.sendMessage(serializedMessage);
                }
            }
        }
        catch (CAException caExcpt) {
            String msg = "Error sending a message to EPICS " +
                "(process variables " + destinationPV + " and " +
                monitoringOutboxPV + ").";
            throw new RmsCommunicationException(msg, caExcpt, dest);
        }
        catch (IOException ioExcpt) {
            String msg = "Error preparing a message to send to EPICS.";
            throw new novadaq.DataException(msg, ioExcpt, message.serialize());
        }

        return;
    }

    /**
     * This operation is not supported by an EpicsConnection.  It is only
     * provided to fulfill the RmsConnection interface.
     *
     * @throws RmsCommunicationException always
     */
    public <T> void sendMessage(Class<T> messageDataType, RmsDestination dest,
                                T theMessage)
        throws RmsCommunicationException {

        String msg = "This version of the sendMessage() method is not " +
            "currently supported in the EpicsConnection class";
        throw new RmsCommunicationException(msg, null);
    }

    /**
     * This operation is not supported by an EpicsConnection.  It is only
     * provided to fulfill the RmsConnection interface.
     *
     * @throws RmsCommunicationException always
     */
    public <T> Vector<T> receiveMessage(Class<T> messageDataType, RmsDestination dest)
        throws RmsCommunicationException {

        String msg = "The receiveMessage() method is not currently " +
            "supported in the EpicsConnection class";
        throw new RmsCommunicationException(msg, null);
    }

    /**
     * This operation is not supported by an EpicsConnection.  It is only
     * provided to fulfill the RmsConnection interface.
     *
     * @throws RmsCommunicationException always
     */
    public <T> Vector<T> receiveMessage(Class<T> messageDataType, RmsDestination dest,
                                long timeout)
        throws RmsCommunicationException {

        String msg = "The receiveMessage() method is not currently " +
            "supported in the EpicsConnection class";
        throw new RmsCommunicationException(msg, null);
    }

    /**
     * Adds a listener to a particular destination.
     *
     * @param dest The destination that we will be listening to.
     * @param listener The listener that will be called when messages
     *        are posted to the destination.
     * @throws RmsCommunicationException If an error occurs when trying
     *         to add the listener to the destination.
     */
    public void addListener(RmsDestination dest, ProviderListener listener)
        throws RmsCommunicationException {

        // catch exceptions and re-package them
        try {
            String destinationPV = getListenerPVFromDest(dest);
            EpicsMessenger pvHandle = getPVHandle(destinationPV);
            synchronized (_jcaContext) {
                pvHandle.addListener(listener);
            }
        }
        catch (CAException caExcpt) {
            String msg = "Error adding a listener to EPICS.";
            throw new RmsCommunicationException(msg, caExcpt, dest);
        }
    }

    /**
     * Removes a listener from a particular destination.
     *
     * @param dest The destination that we no longer want to listen to.
     * @param listener The listener that will no longer be called.
     * @throws RmsCommunicationException If an error occurs when trying
     *         to remove the listener from the destination.
     */
    public void removeListener(RmsDestination dest, ProviderListener listener)
        throws RmsCommunicationException {

        // catch exceptions and re-package them
        try {
            String destinationPV = getListenerPVFromDest(dest);
            EpicsMessenger pvHandle = getPVHandle(destinationPV);
            synchronized (_jcaContext) {
                pvHandle.removeListener(listener);
            }
            // TODO - consider whether we want to remove an EpicsMessenger
            // from the cache when it no longer has any listeners.  The
            // primary case for that might be run control, which over time
            // may end up talking to several hundred PVs.  If we decide
            // to do that, we should double-check that the loop over the
            // cached PVs in the close() method doesn't go crazy if
            // this method deletes entries from the cache out from under it.
        }
        catch (CAException caExcpt) {
            String msg = "Error removing a listener from EPICS.";
            throw new RmsCommunicationException(msg, caExcpt, dest);
        }
    }

    /**
     * Tests whether the specified destination is reachable (messages
     * could successfully be sent).
     *
     * @param pingDestination The destination to be tested.
     * @return True if messages can be sent to the destination or 
     *         false if the destination is invalid or unreachable.
     */
    public boolean ping(RmsDestination pingDestination) {
        String destinationPV = getDestinationPVFromDest(pingDestination);
        Channel pvChannel;

        try {
            pvChannel = _jcaContext.createChannel(destinationPV);
            _jcaContext.pendIO(0.5d);
            pvChannel.destroy();
        }
        catch (Exception excp) {
            return false;
        }

        return true;
    }

    /**
     * Closes the connection.  After calling this method, the
     * connection should no longer be used since it will no
     * longer be connected to the underlying provider system.
     *
     * @throws RmsCommunicationException If a problem occurs when trying
     *         to disconnect from the provider.
     */
    public void close() throws RmsCommunicationException {
        if (_jcaContext == null) {return;}

        CAException firstException = null;
        synchronized (_cachedPVs) {
            Enumeration<EpicsMessenger> messengerEnum = _cachedPVs.elements();
            while (messengerEnum.hasMoreElements()) {
                EpicsMessenger messenger = messengerEnum.nextElement();
                // we want to close as much as possible, so we keep going
                // when exceptions are encountered here
                try {
                    messenger.close();
                }
                catch (CAException caExcpt) {
                    if (firstException == null) {
                        firstException = caExcpt;
                    }
                }
            }
            _cachedPVs.clear();
        }

        try {
            _jcaContext.destroy();
        }
        catch (CAException caExcpt) {
            if (firstException == null) {
                firstException = caExcpt;
            }
        }
        _jcaContext.dispose();
        _jcaContext = null;

        // re-package the first exception and report it
        if (firstException != null) {
            String msg = "Error closing EpicsConnection.";
            throw new RmsCommunicationException(msg, firstException);
        }
        return;
    }

    /**
     * Handles cleanup when garbage collected.
     */
    protected void finalize() {
        try {
            this.close();
        }
        catch (RmsCommunicationException rcExcpt) {
            rcExcpt.printStackTrace(System.err);
        }
    }

    /**
     * Get a reference to the message source.
     *
     * @return A destination object that contains the source of messages
     * for this connection.
     */    
    public RmsDestination getMessageSource() {
        return new RmsDestination(_messageSource);
    }

    /**
     * Returns a universally unique identifier.  This can be used to uniquely
     * identify a message.
     *
     * @return The unique ID string.
     * @throws novadaq.DataException If an error occurs while creating the UUID.
     */
    public String getUUID() throws novadaq.DataException {
        try {
            return _uuidGenerator.getUUID().toString();
        }
        catch (RmsException rmsExcpt) {
            String msg = "Unable to create unique identifier.";
            throw new novadaq.DataException(msg, rmsExcpt);
        }
    }

    /**
     * Returns a universally unique identifier.  This can be used to uniquely
     * identify a message.  This version of the method will not throw a checked
     * exception.  Instead it will make a best-effort attempt to create a
     * unique ID.
     *
     * @return The unique ID string.
     */
    public String getUUIDNoThrow() {
        try {
            return _uuidGenerator.getUUID().toString();
        }
        catch (RmsException rmsExcpt) {
            return _messageSource.getProperty(RmsDestination.TARGET_PROPERTY_NAME) +
                System.currentTimeMillis();
        }
    }

    /**
     * Returns the default message expiration interval, in milliseconds
     * (also known as the time-to-live).
     *
     * @return The default time-to-live for messages.
     */
    public long getDefaultTimeToLive() {
        return _defaultTimeToLive;
    }

    /**
     * Tests whether the specified destination is supported by this
     * connection type.
     *
     * @param candidateDestination The destination that should be tested.
     * @return True if the destination is supported, false otherwise.
     */
    public boolean supportsDestination(RmsDestination candidateDestination) {

        // special handing for monitoring of "source" destinations
        RmsDestination workDest = new RmsDestination(candidateDestination);
        if (candidateDestination.getProperty(RmsDestination.TARGET_PROPERTY_NAME) == null &&
            candidateDestination.getProperty(RmsDestination.SOURCE_PROPERTY_NAME) != null) {
            workDest.setProperty(RmsDestination.TARGET_PROPERTY_NAME,
                                 candidateDestination.getProperty(RmsDestination.SOURCE_PROPERTY_NAME));
        }

        // check if the destination is a valid target or source
        // of messages...
        String sendPVName = getDestinationPVFromDest(workDest);
        String listenPVName = getListenerPVFromDest(workDest);
        return (sendPVName != null && sendPVName.trim().length() > 0) ||
            (listenPVName != null && listenPVName.trim().length() > 0);
    }

    /**
     * Translates a destination object into the EPICS PV name appropriate
     * for a listener.
     *
     * @param dest The destination object that will be translated into a PV.
     * @return The process variable name.
     */
    private String getListenerPVFromDest(RmsDestination dest) {
        String destinationPV;

        if (dest.getProperty(RmsDestination.TARGET_PROPERTY_NAME) != null) {
            destinationPV = getDestinationPVFromDest(dest);
        }
        else {
            destinationPV = getSourcePVFromDest(dest);
        }

        return destinationPV;
    }

    /**
     * The name of the process variable that should be used
     * when watching for messages from a particular source.
     *
     * @param dest The destination that is the source of messages.
     * @return This will return the name of the PV for the
     *         message source.
     */
    private String getSourcePVFromDest(RmsDestination dest) {
        return getMonitoringPVFromDest(dest);
    }

    /**
     * Translates a destination object into an EPICS PV name.
     *
     * @param dest The destination object that will be translated
     *        into a PV.
     * @return The name of the PV for the location that the destination
     *         object represents.
     */
    private String getDestinationPVFromDest(RmsDestination dest) {
        String destinationPV;
        StringBuffer nameBuff = new StringBuffer();

        String destinationTarget = 
            dest.getProperty(RmsDestination.TARGET_PROPERTY_NAME);
        String destinationService = 
            dest.getProperty(RmsDestination.CHANNEL_PROPERTY_NAME); 
        String destinationType = 
            dest.getProperty(RmsDestination.MESSAGE_TYPE_PROPERTY_NAME);
        String destinationEnvironment = 
            dest.getProperty(RmsDestination.ENVIRONMENT_PROPERTY_NAME);
        String destinationPartition =
            dest.getProperty(RmsDestination.PARTITION_PROPERTY_NAME);

        /**
         * The heartbeat service is a special case.  For each target, there 
         * will be one heartbeat service.  All heartbeat messages will be 
         * pushed out to the outbox.
         */
        if (destinationService.equals(RmsDestination.HEARTBEAT_CHANNEL)) {
            if (destinationEnvironment != null) {
                nameBuff.append(destinationEnvironment + "_");
            }
            nameBuff.append(destinationTarget + PV_DELIMITER + 
                            destinationService +
                            PV_DELIMITER + "outbox");
            return nameBuff.toString();
        }

        if (destinationEnvironment != null) {
            nameBuff.append(destinationEnvironment + "_");
        }
        nameBuff.append(destinationTarget + PV_DELIMITER + 
                        destinationPartition + PV_DELIMITER + 
                        destinationService + PV_DELIMITER + "inbox");
        destinationPV = nameBuff.toString();

        return destinationPV;
    }

    /**
     * The name of the monitoring outbox that a message should be 
     * posted to when sending to a particular destination.
     *
     * @param dest The destination that the message is going to
     * @return This will return the name of the PV for the
     *         monitoring outbox that the message should be posted to.
     */
    private String getMonitoringPVFromDest(RmsDestination dest) {
        String targetName = 
            dest.getProperty(RmsDestination.TARGET_PROPERTY_NAME);
        String destinationPartition = 
            dest.getProperty(RmsDestination.PARTITION_PROPERTY_NAME);
        String destinationService = 
            dest.getProperty(RmsDestination.CHANNEL_PROPERTY_NAME); 

        /**
         * There is no monitoring outbox for the heartbeat service
         */
        if (destinationService.equals(RmsDestination.HEARTBEAT_CHANNEL)) {
            return null;
        }


        if (targetName == null) {
            targetName = dest.getProperty(RmsDestination.SOURCE_PROPERTY_NAME);
        }

        StringBuffer nameBuff = new StringBuffer();
        String destinationEnvironment = 
            dest.getProperty(RmsDestination.ENVIRONMENT_PROPERTY_NAME);
        if (destinationEnvironment != null) {
            nameBuff.append(destinationEnvironment + "_");
        }

        nameBuff.append(targetName + PV_DELIMITER + 
                        destinationPartition + PV_DELIMITER +
                        destinationService + PV_DELIMITER + 
                        "monitoringOutbox");

        return nameBuff.toString();
    }

    /**
     * Returns an EpicsMessenger for the specified process variable.
     *
     * @param pvName The name of the PV that we want a handle to.
     * @return An EpicsMessenger class that is connected to the PV.
     * @throws CAException If a problem occurs creating the handle.
     */
    synchronized private EpicsMessenger getPVHandle(String pvName)
        throws CAException {

        // Get a PV handle out of the cachedPVs hashtable if a
        // handle for the PV already exists.  If the PV isn't in the
        // table, create a new handle and add it.
        EpicsMessenger pvHandle = _cachedPVs.get(pvName);

        if (pvHandle == null) {
            synchronized(_jcaContext) {
                if (_jcaContext instanceof SingleThreadedContext) {
                    ((SingleThreadedContext) _jcaContext).attachCurrentThread();
                }
                pvHandle = new EpicsMessenger(pvName, _jcaContext, 1000);
                _cachedPVs.put(pvName, pvHandle);
            }
        }

        return pvHandle;
    }
}
