package gov.fnal.cd.rms.provider;

import java.io.IOException;
import java.util.Enumeration;
import java.util.ListIterator;
import java.util.Vector;

import gov.aps.jca.CAStatus;
import gov.aps.jca.CAException;
import gov.aps.jca.Context;
import gov.aps.jca.Channel;
import gov.aps.jca.JCALibrary;
import gov.aps.jca.Monitor;
import gov.aps.jca.TimeoutException;
import gov.aps.jca.dbr.BYTE;
import gov.aps.jca.dbr.DBR;
import gov.aps.jca.event.MonitorEvent;
import gov.aps.jca.event.MonitorListener;

/**
 * This class provides functionality to communicate with EPICS byte array
 * process variables.  It automatically splits up messages that are too
 * long into the necessary number of fragments and automatically assembles
 * fragments into messages on input.
 *
 * This class uses the EPICS JCA (Java for Channel Access) library for
 * interacting with EPICS.
 *
 * @author Kurt Biery
 * @author Steve Foulkes
 * @version $Revision: 1.1.1.1 $ $Date: 2009/10/09 20:03:00 $
 */
public class EpicsMessenger implements MonitorListener {
    /**
     * The length of time (in seconds) to use when waiting for
     * pending operations to complete.
     */
    private static final double _PEND_OP_TIMEOUT = 2.0d;

    /**
     * The JCA context that we're using for communication.
     */
    private Context _jcaContext;

    /**
     * The JCA channel that we'll use to access the process variable.
     */
    private Channel _pvChannel;

    /**
     * The name of the process variable.
     */
    private String _pvName;

    /**
     * The maximum number of bytes that we can write to an underlying
     * EPICS byte array at one time.
     */
    private int _maxMessageSize;

    /**
     * The list of listeners interested in being notified when
     * the process variable value changes.
     */
    private Vector<ProviderListener> _listenerList =
        new Vector<ProviderListener>();

    /**
     * The JCA monitor instance that is watching for changes in the
     * PV value.
     */
    private Monitor _pvMonitor = null;

    /**
     * The list of message assemblers that are active.
     */
    private Vector<MessageAssembler> _assemblerList =
        new Vector<MessageAssembler>();

    /**
     * Flag to indicate that we're waiting for an EPICS
     * monitor to be attached.
     */
    private boolean _waitingForMonitor = false;

    /**
     * Flag for debug output.
     */
    private boolean _debug = false;

    /**
     * Constructs a instance of this class for the specified
     * EPICS process variable.
     *
     * @param pvName the name of the process variable
     *        that this messenger will be used to communicate with
     * @param jcaContext the JCA context that will be used when creating
     *        needed EPICS/JCA communication objects (e.g. a channel)
     * @param maxStringSize the maximum size of the strings that can be
     *        written to the specified process variable
     * @throws CAException if a channel accesss error occurs while
     *         creating a connection to the process variable
     */
    public EpicsMessenger(String pvName, Context jcaContext, int maxStringSize)
        throws CAException {

        // save items that we'll need later
        _pvName = pvName;
        _jcaContext = jcaContext;
        _maxMessageSize = maxStringSize;

        // create the channel
        _pvChannel = _jcaContext.createChannel(_pvName);

        // wait for a time to allow JCA/EPICS to locate the PV
        try {
            _jcaContext.pendIO(_PEND_OP_TIMEOUT);
        }
        catch (TimeoutException toExcpt) {
            // re-package the timeout exception as a CA exception for easier
            // handling outside this routine (users probably don't care
            // why the creation failed...)
            String msg = "Timout waiting to communicate with " + pvName;
            throw new CAException(msg, toExcpt);
        }
    }

    /**
     * Returns the name of the EPICS process variable that this messenger
     * is communicating with.
     *
     * @return a string containing the process variable name
     */
    public String getName() {
        return _pvName;
    }

    /**
     * Sends the specified message.
     *
     * @param messageText the string text that should be written to the
     *        process variable.
     * @throws CAException if a channel access error occurs while sending
     *         the message data
     * @throws IOException if an error occurs while translating the message
     *         text into a form suitable for transfer
     */
    public void sendMessage(String messageText)
        throws CAException, IOException {

        // split the message into fragments (as many as needed)
        MessageFragment[] fragments =
            MessageSplitter.splitMessage(messageText, _maxMessageSize);

        // send each fragment
        for (int idx = 0; idx < fragments.length; idx++ ) {
            _pvChannel.put(fragments[idx].getBytes());
            _jcaContext.flushIO();
        }
    }

    /**
     * Adds the specified listener to this messenger.
     *
     * @param listener the ProviderListener instance to be added
     *        to the list of listeners that are notified
     *        when the value of the process variable changes
     * @throws CAException if a channel access error occurs while
     *         adding the listener to the process variable
     */
    public void addListener(ProviderListener listener) throws CAException {

        // sanity check
        if (listener == null) {return;}

        // ensure exclusive access to the listener list
        synchronized (_listenerList) {

            // check if the listener is already in the list
            if (! _listenerList.contains(listener)) {

                // when the first listener is added, we need to
                // attach a monitor to the PV (watching VALUE changes)
                if (_listenerList.size() == 0) {

                    // 31-Oct-2007, KAB - we now make use of the "spurious"
                    // callback that EPICS generates when a monitor is first
                    // attached to ensure that we don't miss any messages.
                    // So, we attach the monitor, wait for EPICS to tell us
                    // that it was attached, and only then return from this
                    // method (modulo a timeout).
                    _waitingForMonitor = true;
                    synchronized (this) {
                        _pvMonitor = _pvChannel.addMonitor(Monitor.VALUE, this);
                        _jcaContext.flushIO();

                        long startTime = System.currentTimeMillis();
                        long now = startTime;
                        long timeout = (long) (_PEND_OP_TIMEOUT * 1000);
                        while (_waitingForMonitor &&
                               (now - startTime) <= timeout) {
                            try {
                                wait(timeout);
                            }
                            catch (InterruptedException ie) {}
                            now = System.currentTimeMillis();
                        }
                        if (_waitingForMonitor) {
                            _waitingForMonitor = false;
                            String msg = "Timout waiting to attach a " +
                                "monitor to " + _pvName;
                            throw new CAException(msg);
                        }
                    }
                }

                // add the listener to the list
                _listenerList.addElement(listener);
            }
        }
    }

    /**
     * Removes the specified listener from this messenger.
     *
     * @param listener the ProviderListener instance to be removed
     *        from the list of listeners that are notified
     *        when the value of the process variable changes
     * @throws CAException if a channel access error occurs while
     *         removing the listener from the process variable
     */
    public void removeListener(ProviderListener listener) throws CAException {

        // sanity check
        if (listener == null) {return;}

        // ensure exclusive access to the listener list
        synchronized (_listenerList) {

            // remove the listener from the list
            _listenerList.removeElement(listener);

            // when the last listener is removed, we need to
            // detach the monitor from the PV
            if (_listenerList.size() == 0) {
                if (_pvMonitor != null) {

                    // Monitor.clear() seems to be the correct method
                    // to call here - Monitor.removeMonitorListener()
                    // isn't thorough enough
                    try {
                        _pvMonitor.clear();
                        _jcaContext.flushIO();
                    }
                    finally {
                        _pvMonitor = null;
                    }
                }
            }
        }
    }

    /**
     * Processes changes to the PV value.  This method is part of the
     * JCA MonitorListener interface and is called when the PV value changes.
     *
     * @param monEvent the JCA monitor event that encapsulates the change
     *        in the value of the process variable
     */
    public void monitorChanged(MonitorEvent monEvent) {
        String messageText = null;

        // if we're waiting for a monitor to be attached (and
        // this is the callback that EPICS generates when a monitor is
        // first added), indicate that we're no longer waiting for the
        // monitor to be attached, wake up any waiting threads, and return.
        if (_waitingForMonitor) {
            synchronized (this) {
                _waitingForMonitor = false;
                notifyAll();
                return;
            }
        }

        // sanity check
        if (monEvent == null) {return;}

        // if no listeners are attached, ignore updates
        if (_listenerList.size() == 0) {
            if (_debug) {
                String msg = "WARNING:  EpicsMessenger.monitorChanged " +
                    "method was called for PV " + _pvName +
                    " when no listeners were attached.";
                System.out.println(msg);
            }
            return;
        }

        // check the status of the operation as reported by EPICS
        CAStatus statusObj = monEvent.getStatus();
        if (! statusObj.isSuccessful()) {
            messageText = "Error receiving update from PV (" +
                this.getName() + "): " + statusObj.toString();
        }

        // fetch the updated PV contents
        DBR pvValue = monEvent.getDBR();
        byte[] pvBytes = ((BYTE) pvValue).getByteValue();

        // if everything has been successful so far,
        // convert the data received from EPICS into a string, taking
        // into account multiple fragments, etc.
        if (messageText == null) {
            try {
                messageText = this.convertText(pvBytes);
            }
            catch (IOException ioExcpt) {
                messageText = "Error receiving update from PV (" +
                this.getName() + "): " + ioExcpt.toString();
            }
        }

        // notify all listeners of the message 
        if (messageText != null) {
            this.notifyListeners(messageText);
        }
    }

    /**
     * Closes the messenger.  This takes care of any cleanup that
     * needs to be done after the messenger is through being used.
     * Instances of this class should <b>not</b> be used after this
     * method has been called on them.
     *
     * @throws CAException if a channel access error occurs while
     *         removing listeners from the process variable or shutting
     *         down the connection to the PV
     */
    public void close() throws CAException {

        // we don't want an exception early in the process to prevent
        // later cleanup, so we'll track the first exception, but
        // persevere until the end
        CAException firstProblem = null;

        // remove all listeners (in reverse order to avoid problems
        // with the list elements changing out from under us - the
        // removeListener() call actually removes the listener from the
        // listenerList)
        synchronized (_listenerList) {
            for (int idx = _listenerList.size() - 1; idx >= 0; idx--) {
                ProviderListener listener = _listenerList.elementAt(idx);
                try {
                    this.removeListener(listener);
                }
                catch (CAException caExcpt) {
                    if (firstProblem == null) {
                        firstProblem = caExcpt;
                    }
                }
            }
        }

        // close the channel
        if (_pvChannel != null) {
            try {
                _pvChannel.destroy();
            }
            catch (CAException caExcpt) {
                if (firstProblem == null) {
                    firstProblem = caExcpt;
                }
            }
            finally {
                _pvChannel = null;
            }
        }

        // rethrow the first exception, if any
        if (firstProblem != null) {
            throw firstProblem;
        }
    }

    /**
     * Handles cleanup when garbage collected.
     */
    protected void finalize() {
        try {
            this.close();
        }
        catch (CAException caExcpt) {
            caExcpt.printStackTrace(System.err);
        }
    }

    /**
     * Sets the debug output flag on (true) or off (false).
     *
     * @param value the boolean value to which the debug flag is set
     */
    public void setDebug(boolean value) {
        _debug = value;
    }

    /**
     * "main" method - used for unit testing.
     */
    public static void main(String[] args) {

        if (args.length < 1) {
            System.out.println("No PV specified.");
            System.exit(0);
        }
        String pvName = args[0];

        try {
            final java.text.SimpleDateFormat fmt =
                new java.text.SimpleDateFormat("HH:mm:ss.SSS");

            JCALibrary jcaLib = JCALibrary.getInstance();
            //jcaContext = jcaLib.createContext(JCALibrary.JNI_THREAD_SAFE);
            Context jcaContext =
                jcaLib.createContext(JCALibrary.JNI_SINGLE_THREADED);

            EpicsMessenger eMess =
                new EpicsMessenger(pvName, jcaContext, 1000);
            eMess.setDebug(true);

            ProviderListener listener = new ProviderListener() {
                    public void textReceived(String messageText) {
                        System.out.println(fmt.format(new java.util.Date()) +
                                           " Listener received message \"" +
                                           messageText + "\"");
                    }
                };
            
            System.out.println(fmt.format(new java.util.Date()) +
                               " Adding a listener to " + pvName + "...");
            eMess.addListener(listener);

            System.out.println(fmt.format(new java.util.Date()) +
                               " Sleeping before sending first message...");
            try {
                Thread.sleep(3000);
            }
            catch (Exception excpt) {}

            System.out.println(fmt.format(new java.util.Date()) +
                               " Sending first message...");
            eMess.sendMessage("Hello, world!");

            // sleep a bit to allow the listener to be notified
            System.out.println(fmt.format(new java.util.Date()) +
                               " Sleeping after sending first message...");
            try {
                Thread.sleep(3000);
            }
            catch (Exception excpt) {}

            System.out.println(fmt.format(new java.util.Date()) +
                               " Removing the listener from " + 
                               pvName + "...");
            eMess.removeListener(listener);

            System.out.println(fmt.format(new java.util.Date()) +
                               " Sleeping before sending second message...");
            try {
                Thread.sleep(3000);
            }
            catch (Exception excpt) {}

            System.out.println(fmt.format(new java.util.Date()) +
                               " Sending second message...");
            eMess.sendMessage("Goodbye, for now.");

            // sleep a bit to allow the listener to be notified
            System.out.println(fmt.format(new java.util.Date()) +
                               " Sleeping after sending second message...");
            try {
                Thread.sleep(3000);
            }
            catch (Exception excpt) {}

            System.out.println(fmt.format(new java.util.Date()) +
                               " Before closing the messenger...");
            eMess.close();

            System.out.println(fmt.format(new java.util.Date()) +
                               " Before destroying the context...");
            jcaContext.destroy();
            System.out.println(fmt.format(new java.util.Date()) +
                               " Before disposing of the context...");
            jcaContext.dispose();
            System.out.println(fmt.format(new java.util.Date()) +
                               " All done...");
        }
        catch (Exception excpt) {
            excpt.printStackTrace(System.err);
        }

        System.exit(0);
    }

    /**
     * Converts the specified byte array into a text message that
     * is ready to send to listeners, taking into account that
     * the byte array might contain only one fragment of many.
     * 
     * @return a completely reconstructed text message or null
     *         if the bytes are part of a not-yet-complete message
     * @throws IOException if an error occurs converting the bytes
     *         into a String
     */
    private String convertText(byte[] pvBytes) throws IOException {

        // deserialize the bytes into a MessageFragment
        MessageFragment fragment = new MessageFragment(pvBytes);

        // add the fragment to an existing assembler or create a new one
        boolean foundAssembler = false;
        synchronized (_assemblerList) {
            Enumeration<MessageAssembler> listEnum = _assemblerList.elements();
            while (listEnum.hasMoreElements()) {
                MessageAssembler assembler = listEnum.nextElement();
                if (assembler.addFragment(fragment)) {
                    foundAssembler = true;
                    break;
                }
            }
            if (! foundAssembler) {
                MessageAssembler assembler = new MessageAssembler(fragment);

                // if the message is already complete, return it now
                if (assembler.hasCompleteMessage()) {
                    return assembler.getMessage();
                }

                // otherwise, add this assembler to the active list
                else {
                    _assemblerList.addElement(assembler);
                }
            }
        }

        // check if any assemblers have stale data - remove any that do
        synchronized (_assemblerList) {
            ListIterator<MessageAssembler> listIter =
                _assemblerList.listIterator();
            while (listIter.hasNext()) {
                MessageAssembler assembler = listIter.next();
                if (assembler.hasStaleData()) {

                    // remove the assembler from the active list
                    listIter.remove();
                }
            }
        }

        // check if any assemblers have complete messages
        synchronized (_assemblerList) {
            ListIterator<MessageAssembler> listIter =
                _assemblerList.listIterator();
            while (listIter.hasNext()) {
                MessageAssembler assembler = listIter.next();
                if (assembler.hasCompleteMessage()) {

                    // fetch the reconstructed message
                    String messageText = assembler.getMessage();

                    // remove the assembler from the active list
                    listIter.remove();

                    // return the completed message
                    return messageText;
                }
            }
        }

        // return null to indicate that we don't yet have any fully
        // reconstructed messages
        return null;
    }

    /**
     * Notifies all listeners of the specified message.
     *
     * @param messageText the string text that contains the new value
     *        of the process variable
     */
    private void notifyListeners(String messageText) {
        synchronized (_listenerList) {
            for (int idx = 0; idx < _listenerList.size(); idx++) {
                ProviderListener listener = _listenerList.elementAt(idx);
                listener.textReceived(messageText);
            }
        }
    }
}
