package gov.fnal.cd.rms.provider;

import gov.fnal.cd.rms.base.RmsCommunicationException;
import gov.fnal.cd.rms.base.RmsDestination;
import gov.fnal.cd.rms.base.RmsMessage;
import gov.fnal.cd.rms.util.RmsException;
import gov.fnal.cd.rms.util.UUIDGenerator;
import java.util.LinkedList;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.Vector;

/**
 * A simple provider useful for testing the RMS interface
 * routines.  This provider will route messages between
 * consumers and producers that are connected to it.  Message
 * processing is done in a seperate thread as to simulate the
 * decoupled nature of sending and receiving messages.
 *
 * @author Kurt Biery
 * @author Steve Foulkes
 * @version $Revision: 1.6 $ $Date: 2010/03/15 20:38:46 $
 */
public class LocalhostConnection implements RmsConnection {
    /**
     * The queue that messages are placed onto before they are
     * sent out to listeners.
     */
    private LinkedBlockingQueue <MessageContainer> _inputQueue;

    /**
     * The list of destinations and listeners that have been
     * registered with this provider.
     */
    private LinkedList <ListenerContainer> _destinationList;

    /**
     * Handle to the thread that distributes and processes messages.
     */
    private Thread _messageProcessingThread;

    /**
     * UUID Generator to generate UUIDs for messages.
     */
    private UUIDGenerator _uuidGenerator;

    /**
     * Destination to represent where messages originate from.
     */
    private RmsDestination _messageSource;

    /**
     * Container for a destination and a serialized message. The
     * message text and destination are placed into this object
     * before being put into the input queue for processing.
     */
    private class MessageContainer {
        private RmsDestination _dest;
        private String _messageText;

        /**
         * Create a new MessageContainer with a destination object and a 
         * serialized message.
         *
         * @param dest The destination object
         * @param messageText The serialized message
         *
         */
        public MessageContainer(RmsDestination dest, String messageText) {
            _dest = dest;
            _messageText = messageText;
            
            return;
        }

        /**
         * Retrieve the destination object that resides within the 
         * MessageContainer.
         *
         * @return The destination object
         */
        public RmsDestination getDestination() {
            return _dest;
        }

        /**
         * Retrieve the serialized messages that resides within the
         * MessageContainer.
         *
         * @return The serialized message
         */
        public String getMessageText() {
            return _messageText;
        }
    }
    
    /**
     * Container for a destination and a provider listener. The
     * destination and the provider listener are placed into this
     * object before being put into the destinationList for processing.
     */
    private class ListenerContainer {
        private RmsDestination _dest;
        private ProviderListener _listener;

        /**
         * Create a new ListenerContainer with a destination object
         * and a ProviderListener.
         *
         * @param dest The destination object
         * @param listener The ProviderListener object
         *
         */
        public ListenerContainer(RmsDestination dest, 
                                 ProviderListener listener) {
            _dest = dest;
            _listener = listener;

            return;
        }

        /**
         * Retrieve the destination object that resides within the
         * ListenerContainer.
         *
         * @return The destination object
         */
        public RmsDestination getDestination() {
            return _dest;
        }

        /**
         * Retrieve the listener object that resides within the 
         * ListenerContainer.
         *
         * @return The listener object
         */
        public ProviderListener getListener() {
            return _listener;
        }
    }
    
    /**
     * Initialize the Localhost provider given the connectLocation,
     * reply target and a provider.  Here, we initialize the inputQueue,
     * destinationList and startup our message processing thread.
     *
     * @param applicationName The name of the application generating/receiving
     * messages.
     * @param applicationPartition The partition of the application that is
     * generating/receiving messages.  
     *
     */
    public LocalhostConnection (String applicationName, int applicationPartition) {
        try {
            _uuidGenerator = new UUIDGenerator(java.net.InetAddress.getLocalHost().getHostName());
        }
        catch (java.net.UnknownHostException uhExcpt) {
            _uuidGenerator = new UUIDGenerator("JavaLocalhostConnection");
        }

        _messageProcessingThread = new Thread(new MessageLoop());

        _inputQueue = new LinkedBlockingQueue<MessageContainer>();
        _destinationList = new LinkedList<ListenerContainer>();
        _messageProcessingThread.start();

        _messageSource = new RmsDestination();
        _messageSource.setProperty(RmsDestination.TARGET_PROPERTY_NAME, 
                                   applicationName);
        _messageSource.setPartitionProperty(applicationPartition);
    }

    /**
     * Send a string to a particular destination.
     *
     * @param dest The destination of the message
     * @param messageString The string that we are sending.
     * @throws RmsCommunicationException if an error occurs when trying
     *         to send the message to the destination
     * @throws novadaq.DataException if an error occurs when trying
     *         to prepare the message for sending
     */
    public void sendString(RmsDestination dest, String messageString)
        throws RmsCommunicationException, novadaq.DataException {

        MessageContainer newMessage = new MessageContainer(dest, messageString);
        _inputQueue.add(newMessage);

        return;
    }

    /**
     * Send an RmsMessage to a particular destination.
     *
     * @param dest The destination of the message
     * @param message The RmsMessage that we are sending.
     * @throws RmsCommunicationException if an error occurs when trying
     *         to send the message to the destination
     * @throws novadaq.DataException if an error occurs when trying
     *         to prepare the message for sending
     */
    public void sendMessage(RmsDestination dest, RmsMessage message)
        throws RmsCommunicationException, novadaq.DataException {

        sendString(dest, message.serialize());

        return;
    }

    /**
     * Sends a message of the specified type to the specified
     * destination.
     *
     * @param messageDataType the java.lang.Class of the message to be sent
     * @param dest the destination that the message should be sent to
     * @param theMessage the message to be sent
     * @throws RmsCommunicationException if an error occurs when trying
     *         to send the message to the destination
     */
    public <T> void sendMessage(Class<T> messageDataType, RmsDestination dest,
                                T theMessage)
        throws RmsCommunicationException {

        String msg = "This version of the sendMessage() method is not " +
            "currently supported in the LocalhostConnection class";
        throw new RmsCommunicationException(msg, null);
    }

    /**
     * This operation is not supported by a LocalhostConnection.  It is only
     * provided to fulfill the RmsConnection interface.
     *
     * @throws RmsCommunicationException always
     */
    public <T> Vector<T> receiveMessage(Class<T> messageDataType, RmsDestination dest)
        throws RmsCommunicationException {

        String msg = "The receiveMessage() method is not currently " +
            "supported in the LocalhostConnection class";
        throw new RmsCommunicationException(msg, null);
    }

    /**
     * This operation is not supported by a LocalhostConnection.  It is only
     * provided to fulfill the RmsConnection interface.
     *
     * @throws RmsCommunicationException always
     */
    public <T> Vector<T> receiveMessage(Class<T> messageDataType, RmsDestination dest,
                                long timeout)
        throws RmsCommunicationException {

        String msg = "The receiveMessage() method is not currently " +
            "supported in the LocalhostConnection class";
        throw new RmsCommunicationException(msg, null);
    }

    /**
     * Add a listener to our listener list so that we can route messages
     * to it.
     *
     * @param dest The destination we're interested in
     * @param listener The listener that matching messages will be sent to.
     * @throws RmsCommunicationException if a problem is encountered setting
     *         up the system to monitor the specified destination
     */
    public void addListener(RmsDestination dest, ProviderListener listener)
        throws RmsCommunicationException {

        ListenerContainer newListener = new ListenerContainer(dest, listener);
        _destinationList.add(newListener);

        return;
    }

    /**
     * Remove a listener from the list.
     *
     * @param dest The destination to be removed
     * @param listener The listener that will no longer be called.
     * @throws RmsCommunicationException if an error occurs when trying
     *         to remove the listener from the destination
     */
    public void removeListener(RmsDestination dest, ProviderListener listener)
        throws RmsCommunicationException {

        for (int i = 0; i < _destinationList.size(); i++) {
            if (_destinationList.get(i).getListener() == listener &&
                _destinationList.get(i).getDestination().equals(dest)) {
                _destinationList.remove(i);
                break;
            }
        }

        return;
    }

    /**
     * Closes the connection.  After calling this method, the object
     * should no longer be used since it will no
     * longer be connected to the underlying provider system.
     *
     * @throws RmsCommunicationException if a problem occurs when trying
     *         to disconnect from the provider
     */
    public void close() throws RmsCommunicationException {
        _messageProcessingThread.interrupt();
        return;
    }

    /**
     * Since this provider supports all destinations, and they're 
     * always reachable, we'll just return true all the time.
     *
     * @param pingDestination The destination in the message passing
     * system to ping.
     *
     * @return 
     */
    public boolean ping(RmsDestination pingDestination) {
        return true;
    }

    /**
     * Get a reference to the message source.
     *
     * @return A destination object that contains the source of messages
     * for this connection.
     */    
    public RmsDestination getMessageSource() {
        return new RmsDestination(_messageSource);
    }

    /**
     * Returns a universally unique identifier.  This can be used to uniquely
     * identify a message.
     *
     * @return the unique ID string
     * @throws novadaq.DataException if an error occurs while creating the UUID
     */
    public String getUUID() throws novadaq.DataException {
        try {
            return _uuidGenerator.getUUID().toString();
        }
        catch (RmsException rmsExcpt) {
            String msg = "Unable to create unique identifier.";
            throw new novadaq.DataException(msg, rmsExcpt);
        }
    }

    /**
     * Returns a universally unique identifier.  This can be used to uniquely
     * identify a message.  This version of the method will not throw a checked
     * exception.  Instead it will make a best-effort attempt to create a
     * unique ID.
     *
     * @return The unique ID string.
     */
    public String getUUIDNoThrow() {
        try {
            return _uuidGenerator.getUUID().toString();
        }
        catch (RmsException rmsExcpt) {
            return _messageSource.getProperty(RmsDestination.TARGET_PROPERTY_NAME) +
                System.currentTimeMillis();
        }
    }

    /**
     * Returns the default message expiration interval, in milliseconds
     * (also known as the time-to-live).
     *
     * @return the default time-to-live for messages
     */
    public long getDefaultTimeToLive() {
        long value = 2000;

        return value;
    }

    /**
     * Tests whether the specified destination is supported by this
     * connection type.
     *
     * @param candidateDestination The destination that should be tested.
     * @return True if the destination is supported, false otherwise.
     */
    public boolean supportsDestination(RmsDestination candidateDestination) {
        return true;
    }

    /**
     * Handles cleanup when garbage collected.
     */
    protected void finalize() {
        try {
            this.close();
        }
        catch (RmsCommunicationException rcExcpt) {
            rcExcpt.printStackTrace(System.err);
        }
    }

    /**
     * The message processing loop for the provider, which runs
     * in its own thread.  Messages are pulled off the input queue
     * and then compared to all the destinations in the destinationList.
     * If the destinations match, the message is passed to the listener.
     *
     * @author Kurt Biery
     * @author Steve Foulkes
     * @version $Version$ $Date: 2010/03/15 20:38:46 $
     */
    private class MessageLoop implements Runnable {

        /**
         * Main method in the MessageLoop class that does all the
         * message processing.  
         *
         * @return NONE
         */
        public void run() {
            MessageContainer incomingMessage;
            ListenerContainer aListener;

            while (!Thread.currentThread().isInterrupted()) {
                try {
                    incomingMessage = (MessageContainer)_inputQueue.take();
                }
                catch (InterruptedException excp) {
                    return;
                }

                for (int i = 0; i < _destinationList.size(); i++) {
                    aListener = (ListenerContainer)_destinationList.get(i);
                    
                    if (aListener.getDestination().toString().equals(incomingMessage.getDestination().toString())) {
                        aListener.getListener().textReceived(incomingMessage.getMessageText());
                    }
                }
            }
        }
    }
}
