package gov.fnal.cd.rms.provider;

import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.UUID;

/**
 * Helper class for assembling byte-based message fragments into a single
 * string-based message.
 *
 * @author Kurt Biery
 * @author Steve Foulkes
 * @version $Revision: 1.1.1.1 $ $Date: 2009/10/09 20:03:00 $
 */
public class MessageAssembler {
    /**
     * The time at which we receive the latest non-null fragment.
     */
    private long _lastFragmentTime = 0;

    /**
     * The unique identifier of the message that we are assembling.
     */
    private UUID _messageId = null;

    /**
     * The array of fragments that we are assembling.
     */
    private MessageFragment[] _fragments = null;

    /**
     * The number of fragments that are needed to complete the full message.
     */
    private int _requiredFragmentCount;

    /**
     * The number of fragments that have been received so far.
     */
    private int _assembledFragmentCount;

    /**
     * The amount of time, in milliseconds, after which we consider
     * an assembler instance to be stale.  A reasonable default is
     * specified here, but the environment is also checked for an
     * RMS_ASSEMBLER_TIMEOUT variable to be used instead.
     */
    private static long _freshnessTimeout = 900000;  // 15 minutes
    static {
        // check for a user-specified timeout
        String stringValue = System.getenv("RMS_ASSEMBLER_TIMEOUT");
        if (stringValue != null && stringValue.trim().length() > 0) {
            try {
                _freshnessTimeout = Long.parseLong(stringValue, 10);
            }
            catch (NumberFormatException nfe) {
                // simply leave the default value in use
            }
        }
    }

    /**
     * Creates a new assembler instance with the specified initial fragment.
     *
     * @param initialFragment A message fragment to place into the message
     * assembler.
     */
    public MessageAssembler(MessageFragment initialFragment) {
        if (initialFragment != null) {
            this.processFirstFragment(initialFragment);
        }
    }

    /**
     * Tests whether the specified fragment is needed by this assembler,
     * saves it internally if so, and returns the result of the test.
     *
     * @param fragment The fragment to add to the assembler. Null input
     * will be gracefully handled, and a false result will be returned
     * in that case.
     *
     * @return True if the fragment was successfully merged in,
     * false otherwise.
     */
    public boolean addFragment(MessageFragment fragment) {

        // input fragment can't be null
        if (fragment == null) {return false;}

        // if this is the first non-null fragment that we have received,
        // process it accordingly
        if (_fragments == null) {
            this.processFirstFragment(fragment);
            return true;
        }

        // second, third, etc. input fragments must have non-null UUID
        UUID inputId = fragment.getMessageId();
        if (inputId == null) {
            return false;
        }

        // input fragment must have same number of total fragments as
        // the message we are building
        if (fragment.getFragmentCount() != _requiredFragmentCount) {
            return false;
        }

        // input fragment index must be reasonable
        int inputIndex = fragment.getFragmentIndex();
        if (inputIndex < 0 || inputIndex >= _requiredFragmentCount) {
            return false;
        }

        // our internal array must still need a fragment at the input index
        if (_fragments[inputIndex] != null) {
            return false;
        }

        // input fragment UUID must match ours
        if (_messageId == null) {
            return false;
        }
        if (! inputId.equals(_messageId)) {
            return false;
        }

        // store the fragment if everything has checked out
        _fragments[inputIndex] = fragment;
        _assembledFragmentCount++;
        _lastFragmentTime = System.currentTimeMillis();
        return true;
    }

    /**
     * Tests whether a complete message has been assembled.
     *
     * @return True if the message has been completely assembled,
     * false otherwise.
     */
    public boolean hasCompleteMessage() {
        return (_fragments != null &&
                _assembledFragmentCount > 0 &&
                _assembledFragmentCount == _requiredFragmentCount);
    }

    /**
     * Tests whether the assembler has stale data.  "Stale" in this
     * case means that the no new fragments have been added to the
     * assembler within some (presumably lengthy) time interval.  This
     * interval is set to several minutes by default, but its value can
     * be specified externally by setting the RMS_ASSEMBLER_TIMEOUT
     * environmental variable.
     *
     * @return True if the assembler has stale data, false if the
     * data is fresh.
     */
    public boolean hasStaleData() {
        return ((System.currentTimeMillis() - _lastFragmentTime) >
                _freshnessTimeout);
    }

    /**
     * Returns the string message that has been assembled or null
     * if the message is still incomplete.
     *
     * @return The message if all the fragments have been
     * received, null otherwise.
     * @throws IOException If there is a problem converting the fragments
     * into a string.
     */
    public String getMessage() throws IOException {

        // not much we can do without a full message
        if (! this.hasCompleteMessage()) {
            return null;
        }
        String resultMessage = null;

        // use try and finally blocks to ensure that streams get closed.
        // Also, track the first problem encountered for correct reporting.
        IOException firstProblem = null;
        ByteArrayOutputStream byteStream = null;
        DataOutputStream dataStream = null;
        try {

            // determine a reasonable estimate of the byte count
            int estimatedSize = _fragments[0].getPayloadSize() *
                _requiredFragmentCount;

            // create byte array output stream to hold the generated bytes
            byteStream = new ByteArrayOutputStream(estimatedSize);

            // create a data output stream on top of the byte stream
            dataStream = new DataOutputStream(byteStream);

            // write each fragment to the data stream
            for (int idx = 0; idx < _fragments.length; idx++) {
                _fragments[idx].writePayload(dataStream);
            }

            // convert the byte stream to a string
            resultMessage = byteStream.toString();
        }
        catch (IOException ioExcpt) {
            if (firstProblem == null) {
                firstProblem = ioExcpt;
            }
        }
        finally {
            // close streams, if needed
            if (dataStream != null) {
                try {
                    dataStream.close();
                }
                catch (IOException ioExcpt) {
                    if (firstProblem == null) {
                        firstProblem = ioExcpt;
                    }
                }
                dataStream = null;
            }
            if (byteStream != null) {
                try {
                    byteStream.close();
                }
                catch (IOException ioExcpt) {
                    if (firstProblem == null) {
                        firstProblem = ioExcpt;
                    }
                }
                byteStream = null;
            }
        }
        if (firstProblem != null) {
            throw firstProblem;
        }

        // return the result
        return resultMessage;
    }

    /**
     * Processes the first fragment and initializes internal
     * data accordingly.
     *
     * @param fragment The first message fragment to be added to
     * this class.
     * @throws IndexOutOfBoundsException If there is a problem adding the
     * fragment to the internal list of fragments.
     * @throws NullPointerException If a null fragment is specified as input.
     */
    private void processFirstFragment(MessageFragment fragment) {

        // In this method, we have purposely omitted the handling of certain
        // runtime exceptions.  For example, this method should not be called
        // with a null fragment, but if it is, then the user should have to
        // deal with the resulting NullPointerException.  Also, if the
        // total fragment count or the fragment index are corrupt, throwing an
        // ArrayIndexOutOfBoundsException is probably the right thing to do.

        // fetch the total number of fragments
        _requiredFragmentCount = fragment.getFragmentCount();

        // create the fragment array
        _fragments = new MessageFragment[_requiredFragmentCount];

        // store the first fragment
        int index = fragment.getFragmentIndex();
        _fragments[index] = fragment;
        _messageId = fragment.getMessageId();
        _assembledFragmentCount = 1;
        _lastFragmentTime = System.currentTimeMillis();
    }
}
