package gov.fnal.cd.rms.provider;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.util.UUID;

/**
 * Message fragment class.  This class is used to facilitate the
 * transmission of string messages that need to be broken
 * into multiple parts because of some size constraint in the system.
 * It does not enforce any size constraints itself, but instead relies
 * on users to only create payloads of appropriate sizes.
 *
 * @author Kurt Biery
 * @author Steve Foulkes
 * @version $Revision: 1.1.1.1 $ $Date: 2009/10/09 20:03:00 $
 */
public class MessageFragment {
    // 12-Jun-2007, KAB - switch to header version style.
    // Two header versions are currently supported:
    //
    // Version 0:
    // - fragment prefix = "RMS" -> 3 bytes
    // - header version -> 1 byte
    // - fragment index -> 2 bytes (zero-based)
    // - fragment count -> 2 bytes (one-based)
    // - payload size -> 4 bytes
    // - message UUID -> 16 bytes
    //
    // Version 1:
    // - fragment prefix = "RMS" -> 3 bytes
    // - header version -> 1 byte
    // - fragment index -> 2 bytes (zero-based)
    // - fragment count -> 2 bytes (one-based)
    // - payload size -> 4 bytes

    /**
     * Special prefix string that is used for all fragments to help
     * tag byte streams as coming from this class (or decodable by
     * this class).
     */
    private static final String _FRAGMENT_PREFIX = "RMS";

    /**
     * Byte array corresponding to the special prefix string.
     */
    private static final byte[] _PREFIX_BYTES = _FRAGMENT_PREFIX.getBytes();

    /**
     * Header version used for the fragment.
     */
    private byte _version;

    /**
     * Size of the header (in bytes) for the fragment.
     */
    private short _headerSize;

    /**
     * Index of the fragment within the full message (zero-based).
     */
    private int _fragmentIndex;

    /**
     * Number of fragments in the full message.
     */
    private int _numberOfFragments;

    /**
     * Universally unique identifier for the message that this fragment
     * is part of.
     */
    private UUID _uniqueId;

    /**
     * Data byte array for this fragment.
     */
    private byte[] _dataBytes;

    /**
     * Offset into the data array for this fragment.
     */
    private int _dataOffset;

    /**
     * Size of the data (number of bytes) in this fragment.
     */
    private int _dataSize;

    /**
     * Serialized byte array.
     */
    private byte[] _serializedData = null;

    /**
     * Creates a message fragment with the specified UUID, index, and
     * total fragment count.  The initial payload is empty.
     *
     * @param guid The UUID of the fragment
     * @param index The index of the fragment
     * @param totalCount The total number of fragments
     */
    public MessageFragment(UUID guid, int index, int totalCount) {
        _uniqueId = guid;
        _fragmentIndex = index;
        _numberOfFragments = totalCount;
        if (_numberOfFragments > 1) {
            _version = 0;
            _headerSize = getMaxHeaderSize();
        }
        else {
            _version = 1;
            _headerSize = (short) (getMaxHeaderSize() - 16);
        }
        clearPayload();
    }

    /**
     * Creates a message fragment from the specified byte array.  This
     * constructor automatically deserializes the bytes into the
     * appropriate fragment fields or throws an exception if there is
     * a problem doing that.
     *
     * @param fragmentBytes A serialized message fragment.
     * @throws IOException If there is a problem parsing the data in the
     * byte array.
     */
    public MessageFragment(byte[] fragmentBytes) throws IOException {

        // initialization
        _serializedData = fragmentBytes;
        clearPayload();

        // we'll use try/catch/finally blocks below to ensure that the
        // data streams that we create get closed correctly, and we'll
        // keep track of the first exception encountered (if any) so that 
        // we can correctly report the original error when things fail.
        IOException firstProblem = null;
        ByteArrayInputStream byteStream = null;
        DataInputStream dataStream = null;
        try {

            // create a byte array input stream to provide access to the bytes
            byteStream = new ByteArrayInputStream(fragmentBytes);

            // create a data input stream on top of the byte stream for
            // easier data conversion
            dataStream = new DataInputStream(byteStream);

            // prefix
            int prefixLen = _PREFIX_BYTES.length;
            byte[] prefixBytes = new byte[prefixLen];
            dataStream.read(prefixBytes, 0, prefixLen);
            for (int idx = 0; idx < prefixLen; idx++) {
                if (prefixBytes[idx] != _PREFIX_BYTES[idx]) {
                    StringBuffer buff = new StringBuffer();
                    for (int jdx = 0; jdx < prefixLen; jdx++) {
                        buff.append(Byte.toString(prefixBytes[idx]));
                    }
                    throw new IOException("Fragment prefix mismatch, " +
                                          "received \"\\" +
                                          buff.toString() +
                                          "\", expected \"" +
                                          _FRAGMENT_PREFIX + "\"");
                }
            }

            // version
            _version = dataStream.readByte();

            // header size
            _headerSize = getMaxHeaderSize();
            if (_version == 1) {
                _headerSize -= 16;
            }

            // fragment index and count
            _fragmentIndex = dataStream.readShort();
            _numberOfFragments = dataStream.readShort();

            // payload size
            _dataSize = dataStream.readInt();

            // UUID
            if (_version == 0) {
                long hiBits = dataStream.readLong();
                long loBits = dataStream.readLong();
                _uniqueId = new UUID(hiBits, loBits);
            }

            // set the data block parameters
            if (_dataSize > 0) {
                _dataOffset = _headerSize;
            }
            if (_dataOffset > 0) {
                _dataBytes = _serializedData;
            }
        }
        catch (IOException ioExcpt) {
            if (firstProblem == null) {
                firstProblem = ioExcpt;
            }
        }
        finally {
            // close streams, if needed
            if (dataStream != null) {
                try {
                    dataStream.close();
                }
                catch (IOException ioExcpt) {
                    if (firstProblem == null) {
                        firstProblem = ioExcpt;
                    }
                }
                dataStream = null;
            }
            if (byteStream != null) {
                try {
                    byteStream.close();
                }
                catch (IOException ioExcpt) {
                    if (firstProblem == null) {
                        firstProblem = ioExcpt;
                    }
                }
                byteStream = null;
            }
        }

        // re-throw the original exception, if any
        if (firstProblem != null) {
            throw firstProblem;
        }
    }

    /**
     * Returns the index of this fragment (zero-based).
     *
     * @return The index of this message fragment.
     */
    public int getFragmentIndex() {
        return _fragmentIndex;
    }

    /**
     * Returns the number of fragments in the full message (one-based).
     *
     * @return The total number of fragments that make up the
     * message.
     */
    public int getFragmentCount() {
        return _numberOfFragments;
    }

    /**
     * Returns the unique identifier for the full message that this fragment
     * is part of (may be null if none was specified).
     *
     * @return The UUID of the message.
     */
    public UUID getMessageId() {
        return _uniqueId;
    }

    /**
     * Returns the size of the payload data.
     *
     * @return The total size of the message payload.
     */
    public int getPayloadSize() {
        return _dataSize;
    }

    /**
     * Writes the payload in this fragment to the specified data stream.
     *
     * @param dataStream The data stream that the fragment payload will
     * be written to.
     * @throws IOException If there is a problem writing to the data stream
     */
    public void writePayload(DataOutputStream dataStream) throws IOException {

        // nothing to do if there is no payload
        if (_dataBytes == null) {
            return;
        }

        // write the payload to the stream
        dataStream.write(_dataBytes, _dataOffset, _dataSize);
    }

    /**
     * Specifies the data to be included in this fragment.  If the payload
     * string is null, the fragment data is cleared.
     *
     * @param payloadString The payload that will be associated with
     * this fragment.
     */
    public void setPayload(String payloadString) {

        // pass off the real work
        if (payloadString != null) {
            byte[] payloadBytes = payloadString.getBytes();
            this.setPayload(payloadBytes, 0, payloadBytes.length);
        }
        else {
            this.clearPayload();
        }
    }

    /**
     * Specifies the data to be included in this fragment.  If the payload
     * byte array is null, the fragment data is cleared.
     *
     * @param payloadBytes The payload that will be associated with
     * this fragment.
     */
    public void setPayload(byte[] payloadBytes) {

        // pass off the real work
        if (payloadBytes != null) {
            this.setPayload(payloadBytes, 0, payloadBytes.length);
        }
        else {
            this.clearPayload();
        }
    }

    /**
     * Specifies the data to be included in this fragment.  If the payload
     * byte array is null, the fragment data is cleared and the offset
     * and size are ignored.
     *
     * @param payloadBytes The payload that will be associated with
     * this message fragment.
     * @param offset The offset to place the payload at.
     * @param size The size of the payload.
     */
    public void setPayload(byte[] payloadBytes, int offset, int size) {

        // in an attempt to minimize the number of data copies,
        // we simply store the input byte array, the offset, and the size
        if (payloadBytes != null) {
            _dataBytes = payloadBytes;
            _dataOffset = offset;
            _dataSize = size;
        }
        else {
            this.clearPayload();
        }
        _serializedData = null;
    }

    /**
     * Serializes the fragment, including all necessary header information,
     * into an array of bytes.
     *
     * @return A serialized message fragment.
     * @throws IOException If there is a problem converting the fragments
     * into a byte array.
     */
    public byte[] getBytes() throws IOException {

        // serialize the data, if needed
        if (_serializedData == null) {

            // use try and finally blocks to ensure that streams get closed.
            // Also, track the first problem encountered for correct reporting.
            IOException firstProblem = null;
            ByteArrayOutputStream byteStream = null;
            DataOutputStream dataStream = null;
            try {

                // determine a reasonable estimate of the byte count
                int estimatedSize = _dataSize + getMaxHeaderSize();

                // create byte array output stream to hold the generated bytes
                byteStream = new ByteArrayOutputStream(estimatedSize);

                // create a data output stream on top of the byte stream for
                // easier data conversion
                dataStream = new DataOutputStream(byteStream);

                // prefix
                dataStream.write(_PREFIX_BYTES, 0, _PREFIX_BYTES.length);

                // version
                dataStream.writeByte(_version);

                // fragment index and count
                dataStream.writeShort(_fragmentIndex);
                dataStream.writeShort(_numberOfFragments);

                // payload size
                if (_dataBytes == null) {
                    dataStream.writeInt(0);
                }
                else {
                    dataStream.writeInt(_dataSize);
                }

                // UUID, if needed
                if (_numberOfFragments > 1) {
                    dataStream.writeLong(_uniqueId.getMostSignificantBits());
                    dataStream.writeLong(_uniqueId.getLeastSignificantBits());
                }

                // payload, if available
                if (_dataBytes != null) {
                    dataStream.write(_dataBytes, _dataOffset, _dataSize);
                }

                // flush data stream and return serialized bytes
                dataStream.flush();
                _serializedData = byteStream.toByteArray();
            }
            catch (IOException ioExcpt) {
                if (firstProblem == null) {
                    firstProblem = ioExcpt;
                }
            }
            finally {
                // close streams, if needed
                if (dataStream != null) {
                    try {
                        dataStream.close();
                    }
                    catch (IOException ioExcpt) {
                        if (firstProblem == null) {
                            firstProblem = ioExcpt;
                        }
                    }
                    dataStream = null;
                }
                if (byteStream != null) {
                    try {
                        byteStream.close();
                    }
                    catch (IOException ioExcpt) {
                        if (firstProblem == null) {
                            firstProblem = ioExcpt;
                        }
                    }
                    byteStream = null;
                }
            }
            if (firstProblem != null) {
                throw firstProblem;
            }
        }

        return _serializedData;
    }

    /**
     * Tests if the specified object is equal to this fragment.
     *
     * @param testObject The object that we are comparing this one to.
     * @return True if the objects are equal, false otherwise.
     */
    public boolean equals(Object testObject) {
        if (testObject == null ||
            ! (testObject instanceof MessageFragment)) return false;
        MessageFragment testFragment  = (MessageFragment) testObject;

        // test the UUIDs
        UUID thisUUID = this.getMessageId();
        UUID thatUUID = testFragment.getMessageId();
        if ((thisUUID == null && thatUUID != null) ||
            (thisUUID != null && thatUUID == null)) {
            return false;
        }
        if (thisUUID != null && thatUUID != null &&
            ! thisUUID.equals(thatUUID)) {
            return false;
        }

        // test the indices and counts
        if (this.getFragmentIndex() != testFragment.getFragmentIndex()) {
            return false;
        }
        if (this.getFragmentCount() != testFragment.getFragmentCount()) {
            return false;
        }

        // everything must have matched!
        return true;
    }

    /**
     * Returns a hash code for this fragment.
     * @return A hash code representing the fragment.
     */
    public int hashCode() {
        StringBuffer buff = new StringBuffer();

        UUID thisUUID = this.getMessageId();
        if (thisUUID == null) {
            buff.append("null");
        }
        else {
            buff.append(thisUUID.toString());
        }

        buff.append(this.getFragmentIndex());
        buff.append(this.getFragmentCount());

        return buff.toString().hashCode();
    }

    /**
     * Prints a diagnostic dump of the specified byte array
     * to the specified PrintStream, interpreting the bytes as the contents
     * of a message fragment.
     *
     * @param outStream The stream data will be printed to.
     * @param fragmentBytes The data that will be printed.
     * @param payloadCutoff The maximum number of payload bytes to be
     * printed.  If zero or negative, all payload bytes will be printed.
     */
    public static void dumpBytes(PrintStream outStream,
                                 byte[] fragmentBytes,
                                 int payloadCutoff) {
        int loIdx, hiIdx;
        int version = 0;

        // print full size
        outStream.println("Total number of bytes = " + fragmentBytes.length);

        // print prefix, version, index, and count
        loIdx = Math.min(0, fragmentBytes.length-1);
        hiIdx = Math.min(_PREFIX_BYTES.length+4, fragmentBytes.length-1);
        outStream.print("Prefix, version, index, count:");
        for (int idx = loIdx; idx <= hiIdx; idx++) {
            outStream.printf(" %02x", fragmentBytes[idx]);
            if (idx == _PREFIX_BYTES.length) {version = fragmentBytes[idx];}
        }
        outStream.println("");

        // print data size
        loIdx = hiIdx + 1;
        hiIdx = Math.min(hiIdx+4, fragmentBytes.length-1);
        outStream.print("Payload size:");
        for (int idx = loIdx; idx <= hiIdx; idx++) {
            outStream.printf(" %02x", fragmentBytes[idx]);
        }
        outStream.println("");

        // print uuid, if needed
        if (version == 0) {
            loIdx = hiIdx + 1;
            hiIdx = Math.min(hiIdx+16, fragmentBytes.length-1);
            outStream.print("UUID:");
            for (int idx = loIdx; idx <= hiIdx; idx++) {
                outStream.printf(" %02x", fragmentBytes[idx]);
            }
            outStream.println("");
        }

        // print the data
        byte[] dummyByte = new byte[1];
        loIdx = hiIdx + 1;
        if (payloadCutoff <= 0) {
            hiIdx = fragmentBytes.length-1;
        }
        else {
            hiIdx = Math.min(loIdx+payloadCutoff-1,fragmentBytes.length-1);
        }
        outStream.print("Payload:");
        for (int idx = loIdx; idx <= hiIdx; idx++) {
            if ((idx > loIdx) && ((idx - loIdx) % 8) == 0) {
                System.out.print("\n        ");
            }
            outStream.printf(" %02x", fragmentBytes[idx]);
            dummyByte[0] = fragmentBytes[idx];
            outStream.print(" " +
                            (dummyByte[0] == 0 ? " " : new String(dummyByte))
                            + " ");
        }
        outStream.println("");
    }

    /**
     * Returns the maximum size (in bytes) of the header data that
     * is added to the payload when serializing the fragment.
     *
     * @return The maximum header size.
     */
    public static short getMaxHeaderSize() {
        return (short)
            (_PREFIX_BYTES.length + 1 /* version size */ + 2 /* index */ +
             2 /* fragment count */ + 4 /* payload size */ + 16 /* uuid */);
    }

    /**
     * Main method for unit testing.
     */
    public static void main(String[] args) {

        // test serialization of an empty fragment
        /*
        try {
            UUID msgGuid = UUID.randomUUID();
            MessageFragment frag = new MessageFragment(msgGuid, 1, 1);
            byte[] fragmentBytes = frag.getBytes();
            int loIdx, hiIdx;

            // print prefix and version number (admittedly v1 specific)
            loIdx = Math.min( 0, fragmentBytes.length-1);
            hiIdx = Math.min( 3, fragmentBytes.length-1);
            System.out.print("Prefix, Version:");
            for (int idx = loIdx; idx <= hiIdx; idx++) {
                System.out.printf(" %02x", fragmentBytes[idx]);
            }
            System.out.println("");

            // print uuid (admittedly v1 specific)
            loIdx = Math.min( 4, fragmentBytes.length-1);
            hiIdx = Math.min(19, fragmentBytes.length-1);
            System.out.print("UUID:           ");
            for (int idx = loIdx; idx <= hiIdx; idx++) {
                System.out.printf(" %02x", fragmentBytes[idx]);
            }
            System.out.println("");

            // print index, count, size (admittedly v1 specific)
            loIdx = Math.min(20, fragmentBytes.length-1);
            hiIdx = Math.min(25, fragmentBytes.length-1);
            System.out.print("Idx, Tot, Size: ");
            for (int idx = loIdx; idx <= hiIdx; idx++) {
                System.out.printf(" %02x", fragmentBytes[idx]);
            }
            System.out.println("");
        }
        catch (Exception excpt) {
            excpt.printStackTrace(System.err);
        }
        */

        // test how bogus data size and offset are handled...
    }

    /**
     * Clears the payload data array.
     */
    private void clearPayload() {
        _dataBytes = null;
        _dataOffset = 0;
        _dataSize = 0;
    }
}
