package gov.fnal.cd.rms.provider;

import gov.fnal.cd.rms.util.RmsException;
import gov.fnal.cd.rms.util.UUIDGenerator;

/**
 * Helper class for splitting long string messages into 
 * arrays of smaller byte-based message fragments.
 *
 * @author Kurt Biery
 * @author Steve Foulkes
 * @version $Revision: 1.1.1.1 $ $Date: 2009/10/09 20:03:00 $
 */
public class MessageSplitter {

    /**
     * The UUIDGenerator instance that we'll use to create unique IDs.
     */
    private static UUIDGenerator _uuidGenerator;
    static {
        // seed the hostname with a somewhat random value
        String hostname = Long.toString(System.currentTimeMillis());
        try {
            hostname = java.net.InetAddress.getLocalHost().getHostName();
        }
        catch (java.net.UnknownHostException uhExcpt) {
            // leave default hostname in place
        }
        _uuidGenerator = new UUIDGenerator(hostname);
    }

    /**
     * Splits the specified message text into the necessary number of
     * message fragments of which none will exceed the specified number
     * bytes in size.
     *
     * @param messageText The message to split up.
     * @param maxFragmentSize The maximum size of a message fragment.
     * @throws IllegalArgumentException If invalid input arguments are
     * specified such as a null message text string or a maximum fragment
     * size that is smaller than what is needed for the message header.
     */
    public static MessageFragment[] splitMessage(String messageText,
                                                 int maxFragmentSize) {
        // check for null message
        if (messageText == null) {
            String errMsg = "MessageSplitter.splitMessage is not designed " +
                "to handle a null message.";
            throw new IllegalArgumentException(errMsg);
        }

        // check if the maximum fragment size is too small
        int fragmentHeaderSize = MessageFragment.getMaxHeaderSize();
        if (maxFragmentSize <= fragmentHeaderSize) {
            String errMsg = "MessageSplitter.splitMessage is unable to " +
                "create fragments smaller than " + (fragmentHeaderSize+1) +
                " bytes. (" + maxFragmentSize + " was requested)";
            throw new IllegalArgumentException(errMsg);
        }
        int packetSize = maxFragmentSize - fragmentHeaderSize;

        // convert the message text to a byte array
        byte[] rawMessageBytes = messageText.getBytes();
        int rawMessageSize = rawMessageBytes.length;

        // determine the number of fragments
        int fragmentCount = 1 + (int) ((rawMessageSize - 1) / packetSize);
        if (fragmentCount < 1) {fragmentCount = 1;}

        // create a UUID for this message
        java.util.UUID uniqueId = null;
        try {
            uniqueId = _uuidGenerator.getUUID();
        }
        catch (RmsException rmsExcpt) {
            uniqueId = java.util.UUID.randomUUID();
        }

        // create and fill the fragment array
        MessageFragment[] fragArray = new MessageFragment[fragmentCount];
        for (int idx = 0; idx < fragmentCount; idx++) {
            MessageFragment fragment =
                new MessageFragment(uniqueId, idx, fragmentCount);
            fragArray[idx] = fragment;

            // set the payload of this fragment to the correct subset
            // of the raw message bytes
            int startIdx = packetSize * idx;
            int fragSize = Math.min((packetSize * (idx + 1)),
                                    rawMessageSize) - startIdx;
            fragment.setPayload(rawMessageBytes, startIdx, fragSize);
        }

        // return the result
        return fragArray;
    }

    /**
     * Main method for unit testing of message splitting.
     */
    public static void main(String[] args) {
        int testCount = 0;

        // test handling of invalid arguments
        testCount++;
        System.out.println("\n>>>>> Test " + testCount + " <<<<<");
        System.out.println("Checking whether null message text " +
                           "generates an exception...");
        System.out.println("----------------------------------------");
        try {
            MessageFragment[] fragArray =
                MessageSplitter.splitMessage(null, 100);
            System.out.println("WARNING: a null message text did NOT " +
                               "generate an exception!  This could lead to " +
                               "undefined behavior!");
        }
        catch (Exception excpt) {
            System.out.println("A null message text " +
                               "generated an exception, as expected:\n    " +
                               excpt);
        }
        System.out.println("----------------------------------------");

        testCount++;
        System.out.println("\n>>>>> Test " + testCount + " <<<<<");
        System.out.println("Checking whether a small fragment size " +
                           "generates an exception...");
        System.out.println("----------------------------------------");
        try {
            int fragSize = MessageFragment.getMaxHeaderSize() - 1;
            MessageFragment[] fragArray =
                MessageSplitter.splitMessage("hello", fragSize);
            System.out.println("WARNING: a small fragment size did NOT " +
                               "generate an exception!  This could lead to " +
                               "undefined behavior!");
        }
        catch (Exception excpt) {
            System.out.println("A small fragment size " +
                               "generated an exception, as expected:\n    " +
                               excpt);
        }
        System.out.println("----------------------------------------");

        // walk through variously sized messages to get 1, 2, or 3 fragments
        // (In actual use, we wouldn't need to do anything with packet sizes
        // or getMaxHeaderSize.  We just use them here to generate an
        // known packet size for easier deciphering of the output.)
        int maxPacketSize = 5;
        int maxFragSize = maxPacketSize + MessageFragment.getMaxHeaderSize();
        StringBuffer messageBuffer = new StringBuffer();
        for (int mdx = 0; mdx < (maxPacketSize * 2.5); mdx++) {
            testCount++;
            System.out.println("\n>>>>> Test " + testCount + " <<<<<");
            System.out.println("Input message size = " +
                               messageBuffer.length() +
                               ", fragment size = " + maxFragSize);
            System.out.println("----------------------------------------");
            try {
                MessageFragment[] fragArray =
                    MessageSplitter.splitMessage(messageBuffer.toString(),
                                                 maxFragSize);
                for (int fdx = 0; fdx < fragArray.length; fdx++) {
                    System.out.println("FRAGMENT " + fdx + ":");
                    MessageFragment.dumpBytes(System.out,
                                              fragArray[fdx].getBytes(), -1);
                }
            }
            catch (Exception excpt) {
                System.err.println("Exception caught while splitting message");
                excpt.printStackTrace(System.err);
            }
            System.out.println("----------------------------------------");

            // add another character to the message text
            messageBuffer.append(mdx % 10);
        }

        System.out.println("");
    }
}
