package gov.fnal.cd.rms.provider;

/**
 * @package gov.fnal.cd.rms.provider
 *
 * @brief Provider-specific classes and interfaces.
 *
 * This package contains classes and interfaces for interacting with
 * third-party messaging providers.
 */

import gov.fnal.cd.rms.base.RmsCommunicationException;
import gov.fnal.cd.rms.base.RmsDestination;
import gov.fnal.cd.rms.base.RmsMessage;
import gov.fnal.cd.rms.util.RmsException;
import java.util.Vector;

/**
 * This interface defines the methods that provider-specific connection
 * classes must implement.
 *
 * @author Kurt Biery
 * @author Steve Foulkes
 * @version $Revision: 1.5 $ $Date: 2010/03/15 20:38:46 $
 */
public interface RmsConnection {
    /**
     * Tests whether the specified destination is supported by this
     * provider.
     *
     * @param candidateDestination the destination that should be tested
     * @return true if the destination is supported, false otherwise
     */
    public boolean supportsDestination(RmsDestination candidateDestination);

    /**
     * Sends the specified string to the specified destination.
     *
     * @param dest the destination for the message
     * @param messageString The string to send to the destination
     * @throws RmsCommunicationException if an error occurs when trying
     *         to send the message to the destination
     * @throws novadaq.DataException if an error occurs when trying
     *         to prepare the message for sending
     */
    public void sendString(RmsDestination dest, String messageString)
        throws RmsCommunicationException, novadaq.DataException;

    /**
     * Sends the specified RmsMessage to the specified destination.
     *
     * @param dest the destination for the message
     * @param message The RmsMessage to send to the destination
     * @throws RmsCommunicationException if an error occurs when trying
     *         to send the message to the destination
     * @throws novadaq.DataException if an error occurs when trying
     *         to prepare the message for sending
     */
    public void sendMessage(RmsDestination dest, RmsMessage message)
        throws RmsCommunicationException, novadaq.DataException;

    /**
     * Sends a message of the specified type to the specified
     * destination.
     *
     * @param messageDataType the java.lang.Class of the message to be sent
     * @param dest the destination that the message should be sent to
     * @param theMessage the message to be sent
     * @throws RmsCommunicationException if an error occurs when trying
     *         to send the message to the destination
     */
    public <T> void sendMessage(Class<T> messageDataType, RmsDestination dest,
                                T theMessage)
        throws RmsCommunicationException;

    /**
     * Receives a message of the specified type from the specified
     * destination.  This is a blocking call, so it does not return
     * until a message becomes available.
     *
     * @param messageDataType the java.lang.Class of the message that will
     *        be received
     * @param dest the destination that the message was sent to
     * @throws RmsCommunicationException if an error occurs when trying
     *         to receive the message from the destination
     */
    public <T> Vector<T> receiveMessage(Class<T> messageDataType, RmsDestination dest)
        throws RmsCommunicationException;

    /**
     * Receives a message of the specified type from the specified
     * destination.  If no message is received within the specified timeout,
     * a null message is returned.
     *
     * @param messageDataType the java.lang.Class of the message that will
     *        be received
     * @param dest the destination that the message was sent to
     * @param timeout The maximum amount of time to wait for a message,
     * in milliseconds.
     * @throws RmsCommunicationException if an error occurs when trying
     *         to receive the message from the destination
     */
    public <T> Vector<T> receiveMessage(Class<T> messageDataType, RmsDestination dest,
                                long timeout)
        throws RmsCommunicationException;

    /**
     * Adds the specified listener to the specified destination.  After
     * this method is called, the listener will be notified whenever
     * messages arrive at the destination.
     *
     * @param dest the destination to be monitored
     * @param listener the listener instance to be notified when a 
     *        message is received
     * @throws RmsCommunicationException if a problem is encountered setting
     *         up the system to monitor the specified destination
     */
    public void addListener(RmsDestination dest, ProviderListener listener)
        throws RmsCommunicationException;

    /**
     * Removes a listener from a particular destination.
     *
     * @param dest the destination instance to be removed
     * @param listener the listener instance to be removed
     * @throws RmsCommunicationException if an error occurs when trying
     *         to remove the listener from the destination
     */
    public void removeListener(RmsDestination dest, ProviderListener listener)
        throws RmsCommunicationException;

    /**
     * Get a reference to the message source.
     *
     * @return A destination object that contains the source of messages
     * for this connection.
     */
    public RmsDestination getMessageSource();

    /**
     * Returns a universally unique identifier.  This can be used to uniquely
     * identify a message.
     *
     * @return the unique ID string
     * @throws novadaq.DataException if an error occurs while creating the UUID
     */
    public String getUUID() throws novadaq.DataException;

    /**
     * Returns a universally unique identifier.  This can be used to uniquely
     * identify a message.  This version of the method will not throw a checked
     * exception.  Instead it will make a best-effort attempt to create a
     * unique ID.
     *
     * @return the unique ID string
     */
    public String getUUIDNoThrow();

    /**
     * Returns the default message expiration interval, in milliseconds
     * (also known as the time-to-live).
     *
     * @return the default time-to-live for messages
     */
    public long getDefaultTimeToLive();

    /**
     * Tests whether the specified destination is reachable (messages
     * could successfully be sent).
     *
     * @param pingDestination the destination to be tested
     * @return true if messages can be sent to the destination or 
     *         false if the destination is invalid or unreachable
     */
    public boolean ping(RmsDestination pingDestination);

    /**
     * Closes the connection.  After calling this method on a RmsConnection
     * object, the object should no longer be used since it will no
     * longer be connected to the underlying provider system.
     *
     * @throws RmsCommunicationException if a problem occurs when trying
     *         to disconnect from the provider
     */
    public void close() throws RmsCommunicationException;
}
