package gov.fnal.cd.rms.provider;

import gov.fnal.cd.rms.base.RmsCommunicationException;

// NOTE:  when adding a new provider type, be sure to check for all places
// that existing ones are used and add the new type as needed.  (There is
// more to adding a provider than just adding a new constant string.)

/**
 * This class provides methods for creating connection objects for
 * different providers.
 * (The connection objects implement the RmsConnection interface and
 * are used by consumers and producers to communicate with
 * the underlying messaging provider.)
 *
 * @author Kurt Biery
 * @author Steve Foulkes
 * @version $Revision: 1.2 $ $Date: 2010/02/03 14:30:37 $
 */
public class RmsConnectionFactory {

    /**
     * Constant string used for a EPICS provider connection.
     */
    public static final String EPICS_PROVIDER = "epics";

    /**
     * Constant string used for a DDS provider connection.
     */
    public static final String DDS_PROVIDER = "opensplice_dds";

    /**
     * Constant string used for a localhost provider connection.
     */
    public static final String LOCALHOST_PROVIDER = "localhost";

    /**
     * Creates a connection using the default provider.    
     * To determine the default provider,
     * this method first checks for a choice specified in
     * an environmental variabled named RMS_PROVIDER.  If
     * that env. var. is defined and contains an unempty string, 
     * it is used.  If not, then a hard-coded default is used.
     *
     * @param applicationName The name of the application generating/receiving
     * messages.
     * @param applicationPartition The partition of the application that is
     * generating/receiving messages.  
     *
     * @return an instance of the connection class specified in
     *         the RMS_PROVIDER environmental variable or the
     *         default provider type if the env. var. is not defined.
     * @throws java.lang.IllegalArgumentException if a null or empty
     *         application name or messaging region is specified
     * @throws RmsCommunicationException if an exception occurs creating
     *         the connection object
     */
    public static RmsConnection getConnection(String applicationName,
                                              int applicationPartition) 
        throws RmsCommunicationException {

        // check for a provider type specified in an env var
        // (is it worth making the env var value a class attribute
        // so we don't have to call getenv every time?)
        String selectedProvider = System.getenv("RMS_PROVIDER");

        // use a default type, if needed
        if (selectedProvider == null || selectedProvider.trim().equals("")) {
            selectedProvider = EPICS_PROVIDER;
        }

        // create the connection and return it
        return getConnection(applicationName, applicationPartition, selectedProvider);
    }

    /**
     * Creates a connection using a specified messaging provider.
     *
     * <b>Note</b> that this method not intended for general use!  The
     * methods in this class that automatically determine which
     * provider to use should be used by most, if not all, applications.
     * It has only been declared public to allow unit testing of specific
     * provider implementations.
     *
     * @param applicationName The name of the application generating/receiving
     * messages.
     * @param applicationPartition The partition of the application that is
     * generating/receiving messages.  
     * @param providerType string containing the requested provider type
     * @return an instance of the requested connection class
     * @throws java.lang.IllegalArgumentException if any of the input
     *         parameters are null or have invalid values
     * @throws RmsCommunicationException if an exception occurs creating
     *         the connection object
     */
    public static RmsConnection getConnection(String applicationName,
                                              int applicationPartition,
                                              String providerType)
        throws RmsCommunicationException {

        // check for invalid inputs
        if (providerType == null) {
            String msg = "A non-null connection type must be specified " +
                "when creating a connection.";
            throw new IllegalArgumentException(msg);
        }

        // translate the provider string into our preferred format
        providerType = providerType.trim().toLowerCase();

        // DDS provider type
        if (providerType.equals(DDS_PROVIDER)) {
            return new DDSConnection(applicationName, applicationPartition);
        }

        // EPICS provider type
        else if (providerType.equals(EPICS_PROVIDER)) {
            return new EpicsConnection(applicationName, applicationPartition);
        }

        // localhost provider type
        else if (providerType.equals(LOCALHOST_PROVIDER)) {
            return new LocalhostConnection(applicationName, applicationPartition);
        }

        // unsupported provider type
        else {
            String msg = "Unable to create a(n) " + providerType +
                " connection (unsupported provider type).";
            throw new IllegalArgumentException(msg);
        }
    }
}
