package gov.fnal.cd.rms.util;

/**
 * @package gov.fnal.cd.rms.util
 *
 * @brief A collection of classes that simplify miscellaneous
 * tasks.
 *
 * The util package contains helper classes for managing threads,
 * parsing time and date strings, generating UUIDs as well as
 * thread safe data structures.
 */

import gov.fnal.cd.rms.util.SHA1;

import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.IOException;

import java.util.Date;
import java.util.Random;
import java.util.UUID;

/**
 * This class generates time based type 1 UUIDs, as specified in RFC4122.
 *
 * @author Kurt Biery
 * @author Steve Foulkes
 * @version $Revision: 1.1.1.1 $ $Date: 2009/10/09 20:03:00 $
 */
public class UUIDGenerator {
    /**
     * Time at which the last UUID was generated.
     */
    private Date _previousTime = null;

    /**
     * Counts the number of times a UUID has been asked for during
     * one clock tick.
     */
    private int _overrunCounter = 0;

    /**
     * Name of the node requesting the UUID.
     */
    private String _nodeName = null;

    /**
     * Hashed version of the node name.
     */
    private byte[] _nodeNameHashed;

    /**
     * A 16 bit random number.
     */
    private byte[] _clockSequence;

    /**
     * Create a new UUID generator for a specific node.
     *
     * @param nodeName The name of the node that UUIDs will be generated
     * for.  The nodeName should be unique.
     */
    public UUIDGenerator(String nodeName) {
        _nodeName = new String(nodeName);
        _nodeNameHashed = new byte[20];

        /**
         * Hash the node name, as we're only interested in the first 6 bytes.
         */
        SHA1 s = new SHA1();
        s.init();
        s.updateASCII(_nodeName);
        s.finish();
        _nodeNameHashed = s.digestBits;

        /**
         * The clock sequence helps to guard against the case where the 
         * node ID changes or the time on the host changes.  We'll 
         * implement it just as a 16 bit random number.
         */
        Random numberGenerator = new Random();
        _clockSequence = new byte[2];
        numberGenerator.nextBytes(_clockSequence);
    }

    /**
     * Generate a new UUID.  
     *
     * @return A UUID.
     * @throws RmsException If a problem is encountered when creating
     * the UUID.
     */
    public synchronized UUID getUUID() throws RmsException {
        IOException firstProblem = null;
        ByteArrayOutputStream uuidByteStream = new ByteArrayOutputStream(16);
        DataOutputStream uuidDataStream = new DataOutputStream(uuidByteStream);
        Date currentTime = new Date();

        if (_previousTime == null) {
            _overrunCounter = 0;
        }
        else if (currentTime.equals(_previousTime)) {
            _overrunCounter++;
        }
        else {
            _overrunCounter = 0;
        }

        /**
         * The java timestamp only has a resolution of 1 milisecond,
         * which would limit us to 1000 UUIDs per second.  To combat 
         * this everytime we have a timestamp collision,  The overRunCounter 
         * is incremented.  The timestamp is left shited 10 bits, and the 
         * overRunCounter is ORed onto the timestamp.  Even with the loss 
         * of 10 bits on our timestamp, we won't overrun the timestamp 
         * for ~35,700 years.
         */ 
        long currentTimeLong = currentTime.getTime();
        currentTimeLong = (currentTimeLong << 10) | (_overrunCounter & 0x3ff); 

        /**
         * Add the version tag to the timestamp.
         */
        long version = 0x1;
        currentTimeLong = currentTimeLong | (version << 60);

        try {
            uuidDataStream.write(_nodeNameHashed, 0, 6);
            uuidDataStream.write(_clockSequence, 0, 2);
            uuidDataStream.writeLong(currentTimeLong);
        }
        catch (IOException ioExcpt) {
            if (firstProblem == null) {
                firstProblem = ioExcpt;
            }
        }
        finally {
            try {
                uuidDataStream.close();
            }
            catch (IOException ioExcpt) {
                if (firstProblem == null) {
                    firstProblem = ioExcpt;
                }
            }
            try {
                uuidByteStream.close();
            }
            catch (IOException ioExcpt) {
                if (firstProblem == null) {
                    firstProblem = ioExcpt;
                }
            }
        }
        if (firstProblem != null) {
            String msg = "Error creating the timestamp portion of a UUID.";
            throw new RmsException(msg, firstProblem);
        }

        _previousTime = currentTime;

        ByteArrayInputStream uuidByteInputStream = 
            new ByteArrayInputStream(uuidByteStream.toByteArray());
        DataInputStream uuidDataInputStream = 
            new DataInputStream(uuidByteInputStream);

        long highLong = 0;
        long lowLong = 0;

        try {
            highLong = uuidDataInputStream.readLong();
            lowLong = uuidDataInputStream.readLong();
        }
        catch (IOException ioExcpt) {
            if (firstProblem == null) {
                firstProblem = ioExcpt;
            }
        }
        finally {
            try {
                uuidDataInputStream.close();
            }
            catch (IOException ioExcpt) {
                if (firstProblem == null) {
                    firstProblem = ioExcpt;
                }
            }
            try {
                uuidByteInputStream.close();
            }
            catch (IOException ioExcpt) {
                if (firstProblem == null) {
                    firstProblem = ioExcpt;
                }
            }
        }
        if (firstProblem != null) {
            String msg = "Error converting UUID long values.";
            throw new RmsException(msg, firstProblem);
        }

        return new UUID(highLong, lowLong);
    }

    /**
     * Generate a couple UUIDs.  
     */
    public static void main(String args[]) {
        UUIDGenerator uuidGen = new UUIDGenerator("steve");

        for (int i = 0; i < 20; i++) {
            try {
                System.out.println(uuidGen.getUUID().toString());
            }
            catch (RmsException rmsExcpt) {
                rmsExcpt.printStackTrace(System.err);
            }
        }

        return;
    }

    /**
     * Given a long, print out its contents in hex.
     *
     * @param longData The long that we want to print out.
     */
    private void longToStdout(long longData) {
        ByteArrayOutputStream byteStream = new ByteArrayOutputStream(8);
        DataOutputStream dataStream = new DataOutputStream(byteStream);
        byte[] byteArray;

        try {
            dataStream.writeLong(longData);
        }
        catch (IOException excp) {
            System.out.println("Something happened while I was trying to" +
                               " print a long:");
            System.out.println(excp);
            return;
        }

        System.out.print("0x");
        byteArray = byteStream.toByteArray();

        for (int i = 0; i < 8; i++) {
            System.out.printf("%02x", byteArray[i]);
        }

        System.out.println("");
    }

    /**
     * Given a byte array, print out its contents in hex.
     *
     * @param byteArray The byte array that we want to print out
     */
    private void byteArrayToStdout(byte[] byteArray) {
        System.out.print("0x");

        for (int i = 0; i < byteArray.length; i++) {
            System.out.printf("%02x", byteArray[i]);
        }

        System.out.println("");
    }
}
