import gov.fnal.cd.rms.provider.RmsConnection;
import gov.fnal.cd.rms.provider.RmsConnectionFactory;

import gov.fnal.cd.rms.base.RmsDestination;
import gov.fnal.cd.rms.base.RmsMessage;
import gov.fnal.cd.rms.base.RmsCommunicationException;

import gov.fnal.cd.rms.RmsConsumer;

import gov.fnal.cd.rms.example.RmsExampleMessage;

/**
 * This is simple example showing how to receive messages
 * and send replies using RMS.  It assumes that two
 * destinations exist in the message passing system:
 *   target=RmsConsumerExample, channel=control, message type=request
 *   target=RmsProducerExample, channel=control, message type=request
 * (These will be automatically created when using the EPICS autoCas.)
 *
 * Using the EPICS provider, the following commands need
 * to be run to provide the necessary PVs:
 *   [sfoulkes@cd-90837 rms]$ export RMS_CUSTOM_ENVIRONMENT="<yourEnv>"
 *   [sfoulkes@cd-90837 rms]$ autoCas
 *
 * This program should produce the following output:
 *   [sfoulkes@cd-90837 rms]$ java RmsConsumerExample
 *   Received a message: Hello World!
 *   Sending reply...
 */
public class RmsConsumerExample {
    public static void main(String args[]) {
        RmsConnection rmsConnection = null;
        RmsConsumer exampleConsumer = null;
        RmsDestination exampleDest = null;

        try {
            /**
             * Create an RmsConnection object using the connection factory.
             * We want replies routed to the destination with a target equal
             * to "RmsConsumerExample" so we set that as the applicationName
             * and -1 as the partition as we don't care about partitions.
             */
            rmsConnection = RmsConnectionFactory.getConnection("RmsConsumerExample", -1);

            /**
             * Create a destination to represent where we want to receive messages
             * from.  For the example, we're receiving messages to the "RmsConsumerExample" 
             * target.  The messages will use the control channel and be request messages.
             */
            exampleDest = new RmsDestination("RmsConsumerExample",
                                             RmsDestination.CONTROL_CHANNEL,
                                             RmsDestination.REQUEST_MESSAGE_TYPE);

            /**
             * Create a consumer with the connection object and destination.
             */
            exampleConsumer = new RmsConsumer(rmsConnection, exampleDest);

            /**
             * Wait for a message to show up.
             */
            RmsMessage incomingMessage = null;
            incomingMessage = exampleConsumer.receiveMessage();

            /**
             * Check the message to make sure it isn't empty.
             */
            if (incomingMessage == null) {
                System.out.println("Failed to receive a valid message.");
                exampleConsumer.close();
                rmsConnection.close();
                return;
            }

            /**
             * Verify that the message we received is of the type 
             * that we are expecting.
             */
            String messageClassName = incomingMessage.getBody().getClassName();

            if (messageClassName.compareTo("gov.fnal.cd.rms.example.RmsExampleMessage") != 0) {
                System.out.println("Received an unexpected message:");
                System.out.println(messageClassName);
                return;
            }

            /**
             * Retrieve the message content, and print it to the screen.
             */            
            RmsExampleMessage incomingMessageBody = 
                (RmsExampleMessage) incomingMessage.getBodyObject();

            System.out.print("Received a message: ");
            System.out.println(incomingMessageBody.getMessage());
            System.out.print("The message was sent at ");
            System.out.println(incomingMessage.getSentTime()+" (local time).");

            /**
             * Create a reply and stuff some content into it.
             * This data type was created from the RmsExampleMessages xsd file.
             */            
            RmsMessage exampleReply = incomingMessage.createReply();
            RmsExampleMessage replyBody = new RmsExampleMessage();
            replyBody.setMessage("Hello World Reply!");
            exampleReply.setBody(replyBody);

            /**
             * Send the reply.
             */            
            System.out.println("Sending reply...");
            exampleConsumer.sendReply(exampleReply);

            /**
             * Cleanup, and exit.
             */            
            exampleConsumer.close();
            rmsConnection.close();

            System.exit(0);
        }
        catch (novadaq.DataException dataException) {
            System.out.println("Received a data exception:");
            System.out.println(dataException);
            Throwable cause = dataException.getCause();
            while (cause != null) {
                System.out.println(cause);
                cause = cause.getCause();
            }
        }
        catch (RmsCommunicationException commException) {
            System.out.println("Received a communication exception:");
            System.out.println(commException);
        }
        finally {
            try {
                if (exampleConsumer != null) {
                    exampleConsumer.close();
                }
                if (rmsConnection != null) {
                    rmsConnection.close();
                }
            }
            catch (RmsCommunicationException commException) {
                System.out.println("There was a problem cleaning up:");
                System.out.println(commException);
            }
        }
    }
}
