import gov.fnal.cd.rms.provider.ProviderListener;
import gov.fnal.cd.rms.provider.RmsConnection;
import gov.fnal.cd.rms.provider.RmsConnectionFactory;

import gov.fnal.cd.rms.HeartbeatProducer;
import gov.fnal.cd.rms.base.RmsDestination;
import gov.fnal.cd.rms.base.RmsMessage;
import gov.fnal.cd.rms.base.RmsMessageBody;
import gov.fnal.cd.rms.RmsConsumer;

/**
 * Repeatedly receive messages from a particluar destination.  This has the
 * capability to receive messages as well as strings to avoid the overhead
 * of deserializing messages.  This class will also broadcast heartbeat 
 * messages.
 *
 * @author Kurt Biery
 * @author Steve Foulkes
 * @version $Revision: 1.1.1.1 $ $Date: 2009/10/09 20:03:00 $
 */
public class RmsConsumerTest {
    private RmsConnection _myConnection;
    private RmsConsumer _myConsumer;

    private HeartbeatProducer _myHeartbeat;

    private int _msgsReceived = 0;
    private int _totalMsgsReceived = 0;
    private boolean _superFastMode = false;
    private boolean _sendReplies = false;

    private Thread _receivingThread;
    private String previousMessageText = "";

    private static java.text.SimpleDateFormat sentTimeFmt =
        new java.text.SimpleDateFormat("yyyy-MM-dd HH:mm:ss.SSS");

    /**
     * Create provider, connection and consumer objects based
     * upon what was passed in from the command line.
     *
     * @param args The command line arguments where args[1] is the 
     * application name, args[2] is the target name, and args[3] is
     * the target service.
     */ 
    private void initializeRms(String[] args) {
	RmsDestination myDestination;

	try {
	    _myConnection = RmsConnectionFactory.getConnection(args[1], -1);
	}
	catch (Exception excp) {
	    System.out.println("Had some problems creating the connection");
	    System.out.println(excp);
	
	    return;
	}

	/**
	 * Send out a heartbeat message every 5 seconds...
	 */
	_myHeartbeat = new HeartbeatProducer("RmsConsumerTest", "$Revision: 1.1.1.1 $", 5, _myConnection);

	myDestination = new RmsDestination(args[2], args[3], "request");

	try {
	    _myConsumer = new RmsConsumer(_myConnection, myDestination);
	}
	catch (Exception excp) {
	    System.out.println("Failed to create a consumer:");
	    System.out.println(excp);
	}

	return;
    }

    /**
     * Receive messages from the consumer.  If we are in superfast
     * mode, then we just receive the string.  Otherwise, we deserialize
     * the message.  If the mode is normalReply we will also generate
     * a reply, and send that out.
     */
    private void receiveMessages() {
	RmsMessage incomingMessage = null;
	RmsMessage messageReply;
	RmsMessageBody replyBody;
	String incomingString = null;

	while(true) {
	    if(_superFastMode) {
		try {
		    incomingString = _myConsumer.receiveString();
		}
		catch (Exception excp) {
		    System.out.println("Received an exception while trying to receive a string:");
		    System.out.println(excp);
		}
	    }
	    else {
		incomingMessage = _myConsumer.receiveMessage();
	    }

	    if (incomingMessage != null) {
		String someMessageText = incomingMessage.getId();
		java.util.Date sentTime = incomingMessage.getSentTime();
		if (sentTime != null) {
		    synchronized (sentTimeFmt) {
			someMessageText += " " + sentTimeFmt.format(sentTime);
		    }
		}
		if (someMessageText.equals(previousMessageText)) {
		    System.out.println("WARNING: duplicate message:");
		    try {
			System.out.println(incomingMessage.serialize());
		    }
		    catch (Exception excpt) {
			excpt.printStackTrace(System.out);
		    }
		}
		previousMessageText = someMessageText;
	    }
	    else if (incomingString != null) {
		if (incomingString.equals(previousMessageText)) {
		    System.out.println("WARNING: duplicate message:");
		    System.out.println(incomingString);
		}
		previousMessageText = incomingString;
	    }
	    
	    _msgsReceived++;
	    _totalMsgsReceived++;

	    if(_sendReplies && incomingMessage != null) {
		replyBody = new RmsMessageBody();
		
		replyBody.setContent(incomingMessage.getBody().getContent());
		messageReply = incomingMessage.createReply();
		messageReply.setBody(replyBody);

		try {
		    _myConsumer.sendReply(messageReply);
		}
		catch (Exception excp) {
		    System.out.println("Sending of the reply failed: " + excp);
		}
	    }
	}
    }

    /**
     * Run receiveMessages() in a seperate thread, as it
     * blocks forever.
     */
    private class ReceivingThread implements Runnable {
	public void run() {
	    receiveMessages();
	}
    }

    /**
     * Create a new receiving thread and start it up.
     */
    private void startReceivingThread() {
	_receivingThread = new Thread(new ReceivingThread());
	_receivingThread.start();
    }
      
    /**
     * Look through the command line arguments and set the mode to
     * be whatever the user wants it to be.  Then, create a new
     * RmsConsumer and start it up.  Message rate information will
     * be dumped to the console every five seconds.
     *
     * @param args The command line arguments
     */
    public static void main(String[] args) {
	RmsConsumerTest testObject = new RmsConsumerTest();

	if(args.length != 4) {
	    System.out.println("Usage: RmsConsumerTest [receive mode] [node name] [target name] [target service]");
	    System.out.println("Receive mode can be: normal, normalReply or superfast");
	    
	    System.exit(0);
	}

	if(args[0].equals("superfast")) {
	    System.out.println("Entering superfast mode...");
	    testObject._superFastMode = true;
	}
	if(args[0].equals("normalReply")) {
	    System.out.println("Will send replies...");
	    testObject._sendReplies = true;
	}

	testObject.initializeRms(args);
	testObject.startReceivingThread();

	while(true) {
	    System.out.println("Have received " + testObject._totalMsgsReceived + " messages, " + 
			       testObject._msgsReceived + " in the last 5 seconds.");
	    testObject._msgsReceived = 0;
	    try {
		Thread.sleep(5000);
	    }
	    catch (Exception excp) {
	    }
	}
    }
}
