/**
 * A simple message monitor application.
 *
 * @author Kurt Biery
 * @version $Id: RmsMessageMonitor.java,v 1.3 2009/11/25 15:25:51 biery Exp $
 */

import java.awt.*;
import java.awt.event.*;
import java.io.FileWriter;
import java.text.SimpleDateFormat;
import javax.swing.*;

import gov.fnal.cd.rms.RmsConsumer;
import gov.fnal.cd.rms.RmsMessageListener;
import gov.fnal.cd.rms.base.RmsDestination;
import gov.fnal.cd.rms.base.RmsMessage;
import gov.fnal.cd.rms.provider.RmsConnection;
import gov.fnal.cd.rms.provider.RmsConnectionFactory;
import novadaq.XMLFormatter;
import novadaq.demo.Launchable;
import novadaq.demo.ApplicationLauncher;

public class RmsMessageMonitor implements Launchable, RmsMessageListener {

    /**
     * Application name.
     */
    private static final String APP_NAME = "NOvA Message Monitor";

    /**
     * The name of the application that we're monitoring.
     */
    private String _targetApplication = null;

    /**
     * The channel that we're monitoring.
     */
    private String _channelName = null;

    /**
     * The type of the destination to be monitored (target or source).
     */
    private String _destinationType = null;

    /**
     * The partition number to be used in destinations (may be superfluous
     * if the destination is not for broadcasts.)
     */
    private int _partNum = 0;

    /**
     * Display area for messages received.
     */
    private JTextArea _messageTextArea;
    private JPopupMenu _popupMenu;

    /**
     * Date formatter.
     */
    private static SimpleDateFormat _dateFormatter =
        new SimpleDateFormat("dd-MMM HH:mm:ss");

    /**
     * Optional file writer for saving copies of messages to disk.
     */
    FileWriter _localFileWriter = null;

    /**
     * Attributes for interfacing to RMS.
     */
    private RmsConnection _rmsConnection;
    private RmsConsumer _watcher;

    /**
     * Contructs a monitor application instance.
     */
    public RmsMessageMonitor(String targetApplication, String channelName,
                             String destinationType, int partNum,
                             String diskFileName) {
        _targetApplication = targetApplication;
        _channelName = channelName;
        _destinationType = destinationType;
        _partNum = partNum;

        if (diskFileName != null) {
            try {
                _localFileWriter = new FileWriter(diskFileName);
            }
            catch (java.io.IOException ioExcpt) {
                ioExcpt.printStackTrace(System.err);
            }
        }
    }

    /**
     * Returns the splash screen for this application or null if none.
     */
    public Component getSplashPanel() throws Exception {
        return null;
    }

    /**
     * Performs any long-running initialization that is needed.
     */
    public void initialize() throws Exception {
        // catch and re-throw exceptions so that they appear in the
        // console log during automated testing
        try {

            // create the connection to RMS
            _rmsConnection =
                RmsConnectionFactory.getConnection(APP_NAME,
                                                   RmsDestination.NULL_PARTITION);

            // create the destination to monitor messages
            RmsDestination dest = new RmsDestination();
            dest.setProperty(RmsDestination.CHANNEL_PROPERTY_NAME,
                             _channelName);
            if (_destinationType != null &&
                _destinationType.equalsIgnoreCase("source")) {
                dest.setProperty(RmsDestination.SOURCE_PROPERTY_NAME,
                                 _targetApplication);
            }
            else {
                dest.setProperty(RmsDestination.TARGET_PROPERTY_NAME,
                                 _targetApplication);
            }
            dest.setPartitionProperty(_partNum);

            // create the consumer that will monitor  messages
            _watcher = new RmsConsumer(_rmsConnection, dest);
            _watcher.setMessageListener(this);

        }
        catch (Exception excpt) {
            excpt.printStackTrace(System.err);
            throw excpt;
        }
    }

    /**
     * Processes the message.
     */
    public void messageReceived(RmsMessage receivedMessage) {

        // sanity check
        if (receivedMessage == null) {
            System.err.println((new java.util.Date()) +
                               " ERROR: RMS Message is NULL in " +
                               "RmsMessageMonitor callback!");
            return;
        }

        String msgString = null;
        try {
            msgString = receivedMessage.serialize();
        }
        catch (Exception excpt) {
            excpt.printStackTrace(System.err);
        }

        try {
            String tmpString = XMLFormatter.prettyFormat(msgString,2);
            msgString = tmpString;
        }
        catch (javax.xml.transform.TransformerException excpt) {}

        // SimpleDateFormat is not thread safe
        String dateString = "";
        synchronized (_dateFormatter) {
            dateString = _dateFormatter.format(new java.util.Date()) + ": ";
        }

        if (_messageTextArea != null) {
            synchronized (_messageTextArea) {
                _messageTextArea.append("\n" + dateString + msgString);

                try {
                    int lineCount = _messageTextArea.getLineCount();
                    int newCursorPosition =
                        _messageTextArea.getLineStartOffset(lineCount - 1);
                    _messageTextArea.setCaretPosition(newCursorPosition);
                }
                catch (Exception excpt) {
                    excpt.printStackTrace(System.err);
                }
            }
        }

        if (_localFileWriter != null) {
            synchronized (_localFileWriter) {
                try {
                    _localFileWriter.write(dateString + msgString);
                    _localFileWriter.flush();
                }
                catch (java.io.IOException ioExcpt) {
                    ioExcpt.printStackTrace(System.err);
                }
            }
        }
    }

    /**
     * Returns the main display.
     */
    public Component getMainDisplay(JFrame parentFrame) throws Exception {

        JPanel pane = new JPanel();
        pane.setLayout(new BoxLayout(pane, BoxLayout.Y_AXIS));
        pane.setBorder(BorderFactory.createEmptyBorder(
                                        20, //top
                                        30, //left
                                        20, //bottom
                                        30) //right
                                        );

        _messageTextArea = new JTextArea(24, 80);
        _messageTextArea.setOpaque(true);
        _messageTextArea.setBackground(Color.WHITE);
        _messageTextArea.setLineWrap(true);
        _messageTextArea.setEditable(false);

        _popupMenu = new JPopupMenu();
        JMenuItem menuItem = new JMenuItem("Copy");
        menuItem.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent popupEvent) {
                _messageTextArea.copy();
            }
        });
        _popupMenu.add(menuItem);

        _messageTextArea.addMouseListener(new MouseAdapter() {
            public void mousePressed(MouseEvent pressEvent) {
                if (pressEvent.isPopupTrigger()) {
                    _popupMenu.show(pressEvent.getComponent(),
                                    pressEvent.getX(), pressEvent.getY());
                }
            }
        });

        JScrollPane scroller = new JScrollPane(_messageTextArea);
        pane.add(scroller);

        return pane;
    }

    /**
     * Returns the menu bar for the application or null if no menu bar
     * is needed.
     */
    public JMenuBar getMenuBar() {
        return null;
    }

    /**
     * Cleans up when garbage collected.
     */
    protected void finalize() {
        if (_localFileWriter != null) {
            try {
                _localFileWriter.close();
            }
            catch (java.io.IOException ioExcpt) {
                ioExcpt.printStackTrace(System.err);
            }
        }
    }

    /**
     * Application main method.
     */
    public static void main(String[] args) {

        // verify that all required arguments were specified
        if (args.length < 3) {
            System.err.println("Usage: java RmsMessageMonitor <targetName> " +
                               "<channelName> <\"target\" | \"source\"> " +
                               "[partitionNumber] [diskFileName]");
            System.exit(0);
        }

        // parse the required arguments
        StringBuffer windowName = new StringBuffer(APP_NAME + " watching ");
        windowName.append(args[1] + " messages sent ");
        if (args[2] != null && args[2].equalsIgnoreCase("source")) {
            windowName.append("from");
        }
        else {
            windowName.append("to");
        }
        windowName.append(" \"" + args[0] + "\"");

        // parse the optional arguments
        int partNum = RmsDestination.NULL_PARTITION;
        String diskFileName = null;
        int argIdx = 3;
        final int DUMMY_VALUE = -519;
        while (argIdx < args.length) {
            int tmpInt = DUMMY_VALUE;
            try {tmpInt = Integer.parseInt(args[argIdx], 10);}
            catch (Exception excpt) {}
            if (tmpInt != DUMMY_VALUE) {
                partNum = tmpInt;
            }
            else {
                diskFileName = args[argIdx];
            }
            argIdx++;
        }

        // 26-Sep-2007, KAB - if an output file has been specified,
        // skip the display
        if (diskFileName != null) {
            RmsMessageMonitor msgMon = new RmsMessageMonitor(args[0], args[1],
                                                             args[2], partNum,
                                                             diskFileName);
            try {
                msgMon.initialize();
                while (true) {
                    Thread.sleep(10000);
                }
            }
            catch (Exception excpt) {
                excpt.printStackTrace(System.err);
            }
        }

        else {
            // launch the application display
            new ApplicationLauncher(new RmsMessageMonitor(args[0], args[1],
                                                          args[2], partNum,
                                                          diskFileName),
                                    windowName.toString());
        }
    }
}
