import gov.fnal.cd.rms.provider.RmsConnection;
import gov.fnal.cd.rms.provider.RmsConnectionFactory;

import gov.fnal.cd.rms.base.RmsDestination;
import gov.fnal.cd.rms.base.RmsMessage;
import gov.fnal.cd.rms.base.RmsCommunicationException;

import gov.fnal.cd.rms.RmsProducer;

import gov.fnal.cd.rms.example.RmsExampleMessage;

/**
 * This is simple example showing how to send messages
 * and receive replies using RMS.  It assumes that two
 * destinations exist in the message passing system:
 *   target=RmsConsumerExample, channel=control, message type=request
 *   target=RmsProducerExample, channel=control, message type=request
 * (These will be automatically created when using the EPICS autoCas.)
 *
 * Using the EPICS provider, the following commands need
 * to be run to provide the necessary PVs:
 *   [sfoulkes@cd-90837 rms]$ export RMS_CUSTOM_ENVIRONMENT="<yourEnv>"
 *   [sfoulkes@cd-90837 rms]$ autoCas
 *
 * This program should produce the following output:
 *   [sfoulkes@cd-90837 rms]$ java RmsProducerExample
 *   Sending message...
 *   Received a reply: Hello World Reply!
 */
public class RmsProducerExample {
    public static void main(String args[]) {
        RmsConnection rmsConnection = null;
        RmsProducer exampleProducer = null;
        RmsDestination exampleDest = null;
        RmsMessage exampleMessage = null;

        try {
            /**
             * Create an RmsConnection object using the connection factory.
             * We want replies routed to the destination with a target equal
             * to "RmsProducerExample" so we set that as the applicationName
             * and -1 as the partition as we don't care about partitions.
             */
            rmsConnection = RmsConnectionFactory.getConnection("RmsProducerExample", -1);

            /**
             * Create a destination to represent where the messages that are
             * generated here will go to.  For the example, we're sending messages
             * to the "RmsConsumerExample" target.  The messages will use the control
             * channel and be request messages.
             */
            exampleDest = new RmsDestination("RmsConsumerExample",
                                             RmsDestination.CONTROL_CHANNEL,
                                             RmsDestination.REQUEST_MESSAGE_TYPE);

            /**
             * Create a producer with the connection and the destination.
             */
            exampleProducer = new RmsProducer(rmsConnection, exampleDest);

            /**
             * Create an RmsExampleMessage and stuff some content into it.
             * This data type was created from the RmsExampleMessages xsd file.
             */
            RmsExampleMessage exampleMessageBody = new RmsExampleMessage();
            exampleMessageBody.setMessage("Hello World!");
            exampleMessage = new RmsMessage(exampleMessageBody);

            /**
             * Send the message.
             */
            System.out.println("Sending message...");
            exampleProducer.sendMessage(exampleMessage);

            /**
             * Wait for a reply.
             */
            RmsMessage incomingReply = null;
            incomingReply = exampleProducer.receiveReply();

            /**
             * Verify that the message we received is of the type
             * that we are expecting.
             */
            String messageClassName = incomingReply.getBody().getClassName();

            if (messageClassName.compareTo("gov.fnal.cd.rms.example.RmsExampleMessage") != 0) {
                System.out.println("Received an unexpected message:");
                System.out.println(messageClassName);
                return;
            }

            /**
             * Retrieve the message content.
             */
            RmsExampleMessage incomingReplyBody = null;
            incomingReplyBody = (RmsExampleMessage) incomingReply.getBodyObject();

            System.out.print("Received a reply: ");
            System.out.println(incomingReplyBody.getMessage());

            /**
             * Cleanup, and exit.
             */
            exampleProducer.close();
            rmsConnection.close();

            System.exit(0);
        }
        catch (novadaq.DataException dataException) {
            System.out.println("Received a data exception:");
            System.out.println(dataException);
            Throwable cause = dataException.getCause();
            while (cause != null) {
                System.out.println(cause);
                cause = cause.getCause();
            }
        }
        catch (RmsCommunicationException commException) {
            System.out.println("Received a communication exception:");
            System.out.println(commException);
        }
        finally {
            try {
                if (exampleProducer != null) {
                    exampleProducer.close();
                }
                if (rmsConnection != null) {
                    rmsConnection.close();
                }
            }
            catch (RmsCommunicationException commException) {
                System.out.println("There was a problem cleaning up:");
                System.out.println(commException);
            }
        }
    }
}
