import gov.fnal.cd.rms.provider.RmsConnection;
import gov.fnal.cd.rms.provider.RmsConnectionFactory;

import gov.fnal.cd.rms.base.RmsDestination;
import gov.fnal.cd.rms.base.RmsMessage;
import gov.fnal.cd.rms.base.RmsMessageBody;
import gov.fnal.cd.rms.RmsProducer;
import gov.fnal.cd.rms.HeartbeatProducer;

/**
 * Repeatedly send messages to a particular destination.  This class
 * has the capability to send messages, as well as sending a strings 
 * to avoid having to serialize the message.  This class will also
 * broadcast heartbeat messages.
 *
 * @author Kurt Biery
 * @author Steve Foulkes
 * @version $Revision: 1.1.1.1 $ $Date: 2009/10/09 20:03:00 $
 */
public class RmsProducerTest {
    private RmsConnection _myConnection;
    private RmsProducer _myProducer;
    private RmsDestination _myDestination;

    private HeartbeatProducer _myHeartbeat;

    private boolean _superFastMode = false;
    private boolean _receiveReplies = false;
    private int _messagesToSend = 0;

    private Thread _replyThread;

    /**
     * Create provider, connection and consumer objects based
     * upon what was passed in from the command line.
     *
     * @param args The command line arguments where args[1] is the 
     * application name, args[2] is the target name, and args[3] is
     * the target service.
     */ 
    private void initializeRms(String[] args) {
	try {
	    _myConnection = RmsConnectionFactory.getConnection(args[1], -1);
	}
	catch (Exception excp) {
	    System.out.println("Had some problems creating the connection");
	    System.out.println(excp);
	
	    return;
	}

	/**
	 * Send out a heartbeat message every 5 seconds...
	 */
	_myHeartbeat = new HeartbeatProducer("RmsProducerTest", "$Revision: 1.1.1.1 $", 5, _myConnection);

	_myDestination = new RmsDestination(args[2], args[3], "request");

	try {
	    _myProducer = new RmsProducer(_myConnection, _myDestination);
	}
	catch (Exception excp) {
	    System.out.println("Exception when trying to create a producer: " + excp);
	    return;
	}

	return;
    }

    /**
     * Send a message to the destination that was specified on the command
     * line.  If we're in superfast mode, we just send a string.  Otherwise,
     * create an RmsMessage that can be serialized and then sent out.
     *
     * @param displayValue A number that will be tacked on to the end of
     * the message to the message can be uniquely identified.
     */
    private void sendMessage(int displayValue) {
	System.out.print(".");

	if(_superFastMode) {
	    try {
		_myProducer.sendString("Hello World! (" + displayValue + ")");
	    }
	    catch (Exception excp) {
		System.out.println("Sending of the string failed: " + excp);
	    }

	    return;
	}
	
	RmsMessage myMessage;
	RmsMessageBody myMessageBody;

	myMessageBody = new RmsMessageBody();
	myMessageBody.setContent("Hello World! (" + displayValue + ")");
	myMessage = new RmsMessage(myMessageBody);

	try {
	    _myProducer.sendMessage(myMessage);
	}
	catch (Exception excp) {
	    System.out.println("Sending of the message failed: " + excp);
	}

	return;
    }

    /**
     * Receive a reply for every message that the user wants
     * us to send.  This will only run if we're in normalReply
     * mode.
     */
    private void receiveReplies() {
	RmsMessage reply;

	if(_receiveReplies) {
	    for(int i = 0; i < _messagesToSend; i++) {
		reply = _myProducer.receiveReply();
	    }
	}
    }
    
    /**
     * A thread that will run the receiveReplies() method
     * so that replies can be received as messages are being
     * sent out.
     */
    private class ReplyThread implements Runnable {
	public void run() {
	    receiveReplies();
	}
    }

    /**
     * Start up the reply thread.
     */
    private void startReplyThread() {
	_replyThread = new Thread(new ReplyThread());
	_replyThread.start();
    }

    /**
     * Look through the command line arguments and set the mode to
     * be whatever the user wants it to be.  Then, create a new
     * RmsProducerTest and start it up.  Message rate information will
     * be dumped to the console every five seconds.
     *
     * @param args The command line arguments
     */
    public static void main(String[] args) {
	RmsProducerTest testObject = new RmsProducerTest();

	if(args.length != 6) {
	    System.out.println("Usage:");
	    System.out.println("RmsProducerTest [send mode] [node name] [target name] [target service] [messages to send] [message delay]");
	    System.out.println("Send mode can be: normal, normalReply or superfast");
	    
	    System.exit(0);
	}

	if(args[0].equals("superfast")) {
	    System.out.println("Entering superfast mode...");
	    testObject._superFastMode = true;
	}
	if(args[0].equals("normalReply")) {
	    System.out.println("Will receive replies...");
	    testObject._receiveReplies = true;
	}

	testObject.initializeRms(args);

	int loopCount = 1;
	long delay = 1000;
	try {
	    loopCount = Integer.parseInt(args[4], 10);
	    delay = Long.parseLong(args[5], 10);
	}
	catch (Exception excpt) {}

	testObject._messagesToSend = loopCount;
	testObject.startReplyThread();

	System.out.println("Going to send " + loopCount + " messages with a delay of " + delay + " milliseconds:");

	long startTime = System.currentTimeMillis();
	for (int idx = 0; idx < loopCount; idx++) {
	    testObject.sendMessage(idx);
	    if ((idx+1) < loopCount) {
	    try {Thread.sleep(delay);}
	    catch (Exception excpt) {}
	    }
	}

	long elapsedTime = System.currentTimeMillis() - startTime;
	
	try {
	    testObject._replyThread.join();
	}
	catch (Exception excp) {
	    System.out.println("Receive an exception while trying to join the reply thread:");
	    System.out.println(excp);
	}

	System.out.println("");

	if (loopCount > 0) {
	    java.text.DecimalFormat decFmt =
		new java.text.DecimalFormat("#.####");
	    double avgTime = elapsedTime / (1000.0d * loopCount);
	    System.out.print("Average send time = " + decFmt.format(avgTime) +
			     " seconds");
	    if (avgTime > 0.0d) {
		decFmt = new java.text.DecimalFormat("#.#");
		double avgRate = 1.0d / avgTime;
		System.out.print(" (Average rate = " + decFmt.format(avgRate) +
				 " Hz)");
	    }
	    System.out.println("");
	}

	System.exit(0);
    }
}
