import gov.fnal.cd.rms.provider.RmsConnection;
import gov.fnal.cd.rms.provider.RmsConnectionFactory;

import gov.fnal.cd.rms.base.RmsDestination;
import gov.fnal.cd.rms.base.RmsCommunicationException;
import gov.fnal.cd.rms.util.RmsException;

import gov.fnal.cd.rms.RmsReceiver;
import gov.fnal.cd.rms.RmsSender;

import java.util.*;

/**
 * This is simple example showing how to receive messages
 * and send replies using RMS and DDS.
 */
public class RmsReceiverExample {

    public static void main(String args[]) {
        RmsConnection rmsConnection = null;
        RmsReceiver<rmsexample.RmsExampleMessage> requestReceiver = null;
        RmsDestination receiveDest = null;
        Hashtable<String, RmsSender<rmsexample.RmsExampleMessage> >
            senderTable = new
            Hashtable<String, RmsSender<rmsexample.RmsExampleMessage> >();
        RmsDestination sendDest = null;

        boolean verbosePrintout = false;
        boolean idleStatusPrintout = false;
        int loopCount = 10;
        int argIdx = 0;
        while (argIdx < args.length) {
            String currentArg = args[argIdx];
            if (currentArg.equals("-v")) {
                verbosePrintout = true;
            }
            else if (currentArg.equals("-i")) {
                idleStatusPrintout = true;
            }
            else if (currentArg.equals("-n") && (argIdx+1) < args.length) {
                try {
                    argIdx++;
                    loopCount = Integer.parseInt(args[argIdx]);
                }
                catch (Exception excpt) {
                    excpt.printStackTrace(System.err);
                    System.exit(1);
                }
            }
            else {
                System.out.println("Usage: java RmsReceiverExample [-v] " +
                                   "[-i] [-n loopCount]");
                System.out.println("  where -v enables all diagnostic messages,");
                System.out.println("    and -i just enables the idle status printout.");
                System.exit(1);
            }
            argIdx++;
        }

        try {
            /**
             * Create an RmsConnection object using the connection factory.
             * We want replies routed to the destination with a target equal
             * to "RmsReceiverExample" so we set that as the applicationName
             * and -1 as the partition as we don't care about partitions.
             */
            rmsConnection = RmsConnectionFactory.
                getConnection("RmsReceiverExample", -1,
                              RmsConnectionFactory.DDS_PROVIDER);

            /**
             * Create the destination to represent where we want to receive
             * messages from.  For the example, we want to receive messages
             * send to the "RmsReceiverExample" target.  The messages will
             * use the control channel.
             */
            receiveDest = new RmsDestination("RmsReceiverExample",
                                             RmsDestination.CONTROL_CHANNEL,
                                             RmsDestination.REQUEST_MESSAGE_TYPE);

            /**
             * Create a receiver with the connection and the destination.
             */
            requestReceiver = new RmsReceiver<rmsexample.RmsExampleMessage>
                (rmsConnection, rmsexample.RmsExampleMessage.class,
                 receiveDest);

            for (int idx = 0; idx < loopCount; idx++) {

                /**
                 * Wait for a message to show up.
                 */
                Vector<rmsexample.RmsExampleMessage> msgList =
                    requestReceiver.receiveMessage(2000);
                while (msgList == null || msgList.size() == 0) {
                    if (verbosePrintout || idleStatusPrintout) {
                        System.out.println(idx + " messages received so far.");
                    }
                    msgList = requestReceiver.receiveMessage(2000);
                }
                idx += (msgList.size() - 1);

                for (int jdx = 0; jdx < msgList.size(); jdx++) {
                    rmsexample.RmsExampleMessage incomingMessage =
                        msgList.elementAt(jdx);
                    String replyTarget = incomingMessage.header.replyTarget;
                    if (verbosePrintout) {
                        System.out.print("Received a message: ");
                        System.out.println(idx + " " + incomingMessage.message);
                    }

                    RmsSender<rmsexample.RmsExampleMessage> replySender =
                        senderTable.get(replyTarget);
                    if (replySender == null) {

                        /**
                         * Create the destination to represent where the replies are
                         * to be sent.
                         */
                        sendDest = new RmsDestination(replyTarget,
                                                      RmsDestination.CONTROL_CHANNEL,
                                                      RmsDestination.REPLY_MESSAGE_TYPE);

                        /**
                         * Create a sender with the connection and the destination.
                         */
                        replySender = new RmsSender<rmsexample.RmsExampleMessage>
                            (rmsConnection, rmsexample.RmsExampleMessage.class, sendDest);
                        senderTable.put(replyTarget, replySender);
                    }

                    // Create and send the reply
                    rmsexample.RmsExampleMessage replyMessage =
                        new rmsexample.RmsExampleMessage();
                    replyMessage.message = "Hello World Reply!";
                    replyMessage.header.correlationId = incomingMessage.header.id;
                    replyMessage.header.correlationTime = incomingMessage.header.sentTime;
                    if (verbosePrintout) {
                        System.out.println("Sending reply to " + 
                                           incomingMessage.header.replyTarget + "...");
                    }
                    replySender.sendMessage(replyMessage);
                }
            }
            System.out.println(loopCount + " messages received.");

            /**
             * Cleanup handled in finally block
             */
        }
        catch (RmsCommunicationException commException) {
            System.out.println("Received a communication exception:");
            System.out.println(commException);
        }
        catch (RmsException rmsException) {
            System.out.println("Received an RMS exception:");
            System.out.println(rmsException);
        }
        finally {
            try {
                Enumeration< RmsSender<rmsexample.RmsExampleMessage> > senderEnum =
                    senderTable.elements();
                while (senderEnum.hasMoreElements()) {
                    RmsSender<rmsexample.RmsExampleMessage> sender = senderEnum.nextElement();
                    sender.close();
                }
                if (requestReceiver != null) {
                    requestReceiver.close();
                }
                if (rmsConnection != null) {
                    rmsConnection.close();
                }
            }
            catch (RmsCommunicationException commException) {
                System.out.println("There was a problem cleaning up:");
                System.out.println(commException);
            }
        }

        System.exit(0);
    }
}
