import gov.fnal.cd.rms.provider.RmsConnection;
import gov.fnal.cd.rms.provider.RmsConnectionFactory;

import gov.fnal.cd.rms.base.RmsDestination;
import gov.fnal.cd.rms.base.RmsCommunicationException;
import gov.fnal.cd.rms.util.RmsException;

import gov.fnal.cd.rms.RmsReceiver;
import gov.fnal.cd.rms.RmsSender;

import java.util.Vector;

/**
 * This is simple example showing how to send messages
 * and receive replies using RMS and DDS.
 */

public class RmsSenderExample {

    public static void main(String args[]) {
        RmsConnection rmsConnection = null;
        RmsSender<rmsexample.RmsExampleMessage> requestSender = null;
        RmsDestination sendDest = null;
        RmsReceiver<rmsexample.RmsExampleMessage> replyReceiver = null;
        RmsDestination receiveDest = null;

        String thisAppName = "RmsSenderExample";
        boolean verbosePrintout = false;
        boolean idleStatusPrintout = false;
        int loopCount = 10;
        int argIdx = 0;
        while (argIdx < args.length) {
            String currentArg = args[argIdx];
            if (currentArg.equals("-v")) {
                verbosePrintout = true;
            }
            else if (currentArg.equals("-i")) {
                idleStatusPrintout = true;
            }
            else if (currentArg.equals("-n") && (argIdx+1) < args.length) {
                try {
                    argIdx++;
                    loopCount = Integer.parseInt(args[argIdx]);
                }
                catch (Exception excpt) {
                    excpt.printStackTrace(System.err);
                    System.exit(1);
                }
            }
            else if (currentArg.equals("-a") && (argIdx+1) < args.length) {
                argIdx++;
                thisAppName = args[argIdx];
            }
            else {
                System.out.println("Usage: java RmsReceiverExample [-v] " +
                                   "[-i] [-n loopCount] [-a appName]");
                System.out.println("  where -v enables all diagnostic messages,");
                System.out.println("        -i just enables the idle status printout,");
                System.out.println("    and the application name is used in the " +
                                   "routing of the messages.");
                System.exit(1);
            }
            argIdx++;
        }
        System.out.println("Starting sender app with name \"" +
                           thisAppName + "\".");

        try {
            /**
             * Create an RmsConnection object using the connection factory.
             * We want replies routed to the destination with a target equal
             * to "RmsSenderExample" so we set that as the applicationName
             * and -1 as the partition as we don't care about partitions.
             */
            rmsConnection = RmsConnectionFactory.
                getConnection(thisAppName, -1,
                              RmsConnectionFactory.DDS_PROVIDER);

            /**
             * Create the destination to represent where the messages are
             * to be sent.  For the example, we're sending messages to the
             * "RmsReceiverExample" target.  The messages will use the
             * control channel.
             */
            sendDest = new RmsDestination("RmsReceiverExample",
                                          RmsDestination.CONTROL_CHANNEL,
                                          RmsDestination.REQUEST_MESSAGE_TYPE);

            /**
             * Create the destination to represent where the replies are
             * to be received from.
             */
            receiveDest = new RmsDestination(thisAppName,
                                             RmsDestination.CONTROL_CHANNEL,
                                             RmsDestination.REQUEST_MESSAGE_TYPE);

            /**
             * Create a sender with the connection and the destination.
             */
            requestSender = new RmsSender<rmsexample.RmsExampleMessage>
                (rmsConnection, rmsexample.RmsExampleMessage.class, sendDest);

            /**
             * Create a receiver with the connection and the destination.
             */
            replyReceiver = new RmsReceiver<rmsexample.RmsExampleMessage>
                (rmsConnection,rmsexample.RmsExampleMessage.class,receiveDest);

            for (int idx = 0; idx < loopCount; idx++) {

                /**
                 * Create an RmsExampleMessage and stuff some content into it.
                 */
                rmsexample.RmsExampleMessage exampleMessage =
                    new rmsexample.RmsExampleMessage();
                exampleMessage.message = "Hello World!";

                /**
                 * Send the message.
                 */
                if (verbosePrintout) {
                    System.out.println("Sending message...");
                }
                requestSender.sendMessage(exampleMessage);

                /**
                 * Wait for a reply.
                 */
                Vector<rmsexample.RmsExampleMessage> msgList =
                    replyReceiver.receiveMessage(2000);
                while (msgList == null || msgList.size() == 0) {
                    if (verbosePrintout || idleStatusPrintout) {
                        System.out.println(idx + " replies received so far.");
                    }
                    msgList = replyReceiver.receiveMessage(2000);
                }
                if (verbosePrintout) {
                    for (int msgIdx = 0; msgIdx < msgList.size(); msgIdx++) {
                        System.out.print("Received a reply: ");
                        System.out.println((idx+msgIdx) + " " +
                                           msgList.elementAt(msgIdx).message);
                    }
                }
                idx += (msgList.size() - 1);
            }
            System.out.println(loopCount + " messages sent.");

            /**
             * Cleanup handled in the finally block
             */
        }
        catch (RmsCommunicationException commException) {
            System.out.println("Received a communication exception:");
            System.out.println(commException);
        }
        catch (RmsException rmsException) {
            System.out.println("Received an RMS exception:");
            System.out.println(rmsException);
        }
        finally {
            try {
                if (replyReceiver != null) {
                    replyReceiver.close();
                }
                if (requestSender != null) {
                    requestSender.close();
                }
                if (rmsConnection != null) {
                    rmsConnection.close();
                }
            }
            catch (RmsCommunicationException commException) {
                System.out.println("There was a problem cleaning up:");
                System.out.println(commException);
            }
        }

        System.out.println("Exiting sender app with name \"" +
                           thisAppName + "\".");
            System.exit(0);
    }
}
